/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.awt.BasicStroke;
import java.io.*;


/**
* The sole purpose of this class is to aid classes which have a BasicStroke field that needs to get serialized.
* <p>
* <i>This class would be unnecessary of Sun were to merely make BasicStroke implement Serializable.</i>
* I filed a bug report with Sun about this; it has been assigned an internal review ID of: 383603
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @author Brent Boyer
*/
public class BasicStrokeSerializer implements Serializable {


	// -------------------- constants --------------------


	private static final long serialVersionUID = 1;


	// -------------------- instance fields --------------------
	
	
	/** @serial */
	private final float width;


	/** @serial */
	private final int cap;


	/** @serial */
	private final int join;


	/** @serial */
	private final float miterlimit;


	/** @serial */
	private final float[] dash;


	/** @serial */
	private final float dash_phase;
	
	
	// -------------------- writeBasicStroke, readBasicStroke --------------------


	/**
	* Constructs a new {@link #BasicStrokeSerializer} instance from basicStroke
	* and serializes it to objectOutputStream.
	* <p>
	* @throws IllegalArgumentException if basicStroke == null; objectOutputStream == null
	* @throws InvalidClassException if something is wrong with a class used by serialization
	* @throws NotSerializableException if some object to be serialized does not implement the java.io.Serializable interface
    * @throws IOException if any is thrown by the underlying OutputStream
	*/
	public static void writeBasicStroke(BasicStroke basicStroke, ObjectOutputStream objectOutputStream) throws IllegalArgumentException, InvalidClassException, NotSerializableException, IOException {
		if (basicStroke == null) throw new IllegalArgumentException("basicStroke == null");
		if (objectOutputStream == null) throw new IllegalArgumentException("objectOutputStream == null");
		
		BasicStrokeSerializer serializer = new BasicStrokeSerializer(basicStroke);
		objectOutputStream.writeObject(serializer);
	}


	/**
	* Deserializes a BasicStrokeSerializer object from objectInputStream,
	* {@link #convert converts} it to a new BasicStroke instance, and returns it.
	* <p>
	* @throws IllegalArgumentException if objectInputStream == null
	* @throws ClassNotFoundException if a class of a serialized object cannot be found
	* @throws InvalidClassException if something is wrong with a class used by serialization
	* @throws StreamCorruptedException if control information in the stream is inconsistent
	* @throws OptionalDataException if primitive data was found in the stream instead of objects
    * @throws IOException if any is thrown by the underlying InputStream
    * @throws ClassCastException if the next Object on objectInputStream is not a BasicStrokeSerializer instance
	*/
	public static BasicStroke readBasicStroke(ObjectInputStream objectInputStream) throws IllegalArgumentException, ClassNotFoundException , InvalidClassException, StreamCorruptedException, OptionalDataException, IOException, ClassCastException {
		if (objectInputStream == null) throw new IllegalArgumentException("objectInputStream == null");
		
		BasicStrokeSerializer serializer = (BasicStrokeSerializer) objectInputStream.readObject();
		return serializer.convert();
	}
	
	
	// -------------------- constructor --------------------
	
	
	/**
	* Creates a new BasicStrokeSerializer instance.
	* <p>
	* @throws IllegalArgumentException if basicStroke == null
	*/
	public BasicStrokeSerializer(BasicStroke basicStroke) throws IllegalArgumentException {
		if (basicStroke == null) throw new IllegalArgumentException("basicStroke == null");

		width = basicStroke.getLineWidth();
		cap = basicStroke.getEndCap();
		join = basicStroke.getLineJoin();
		miterlimit = basicStroke.getMiterLimit();
		dash = basicStroke.getDashArray();
		dash_phase = basicStroke.getDashPhase();
	}


	// -------------------- convert --------------------
	
	
	/** Converts this instance into a new BasicStroke instance and returns it. */
	public BasicStroke convert() {
		return new BasicStroke(width, cap, join, miterlimit, dash, dash_phase);
	}


	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		public static void main(String arg[]) throws Exception {
			BasicStroke strokeOriginal = new BasicStroke(1.0f);
			
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			ObjectOutputStream oos = new ObjectOutputStream( baos );
			writeBasicStroke(strokeOriginal, oos);
			ObjectInputStream ois = new ObjectInputStream( new ByteArrayInputStream( baos.toByteArray() ) );
			BasicStroke strokeDeserialized = readBasicStroke(ois);

			System.out.println("strokeOriginal.equals(strokeDeserialized): " + strokeOriginal.equals(strokeDeserialized));
		}
		
	}


}
