/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*

Programmer Notes:

--is this class needed?  or can you just use flow layout and set a components x & y alignments?

*/


package bb.gui;


import java.awt.*;
import java.awt.event.*;
import java.io.Serializable;

import javax.swing.*;


/**
* This LayoutManager always draws each Component centered inside the target Container,
* and at its preferred size. Here, "centered" means that the Component's bounding rectangle's
* center is the same as the center of the Container.
* <p>
* Typically, this LayoutManager will be used for Containers which have only a single
* Component inside, because multiple Components will get drawn over each other.
* But if overdrawing is satisfactory or desired, then this LayoutManager
* is also suitable for Containers with multiple Components inside.
* <p>
* Like typical java gui code, this class is not multithread safe:
* it expects to only be called by {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread}.
* <p>
* @author Brent Boyer
*/
public class CenterLayout implements LayoutManager, LayoutManager2, Serializable {


	// -------------------- constants --------------------


	private static final long serialVersionUID = 1;


	// -------------------- constructor --------------------


	/** Creates a new CenterLayout instance. */
	public CenterLayout() {}


	// -------------------- methods required by LayoutManager and LayoutManager2 --------------------


// +++ are the 2 methods below correct?
// Note that BorderLayout and CardLayout, GridBagLayout always return 0.5f for both x and y
// while BoxLayout, OverlayLayout do more complicated stuff

	/**
	* Returns the alignment along the X axis for the Container.
	* <p>
	* @param target a Container
	* @return the alignment (a value >= 0.0f && <= 1.0f)
	*/
	public float getLayoutAlignmentX(Container target) {
return 0.5f;
	}


	/**
	* Returns the alignment along the Y axis for the Container.
	* <p>
	* @param target a Container
	* @return the alignment (a value >= 0.0f && <= 1.0f)
	*/
	public float getLayoutAlignmentY(Container target) {
return 0.5f;
	}


	/**
	* Returns <code>preferredLayoutSize(target)</code>.
	* <p>
	* @param target the Container that needs to be laid out
	* @return a new Dimension instance (both the height and width are >= 0 && <= Integer.MAX_VALUE)
	* @see #preferredLayoutSize
	* @see #maximumLayoutSize
	*/
	public Dimension minimumLayoutSize(Container target) {
		return preferredLayoutSize(target);
	}


	/**
	* Returns the preferred dimensions needed to lay out the components contained
	* in the specified target container.
	* Invisible child Components are ignored in the calculation.
	* The result includes the effects of the target container's insets.
	* <p>
	* @param target the Container that needs to be laid out
	* @return a new Dimension instance (both the height and width are >= 0 && <= Integer.MAX_VALUE)
	* @see #minimumLayoutSize
	* @see #maximumLayoutSize
	*/
	public Dimension preferredLayoutSize(Container target) {
		Dimension d = new Dimension(0, 0);

		int nChildren = target.getComponentCount();
		for (int i = 0; i < nChildren; i++) {
			Component c = target.getComponent(i);
			if ( !c.isVisible() ) continue;

			Dimension childPrefSize = c.getPreferredSize();

			if (d.width < childPrefSize.width) d.width = childPrefSize.width;
			if (d.height < childPrefSize.height) d.height = childPrefSize.height;
		}

		Insets insets = target.getInsets();
		d.width += insets.left + insets.right;
		d.height += insets.top + insets.bottom;

		return d;
	}


	/**
	* Returns <code>target.getMaximumSize()</code>.
	* <p>
	* @param target the Container that needs to be laid out
	* @return a new Dimension instance (both the height and width are >= 0 && <= Integer.MAX_VALUE)
	* @see #minimumLayoutSize
	* @see #preferredLayoutSize
	*/
	public Dimension maximumLayoutSize(Container target) {
		return target.getMaximumSize();
	}
// +++ can it ever happen that target.getMaximumSize() is smaller than the result of preferredLayoutSize(Container)?


	/**
	* Not used by this class (is implemented as an empty method).
	* <p>
	* @param name the name of the component
	* @param comp the component
	*/
	public void addLayoutComponent(String name, Component comp) {}


	/**
	* Not used by this class (is implemented as an empty method).
	* <p>
	* @param comp the component
	* @param constraints constraints
	*/
	public void addLayoutComponent(Component comp, Object constraints) {}


	/**
	* Not used by this class (is implemented as an empty method).
	* <p>
	* @param target the affected container
	*/
	public void invalidateLayout(Container target) {}


	/**
	* Called by the AWT when the specified Container needs to be laid out.
	* <p>
	* @param target the container to lay out
	*/
	public void layoutContainer(Container target) {
		Dimension containerSize = target.getSize();

		int nChildren = target.getComponentCount();
		for (int i = 0; i < nChildren; i++) {
			Component c = target.getComponent(i);
			//if ( !c.isVisible() ) continue;		// go ahead and set the size and locations even of invisible ones, so that if they are turned back on, they will be correctly positioned

			Dimension cSize = c.getPreferredSize();

			c.setSize( cSize );

			int xOffset = (containerSize.width - cSize.width)/2;
			int yOffset = (containerSize.height - cSize.height)/2;
			c.setLocation( xOffset, yOffset );
		}
	}


	/**
	* Not used by this class (is implemented as an empty method).
	* <p>
	* @param comp the component
	*/
	public void removeLayoutComponent(Component comp) {}


	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {

		public static void main(String arg[]) {
				// see http://java.sun.com/docs/books/tutorial/uiswing/misc/threads.html for why must do ALL swing gui in event dispatch thread (the earlier advice given in http://java.sun.com/developer/JDCTechTips/2003/tt1208.html#1 to only do the displaying in the event dispatch thread is actually wrong)
			EventQueue.invokeLater( new Runnable() {
				public void run() {
					JFrame jframe1 = new JFrame("CenterLayout For A Single Child Component");
					jframe1.setLocation(20, 0);
					jframe1.setSize(500, 300);
					jframe1.getContentPane().add( buildSingleChildComponent() );
					jframe1.addWindowListener( new WindowAdapter() {
						public void windowClosing(WindowEvent e) { System.exit(0); }
					} );
					jframe1.setVisible(true);

					JFrame jframe2 = new JFrame("CenterLayout For Two Child Components");
					jframe2.setLocation(20, 320);
					jframe2.setSize(500, 300);
					jframe2.getContentPane().add( buildTwoChildComponents() );
					jframe2.addWindowListener( new WindowAdapter() {
						public void windowClosing(WindowEvent e) { System.exit(0); }
					} );
					jframe2.setVisible(true);
				}
			} );
		}

		protected static JPanel buildSingleChildComponent() {
			JPanel jpanel = new JPanel( new CenterLayout() );
			jpanel.add( new JLabel("Any text could be here") );
			return jpanel;
		}

		protected static JPanel buildTwoChildComponents() {
			JPanel jpanel = new JPanel( new CenterLayout() );
			jpanel.add( new JLabel("This label should be foremost, on top of the button") );
			jpanel.add( new JButton("OK") );
			return jpanel;
		}

	}


}
