/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.awt.*;
import java.awt.event.*;
import java.io.*;

import javax.swing.*;

import bb.util.StringUtil;


/**
* Subclass of {@link JDialog} which lets the user input some highly sensitive piece of text (e.g. a password).
* <p>
* The text is entered in a {@link TextFieldSecure} instance inside the dialog.
* This class allows optional header text above the TextFieldSecure (e.g. for instructions, a password hint, etc),
* as well as requires prompt text on the left to label the TextFieldSecure.
* It automatically puts a label on the right which specifies the maximum number of chars that may be entered.
* <p>
* The only public member is the static {@link #getInputSecure getInputSecure} method, which is all that normal users will use.
* For subclassing purposes, the other members of this class have protected access.
* <p>
* Like typical java gui code, this class is not multithread safe:
* it expects to only be called by {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread}.
* <p>
* @author Brent Boyer
*/
public class DialogInputSecure extends JDialog {


	// -------------------- constants --------------------


	private static final long serialVersionUID = 1;


	// -------------------- instance fields --------------------	
	
	
	/**
	* A {@link TextFieldSecure} instance that is used by this class to input highly sensitive text.
	* <p>
	* @serial
	*/
	private final TextFieldSecure textFieldSecure;


	// -------------------- getInputSecure --------------------


	/**
	* Reads and returns some sensitive piece of information (e.g. a password) from a new DialogInputSecure instance.
	* The result is a char[], not a String, so that it can be zeroed out after use.
	* <p>
	* See the {@link #DialogInputSecure constructor} for documentation of all the params and throws.
	*/
	public static final char[] getInputSecure(Frame parent, String title, boolean modal, String header, String prompt, int numberCharsMax) throws IllegalArgumentException {
		DialogInputSecure dialog = null;
		try {
			dialog = new DialogInputSecure(parent, title, modal, header, prompt, numberCharsMax);
			return dialog.textFieldSecure.getInput();
		}
		finally {
			if (dialog != null) dialog.textFieldSecure.zeroOutInput();
		}
	}


	// -------------------- constructor --------------------


	/**
	* Constructs a new DialogInputSecure instance.
	* <p>
	* @param parent the parent Frame that this JDialog will be attached to; may be null
	* @param title the dialog's title; may be null
	* @param modal specifies whether or not this JDialog is a modal dialog
	* @param header optional text (e.g. instructions, a password hint) displayed above the TextFieldSecure;
	* this text will be put into a JLabel, so it may be html formatted; may be null
	* @param prompt mandatory text displayed next to the TextFieldSecure;
	* this text will be put into a JLabel, so it may be html formatted; must be non-blank
	* @param numberCharsMax the maximum number of chars that may be typed in the TextFieldSecure
	* as well as the number of columns in the TextFieldSecure (this affects its displayed width); must be > 0
	* @throws IllegalArgumentException if prompt is blank; numberCharsMax <= 0
	*/
	protected DialogInputSecure(Frame parent, String title, boolean modal, String header, String prompt, int numberCharsMax) throws IllegalArgumentException {
		super(parent, title, modal);
		
		if (StringUtil.isBlank(prompt)) throw new IllegalArgumentException("prompt is blank");
		if (numberCharsMax <= 0) throw new IllegalArgumentException("numberCharsMax = " + numberCharsMax + " <= 0");
		
		textFieldSecure = new TextFieldSecure(numberCharsMax);
		
		setDefaultCloseOperation( DISPOSE_ON_CLOSE );
		getContentPane().add( buildGui(header, prompt, numberCharsMax) );
		pack();
		setLocationRelativeTo(parent);	// MUST do this call after have properly sized the dialog with pack
		setVisible(true);
	}
	
	
	// -------------------- build methods --------------------


	protected JComponent buildGui(String header, String prompt, int numberCharsMax) {
		JPanel jpanel = new JPanel( LineLayout.makeVertical() );
		jpanel.setBorder( BorderFactory.createEmptyBorder(20, 20, 20, 20) );
		if (header != null) {
			jpanel.add( new JLabel(header) );
			jpanel.add( Box.createVerticalStrut(20) );
		}		
		jpanel.add( buildSecureEntry(prompt, numberCharsMax) );
		jpanel.add( Box.createVerticalStrut(20) );
		jpanel.add( buildButtons() );
		return jpanel;
	}


	protected JComponent buildSecureEntry(String prompt, int numberCharsMax) {
		JPanel jpanel = new JPanel( LineLayout.makeHorizontal() );
		jpanel.add( new JLabel(prompt) );
		jpanel.add( Box.createHorizontalStrut(10) );
		jpanel.add( textFieldSecure );
		jpanel.add( Box.createHorizontalStrut(10) );
		jpanel.add( new JLabel("(" + numberCharsMax + " chars max)") );
		return jpanel;
	}
	
	
	protected JComponent buildButtons() {
		JPanel jpanel = new JPanel( LineLayout.makeHorizontal() );

		JButton helpButton = new JButton("Help");
		helpButton.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				String helpText =
					"This dialog allows you to enter text (e.g. a password) in a very secure manner." + "\n" +
					"\n" +
					"For top security, when you type in this dialog's text field, no visible characters are echoed." + "\n" +
					"Unlike some other interfaces, which at least echo some generic character like an asterisk ('*')," + "\n" +
					"this dialog echoes nothing, since that would give away the length of the text." + "\n" +
					"Furthermore, there is no caret (vertical line indicating text position) in the text field either," + "\n" +
					"since that too might give away the length of the text." + "\n" +
					"\n" +
					"Because of the above, it is important to give the user some visual cue when the text field has keyboard focus." + "\n" +
					"(Keyboard focus means that if you type on the keyboard, the text is entered in the component" + "\n" +
					" which has the focus, as opposed to some other text component which might be present in the interface.)" + "\n" +
					"This dialog indicates when its text field has keyboard focus by setting the field's background color to white," + "\n" +
					"and it indicates when it lacks keyboard focus by setting its background color to light red." + "\n" +
					"\n" +
					"If you type in the text field and then click on the OK button" + "\n" +
					"(or type the enter/return key on some platforms), then this dialog disappears" + "\n" +
					"but the text that you typed in (except for any possible enter/return key)" + "\n" +
					"is retained (and should be read by some other part of the program)." + "\n" +
					"\n" +
					"Else, if you type in the text field and then click on the Cancel button," + "\n" +
					"then this dialog disappears and the text that you typed in is lost.";
				JOptionPane.showMessageDialog(DialogInputSecure.this, helpText, "DialogInputSecure Help", JOptionPane.INFORMATION_MESSAGE);				
			}
		} );
		jpanel.add( helpButton );

		jpanel.add( Box.createHorizontalStrut(20) );

		JButton okButton = new JButton("OK");
		okButton.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				dispose();
			}
		} );
		this.getRootPane().setDefaultButton(okButton);
		jpanel.add( okButton );

		jpanel.add( Box.createHorizontalStrut(20) );

		JButton cancelButton = new JButton("Cancel");
		cancelButton.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				textFieldSecure.zeroOutInput();
				textFieldSecure.setText("");	// need to call this as well, since the above does not affect the position indices or anything
				dispose();
			}
		} );
		jpanel.add( cancelButton );

		return jpanel;
	}


	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		public static void main(String[] args) throws Exception {
			test_getInputSecure();
//			test_serialization();
// +++ the method above fails with the message
//	Exception in thread "main" java.io.NotSerializableException: javax.swing.plaf.basic.BasicTextUI$UpdateHandler
// This appears to be a known bug that sun will be fixing in 1.5_02 which is not out yet; see
//	http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6190713
		}
		
		private static void test_getInputSecure() {
			String instructions =
				"<html>" +
					"<p>" +
						"Type <i>anything</i> that you wish in the text field below." + "<br>" +
						"When you hit the OK button, this dialog will disappear" + "<br>" +
						"and the text that you typed will be printed to System.out." +
					"</p>" +
				"</html>";
			char[] chars = DialogInputSecure.getInputSecure(null, "UnitTest Of DialogInputSecure.getInputSecure", true, instructions, "Text input:", 5);
			String inputText = new String( chars );
			System.out.println("The text that was input to the DialogInputSecure is:");
			System.out.println(inputText);
		}
		
// +++ what am hoping to confirm in this method is that the deserialized dialog contains no text initially
// (i.e. that the serialization features promised in TextFieldSecure/ContentSecure work as described)
// and that the deserialized dialog otherwise works as before (that it has an invisible caret etc)
		private static void test_serialization() throws Exception {
			DialogInputSecure dialog = new DialogInputSecure(null, "UnitTest Of DialogInputSecure serialization", true, null, "Text input:", 5);
			String text = new String( dialog.textFieldSecure.getPassword() );
			System.out.println("text entered in the unserialized dialog: " + text);
			dialog.dispose();
			
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			ObjectOutputStream oos = new ObjectOutputStream( baos );
			oos.writeObject( dialog );
			ObjectInputStream ois = new ObjectInputStream( new ByteArrayInputStream( baos.toByteArray() ) );
			DialogInputSecure dialogDeserialized = (DialogInputSecure) ois.readObject();
			dialogDeserialized.setTitle("This is the deserialized dialog");
			dialogDeserialized.setVisible(true);
			String textDeserialized = new String( dialogDeserialized.textFieldSecure.getPassword() );
			System.out.println("text present inside the deserialized dialog: " + textDeserialized);
			dialogDeserialized.dispose();
		}
		
	}


}
