/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*

--an alternative to this class which is xml driven is the JDNC project:
	https://jdnc.dev.java.net/

*/


package bb.gui;


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


/**
* This class simply pops up a new JFrame that contains the JComponent supplied to its constructor.
* It is mainly used as a GUI development aid to quickly show stuff.
* <p>
* This class has the convenience feature that the location of the last instance which was moved or had its location changed
* defines where all subsequent instances will initially be located at.
* <p>
* Like all java gui code, this class is generally not multithread safe:
* it expects to only be called by {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread}.
* The sole exception is the {@link #show show} method, which can safely be called by any method.
* <p>
* @author Brent Boyer
*/
public class Displayer extends JFrame {


	// -------------------- constants --------------------


	private static final long serialVersionUID = 1;


	// -------------------- fields --------------------


	/** Every time any Displayer instance is moved, this field stores its latest x position. */
	private static int latestX = 0;
	/** Every time any Displayer instance is moved, this field stores its latest y position. */
	private static int latestY = 0;
// +++ use the new java preferences api to store this data between jvm sessions
	
	
	// -------------------- show --------------------
	
	
	/**
	* Constructs a new Displayer instance from the args.
	* <p>
	* This method may be safely called by any thread because
	* the work is asynchronously done by a task submitted to EventQueue's dispatch thread.
	* <i>One consequence is that this method will likely return before the new Displayer instance is visible.</i>
	* <p>
	* @throws IllegalArgumentException if jcomponent == null
	*/
	public static void show(final JComponent jcomponent, final String title) throws IllegalArgumentException {
		if (jcomponent == null) throw new IllegalArgumentException("jcomponent == null");
		// no constraint on title
		// no constraint on calling thread

		EventQueue.invokeLater( new Runnable() {
			public void run() {
				new Displayer(jcomponent, title);
			}
		} );			
	}
	
	
	// -------------------- constructor --------------------
	
	
	/**
	* Constructor.
	* Shows the specified JComponent in a new JFrame.
	* <p>
	* @throws IllegalArgumentException if jcomponent == null
	* @throws IllegalStateException if calling thread is not EventQueue's dispatch thread
	*/
	public Displayer(JComponent jcomponent, String title) throws IllegalArgumentException, IllegalStateException {
		super();
		
		if (jcomponent == null) throw new IllegalArgumentException("jcomponent == null");
		// no constraint on title
		if (!EventQueue.isDispatchThread()) throw new IllegalStateException("calling thread (" + Thread.currentThread().toString() + ") is not EventQueue's dispatch thread");
		
		getContentPane().setLayout( new CenterLayout() );	// after call pack below, there will be no space left over, HOWEVER, the user can manually set the size of this instance to something bigger, in which case having this layout is nice
		getContentPane().setBackground( new Color(225, 225, 225) );	// after call pack below, there will be no background left, HOWEVER, the user can manually set the size of this instance to something bigger, in which case having this bg color is nice
		
		if (title != null) setTitle( title );
		setLocation(latestX, latestY);
		//setSize(600, 500);	// obsolete -- will call pack below (will do an optimal setting)
		
		addComponentListener( new ComponentAdapter() {
			public void componentMoved(ComponentEvent ce) {
				Component c = ce.getComponent();
				Displayer.latestX = c.getX();
				Displayer.latestY = c.getY();
			}
		} );
		
		setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		
		getContentPane().add( jcomponent );
		pack();
		setVisible(true);
	}


	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {

		public static void main(String[] args) throws Exception {
			Displayer.show( new JLabel("TRY MOVING THIS WINDOW IN THE NEXT 5 SECONDS!"), "First instance" );

			Thread.sleep(5*1000);

			Displayer.show(
				new JLabel(
					"<html>" +
						"<p>" +
							"If you moved the first window within 5 seconds," + "<br>" +
							"this one should be created in its last location" + "<br>" +
							"(and will then need to move this to see the first one again)" + "<br>" +
						"</p>" +
					"</html>"
				),
				"Second instance"
			);
//			displayer2.setSize(400, 400);
		}

	}


}
