/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.awt.EventQueue;
import javax.swing.UIManager;
import javax.swing.text.*;

import bb.util.ThrowableUtil;


/**
* Subclass of {@link PlainDocument} which limits the number of chars that it will contain.
* <p>
* One use of this class is to serve as the model of a JTextComponent like JTextField
* when the text needs to be constrained to a maximum length.
* <p>
* Like typical java gui code, this class is not multithread safe:
* it expects to only be called by {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread}.
* <p>
* @see <a href="http://forum.java.sun.com/thread.jspa?forumID=57&threadID=404834">Java forum posting #1</a>
* @see <a href="http://forum.java.sun.com/thread.jspa?forumID=57&threadID=340706">Java forum posting #2</a>
* @author Brent Boyer
*/
public class DocumentLimitedLength extends PlainDocument {


	// -------------------- constants --------------------


	private static final long serialVersionUID = 1;


	// -------------------- fields --------------------	
	
	
	/**
	* Records the maximum number of chars that this instance will hold.
	* <p>
	* Contract: is > 0.
	* <p>
	* @serial
	*/
	private final int numberCharsMax;
	
	
	// -------------------- constructor --------------------
	
	
	/**
	* Constructs a new DocumentLimitedLength instance.
	* <p>
	* @param numberCharsMax specifies the maximum number of chars that this instance will hold; must be > 0
	* @throws IllegalArgumentException if numberCharsMax <= 0
	*/
	public DocumentLimitedLength(int numberCharsMax) throws IllegalArgumentException {
		this( new GapContent(), numberCharsMax );
	}
	
	
	/**
	* Constructs a new DocumentLimitedLength instance.
	* <p>
	* @param content a {@link javax.swing.AbstractDocument.Content} instance that contains initial text content
	* @param numberCharsMax specifies the maximum number of chars that this instance will hold; must be > 0
	* @throws IllegalArgumentException if content == null; numberCharsMax <= 0; content.length() > numberCharsMax
	*/
	public DocumentLimitedLength(AbstractDocument.Content content, int numberCharsMax) throws IllegalArgumentException {
		super(content);
		
		if (content == null) throw new IllegalArgumentException("content == null");
		if (numberCharsMax <= 0) throw new IllegalArgumentException("numberCharsMax = " + numberCharsMax + " <= 0");
		if (content.length() > numberCharsMax) throw new IllegalArgumentException("content.length() = " + content.length() + " > numberCharsMax = " + numberCharsMax);

		this.numberCharsMax = numberCharsMax;
	}
	

	// -------------------- accessors and mutators --------------------


	/** Returns the maximum number of chars that may be typed in this instance. */
	public int getNumberCharsMax() { return numberCharsMax; }
	
	
	/**
	* Returns all the text contained in this instance.
	* <p>
	* @throws RuntimeException (or some subclass) if any error occurs; this may merely wrap some other underlying Throwable
	*/
	public String getText() throws RuntimeException {
		try {
			return getText(0, getLength());
		}
		catch (Throwable t) {
			throw ThrowableUtil.toRuntimeException(t);	// this line should never get called, but need the catch block to suppress the compiler from forcing a throws BadLocationException declaration
		}
	}
// +++ why does Document lack this method?  must have been an oversight by Sun...


	// -------------------- Document methods --------------------
	
	
	/**
	* First calls <code>super.insertString(offset, s, attributeSet)</code>.
	* Then checks to see if the new length exceeds {@link #numberCharsMax}, and if it does,
	* sounds a beep and then truncates the final chars so that the length ends up equaling numberCharsMax.
	*/
	public void insertString(int offset, String s, AttributeSet attributeSet) throws BadLocationException {
		// all args checked by the call to super.insertString below

		super.insertString(offset, s, attributeSet);
		
		if (getLength() > numberCharsMax) {	
			UIManager.getLookAndFeel().provideErrorFeedback(null);
			remove(numberCharsMax, getLength() - numberCharsMax);
		}
	}

	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {

		public static void main(String[] args) throws Exception {
			DocumentLimitedLength doc = new DocumentLimitedLength(10);
			doc.insertString(0, "abcdefghij", null);	// add first 10 chars of the alphabet
			System.out.println("doc after step #1: " + doc.getText());
			doc.insertString(5, "123", null);	// add first 3 natural numbers starting at index 5
			System.out.println("doc after step #2: " + doc.getText());
			System.out.println("did insertString work as expected: " + doc.getText().equals("abcde123fg") );
		}

	}


}
