/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.awt.*;
import java.util.*;

import bb.util.Benchmark;


/**
* This class provides miscellaneous static Font utility methods.
* <p>
* This class is multithread safe.
* All non-multithread safe fields are only accessed inside synchronized methods.
* <p>
* @author Brent Boyer
*/
public class FontUtil {
	
	
	// -------------------- constants --------------------
	
	
	/** The default Font point size. */
	private static final float fontSize_default = 12.0f;
	
	
	// -------------------- static fields --------------------
	
	
	/** Records all the Fonts available on this local system. */
	private static Font[] fonts = null;
	
	
	// -------------------- static accessors & mutators --------------------
	
	
	/**
	* Returns all the Fonts available on this system <i>since the last call to {@link #resetFonts resetFonts}</i>.
	* <p>
	* A call to resetFonts will automaticly be done the first time this method is called
	* in order to perform initialization.
	* The user never need call resetFonts <i>unless Fonts are added/removed</i>.
	* <p>
	* Contract: the result will have a length >= 20,
	* and each Font's size will be {@link #fontSize_default fontSize_default}; see {@link #resetFonts resetFonts}.
	* <p>
	* @throws IllegalStateException if GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts().length < 20
	*/
	public static synchronized Font[] getFonts() throws IllegalStateException {
		if (fonts == null) resetFonts();
		return fonts.clone();
	}
	
	
	/**
	* Calls {@link GraphicsEnvironment#getAllFonts GraphicsEnvironment.getLocalGraphicsEnvironment.getAllFonts}
	* to determine all the Fonts which are currently available on the running system,
	* and stores them in an internal field.
	* <p>
	* <i>Note</i>: the number of Fonts returned by GraphicsEnvironment.getLocalGraphicsEnvironment.getAllFonts
	* should be at least 20, because it should always return, at a minimum, Java's 5 logical Fonts
	* (i.e. Serif, SansSerif, Monospaced, Dialog, DialogInput)
	* as well as the 4 style variants (i.e. plain, bold, italic, bolditalic) of each of those Fonts.
	* This method will confirm this minimum length.
	* <p>
	* <i>Note</i>: since the Fonts returned by GraphicsEnvironment.getLocalGraphicsEnvironment.getAllFonts
	* always have a size of 1 point, which is almost invisible on most systems,
	* they are resized to {@link #fontSize_default fontSize_default}
	* before this method returns for the benefit of subsequent users.
	* <p>
	* @throws IllegalStateException if GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts().length < 20
	*/
	public static synchronized void resetFonts() throws IllegalStateException {
		fonts = GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts();
		if (fonts.length < 20) throw new IllegalStateException("GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts().length = " + fonts.length + " which is < 20");
		for (int i = 0; i < fonts.length; i++) {
			fonts[i] = fonts[i].deriveFont(fontSize_default);
		}
	}
	
	
	// -------------------- canHandle, requiresGlyph --------------------
	
	
	/**
	* Determines whether or not the {@link Component#getFont component's Font}
	* is capable of displaying all of the chars of s that {@link #requiresGlyph require a glyph}.
	* <p>
	* This method is needed because the {@link Font#canDisplayUpTo Font.canDisplayUpTo} method
	* does not exclude chars that are highly unlikely to have a glyph in any Font.
	* <p>
	* To see why this method is needed,
	* suppose you want to determine if a JTextArea, say, is currently set to a Font
	* that is capable of displaying some text that contains line returns.
	* Here, you do not want the line returns to count in the Font determination because they will require no glyphs.
	* <p>
	* @throws IllegalArgumentException if component == null; s == null
	*/
	public static boolean canHandle(Component component, String s) throws IllegalArgumentException {
		if (component == null) throw new IllegalArgumentException("component == null");
		if (s == null) throw new IllegalArgumentException("s == null");
		
		Font font = component.getFont();
		for (int i = 0; i < s.length(); i++) {
			char c = s.charAt(i);
			if (requiresGlyph(c) && !font.canDisplay(c)) return false;
		}
		return true;
	}
	
	
	/**
	* Determines whether or not c requires a glyph from a Font or not.
	* This method assumes that all chars which are not {@link Character#isWhitespace whitespace characters} or
	* {@link Character#isISOControl control characters} require glyphs.
	*/
	public static boolean requiresGlyph(char c) {
		return
			!Character.isWhitespace(c) &&	// test for whitespace first, as is more common so can benefit from short-circuit evaluation
			!Character.isISOControl(c);
	}
	
	
	// -------------------- findBestFont, rankFonts --------------------
	
	
	/**
	* Determines that Font in the local system which can display the highest percentage of s's chars.
	* In the event of ties (e.g. several Fonts can display 100% of s's chars),
	* it returns that Font which was encountered first.
	* <p>
	* Contract: the result is never null, and will have a Font size of {@link #fontSize_default}.
	* <p>
	* @throws IllegalArgumentException if s == null
	* @throws IllegalStateException if GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts().length < 20
	*/
	public static Font findBestFont(String s) throws IllegalArgumentException, IllegalStateException {
		//return rankFonts(s)[0].getFont();	// elegant, but lacks the early return possibility of the code below which can be an enormous speed improvement (see benchmark results in UnitTest)
		
		Font bestFont = null;
		double bestRank = -1.0;
		for (Font font : getFonts()) {
			double rank = findDisplayableCharPercent(font, s);
			if (rank == 1.0) return font;	// can immediately return if find a font that can display 100% of s's chars
			if (rank > bestRank) {
				bestFont = font;
				bestRank = rank;
			}
		}
		if (bestFont == null) throw new IllegalStateException("failed to find a best Font; this should never happen");
		return bestFont;
	}
	
	
	/**
	* Retrieves every Font in the local system by calling {@link #getFonts getFonts}.
	* Determines the {@link FontRank} for each
	* using {@link FontUtil.FontRank#FontUtil.FontRank(Font, String) this constructor}.
	* Returns these FontRanks as an array, sorted by FontRank's {@link FontRank#compareTo natural ordering}.
	* <p>
	* Contract: see getFonts for guarantees on the Fonts in the result.
	* <p>
	* @throws IllegalArgumentException if s == null
	* @throws IllegalStateException if GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts().length < 20
	*/
	public static FontRank[] rankFonts(String s) throws IllegalArgumentException, IllegalStateException {
		if (s == null) throw new IllegalArgumentException("arg s is null");
		
		SortedSet<FontRank> fontRanks = new TreeSet<FontRank>();
		for (Font font : getFonts()) {
			fontRanks.add( new FontRank(font, s) );
		}
		return fontRanks.toArray( new FontRank[fontRanks.size()] );
	}
	
	
	// -------------------- findDisplayableCharPercent --------------------
	
	
	/**
	* Returns the percent of chars in the supplied String that can be displayed by the supplied Font.
	* <p>
	* @throws IllegalArgumentException if font == null; s == null or s.length() == 0
	*/
	public static double findDisplayableCharPercent(Font font, String s) throws IllegalArgumentException {
		if (font == null) throw new IllegalArgumentException("font == null");
		if (s == null) throw new IllegalArgumentException("s == null");
		if (s.length() == 0) throw new IllegalArgumentException("s.length() == 0");
		
		int displayCount = 0;
		for (int i = 0; i < s.length(); i++) {
			if (font.canDisplay( s.charAt(i) )) ++displayCount;
		}
		
		return displayCount / ((double) s.length());
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private FontUtil() {}
	
	
	// -------------------- FontRank (static inner class) --------------------
	
	
	/**
	* The main purpose of this class is simply to record a Font and an associated rank of that Font.
	* The semantic meaning of that rank is determined by the user.
	* <p>
	* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
<!--
+++ umm, is that deep state comment really true for Font?  Its javadocs never state this...
-->
	*/
	public static class FontRank implements Comparable<FontRank> {
		
		private final Font font;
		private final double rank;
		
		/**
		* Simply calls <code>this( font, {@link #findDisplayableCharPercent findDisplayableCharPercent}(font, s) )</code>.
		* In other words, it ranks font based on what percent of characters it can display.
		*/
		public FontRank(Font font, String s) {
			this( font, findDisplayableCharPercent(font, s) );
		}
		
		/**
		* Constructs a new FontRank instance from the supplied arguments.
		* The font arg may be any Font value except null; the rank arg may be any double value (including NaN and infinity).
		* <p>
		* @throws IllegalArgumentException if arg font is null
		*/
		public FontRank(Font font, double rank) throws IllegalArgumentException {
			if (font == null) throw new IllegalArgumentException("arg font is null");
			
			this.font = font;
			this.rank = rank;
		}
		
		public Font getFont() { return font; }
		public double getRank() { return rank; }
		
		/**
		* Only returns true if obj is another FontRank instance whose font and rank fields equal this.
		* <p>
		* @throws IllegalStateException if obj is another FontRank instance whose font field equals this, but whose rank field differs; this is meant to detect subtle bugs
		*/
		@Override public final boolean equals(Object obj) {	// for why is final, see the essay stored in the file equalsImplementation.txt
			if (this == obj) return true;
			if (!(obj instanceof FontRank)) return false;
			
			FontRank other = (FontRank) obj;
			if (this.font.equals(other.font)) {
				if (this.rank != other.rank) throw new IllegalStateException("this.font.equals(other.font) = true, but this.rank != other.rank");
				return true;
			}
			else {
				return false;
			}
		}
		
		@Override public final int hashCode() {	// for why is final, see the essay stored in the file equalsImplementation.txt
			return font.hashCode();	// just font suffices
		}
		
		@Override public String toString() { return "font = " + getFont().toString() + ", rank = " + getRank(); }
		
		/**
		* Attempts to orders two FontRanks based on comparing their ranks: larger rank is ordered before smaller rank.
		* If both ranks equal, then attempts to order them based on their Font names: normal alphabetical ordering used.
		* If both Font names equal, then attempts to order them based on their Font sizes.
		* If both Font sizes equal, then attempts to order them based on their Font styles.
		* If both Font styles equal, then returns 0 if their fonts are equal.
		* Otherwise throws an IllegalStateException.
		* <p>
		* The ordering imposed by this method is consistent with equals.
		* <p>
		* @throws IllegalArgumentException if other == null
		* @throws IllegalStateException if unable to order this with respect to other
		*/
		public int compareTo(FontRank other) throws IllegalArgumentException, IllegalStateException {
			if (other == null) throw new IllegalArgumentException("other == null");
			
			int rankComparison = Double.compare( other.rank, this.rank );	// CRITICAL: note the argument order inside Double.compare; also, use Double.compare because of how it handle special values
			if (rankComparison != 0) return rankComparison;
			
			int nameComparison = this.font.getName().compareTo( other.font.getName() );
			if (nameComparison != 0) return nameComparison;
			
			int sizeComparison = this.font.getSize() - other.font.getSize();
			if (sizeComparison != 0) return sizeComparison;
			
			int styleComparison = this.font.getStyle() - other.font.getStyle();
			if (styleComparison != 0) return styleComparison;
			
			// if look inside Font.equals, there are some other attributes that could possibly test, but will give up for now...
			
			if (this.font.equals(other.font)) return 0;
			
			//int identityHashComparison = System.identityHashCode(this) - System.identityHashCode(other);
			//if (identityHashComparison != 0) return identityHashComparison;
			
			throw new IllegalStateException("unable to order this = " + this + " with respect to other = " + other);
		}
		
	}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		/** The first letters of the English alphabet. */
		private static final String englishLettersFirst = "abcd";
		
		/**
		* The unique letters of the German alphabet (relative to English).
		* <p>
		* @see <a href="http://german.about.com/library/anfang/blanfang_abc.htm">German alphabet</a>
		* @see <a href="http://www.unicode.org/charts/PDF/U0080.pdf">The relevant Unicode code chart</a>
		*/
		private static final String germanLettersUnique = "\u00c4\u00d6\u00df\u00dc";
		
		/**
		* The first letters of the Chinese-Japanese-Korean (unified ideographs) alphabet.
		* <p>
		* @see <a href="http://www.unicode.org/charts/PDF/U4E00.pdf">The relevant Unicode code chart</a>
		*/
		private static final String cjkLettersFirst = "\u4e00\u4e01\u4e02\u4e03";
		
		/** Concatentation of {@link #englishLettersFirst}, {@link #germanLettersUnique}, {@link #cjkLettersFirst}. */
		private static final String charsFromManyLanguages = englishLettersFirst + germanLettersUnique + cjkLettersFirst;
		
		public static void main(String[] args) throws Exception {
// +++ add a test for getFonts in which display in a JTextArea or something a list of all the font names using the font itself
// +++ also need to test canHandle, requiresGlyph
			test_findDisplayableCharPercent();
			test_findBestFont();
			test_rankFonts();
		}
		
		private static void test_findDisplayableCharPercent() {
			System.out.println();
			System.out.println("test_findDisplayableCharPercent:");
			
			Font aLogicalFont = new Font("SansSerif", Font.PLAIN, 12);
			System.out.println("findDisplayableCharPercent(aLogicalFont, charsFromManyLanguages) = " + findDisplayableCharPercent(aLogicalFont, charsFromManyLanguages) );
		}
		
		/**
		* Results on 2007/11/27 on my 2.0 GHz Core 2 Duo laptop (server jvm):
		* <pre><code>
			findBestFont(charsFromManyLanguages): java.awt.Font[family=Arial Unicode MS,name=Arial Unicode MS,style=plain,size=12]
			findBestFont individual execution time statistics: mean = 22.659 us (CI deltas: -14.549 ns, +16.391 ns), sd = 11.032 us (CI deltas: -1.648 us, +2.500 us)</code></pre>
		*/
		private static void test_findBestFont() throws Exception {
			System.out.println();
			System.out.println("test_findBestFont:");
			
			System.out.println( "findBestFont(charsFromManyLanguages): " + findBestFont(charsFromManyLanguages) );
			
			Runnable task = new Runnable() { public void run() { findBestFont(charsFromManyLanguages); } };
			System.out.println("findBestFont: " + new Benchmark(task));
		}
		
		/**
		* (Abbreviated) Results on 2007/11/27 on my 2.0 GHz Core 2 Duo laptop (server jvm):
		* <pre><code>
			rankFonts(charsFromManyLanguages):
					font = java.awt.Font[family=Arial Unicode MS,name=Arial Unicode MS,style=plain,size=12], rank = 1.0
					font = java.awt.Font[family=Dialog,name=Dialog.bold,style=plain,size=12], rank = 1.0
					font = java.awt.Font[family=Dialog,name=Dialog.bolditalic,style=plain,size=12], rank = 1.0
					font = java.awt.Font[family=Dialog,name=Dialog.italic,style=plain,size=12], rank = 1.0
					...
					font = java.awt.Font[family=Webdings,name=Webdings,style=plain,size=12], rank = 0.0
					font = java.awt.Font[family=Wingdings,name=Wingdings,style=plain,size=12], rank = 0.0
					font = java.awt.Font[family=Wingdings 2,name=Wingdings 2,style=plain,size=12], rank = 0.0
					font = java.awt.Font[family=Wingdings 3,name=Wingdings 3,style=plain,size=12], rank = 0.0
			rankFonts individual execution time statistics: mean = 437.589 us (CI deltas: -156.454 ns, +194.276 ns), sd = 30.997 us (CI deltas: -7.322 us, +10.583 us)</code></pre>
		*/
		private static void test_rankFonts() throws Exception {
			System.out.println();
			System.out.println("test_rankFonts:");
			
			System.out.println("rankFonts(charsFromManyLanguages):");
			for (FontRank fontRank : rankFonts(charsFromManyLanguages)) {
				System.out.println("\t" + fontRank);
			}
			
			Runnable task = new Runnable() { public void run() { rankFonts(charsFromManyLanguages); } };
			System.out.println("rankFonts: " + new Benchmark(task));
		}
		
	}
	
	
}