/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.awt.*;

import javax.swing.*;
import javax.swing.event.*;


/**
* This subclass of JLabel adds exactly one new bit of functionality to its superclass:
* it is interested in mouse events.
* <p>
* This class was originally written to enable popup menus to be associated with JLabels.
* <a href="http://forum.java.sun.com/thread.jspa?threadID=590867&tstart=0">This forum posting</a> provides more discussion.
* <p>
* The fundamental reason why this class needed to be written was because, at present,
* Swing is designed to not pass mouse events to any lightweight component
* that does not have at least registered mouse listener,
* a design flaw that has been <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4966109">reported as a bug</a>
* but which Sun has yet to fix.
* <p>
* Like typical java gui code, this class is not multithread safe:
* it expects to only be called by {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread}.
* <p>
* @author Brent Boyer
*/
public class JLabel2 extends JLabel {


	// -------------------- constants --------------------


	private static final long serialVersionUID = 1;


	// -------------------- constructors --------------------

	
	/**
	* Creates a <code>JLabel2</code> instance with with an empty string for the title and a null image.
	* The label is aligned against the leading edge of its display area, and centered vertically.
	*/
	public JLabel2() {
		this("");
	}
	
	
	/**
	* Creates a <code>JLabel2</code> instance with the specified text and a null image.
	* The label is aligned against the leading edge of its display area, and centered vertically.
	* <p>
	* @param text The text to be displayed by the label
	*/
	public JLabel2(String text) {
		this(text, LEADING);
	}
	
	
	/**
	* Creates a <code>JLabel2</code> instance with the specified text and horizontal alignment, along with a null image.
	* The label is centered vertically in its display area.
	* <p>
	* @param text The text to be displayed by the label
	* @param horizontalAlignment One of the following constants defined in {@link SwingConstants}:
	* <code>LEFT</code>, <code>CENTER</code>, <code>RIGHT</code>, <code>LEADING</code>, or <code>TRAILING</code>
	*/
	public JLabel2(String text, int horizontalAlignment) {
		this(text, null, horizontalAlignment);
	}
	
	
	/**
	* Creates a <code>JLabel2</code> instance with the specified image.
	* The label is centered vertically and horizontally in its display area.
	* <p>
	* @param image The image to be displayed by the label
	*/
	public JLabel2(Icon image) {
		this(image, CENTER);
	}
	
	
	/**
	* Creates a <code>JLabel2</code> instance with the specified image and horizontal alignment.
	* The label is centered vertically in its display area.
	* <p>
	* @param image The image to be displayed by the label
	* @param horizontalAlignment One of the following constants defined in {@link SwingConstants}:
	* <code>LEFT</code>, <code>CENTER</code>, <code>RIGHT</code>, <code>LEADING</code> or, <code>TRAILING</code>
	*/
	public JLabel2(Icon image, int horizontalAlignment) {
		this(null, image, horizontalAlignment);
	}
	
	
	/**
	* Creates a <code>JLabel2</code> instance with the specified text, image, and horizontal alignment.
	* The label is centered vertically in its display area.
	* The text is on the trailing edge of the image.
	* <p>
	* @param text The text to be displayed by the label
	* @param icon The image to be displayed by the label
	* @param horizontalAlignment One of the following constants defined in {@link SwingConstants}:
	* <code>LEFT</code>, <code>CENTER</code>, <code>RIGHT</code>, <code>LEADING</code>, or <code>TRAILING</code>
	*/
	public JLabel2(String text, Icon icon, int horizontalAlignment) {
		super(text, icon, horizontalAlignment);
		
		enableEvents(AWTEvent.MOUSE_EVENT_MASK);
	}


	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {

		public static void main(String[] args) throws Exception {
			buildGui();
		}

		private static void buildGui() {
			JFrame frame = new JFrame();
			frame.setSize(500, 500);
			frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
			frame.getContentPane().add( buildItems() );
			frame.setVisible(true);
		}
		
		private static JComponent buildItems() {
			JPanel jpanel = new JPanel();
			jpanel.add( buildJLabel() );
			jpanel.add( Box.createHorizontalStrut(20) );
			jpanel.add( buildJLabel2() );
			return jpanel;
		}
		
		private static JComponent buildJLabel() {
			JLabel label = new JLabel(
				"<html>" +
					"<p>" +
						"JLabel:" + "<br>" +
						"popup menu events should fail" + "<br>" +
						"for this label" +
					"</p>" +
				"</html>"
			);
			label.setForeground( Color.RED );
			label.setComponentPopupMenu( new PopupMenuLabeled("Event from JLabel") );
			return label;
		}
		
		private static JComponent buildJLabel2() {
			JLabel2 label2 = new JLabel2(
				"<html>" +
					"<p>" +
						"JLabel2:" + "<br>" +
						"popup menu events should work" + "<br>" +
						"for this label" +
					"</p>" +
				"</html>"
			);
			label2.setForeground( Color.GREEN );
			label2.setComponentPopupMenu( new PopupMenuLabeled("Event from JLabel2") );
			return label2;
		}
		
		
		private static class PopupMenuLabeled extends JPopupMenu {
			
			private static final long serialVersionUID = 1;

			private PopupMenuLabeled(final String label) {
				add( new JMenuItem( label ) );

				addPopupMenuListener( new PopupMenuListener() {
					public void popupMenuWillBecomeVisible(PopupMenuEvent e) { System.out.println( label + ": popupMenuWillBecomeVisible" ); }
					public void popupMenuWillBecomeInvisible(PopupMenuEvent e) { System.out.println( label + ": popupMenuWillBecomeInvisible" + "\n" ); }
					public void popupMenuCanceled(PopupMenuEvent e) { System.out.println( label + ": popupMenuCanceled" ); }
				} ); 
			}
			
		}
		
	}
	
	
}
