/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.awt.*;

import javax.swing.*;
import javax.swing.event.*;


/**
* This subclass of JPanel adds exactly one new bit of functionality to its superclass:
* it is interested in mouse events.
* <p>
* This class was originally written to enable popup menus to be associated with JPanels.
* <a href="http://forum.java.sun.com/thread.jspa?threadID=590867&tstart=0">This forum posting</a> provides more discussion.
* <p>
* The fundamental reason why this class needed to be written was because, at present,
* Swing is designed to not pass mouse events to any lightweight component
* that does not have at least registered mouse listener,
* a design flaw that has been <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4966109">reported as a bug</a>
* but which Sun has yet to fix.
* <p>
* Like typical java gui code, this class is not multithread safe:
* it expects to only be called by {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread}.
* <p>
* @author Brent Boyer
*/
public class JPanel2 extends JPanel {


	// -------------------- constants --------------------


	private static final long serialVersionUID = 1;


	// -------------------- constructors --------------------
	
	
	/** Creates a new <code>JPanel2</code> with a double buffer and a flow layout. */
	public JPanel2() {
		this(true);
	}
	
	
	/**
	* Creates a new <code>JPanel2</code> with <code>FlowLayout</code> and the specified buffering strategy.
	* If <code>isDoubleBuffered</code> is true, the <code>JPanel2</code> will use a double buffer.
	* <p>
	* @param isDoubleBuffered a boolean, true for double-buffering, which uses additional memory space to achieve fast, flicker-free updates
	*/
	public JPanel2(boolean isDoubleBuffered) {
		this(new FlowLayout(), isDoubleBuffered);
	}
	
	
	/**
	* Create a new buffered JPanel2 with the specified layout manager and a double buffer.
	* <p>
	* @param layout the LayoutManager to use
	*/
	public JPanel2(LayoutManager layout) {
		this(layout, true);
	}
	
	
	/**
	* Creates a new JPanel2 with the specified layout manager and buffering strategy.
	* <p>
	* @param layout the LayoutManager to use
	* @param isDoubleBuffered a boolean, true for double-buffering, which uses additional memory space to achieve fast, flicker-free updates
	*/
	public JPanel2(LayoutManager layout, boolean isDoubleBuffered) {
		super(layout, isDoubleBuffered);
		
		enableEvents(AWTEvent.MOUSE_EVENT_MASK);
	}
	
	
	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {

		public static void main(String[] args) throws Exception {
			buildGui();
		}

		private static void buildGui() {
			JFrame frame = new JFrame();
			frame.setSize(500, 500);
			frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
			frame.getContentPane().add( buildItems() );
			frame.setVisible(true);
		}
		
		private static JComponent buildItems() {
			JPanel jpanel = new JPanel( new GridLayout(2, 1) );
			jpanel.add( buildJPanel() );
			jpanel.add( buildJPanel2() );
			return jpanel;
		}

		
		private static JComponent buildJPanel() {
			JPanel panel = new JPanel();
			panel.setBackground( Color.RED );
			panel.setComponentPopupMenu( new PopupMenuLabeled("Event from JPanel") );
			panel.add( new JLabel("JPanel: popup menu events should fail for this panel") );
			return panel;
		}
		
		private static JComponent buildJPanel2() {
			JPanel2 panel2 = new JPanel2();
			panel2.setBackground( Color.GREEN );
			panel2.setComponentPopupMenu( new PopupMenuLabeled("Event from JPanel2") );
			panel2.add( new JLabel("JPanel2: popup menu events should work for this panel") );
			return panel2;
		}
		
		
		private static class PopupMenuLabeled extends JPopupMenu {
			
			private static final long serialVersionUID = 1;

			private PopupMenuLabeled(final String label) {
				add( new JMenuItem( label ) );

				addPopupMenuListener( new PopupMenuListener() {
					public void popupMenuWillBecomeVisible(PopupMenuEvent e) { System.out.println( label + ": popupMenuWillBecomeVisible" ); }
					public void popupMenuWillBecomeInvisible(PopupMenuEvent e) { System.out.println( label + ": popupMenuWillBecomeInvisible" + "\n" ); }
					public void popupMenuCanceled(PopupMenuEvent e) { System.out.println( label + ": popupMenuCanceled" ); }
				} ); 
			}
			
		}
		
	}
	
	
}
