/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.awt.*;
import java.awt.event.*;

import javax.swing.*;


/**
* This subclass of JDialog presents a GUI widget that allows the user to specify the
* Swing Look and Feel.
* The Look and Feels presented to the user are determined by the LookAndFeelUtil class.
* <p>
* Like typical java gui code, this class is not multithread safe:
* it expects to only be called by {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread}.
* <p>
* @see LookAndFeelUtil
* @author Brent Boyer
*/
public class LookAndFeelDialog extends JDialog {


	// -------------------- constants --------------------


	private static final long serialVersionUID = 1;


	protected static final String DIALOG_TITLE = "Look And Feel Chooser";

	protected static final String DEFAULT_LOOK_AND_FEEL = "Metal";


	// -------------------- instance Fields --------------------


	/**
	* Stores the Object to call back to indicate the new Look and Feel.
	* @serial
	*/
	private final UpdatableLookAndFeel target;

	/**
	* Is the ButtonGroup for all the Look and Feel choices.
	* @serial
	*/
	private final ButtonGroup buttonGroup = new ButtonGroup();


	// -------------------- constructor --------------------


	/**
	* Constructs a new LookAndFeelDialog instance
	* @param parent the parent Frame that this JDialog will be attached to; may be null
	* @param modal specifies whether or not this JDialog is a modal dialog
	* @param target specifies the Object that is to update its Look And Feel for the new choice
	*/
	public LookAndFeelDialog(Frame parent, boolean modal, UpdatableLookAndFeel target) {
		super(parent, DIALOG_TITLE, modal);

		this.target = target;

		this.setDefaultCloseOperation( DISPOSE_ON_CLOSE );
		this.getContentPane().add( buildLAFSelectorGui() );
this.pack();
// +++ a defect of the pack method is that it does not take the JDialog's title into account (it only considers the window's contents)
// See if can find a way to set the window size that takes all items into account!

//		this.setLocationRelativeTo(parent);	// MUST do this call after have properly sized the dialog with pack
		if (parent != null) this.setLocation( parent.getX() + 100, parent.getY() + 100 );
		this.setVisible(true);
	}


	// -------------------- accessors and mutators --------------------


	protected UpdatableLookAndFeel getTarget() { return target; }

	protected ButtonGroup getButtonGroup() { return buttonGroup; }


	// -------------------- build methods --------------------


	protected JPanel buildLAFSelectorGui() {
		JPanel jpanel = new JPanel( LineLayout.makeVertical() );
		jpanel.setBorder( BorderFactory.createEmptyBorder(20, 20, 20, 20) );

		jpanel.add( new JLabel("Please select a Look and Feel:") );
		jpanel.add( buildLAFChoices() );
		jpanel.add( Box.createVerticalStrut(20) );	// puts blank vertical space to separate the radio buttons from the ok and cancel buttons
		jpanel.add( buildOKAndCancelButtons() );

		return jpanel;
	}


	protected JPanel buildLAFChoices() {
		JPanel jpanel = new JPanel( LineLayout.makeVertical() );
		
		for (String lafName : LookAndFeelUtil.lookAndFeels) {
			JRadioButton button = new JRadioButton(lafName);
			button.setActionCommand(lafName);

			//preselectIfIsDefault(button);
			preselectIfIsCurrent(button);

			getButtonGroup().add( button );
			jpanel.add( button );
		}

		return jpanel;
	}


	protected JPanel buildOKAndCancelButtons() {
		JPanel jpanel = new JPanel( LineLayout.makeHorizontal() );

		JButton okButton = new JButton("OK");
		okButton.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				LookAndFeelDialog.this.dispose();

				String newLAF = getButtonGroup().getSelection().getActionCommand();
				getTarget().updateLookAndFeel( LookAndFeelUtil.getClassName(newLAF) );
			}
		} );
		this.getRootPane().setDefaultButton(okButton);
		jpanel.add( okButton );

		jpanel.add( Box.createHorizontalStrut(20) );

		JButton cancelButton = new JButton("Cancel");
		cancelButton.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				LookAndFeelDialog.this.dispose();
			}
		} );
		jpanel.add( cancelButton );

		return jpanel;
	}


	// -------------------- preselectIfIsXXX --------------------


	protected void preselectIfIsDefault(JRadioButton button) {
		if ( button.getActionCommand().equals( DEFAULT_LOOK_AND_FEEL ) )
			button.setSelected(true);
	}


	protected void preselectIfIsCurrent(JRadioButton button) {
		if ( button.getActionCommand().equals( UIManager.getLookAndFeel().getName() ) )
			button.setSelected(true);
	}


	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {

		public static void main(String[] args) throws Exception {
			
				// see http://java.sun.com/docs/books/tutorial/uiswing/misc/threads.html for why must do ALL swing gui in event dispatch thread (the earlier advice given in http://java.sun.com/developer/JDCTechTips/2003/tt1208.html#1 to only do the displaying in the event dispatch thread is actually wrong)
			EventQueue.invokeLater( new Runnable() {
				public void run() {
					UpdatableLookAndFeel target = new TestDialog();
					new LookAndFeelDialog(null, false, target);
				}
			} );
		}
		
		private static class TestDialog extends ThrowableDialog implements UpdatableLookAndFeel {
			private static final long serialVersionUID = 1;
			
			private static final String message =
				"<html>" +
					"LookAndFeelDialog.UnitTest encountered an error." + "<br>" +
					"<i>This was deliberately programmed:</i> can reasonably click any button below." +
				"</html>";

			private TestDialog() {
				super((Frame) null, "TestDialog", false, message, new RuntimeException("purposely generated RuntimeException..."));
			}
			
			public void updateLookAndFeel(String lookAndFeelClassName) {
				try {
					UIManager.setLookAndFeel( lookAndFeelClassName );
					SwingUtilities.updateComponentTreeUI( this );
					//pack();		// causes the JFrame to be sized so that all of its children can be set to their preferred sizes
					refreshDisplay();
				}
				catch (Throwable t) {
					t.printStackTrace(System.err);
					JOptionPane.showMessageDialog(this, "Failed to change the Look and Feel to " + lookAndFeelClassName, "Look and Feel Failure", JOptionPane.ERROR_MESSAGE);
				}
			}
			
			private void refreshDisplay() {
				this.invalidate();
				this.validate();
				this.repaint();
			}
		};
		
	}


}