/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.awt.*;
import java.awt.event.*;


/**
* Subclass of Dialog which displays a title & message(s), as well as an "OK" button.
* <p>
* This class was written because AWT does not have an equivalent to {@link javax.swing.JOptionPane}.
* <p>
* Like typical java gui code, this class is not multithread safe:
* it expects to only be called by {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread}.
* <p>
* @author Brent Boyer
*/
public class MessageDialog extends Dialog {


	// -------------------- constants --------------------


	private static final long serialVersionUID = 1;


	// -------------------- constructors --------------------


	/**
	* Calls <code>super(owner, title, modal)</code>, then adds each message in the array as a separate Label
	* and an "OK" Button on the bottom, then shows this instance.
	*/
	public MessageDialog(Frame owner, String title, boolean modal, String[] messages) {
		super(owner, title, modal);
		add( makeMessageLines(messages), BorderLayout.NORTH );
		add( makeOkButton(), BorderLayout.SOUTH );
		pack();
		setVisible(true);
	}

	/** Convenience constructor which merely calls <code>this(owner, title, modal, new String[] {message})</code>. */
	public MessageDialog(Frame owner, String title, boolean modal, String message) {
		this(owner, title, modal, new String[] {message});
	}


	// -------------------- makeXXX --------------------


	protected Component makeMessageLines(String[] messages) {
		Panel panel = new Panel();
		panel.setLayout( new GridLayout(messages.length, 1) );
		for (String message : messages) {
			panel.add( new Label(message) );
		}
		return panel;
	}


	protected Component makeOkButton() {
		Button b = new Button("OK");
		b.addActionListener(
			new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					MessageDialog.this.dispose();
				}
			}
		);
		//return b;
			// Buttons get stretched by many layouts, so return a flexible panel instead:
		Panel panel = new Panel();
		panel.add(b);
		return panel;
	}


}
