/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.net.URL;
import java.io.*;
import java.util.*;

import javax.sound.sampled.*;

import bb.io.FileUtil;
import bb.util.ThrowableUtil;


/**
* Provides static methods relating to sound.
* <p>
* <b>Warning: there is apparently a <a href="http://forum.java.sun.com/thread.jsp?forum=406&thread=508583">bug</a>
* in all JDKs at least thru 1.4.x, in which the JVM will not naturally terminate if the JavaSound API is used;
* System.exit must be used to force shutdown.</b>
* Note: this bug has been fixed in 1.5.0.
* <p>
* This class is multithread safe: it is stateless.
* <p>
* @author Brent Boyer
* @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/sound/programmer_guide/contents.html">Java Sound Programmer Guide</a>
* @see "Brent Boyer's SimplePlayer class written when at IAM, which uses JavaSound"
* @see <a href="http://java.sun.com/products/java-media/jmf/mp3/download.html">Java MP3 PlugIn Manual Installation</a>
* @see <a href="http://www.javaworld.com/javaworld/jw-11-2000/jw-1103-mp3_p.html">Adding mp3 support to JavaSound via SPI</a>
* @see <a href="http://www.eventsounds.com/newsrad.htm">Free program sounds #1</a>
* @see <a href="http://www.space1999.net/~moonbase99/sound.htm">Free program sounds #2</a>
* @see <a href="http://java.sun.com/docs/books/tutorial/sound/playing.html">Using AudioClip to playing sounds (this is now mostly obsolete)</a>
*/
public class SoundUtil {


	// -------------------- playSynch --------------------


	/**
	* Simply calls <code>{@link #playSynch(File, int) playSynch}(file, 1)</code>.
	* <p>
	* @throws IllegalArgumentException if {@link FileUtil#check FileUtil.check} objects
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static void playSynch(File file) throws IllegalArgumentException, RuntimeException {
		playSynch(file, 1);
	}


	/**
	* Creates a URL for file and supplies this to {@link #playSynch(URL, int) playSynch}.
	* <p>
	* @throws IllegalArgumentException if {@link FileUtil#check FileUtil.check} objects; numberLoops < 1
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static void playSynch(File file, int numberLoops) throws IllegalArgumentException, RuntimeException {
		try {
			FileUtil.check(file);
			// numberLoops checked by the playSynch call below
			
			URL url = new URL( "file:" + file.getPath() );
			playSynch( url, numberLoops );
		}
		catch (Throwable t) {
			throw ThrowableUtil.toRuntimeException(t);
		}
	}


	/**
	* Simply calls <code>{@link #playSynch(URL, int) playSynch}(url, 1)</code>.
	* <p>
	* @throws IllegalArgumentException if url == null
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static void playSynch(URL url) throws IllegalArgumentException, RuntimeException {
		playSynch(url, 1);
	}


	/**
	* Simply calls <code>{@link #play play}(url, numberLoops, true)</code>.
	* <p>
	* Because the synchronous parameter supplied to play is true,
	* the playing is done "synchronously", that is, this method does not return until the sound has finished playing.
	* <p>
	* @throws IllegalArgumentException if url == null; numberLoops < 1
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static void playSynch(URL url, int numberLoops) throws IllegalArgumentException, RuntimeException {
		play(url, numberLoops, true);
	}


	// -------------------- playAsynch --------------------


	/**
	* Simply calls <code>{@link #playAsynch(File, int) playAsynch}(file, 1)</code>.
	* <p>
	* @throws IllegalArgumentException if {@link FileUtil#check FileUtil.check} objects
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static void playAsynch(File file) throws IllegalArgumentException, RuntimeException {
		playAsynch(file, 1);
	}


	/**
	* Creates a URL for file and supplies this to {@link #playAsynch(URL, int) playAsynch}.
	* <p>
	* @throws IllegalArgumentException if {@link FileUtil#check FileUtil.check} objects; numberLoops < 1
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static void playAsynch(File file, int numberLoops) throws IllegalArgumentException, RuntimeException {
		try {
			FileUtil.check(file);
			// numberLoops checked by the playAsynch call below
			
			URL url = new URL( "file:" + file.getPath() );
			playAsynch( url, numberLoops );
		}
		catch (Throwable t) {
			throw ThrowableUtil.toRuntimeException(t);
		}
	}


	/**
	* Simply calls <code>{@link #playAsynch(URL, int) playAsynch}(url, 1)</code>.
	* <p>
	* @throws IllegalArgumentException if url == null
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static void playAsynch(URL url) throws IllegalArgumentException, RuntimeException {
		playAsynch(url, 1);
	}


	/**
	* Simply calls <code>{@link #play play}(url, numberLoops, false)</code>.
	* <p>
	* Because the synchronous parameter supplied to play is false,
	* the playing is done "asynchronously", that is, this method immediately returns
	* while another thread carries out the playing of the sound.
	* <p>
	* @throws IllegalArgumentException if url == null; numberLoops < 1
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static void playAsynch(URL url, int numberLoops) throws IllegalArgumentException, RuntimeException {
		play(url, numberLoops, false);
	}
	
	
	// -------------------- play --------------------


	/**
	* Plays the sound located at url in a loop that will execute numberLoops times.
	* <p>
	* The actual playing of the sound is always done by some internal thread in the sound system.
	* If the synchronous parameter is true, however, then this method behaves "synchronously", that is,
	* it does not return until that other thread has finished playing the sound.
	* Otherwise, this method behaves "asynchronously", that is,
	* it immediately returns while that other thread concurrently plays the sound.
	* <p>
	* <b>Warning: this method does nothing to prevent another thread from simultaneously using the sound output system</b>
	* (e.g. by calling some method of this class), in which case multiple sounds may be playing simultaneously.
	* <p>
	* @throws IllegalArgumentException if url == null; numberLoops < 1
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static void play(URL url, int numberLoops, boolean synchronous) throws IllegalArgumentException, RuntimeException {
		if (url == null) throw new IllegalArgumentException("url == null");
		if (numberLoops < 1) throw new IllegalArgumentException("numberLoops = " + numberLoops + " < 1");
		
		try {
			Player player = new Player(url, numberLoops);
			player.play();
			
			if (synchronous) {
				synchronized (player) {
					player.wait( player.getDuration() );	// wait only the expected amount of time, else there was some other problem so abort
				}
			}
		}
		catch (Throwable t) {
			throw ThrowableUtil.toRuntimeException(t);
		}
	}
	
	
	// -------------------- timeToPlay --------------------


	/**
	* Returns <code>{@link #timeToPlay(Clip, int) timeToPlay}(clip, 1)</code>.
	* <p>
	* @throws IllegalArgumentException if clip == null
	* @throws IllegalStateException if the line for clip is not open
	*/
	public static long timeToPlay(Clip clip) throws IllegalArgumentException, IllegalStateException {
		return timeToPlay(clip, 1);
	}


	/**
	* Returns the number of milliseconds that it should take to play clip in a loop that will execute numberLoops times.
	* The result is always an upper bound because it comes from rounding up a microsecond time.
	* <p>
	* @throws IllegalArgumentException if clip == null; numberLoops < 1
	* @throws IllegalStateException if the line for clip is not open
	*/
	public static long timeToPlay(Clip clip, int numberLoops) throws IllegalArgumentException, IllegalStateException {
		if (clip == null) throw new IllegalArgumentException("clip == null");
		if (numberLoops < 1) throw new IllegalArgumentException("numberLoops = " + numberLoops + " < 1");
		if (!clip.isOpen()) throw new IllegalStateException("the line for clip is not open");
		
		long timeMicroseconds = clip.getMicrosecondLength();
		if (timeMicroseconds == AudioSystem.NOT_SPECIFIED) throw new IllegalStateException("the line for clip must not be open, because calling clip.getMicrosecondLength() returned AudioSystem.NOT_SPECIFIED");
		
		double timeMilliseconds = Math.ceil( (numberLoops * timeMicroseconds) / 1000.0 );	// CRITICAL: for best accuracy, do all the arithmetic inside the Math.ceil()
		if (timeMilliseconds > Long.MAX_VALUE) throw new IllegalStateException("the result (approximately " + timeMilliseconds + " milliseconds) is greater then the max value of a long");
		
		return (long) timeMilliseconds;
	}


	// -------------------- constructor --------------------


	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private SoundUtil() {}


	// -------------------- Player (static inner class) --------------------
	
	
	/**
	* This class generates a Clip from a {@link URL} and can play it,
	* but most importantly it robustly ensures that the clip is eventually closed
	* and that this close event gets propagated.
	* <p>
	* There seems to be a common code idiom for working with Java Sound that is not robust against known bugs and defects.
	* In this idiom, a LineListener alone listens for the Clip's STOP event and either closes the Clip at that point
	* and/or notifies the calling thread (which is waiting on the Clip instance, if synchronous sound playback is desired).
	* Examples of this code idiom include:
	* <ol>
	*  <li><a href="http://www.jsresources.org/examples/ClipPlayer.java.html">http://www.jsresources.org/examples/ClipPlayer.java.html</a></li>
	*  <li><a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4434125">http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4434125</a> (see the source code in the initial posting)</li>
	*  <li><a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4631192">http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4631192</a> (see the source code in the initial posting)</li>
	* </ol>
	* <p>
	* Unfortunately, as the bug reports cited above illustrate, there are many problems with Java Sound:
	* <ol>
	*  <li>it is a very bad idea to use the Clip instance itself as the wait/notify object, since it is used internally</li>
	*  <li>STOP events may fail to get generated</li>
	* </ol>
	* <p>
	* This class avoids all these problems.
	* First, it uses the Player instance itself as the wait/notify object.
	* Second, it does not solely rely on the event mechanism to close the Clip.
	* Instead, a task that executes in a background thread (a {@link Timer}) is used to guarantee that {@link #close close}
	* is called after the expected amount of playing time has elapsed, which guards against dropped events.
	*/
	private static class Player implements LineListener {
		
		private static final Timer timer = new Timer("SoundUtil.Player_timer", true);
		
		private final int numberLoops;
		private final Clip clip;
		private boolean closeExecuted = false;
		
		private Player(URL url, int numberLoops) throws UnsupportedAudioFileException, IOException, LineUnavailableException, IllegalArgumentException, IllegalStateException, SecurityException {
			this.numberLoops = numberLoops;
			
			AudioInputStream ais = AudioSystem.getAudioInputStream(url);
			DataLine.Info info = new DataLine.Info(Clip.class, ais.getFormat());
			this.clip = (Clip) AudioSystem.getLine(info);
			
			clip.addLineListener( this );
			clip.open( ais );
		}
		
		private long getDuration() {
			return timeToPlay(clip, numberLoops);
		}
		
		/**
		* Initiates the playing of clip in a loop that will execute numberLoops times.
		* (Some other thread in the sound system actually does the playing, not the calling thread,
		* so this method executes very quickly).
		* <p>
		* Schedules a background task which will call {@link #close close}
		* after a delay of {@link #getDuration getDuration}.
		* <p>
		* @throws IllegalStateException if clip is currently running
		*/
		private void play() throws IllegalStateException {
			if (clip.isRunning()) throw new IllegalStateException("clip is currently running");
			
			clip.loop(numberLoops - 1);	// - 1 because of the weird way that loop operates
			
			timer.schedule(
				new TimerTask() {
					public void run() {
						Player.this.close();
					}
				},
				getDuration()
			);
		}
		
		/**
		* If this method has never executed without error before, then closes clip and then notifies this instance.
		* Otherwise, this method does nothing.
		*/
		private void close() {
			synchronized (this) {
				if (!closeExecuted) {
					clip.close();
					this.notify();
					closeExecuted = true;
				}
			}
		}
		
		public void update(LineEvent event) {
			if (event.getType().equals(LineEvent.Type.STOP)) {
				close();
			}
		}
	
	}
	
	
	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {

		public static void main(String[] args) throws Exception {
				// should hear the sounds played serially:
			SoundUtil.playSynch( new File("D:/ellipticGroup/software/dev/sound/alert.wav") );
			SoundUtil.playSynch( new URL("http://www.eventsounds.com/wav/klrsanta.wav") );

			Thread.sleep(1*1000);
			
				// should hear all the sounds played simultaneously:
			SoundUtil.playAsynch( new File("D:/ellipticGroup/software/dev/sound/alert.wav") );
			SoundUtil.playAsynch( new URL("http://www.eventsounds.com/wav/klrsanta.wav") );

			Thread.sleep(3*1000);
			System.exit(0);
		}
	
	}


}
