/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.awt.*;


/**
* This class is a Component that is always a {@link #strutLength specified length}
* along a {@link #axis specified direction}.
* Unless it is explicitly changed by the programmer
* (via the {@link #setAxis setAxis} and {@link #setStrutLength setStrutLength} methods),
* the strut's orientation and length will always be kept fixed.
* In particular, attempts to change an instance's length (e.g. via the {@link #reshape reshape} method) will be rebuffed.
* <p>
* The strut's behavior in the direction perpendicular to the orientation axis is more flexible.
* First, its extent (max, preferred, and min) along this direction may be specified (either in a constructor or changed later).
* Secondly, this extent may be changed via the {@link #reshape reshape} method.
* <p>
* This Component is useful in adding space between other visible Components.
* The sole purpose of this class is to offer an AWT alternative to
* the strut methods available in Swing's {@link javax.swing.Box Box} class.
* <p>
* This class was designed to work with Java 1.1 and forward.
<!--
* This class was designed to work with the oldest JVMs (even 1.0), so it uses some deprecated APIs.

Have commented out the above, as well as several methods below, because do not want to generate compiler warnings.
-->
* <p>
* Like typical java gui code, this class is not multithread safe:
* it expects to only be called by {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread}.
* <p>
* @author Brent Boyer
*/
public class Strut extends Component {


	// -------------------- constants --------------------


	private static final long serialVersionUID = 1;


	/** Specifies the x-axis (i.e. horizontal dimension). */
	public static final int X_AXIS = 0;	// is same as BoxLayout.X_AXIS (from looking at its source code); we do not use it, however, since we want to use this class if swing is absent

	/** Specifies the y-axis (i.e. vertical dimension). */
	public static final int Y_AXIS = 1;	// is same as BoxLayout.Y_AXIS (from looking at its source code); we do not use it, however, since we want to use this class if swing is absent


	// -------------------- instance fields --------------------


	/**
	* Stores the orientation axis of this strut.
	* @serial
	*/
	protected int axis;


	/**
	* Stores the length of this strut along the {@link #axis orientation axis}.
	* @serial
	*/
	protected int strutLength;


	/**
	* Stores the maximum transverse (i.e. perpendicular direction) extent of this strut.
	* @serial
	*/
	protected int maxTransverse;


	/**
	* Stores the preferred transverse (i.e. perpendicular direction) extent of this strut.
	* @serial
	*/
	protected int preferredTransverse;


	/**
	* Stores the minimum transverse (i.e. perpendicular direction) extent of this strut.
	* @serial
	*/
	protected int minTransverse;


	// -------------------- constructors --------------------


	/**
	* Constructs a new Strut instance with the specified
	* strut orientation axis, length, max-preferred-minimum transverse sizes, and color.
	* <p>
	* @param axis the orientation axis of this strut (i.e. the direction in which it will be held rigid)
	* @param strutLength the length of the strut along the orientation axis
	* @param maxTransverse the maximum transverse (i.e. perpendicular direction) extent of this strut
	* @param preferredTransverse the preferred transverse (i.e. perpendicular direction) extent of this strut
	* @param minTransverse the minimum transverse (i.e. perpendicular direction) extent of this strut
	* @param color the foreground Color of this strut; if null its parent's foreground Color will be used
	* @throws IllegalArgumentException if arg axis is neither X_AXIS nor Y_AXIS, or
	* if args strutLength, maxTransverse, preferredTransverse, or minTransverse are < 0
	*/
	public Strut(int axis, int strutLength, int maxTransverse, int preferredTransverse, int minTransverse, Color color) {
		setAxis(axis);
		setStrutLength(strutLength);
		setMaxTransverse(maxTransverse);
		setPreferredTransverse(preferredTransverse);
		setMinTransverse(minTransverse);
		setForeground(color);
	}


	/** Simply calls <code>this(axis, strutLength, Integer.MAX_VALUE, preferredTransverse, 0, color)</code>. */
	public Strut(int axis, int strutLength, int preferredTransverse, Color color) {
		this(axis, strutLength, Integer.MAX_VALUE, preferredTransverse, 0, color);
	}


	/** Simply calls <code>this(axis, strutLength, 0, null)</code>. */
	public Strut(int axis, int strutLength) {
		this(axis, strutLength, 0, null);
	}


	// -------------------- factory methods --------------------


	/**
	* Reurns a new horizontal (x-axis oriented) Strut instance with the specified length, preferred transverse size, and color.
	* @return <code>new Strut(X_AXIS, strutLength, Integer.MAX_VALUE, preferredTransverse, 0, color)</code>
	*/
	public static Strut inX(int strutLength, int preferredTransverse, Color color) {
		return new Strut(X_AXIS, strutLength, Integer.MAX_VALUE, preferredTransverse, 0, color);
	}


	/**
	* Reurns a new horizontal (x-axis oriented) Strut instance with the specified length.
	* @return <code>inX(strutLength, 0, null)</code>
	*/
	public static Strut inX(int strutLength) {
		return inX(strutLength, 0, null);
	}


	/**
	* Reurns a new vertical (y-axis oriented) Strut instance with the specified length, preferred transverse size, and color.
	* @return <code>new Strut(Y_AXIS, strutLength, Integer.MAX_VALUE, preferredTransverse, 0, color)</code>
	*/
	public static Strut inY(int strutLength, int preferredTransverse, Color color) {
		return new Strut(Y_AXIS, strutLength, Integer.MAX_VALUE, preferredTransverse, 0, color);
	}


	/**
	* Reurns a new vertical (y-axis oriented) Strut instance with the specified length.
	* @return <code>inY(strutLength, 0, null)</code>
	*/
	public static Strut inY(int strutLength) {
		return inY(strutLength, 0, null);
	}


	// -------------------- accessors & mutators --------------------


	/**
	* Mutator for the axis field.
	* <p>
	* @throws IllegalArgumentException if arg axis is neither X_AXIS nor Y_AXIS
	*/
	public void setAxis(int axis) throws IllegalArgumentException {
		if ((axis != X_AXIS) && (axis != Y_AXIS)) throw new IllegalArgumentException("arg axis = " + axis + " is neither X_AXIS nor Y_AXIS");

		this.axis = axis;
	}


	/**
	* Mutator for the strutLength field.
	* <p>
	* @throws IllegalArgumentException if arg strutLength is < 0
	*/
	public void setStrutLength(int strutLength) throws IllegalArgumentException {
		if (strutLength < 0) throw new IllegalArgumentException("arg strutLength = " + strutLength + " is < 0");

		this.strutLength = strutLength;
	}


	/**
	* Mutator for the maxTransverse field.
	* <p>
	* @throws IllegalArgumentException if arg maxTransverse is < 0
	*/
	public void setMaxTransverse(int maxTransverse) throws IllegalArgumentException {
		if (maxTransverse < 0) throw new IllegalArgumentException("arg maxTransverse = " + maxTransverse + " is < 0");

		this.maxTransverse = maxTransverse;
	}


	/**
	* Mutator for the preferredTransverse field.
	* <p>
	* @throws IllegalArgumentException if arg preferredTransverse is < 0
	*/
	public void setPreferredTransverse(int preferredTransverse) throws IllegalArgumentException {
		if (preferredTransverse < 0) throw new IllegalArgumentException("arg preferredTransverse = " + preferredTransverse + " is < 0");

		this.preferredTransverse = preferredTransverse;
	}


	/**
	* Mutator for the minTransverse field.
	* <p>
	* @throws IllegalArgumentException if arg minTransverse is < 0
	*/
	public void setMinTransverse(int minTransverse) throws IllegalArgumentException {
		if (minTransverse < 0) throw new IllegalArgumentException("arg minTransverse = " + minTransverse + " is < 0");

		this.minTransverse = minTransverse;
	}


	// -------------------- size methods --------------------


	/**
	* Returns <code>new Dimension(strutLength, maxTransverse)</code> (if axis == X_AXIS)
	* or <code>new Dimension(maxTransverse, strutLength)</code> (if axis == Y_AXIS).
	*/
	public Dimension getMaximumSize() {
		switch (axis) {
			case X_AXIS:
				return new Dimension(strutLength, maxTransverse);
			case Y_AXIS:
				return new Dimension(maxTransverse, strutLength);
			default:
				throw new IllegalStateException("the field axis = " + axis + " has a value that is illegal; it should never have been assigned this value");
		}
	}


/*
	**
	* Returns <code>new Dimension(preferredTransverse, strutLength)</code>.
	* (Looking at the source code for Component, this method seems to be the fundamental method for preferred size.
	* It is called by getPreferredSize.)
	*
	public Dimension preferredSize() {
		switch (axis) {
			case X_AXIS:
				return new Dimension(strutLength, preferredTransverse);
			case Y_AXIS:
				return new Dimension(preferredTransverse, strutLength);
			default:
				throw new IllegalStateException("the field axis = " + axis + " has a value that is illegal; it should never have been assigned this value");
		}
	}
*/
	/** Returns <code>new Dimension(preferredTransverse, strutLength)</code>. */
	public Dimension getPreferredSize() {
		switch (axis) {
			case X_AXIS:
				return new Dimension(strutLength, preferredTransverse);
			case Y_AXIS:
				return new Dimension(preferredTransverse, strutLength);
			default:
				throw new IllegalStateException("the field axis = " + axis + " has a value that is illegal; it should never have been assigned this value");
		}
	}


/*
	**
	* Returns <code>new Dimension(strutLength, minTransverse)</code> (if axis == X_AXIS)
	* or <code>new Dimension(minTransverse, strutLength)</code> (if axis == Y_AXIS).
	* (Looking at the source code for Component, this method seems to be the fundamental method for minimum size.
	* It is called by getMinimumSize.)
	*
	public Dimension minimumSize() {
		switch (axis) {
			case X_AXIS:
				return new Dimension(strutLength, minTransverse);
			case Y_AXIS:
				return new Dimension(minTransverse, strutLength);
			default:
				throw new IllegalStateException("the field axis = " + axis + " has a value that is illegal; it should never have been assigned this value");
		}
	}
*/
	/**
	* Returns <code>new Dimension(strutLength, minTransverse)</code> (if axis == X_AXIS)
	* or <code>new Dimension(minTransverse, strutLength)</code> (if axis == Y_AXIS).
	*/
	public Dimension getMinimumSize() {
		switch (axis) {
			case X_AXIS:
				return new Dimension(strutLength, minTransverse);
			case Y_AXIS:
				return new Dimension(minTransverse, strutLength);
			default:
				throw new IllegalStateException("the field axis = " + axis + " has a value that is illegal; it should never have been assigned this value");
		}
	}


	// -------------------- isLightweight --------------------


	/** Returns true: this Component is peerless. */
	public boolean isLightweight() { return true; }


	// -------------------- paint --------------------


	/**
	* Since this is a lightweight (has no peer) gui object that is a direct subclass of Component,
	* this method must be implemented.
	* The code here merely fills in this Component's rectangular (length & width) dimensions with its foreground Color.
	* <p>
	* @see <a href="http://java.sun.com/products/jfc/tsc/articles/painting/index.html">Painting in AWT and Swing</a>
	*/
	public void paint(Graphics g) {
		Dimension size = getSize();
		g.fillRect(0, 0, size.width, size.height);
	}


	// -------------------- reshape --------------------


/*
	**
	* Overrides superclass implementation to enforce the fixed strut length requirement:
	* if axis == X_AXIS, then the width arg is replaced by strutLength in a call to super.reshape;
	* else if axis == Y_AXIS, then the height arg is replaced by strutLength in a call to super.reshape.
	* (Looking at the source code for Component, this method seems to be the fundamental method for shape control.
	* It is eventually called by all the setSize, resize, setBounds methods.)
	*
	public void reshape(int x, int y, int width, int height) {
		switch (axis) {
			case X_AXIS:
				super.reshape(x, y, strutLength, height);
				break;
			case Y_AXIS:
				super.reshape(x, y, width, strutLength);
				break;
			default:
				throw new IllegalStateException("the field axis = " + axis + " has a value that is illegal; it should never have been assigned this value");
		}
	}
*/
	/**
	* Overrides superclass implementation to enforce the fixed strut length requirement:
	* if axis == X_AXIS, then the width arg is replaced by strutLength in a call to super.reshape;
	* else if axis == Y_AXIS, then the height arg is replaced by strutLength in a call to super.reshape.
	*/
	public void setBounds(int x, int y, int width, int height) {
		switch (axis) {
			case X_AXIS:
				super.setBounds(x, y, strutLength, height);
				break;
			case Y_AXIS:
				super.setBounds(x, y, width, strutLength);
				break;
			default:
				throw new IllegalStateException("the field axis = " + axis + " has a value that is illegal; it should never have been assigned this value");
		}
	}


}
