/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.gui;


import java.awt.*;
import java.lang.reflect.InvocationTargetException;
import javax.swing.*;

import bb.util.NumberUtil;
import bb.util.StringUtil;


/**
* This class provides miscellaneous static utility methods that are useful for Swing programming.
* <p>
* It might appear that this class is multithread safe because it is stateless.
* However, typical Java gui code is not multithread safe, in particular, once a component has been realized
* it should only be accessed by the event dispatch thread
* (see <a href="http://java.sun.com/developer/JDCTechTips/2003/tt1208.html#1">Multithreading In Swing</a>
* and <a href="http://java.sun.com/developer/JDCTechTips/2004/tt0611.html#1">More Multithreading In Swing</a>).
* So, in order to enforce that requirement, all methods of this class which deal with components
* require the calling thread to be the event dispatch thread.
* See the javadocs of each method for its thread requirements.
* <p>
* @author Brent Boyer
*/
public class SwingUtil {


	// -------------------- constructor --------------------


	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private SwingUtil() {}
	
	
	// -------------------- invoke --------------------


	/**
	* If synchronous is true (i.e. the caller does not want this method to return until task has finished running),
	* then this method checks whether or not the calling thread is EventQueue's dispatch thread:
	* if it is, then the calling thread itself directly calls task.run,
	* else for all other threads, task is submitted to {@link EventQueue#invokeAndWait EventQueue.invokeAndWait}.
	* If synchronous is false, then task is always submitted to {@link EventQueue#invokeLater EventQueue.invokeLater}
	* for future execution by EventQueue's dispatch thread and this method immediately returns.
	* Thus, under all conditions, task is always executed by the dispatch thread,
	* while guaranteeing that from the caller's perspective this method obeys synchronous or asynchronous execution semantics.
	* <p>
	* Any thread may call this method.
	* <p>
	* @throws IllegalArgumentException if task == null
	* @throws InterruptedException if calling thread is interrupted while waiting for the dispatch thread to finish excecuting task.run 
	* @throws InvocationTargetException if an exception is thrown by task.run
	*/
	public static void invoke(Runnable task, boolean synchronous) throws IllegalArgumentException, InterruptedException, InvocationTargetException {
		if (task == null) throw new IllegalArgumentException("task == null");

		if (synchronous) {
			if (EventQueue.isDispatchThread()) {
				task.run();			
			}
			else {
				EventQueue.invokeAndWait( task );
			}
		}
		else {
			EventQueue.invokeLater( task );
		}
	}
// +++ any use for the new SwingWorker class introduced in 1.6?
// http://javaboutique.internet.com/tutorials/swingworker/


	// -------------------- optimizeJScrollPanePreferredSize --------------------


	/**
	* Simply calls <code>{@link #optimizeJScrollPanePreferredSize(JScrollPane, Dimension) optimizeJScrollPanePreferredSize}(jScrollPane, null)</code>.
	* <p>
	* Only {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread} may call this method.
	* <p>
	* @throws IllegalArgumentException if jScrollPane == null
	* @throws IllegalStateException if calling thread is not EventQueue's dispatch thread
	*/
	public static void optimizeJScrollPanePreferredSize(JScrollPane jScrollPane) throws IllegalArgumentException, IllegalStateException {
		optimizeJScrollPanePreferredSize(jScrollPane, null);
	}


	/**
	* Sets the preferred size of the supplied JScrollPane to an optimal value.
	* <p>
	* This method solves a serious display issue with JScrollPane.
	* A problem with typical uses of JScrollPane that wrap, say, a JTextArea is that if
	* one scroll bar has to appear, it will force the appearance of the other one even tho the other is not necessary.
	* To see why this can happen, suppose that the vertical scroll bar appears.
	* Then it will eat some of the horizontal space that the view Component would otherwise have had for itself.
	* To compensate, the JScrollPane will then bring up the horizontal bar.
	* This is visually annoying as well as useless because there is little or nothing for the horizontal bar to scroll over.
	* <p>
	* This method intelligently adjusts the JScrollPane's preferred size to account for
	* the dimensions that will be taken up by its scroll bars and Insets as well as its view.
	* This may allow either scroll bar (horizontal or vertical) to appear without necessarily bringing up the other.
	* Of course, if the view component itself is too large in both dimensions, there is nothing that can be done.
	* <p>
	* The limit param, if non-null, may be further used to set upperbounds on the width and height
	* used for jScrollPane's preferred size.
	* A common choice to supply for limit is the result of calling {@link #fractionOfScreenSize fractionOfScreenSize}
	* <p>
	* Only {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread} may call this method.
	* <p>
	* @param jScrollPane the JScrollPane whose preferred size is to be optimized; must not be null
	* @param limit a Dimension object that sets upperbounds on the width and height used for jScrollPane's preferred size; may be null
	* @throws IllegalArgumentException if jScrollPane == null
	* @throws IllegalStateException if calling thread is not EventQueue's dispatch thread
	*/
	public static void optimizeJScrollPanePreferredSize(JScrollPane jScrollPane, Dimension limit) throws IllegalArgumentException, IllegalStateException {
		if (jScrollPane == null) throw new IllegalArgumentException("jScrollPane == null");
		if (!EventQueue.isDispatchThread()) throw new IllegalStateException("calling thread (" + Thread.currentThread().toString() + ") is not EventQueue's dispatch thread");
		
		Dimension viewPreferredSize = jScrollPane.getViewport().getView().getPreferredSize();
		
		int width =
			viewPreferredSize.width +
			jScrollPane.getVerticalScrollBar().getPreferredSize().width +
			jScrollPane.getInsets().left +
			jScrollPane.getInsets().right;
			
		int height =
			viewPreferredSize.height +
			jScrollPane.getHorizontalScrollBar().getPreferredSize().height +
			jScrollPane.getInsets().top +
			jScrollPane.getInsets().bottom;

		// Note: the Viewport itself never contributes anything to the layout (e.g. no Border or Insets)
		
		if (limit != null) {
			width = Math.min( width, limit.width );
			height = Math.min( height, limit.height );
		}
		
		jScrollPane.setPreferredSize( new Dimension( width, height) );
	}


	// -------------------- addLabel --------------------
	
	
	/**
	* Simply calls <code>{@link #addLabel(String, JComponent, int, float) addLabel}(labelText, component, 10, alignment)</code>.
	* <p>
	* Only {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread} may call this method.
	* <p>
	* @throws IllegalArgumentException if jScrollPane == null
	* @throws IllegalStateException if calling thread is not EventQueue's dispatch thread
	*/
	public static JPanel addLabel(String labelText, JComponent component, float alignment) throws IllegalArgumentException, IllegalStateException {
		return addLabel(labelText, component, 10, alignment);
	}
	
	
	/**
	* Returns a new JPanel instance which contains a single row of children:
	* a JLabel for labelText on the left, followed by component on the right.
	* These two children are spaced apart by separation, and the overall JPanel that is returned is assigned alignment.
	* <p>
	* Only {@link EventQueue}'s {@link EventQueue#isDispatchThread dispatch thread} may call this method.
	* <p>
	* @throws IllegalArgumentException if labelText is blank; component == null;
	* {@link NumberUtil#checkNonNegative NumberUtil.checkNonNegative}(separation, "separation") fails;
	* {@link NumberUtil#checkNormalAndNonNegative NumberUtil.checkNormalAndNonNegative}(alignment, "alignment") fails;
	* alignment > 1.0f.
	* @throws IllegalStateException if calling thread is not EventQueue's dispatch thread
	*/
	public static JPanel addLabel(String labelText, JComponent component, int separation, float alignment) throws IllegalArgumentException, IllegalStateException {
		if (StringUtil.isBlank(labelText)) throw new IllegalArgumentException("labelText is blank");
		if (component == null) throw new IllegalArgumentException("component == null");
		NumberUtil.checkNonNegative(separation, "separation");
		NumberUtil.checkNormalAndNonNegative(alignment, "alignment");
		if (alignment > 1.0f) throw new IllegalArgumentException("alignment = " + alignment + " > 1.0f");
		if (!EventQueue.isDispatchThread()) throw new IllegalStateException("calling thread (" + Thread.currentThread().toString() + ") is not EventQueue's dispatch thread");

		JPanel jpanel = new JPanel( LineLayout.makeHorizontal() );
		jpanel.add( new JLabel(labelText) );
		jpanel.add( Box.createHorizontalStrut( separation ) );
		jpanel.add( component );
		jpanel.setAlignmentX( alignment );
		return jpanel;
	}


	// -------------------- screenSize, fractionOfScreenSize --------------------
	
	
	/**
	* Returns the screen's size.
	* On systems with multiple displays, <i>just the primary display is used</i>.
	* <p>
	* Any thread may call this method.
	* <p>
	* @throws HeadlessException if GraphicsEnvironment.isHeadless() returns true
	*/
	public static Dimension screenSize() throws HeadlessException {
		return Toolkit.getDefaultToolkit().getScreenSize();
	}
	
	
	/**
	* Simply returns <code>{@link #fractionOfScreenSize(double, double) fractionOfScreenSize}(fraction, fraction)</code>.
	* <p>
	* @throws IllegalArgumentException if {@link NumberUtil#checkNormalAndNonNegative NumberUtil.checkNormalAndNonNegative}(fraction, "fraction") fails
	* @throws HeadlessException if GraphicsEnvironment.isHeadless() returns true
	*/
	public static Dimension fractionOfScreenSize(double fraction) throws IllegalArgumentException, HeadlessException {
		return fractionOfScreenSize(fraction, fraction);
	}
	
	
	/**
	* First gets the screen's size by calling <code>{@link #screenSize screenSize}</code>.
	* Then returns a new Dimension instance
	* whose width is the screen's width multiplied by fractionX
	* and whose height is the screen's height multiplied by fractionY.
	* <p>
	* Any thread may call this method.
	* <p>
	* @throws IllegalArgumentException if {@link NumberUtil#checkNormalAndNonNegative NumberUtil.checkNormalAndNonNegative}(fractionX, "fractionX") fails;
	* {@link NumberUtil#checkNormalAndNonNegative NumberUtil.checkNormalAndNonNegative}(fractionY, "fractionY") fails
	* @throws HeadlessException if GraphicsEnvironment.isHeadless() returns true
	*/
	public static Dimension fractionOfScreenSize(double fractionX, double fractionY) throws IllegalArgumentException, HeadlessException {
		NumberUtil.checkNormalAndNonNegative(fractionX, "fractionX");
		NumberUtil.checkNormalAndNonNegative(fractionY, "fractionY");
		
		Dimension ss = screenSize();
		return new Dimension(
			(int) Math.round( fractionX * ss.width ),
			(int) Math.round( fractionY * ss.height )
		);
	}
	
	
}
