/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.io;


import java.io.*;
import java.util.*;

import org.junit.Assert;
import org.junit.Test;

import bb.util.OsUtil;


/**
* This class provides static utility methods for dealing with directories.
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @see FileUtil
* @author Brent Boyer
*/
public class DirUtil {
	
	
	// -------------------- constants --------------------
	
	
	/**
	* This constant records the initial Working Directory.
	* Specificly, it is the value returned by a call to the getWorkingDirectory method
	* <i>at the time this class was loaded</i>.
	* <p>
	* This constant is needed because using "./" for the current working directory
	* will not necessarily work on non-Unix Operating Systems.
	* (There are other problems associated with using "./" e.g. see http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4483097).
	*/
	public static final File WD = getWorkingDirectory();
	
	
	// -------------------- check --------------------
	
	
	/**
	* Checks if directory:
	* <ol>
	*  <li>is non-null</li>
	*  <li>resolves to a path that exists</li>
	*  <li>resolves to an actual directory, and not some other type of file (e.g. a normal file)</li>
	* </ol>
	* Failure to pass any test results in an IllegalArgumentException being thrown.
	* If all tests pass, the method simply returns normally.
	* <p>
	* @throws IllegalArgumentException if directory is null, non-existent, or is not an actual directory
	* @throws SecurityException if a security manager exists and denies read access to directory
	*/
	public static void check(File directory) throws IllegalArgumentException, SecurityException {
		if (directory == null) throw new IllegalArgumentException("directory == null");
		if ( !directory.exists() ) throw new IllegalArgumentException("directory = " + directory.getPath() + " is a non-existent path");
		if ( !directory.isDirectory() ) throw new IllegalArgumentException("directory = " + directory.getPath() + " is not a directory");
	}
	
	
	// -------------------- ensureExists --------------------
	
	
	/**
	* Makes sure that directory exists, creating it and any parent directories if necessary.
	* <p>
	* The usefulness of this method (compared to simply calling {@link File#mkdirs File.mkdirs}) is:
	* <ol>
	*  <li>this method only calls File.mkdirs if directory fails to exist in the first place</li>
	*  <li>it checks that the call to File.mkdirs succeeded</li>
	*  <li>it checks that an actual directory was created, not some other type of file</li>
	* </ol>
	* <p>
	* @return directory, which enables method call chaining
	* @throws IllegalArgumentException if directory is null
	* @throws SecurityException if a security manager exists
	* and does not permit verification of the existence of directory and all necessary parent directories;
	* or if it does not permit directory and all necessary parent directories to be created
	* @throws IllegalStateException if directory failed to be created or is not an actual directory but is some other type of file
	*/
	public static File ensureExists(File directory) throws IllegalArgumentException, SecurityException, IllegalStateException {
		if (directory == null) throw new IllegalArgumentException("directory == null");
		
		if (!directory.exists()) {
			boolean created = directory.mkdirs();
			if (!created) throw new IllegalStateException("directory = " + directory.getPath() + " failed to exist and could not be created");
		}
		
		if (!directory.isDirectory()) throw new IllegalArgumentException("directory = " + directory.getPath() + " is not a directory");
		
		return directory;
	}
	
	
	// -------------------- contains --------------------
	
	
	/**
	* Determines whether or not directory contains file (which could be either a normal file or a directory).
	* <p>
	* For the purposes of this method, "contains" means "equals or is found inside".
	* Consequently, this method returns true if directory equals file.
	* <p>
	* This method works correctly regardless of whether directory or file
	* were initially specified as relative, absolute, or canonical paths
	* because it first internally converts them both to canonical paths.
	* <p>
	* <b>Warning:</b> directory must actually exist, since a call will be made to its {@link File#isDirectory isDirectory} method.
	* In contrast, file need not actually exist.
	* <p>
	* @throws IllegalArgumentException if directory == null; directory is not an actual directory; file == null
	* @throws SecurityException if a security manager exists and denies read access to directory
	* @throws IOException if an I/O problem occurs
	*/
	public static boolean contains(File directory, File file) throws IllegalArgumentException, SecurityException, IOException {
		check(directory);
		if (file == null) throw new IllegalArgumentException("file == null");
		
		return getFullPath(file).startsWith( getFullPath(directory) );
	}
	
	
	/**
	* Returns a String that always starts with {@link File#getCanonicalPath getCanonicalPath}.
	* If file is a directory, then further ensures that the result ends with {@link File#separatorChar}.
	* <p>
	* This method was written to support the path comparison logic inside {@link #contains}
	* because getCanonicalPath does not guarantee the above behavior for directories.
	* <p>
	* To see why this behavior is critical for the logic inside the current implementation of contains,
	* consider the case that on some platform getCanonicalPath returns "a/b/someName" for directory
	* and "a/b/someNamePlusMore" for file:
	* if simply used getCanonicalPath for the path determination, would wrongly return true
	* but if use this method, will correctly return false.
	*/
	private static String getFullPath(File file) throws IOException {
		if (file.isDirectory()) {
			String directoryPath = file.getCanonicalPath();
			if (directoryPath.endsWith(File.separator)) return directoryPath;
			else return directoryPath + File.separatorChar;
		}
		else {
			return file.getCanonicalPath();
		}
	}
	
	
	// -------------------- getWorkingDirectory --------------------
	
	
	/**
	* This method returns a new File representing the current working directory
	* (that is, the path returned by a call to <code>System.getProperty("user.dir")</code>).
	* <p>
	* @throws RuntimeException if any problem occurs
	*/
	public static File getWorkingDirectory() throws RuntimeException {
		try {
			File wd = new File( System.getProperty("user.dir") );
			check(wd);
			return wd;
		}
		catch (Throwable t) {
			throw new RuntimeException("encountered a problem (see cause) trying to determine the working directory from System.getProperty(\"user.dir\") = " + System.getProperty("user.dir"), t);			
		}
	}
	
	
	// -------------------- getContents, getTree, getFilesInTree --------------------
	
	
	/** Simply returns {@link #getContents(File, FileFilter) getContents(directory, null)}. */
	public static File[] getContents(File directory) throws IllegalArgumentException, SecurityException, IOException {
		return getContents(directory, null);
	}
	
	
	/** Simply returns {@link #getContents(File, FileFilter, boolean) getContents(directory, fileFilter, true)}. */
	public static File[] getContents(File directory, FileFilter fileFilter) throws IllegalArgumentException, SecurityException, IOException {
		return getContents(directory, fileFilter, true);
	}
	
	
	/**
	* This utility method returns the contents of a directory.
	* <p>
	* The usefulness of this method (compared to the File.listFiles methods) is:
	* <ol>
	*  <li>first calls {@link #check check} on directory to test for problems</li>
	*  <li>guarantees that the result returned is non-null (tho may be zero length)</li>
	*  <li>has a boolean arg that can specify that the result is guaranteed to be sorted in ascending order</li>
	* </ol>
	* <p>
	* @param directory a File for the directory whose contents are to be retrieved
	* @param fileFilter a java.io.FileFilter instance for filtering the returned result; may be null (i.e. no filtering is done)
	* @param sortResult specifies whether or not the result should be sorted in ascending order (if false, the order returned is unspecified)
	* @throws IllegalArgumentException if directory is null, non-existent, or is not an actual directory
	* @throws SecurityException if a security manager exists and denies read access to some file
	* @throws IOException if an I/O problem occurs
	*/
	public static File[] getContents(File directory, FileFilter fileFilter, boolean sortResult) throws IllegalArgumentException, SecurityException, IOException {
		check(directory);
		
		File[] files = directory.listFiles( fileFilter );
		if (files == null) {
if (OsUtil.isMicrosoft() && directory.getName().equals("System Volume Information")) return new File[0];
// handles a bug with java on windoze: http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4803836
			throw new IOException("the listFiles method returned null for this directory: " + directory.getPath());	// this can happen -- see javadocs on the listFiles method (have seen this on NTFS file systems when encountered hidden system directories like "System Volume Information")
		}
		
		if (sortResult) Arrays.sort(files);
		return files;
	}
	
	
	/** Simply returns {@link #getTree(File, FileFilter) getTree(directory, null)}. */
	public static File[] getTree(File directory) throws IllegalArgumentException, SecurityException, IOException {
		return getTree(directory, null);
	}
	
	
	/** Simply returns {@link #getTree(File, FileFilter, boolean) getTree(directory, fileFilter, true)}. */
	public static File[] getTree(File directory, FileFilter fileFilter) throws IllegalArgumentException, SecurityException, IOException {
		return getTree(directory, fileFilter, true);
	}
	
	
	/**
	* This utility method returns all the Files in the complete directory tree.
	* The result will include both normal files and subdirectories.
	* Subdirectories in the tree are traced out depth first.
	* The result is never null, and is guaranteed to be sorted in ascending order if sortResult is true.
	* <p>
	* @param directory a File for the directory whose tree files are to be retrieved
	* @param fileFilter a java.io.FileFilter instance for filtering the returned result; may be null (i.e. no filtering is done);
	* note that if fileFilter rejects a subdirectory, then that entire subtree is excluded
	* @param sortResult specifies whether or not the result should be sorted in ascending order (if false, the order returned is unspecified)
	* @throws IllegalArgumentException if directory is null, non-existent, or is not an actual directory
	* @throws SecurityException if a security manager exists and denies read access to some file
	* @throws IOException if an I/O problem occurs
	*/
	public static File[] getTree(File directory, FileFilter fileFilter, boolean sortResult) throws IllegalArgumentException, SecurityException, IOException {
		List<File> fileList = new ArrayList<File>(128);
		getTree(directory, fileFilter, fileList, sortResult, true);
		return fileList.toArray( new File[fileList.size()] );
	}
	
	
	/** Simply returns {@link #getFilesInTree(File, FileFilter) getFilesInTree(directory, null)}. */
	public static File[] getFilesInTree(File directory) throws IllegalArgumentException, SecurityException, IOException {
		return getFilesInTree(directory, null);
	}
	
	
	/** Simply returns {@link #getFilesInTree(File, FileFilter, boolean) getFilesInTree(directory, fileFilter, true)}. */
	public static File[] getFilesInTree(File directory, FileFilter fileFilter) throws IllegalArgumentException, SecurityException, IOException {
		return getFilesInTree(directory, fileFilter, true);
	}
	
	
	/**
	* This utility method returns all the normal Files in the complete directory tree.
	* The result will never include subdirectories, altho these will be recursively examined depth first.
	* The result is never null, and is guaranteed to be sorted in ascending order if sortResult is true.
	* <p>
	* @param directory a File for the directory whose tree files are to be retrieved
	* @param fileFilter a java.io.FileFilter instance for filtering the returned result; may be null (i.e. no filtering is done);
	* note that if fileFilter rejects a subdirectory, then that entire subtree is excluded
	* @param sortResult specifies whether or not the result should be sorted in ascending order (if false, the order returned is unspecified)
	* @throws IllegalArgumentException if directory is null, non-existent, or is not an actual directory
	* @throws SecurityException if a security manager exists and denies read access to some file
	* @throws IOException if an I/O problem occurs
	*/
	public static File[] getFilesInTree(File directory, FileFilter fileFilter, boolean sortResult) throws IllegalArgumentException, SecurityException, IOException {
		List<File> fileList = new ArrayList<File>(128);
		getTree(directory, fileFilter, fileList, sortResult, false);
		return fileList.toArray( new File[fileList.size()] );
	}
	
	
	private static void getTree(File directory, FileFilter fileFilter, List<File> fileList, boolean sortResult, boolean addAllTypes) throws IllegalArgumentException, SecurityException, IOException {
		// directory checked by call to getContents below
		
		for (File path : getContents(directory, fileFilter, sortResult)) {
			if ( addAllTypes || path.isFile() ) {
				fileList.add(path);
			}
			if ( path.isDirectory() ) {
				getTree(path, fileFilter, fileList, sortResult, addAllTypes);
			}
		}
	}
	
	
	// -------------------- compareContents --------------------
	
	
	// see also .../filePrograms/src/DirectoryComparer.java
	
	
	/**
	* Compares the contents of the two supplied directories to see if they are exactly equal or somehow differ.
	* <p>
	* The current implementation retrieves the contents of directory1 and directory2
	* using calls to {@link #getTree(File) getTree}.
	* Note that these content results are always sorted in alphabetical order.
	* It then compares these results element by element (i.e. elements at the same index in each result),
	* for as many indices are shared by both results.
	* If a pair of elements are in any way different
	* (have different names, different types, or different contents), then the one from directory2 is returned.
	* In particular, if both are directories, then their contents are examined by recursively calling this method,
	* and if both are normal files, a call is made to {@link FileUtil#compareContents FileUtil.compareContents}.
	* If this common element comparison test is passed, then the first uncommon element is returned.
	* If there is no such uncommon element, then the directories are exactly equal, and null is returned
	* (and this is the <i>only</i> circumstance under which null will be returned).
	* <p>
	* @throws IllegalArgumentException if {@link #check check} objects to directory1 or directory2
	* @throws SecurityException if a security manager exists and denies read access to either directory
	* @throws IOException if an I/O problem occurs
	*/
	public static File compareContents(File directory1, File directory2) throws IllegalArgumentException, SecurityException, IOException {
		// directory1 and directory2 checked by getTree below
		
		File[] contents1 = getTree(directory1);
		File[] contents2 = getTree(directory2);
		int n = Math.min(contents1.length, contents2.length);
		for (int i = 0; i < n; i++) {
			File file1 = contents1[i];
			File file2 = contents2[i];
			if (!file1.getName().equals(file2.getName())) return file2;
			else if (file1.isDirectory()) {
				if (file2.isDirectory()) {
					File firstDifferent = compareContents(file1, file2);
					if (firstDifferent != null) return firstDifferent;	// if both are directories, recursively drill down
					
// +++ is there any directory meta data that should also do a comparison on?  do NOT want modification times, but is there anything else?
				}
				else {
					return file2;	// file1 is a directory but file2 is not, therefore there is no way that they can be the same
				}
			}
			else if (file1.isFile()) {
				if (file2.isFile()) {
					long lengthDiff = file1.length() - file2.length();
					if (lengthDiff != 0) return file2;	// this is a performance optimization: it is really quick to check this before make an expensive call to FileUtil.compareContents below
					
					long indexDifferent = FileUtil.compareContents(file1, file2);
					if (indexDifferent != -1) return file2;	// if both are normal files, examine the contents of the files themselves
					
// +++ is there any normal file meta data that should also do a comparison on?  do NOT want modification times, but is there anything else?
				}
				else {
					return file2;	// file1 is a normal file but file2 is not, therefore there is no way that they can be the same
				}
			}
		}
		if (contents1.length < contents2.length) return contents2[n];
		else if (contents1.length > contents2.length) return contents1[n];
		else return null;
	}
	
	
	// -------------------- delete, deleteIfExists, gut, erase --------------------
	
	
	/**
	* This utility method deletes a directory along with all of its contents.
	* It is useful because the delete method of File fails if the directory has any content.
	* <p>
	* Note: {@link FileUtil#delete FileUtil.delete} should be used to delete a normal file.
	* <p>
	* @param directory a File for the directory that is to be deleted
	* @throws IllegalArgumentException if directory is null, non-existent, or is not a directory
	* @throws SecurityException if a security manager exists and denies read access to some file
	* @throws IOException if an I/O problem occurs
	*/
	public static void delete(File directory) throws IllegalArgumentException, SecurityException, IOException {
		check(directory);
		erase(directory, true);
	}
	
	
	/**
	* This method calls <code>{@link #delete delete}(directory)</code> if directory != null and directory exists.
	* <p>
	* Note: {@link FileUtil#deleteIfExists FileUtil.deleteIfExists} should be used to delete a normal file if it exists.
	* <p>
	* @param directory a File for the directory that is to be deleted; may be null or non-existent
	* @throws IllegalArgumentException if directory is not  a directory
	* @throws SecurityException if a security manager exists and denies read access to some file
	* @throws IOException if an I/O problem occurs; this includes failure of the directory to be deleted
	*/
	public static void deleteIfExists(File directory) throws IllegalArgumentException, SecurityException, IOException {
		if ( (directory != null) && directory.exists() ) delete(directory);
	}
	
	
	/**
	* This utility method deletes <i>only the contents</i> of a directory; the directory itself will remain.
	* It is useful because the File class has no such method.
	* <p>
	* @param directory a File for the directory whose contents are to be deleted
	* @throws IllegalArgumentException if directory is null, non-existent, or is not an actual directory
	* @throws SecurityException if a security manager exists and denies read access to some file
	* @throws IOException if an I/O problem occurs
	*/
	public static void gut(File directory) throws IllegalArgumentException, SecurityException, IOException {
		check(directory);
		erase(directory, false);
	}
	
	
	/**
	* This utility method always deletes the contents of directory;
	* the fate of the directory itself is specified by the deleteDirectory arg.
	* This method is useful because the File class has no such method.
	* <p>
	* This method marches down the directory tree depth first, deleting all normal files,
	* and then deleting directories as they become empty.
	* <p>
	* @param directory a File for the directory whose contents are to be deleted
	* @throws IllegalArgumentException if directory is null, non-existent, or is not an actual directory
	* @throws SecurityException if a security manager exists and denies read access to some file
	* @throws IOException if an I/O problem occurs
	*/
	protected static void erase(File directory, boolean deleteDirectory) throws IllegalArgumentException, SecurityException, IOException {
		// no need to arg check here, as the the public methods which calls this method should do so already, plus the call to getContents below will as well
		
			// first delete the contents of the directory:
		for (File file : getContents( directory, null )) {
			if ( file.isDirectory() )
				erase(file, true);
			else {
				FileUtil.delete( file );
			}
		}
		
			// and then delete the directory itself, if this is called for:
		if (deleteDirectory) {
			boolean deleted = directory.delete();
			if (!deleted) throw new IOException("Failed to delete " + directory.getPath());
		}
	}
	
	
	/**
	* This method trys to rename directory1 to directory2, and throws an IOException if the rename failed.
	* It is useful because {@link File#renameTo File.renameTo} unfortunately merely returns a boolean indicating
	* the success of the operation, rather than throwing an Exception.
	* <p>
	* Note: {@link FileUtil#rename FileUtil.rename} should be used to rename normal files.
	* <p>
	* @param directory1 the currently existing directory
	* @param directory2 the directory that is to be renamed to
	* @throws IllegalArgumentException if {@link #check check} objects to directory1; directory2 == null or already exists
	* @throws SecurityException if a security manager exists and denies write access to directory1 or directory2
	* @throws IOException if an I/O problem occurs; this includes failure of the directory to be renamed
	*/
	public static void rename(File directory1, File directory2) throws IllegalArgumentException, SecurityException, IOException {
		check(directory1);
		if (directory2 == null) throw new IllegalArgumentException("directory2 == null");
		if (directory2.exists()) throw new IllegalArgumentException("directory2 = " + directory2.getPath() + " already exists");
		
		boolean renamed = directory1.renameTo( directory2 );
		if (!renamed) throw new IOException("failed to rename directory1 = " + directory1.getPath() + " to directory2 = " + directory2.getPath());
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private DirUtil() {}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_all() throws Exception {
				// create a test directory tree:
			String shouldBeUnusedName = "aaaabbbb";
			File dir1 = new File(WD, shouldBeUnusedName);
			ensureExists(dir1);
			File dir2 = new File(dir1, shouldBeUnusedName);
			ensureExists(dir2);
			File dir3 = new File(dir2, shouldBeUnusedName);
			ensureExists(dir3);
			File file3 = new File(dir3, "test.tmp");
			if ( !file3.createNewFile() ) throw new IOException("failed to create " + file3.getPath());
			
				// test the getTree method (which in turn tests getContents):
			Assert.assertTrue( getTree(dir1).length == 3 );
			
				// test the getFilesInTree method (which in turn tests getContents):
			Assert.assertTrue( getFilesInTree(dir1).length == 1 );
			
				// now try and delete the test directory tree:
			delete(dir1);
			Assert.assertFalse( dir1.exists() );
			
// +++ need to test all methods, such as compareContents
		}
		
	}
	
	
}
