/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.io;


import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.security.*;
import java.security.spec.*;
import java.util.*;

import javax.crypto.*;
import javax.crypto.spec.*;
import javax.swing.*;

import org.junit.Assert;
import org.junit.Test;

import bb.gui.DialogInputSecure;
import bb.util.CommandLineUtil;
import bb.util.DateUtil;
import bb.util.Properties2;
import bb.util.StringUtil;


/**
* This class provides static utility methods for encryption.
* A {@link #main main} allows this program to also be used as a standalone command line application,
* especially useful for batch encryption and decryption operations (e.g. as part of a nightly backup process).
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @see <a href = "http://www.rsasecurity.com/products/bsafe/whitepapers/Article3-PBE.pdf">RSA paper on PBE</a>
* @see <a href = "http://developers.slashdot.org/developers/04/09/01/1517242.shtml?tid=172&tid=8&tid=218">Many hash functions like MD5 and SHA0 are now compromised</a>
* @see <a href = "http://www.ftponline.com/javapro/2004_03/online/security_kjones_03_31_04/default_pf.aspx">Java for Symmetric Cryptography</a>
* @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jce/JCERefGuide.html">JCE documentation</a>
* @see <a href="http://www.bouncycastle.org/specifications.html">Bouncy Castle</a>
* @see <a href="http://forum.java.sun.com/thread.jsp?forum=9&thread=542790">What is best algorithm for PBE? Does 3DES still cut it? AES avail?</a>
* @see <a href="http://forum.java.sun.com/thread.jsp?forum=9&thread=490728&tstart=0&trange=15">The DEFINITIVE thread on secure password entry (console, gui, code reviews)</a>
* @see "bbLibrary/doc/encryptionNotes.txt"
* @author Brent Boyer
*/
public class EncryptUtil {


	// -------------------- encryption constants --------------------


	public static final String ENCRYPTION_ALGORITHM_DEFAULT = "PBEWithMD5AndTripleDES";	// NEED unlimited strength policy files installed in order to use this; see http://java.sun.com/products/jce/index-14.html#UnlimitedDownload
// +++ above is BAD: MD5 is now cracked (http://developers.slashdot.org/developers/04/09/01/1517242.shtml?tid=172&tid=8&tid=218), need sha-512, and TripleDES is very slow and less secure than AES
//	public static final String ENCRYPTION_ALGORITHM_DEFAULT = "PBEWithMD5AndDES";	// this is really weak, but will work if only have the default "strong" policy files
	// Note: can use different algorithms for key & cipher generation?  See Reply 7 of 9 of http://forum.java.sun.com/thread.jsp?forum=9&thread=277910

	public static final String SECURE_RANDOM_ALGORITHM = "SHA1PRNG";


	// -------------------- switch constants --------------------


		// high level:
	public static final String fileInput_key = "-fileInput";
	public static final String fileOutput_key = "-fileOutput";
	public static final String operation_key = "-operation";
	
		//algorithm related:
	public static final String algorithm_key = "-algorithm";
	public static final String fileSaltSource_key = "-fileSaltSource";
	public static final String fileSaltTarget_key = "-fileSaltTarget";
	public static final String fileIterationCount_key = "-fileIterationCount";
	
		// password:
	public static final String passwordInstruction_key = "-passwordInstruction";
	public static final String password_key = "-password";
	
		// final commands:
	public static final String ensureExit_key = "-ensureExit";

	/** Specifies all the switch keys which, in theory, can legally appear as command line arguments to {@link #main main}. */
	private static final Set<String> EncryptUtil_keys = new HashSet<String>( Arrays.asList(
		fileInput_key, fileOutput_key, operation_key,
		algorithm_key, fileSaltSource_key, fileSaltTarget_key, fileIterationCount_key,
		passwordInstruction_key, password_key,
		ensureExit_key
	) );
	
		// values for the property corresponding to operation_key:
	public static final String encryptOperation = "encrypt";
	public static final String decryptOperation = "decrypt";
	
	
	// -------------------- constructor --------------------


	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private EncryptUtil() {}


	// -------------------- main --------------------


	/**
	* This main encrypts or decrypts a target File into a destination file.
	* <p>
	* The File to be encrypted/decrypted must have its path specified on the command line in the usual switch format
	* as <code>-fileInput <i>insertPathHere</i></code>.
	* The destination file must have its path specified on the command line in the usual switch format
	* as <code>-fileOutput <i>insertPathHere</i></code>.
	* <p>
	* The operation to perform (encryption or decryption) must be specified on the command line in the usual switch format
	* as <code>-operation encrypt</code> or <code>-operation decrypt</code>.
	* <p>
	* The Password Based Encryption algorithm name must be specified on the command line in the usual switch format
	* as <code>-algorithm <i>insertAlgorithmNameHere</i></code>.
	* <p>
	* The salt (a random byte[] which is used as part of Password Based Encryption) must be specified in one of the following ways:
	* <ol>
	*  <li>
	*		the command line switch <code>-fileSaltSource <i>insertPathHere</i></code> means that the random salt bytes
	*		have previously been generated and stored in the specified file; they are to be read in and used here;
	*		usually this option is used for decryption, with the salt file having been generated as part of encryption
	*  </li>
	*  <li>
	*		the command line switch <code>-fileSaltTarget <i>insertPathHere</i></code> means that the random salt bytes
	*		need to be generated and then stored in the specified file before being used here;
	*		usually this option is used as part of encryption--the salt must be saved in a file for future decryption
	*  </li>
	* </ol>
	* <i>Note that there is no default salt stored in this class; this is absolutely required to thwart dictionary attacks.
	* Instead, the above options strongly encourage new salt to be used for each round of encryption.</i>
	* <p>
	* The iteration count (also part of Password Based Encryption) must be stored in a file whose path
	* is specified on the command line in the usual switch format as <code>-fileIterationCount <i>insertPathHere</i></code>.
	* This file should be a text file which <i>only</i> contains the digits of the iteration count
	* (e.g. "1000000" is valid contents; "1000000\n" is not as it contains a line end).
	* Unlike the salt, it is not critical that the iteration count change, just that it be a sufficiently high value (e.g. > 1,000).
	* <p>
	* A password may optionally be specified on the command line in the usual switch format
	* as <code>-password <i>insertPasswordHere</i></code>; if this is not specified then one will be prompted for.
	* <p>
	* <b>Warning:</b> if supply the password via the command line in this manner, there is a loss in security
	* (the password will be stored at one point in a String, not a char[], and so can never be blanked out after use).
	* Supplying the password via command line is only advised when doing batch encryptions from a trusted machine
	* (e.g. as part of an automated backed process).
	* <p>
	* Finally, it may be necessary to force the JVM to exit upon completion of the main method.
	* (One use is if main uses any gui code: since Java gui code starts all kinds of persistent threads
	* such as event handlers, the program will never die unless an explicit call to System.exit is made.)
	* This guaranteed exit may optionally be specified on the command line in the usual switch format
	* as <code>-ensureExit</code>.
	* If any Throwable is caught by main, this too will force the JVM to exit with an exit code of 1.
	*/
	public static void main(String[] args) {
		Properties2 switches = CommandLineUtil.parseArgs(args);
		CommandLineUtil.checkSwitchKeys(switches, EncryptUtil_keys, "for " + EncryptUtil.class.getName());
		
		int exitCode = 0;
		PBEKeySpec pbeKeySpec = null;
		InputStream in = null;
		OutputStream out = null;
		try {
			File fileInput = switches.getMandatoryFile(fileInput_key);
			File fileOutput = switches.getMandatoryFile(fileOutput_key);
			int operation = extractOperation(switches);
			String algorithm = switches.getMandatoryProperty(algorithm_key);
			PBEParameterSpec pbeParamSpec = extractPbeParameterSpec(switches);
			
			System.out.print("Key & Cipher generation start: " + DateUtil.getTimeStamp());
			pbeKeySpec = extractPbeKeySpec(switches, pbeParamSpec);
			Cipher cipher = getPbeCipher(algorithm, pbeKeySpec, pbeParamSpec, operation);
			System.out.println(" end: " + DateUtil.getTimeStamp());
			
			in = new FileInputStream(fileInput);
			out = new CipherOutputStream( new FileOutputStream(fileOutput), cipher );
			
			System.out.print("Crypto start: " + DateUtil.getTimeStamp());
			StreamUtil.transfer(in, out);
			System.out.println(" end: " + DateUtil.getTimeStamp());
		}
		catch (Throwable t) {
			exitCode = 1;
			t.printStackTrace(System.err);
		}
		finally {
			if (pbeKeySpec != null) pbeKeySpec.clearPassword();
			StreamUtil.close(in);
			StreamUtil.close(out);
			
			if (switches.containsKey(ensureExit_key) || (exitCode != 0)) System.exit(exitCode);
		}
	}
	
	
	private static int extractOperation(Properties2 switches) throws IllegalArgumentException {
		String operation = switches.getMandatoryProperty(operation_key);
		if (encryptOperation.equals(operation)) return Cipher.ENCRYPT_MODE;
		else if (decryptOperation.equals(operation)) return Cipher.DECRYPT_MODE;
		else throw new IllegalArgumentException("the " + operation_key + " value " + operation + " is unsupported");
	}


	private static PBEParameterSpec extractPbeParameterSpec(Properties2 switches) throws IllegalArgumentException, InterruptedException, IllegalStateException, IOException, NoSuchAlgorithmException {
		byte[] salt;
		if (switches.containsKey(fileSaltSource_key)) {
			File file = new File( switches.getProperty(fileSaltSource_key) );
			salt = FileUtil.readBytes(file);
		}
		else if (switches.containsKey(fileSaltTarget_key)) {
			salt = randomBytes(8);	// Sun's JCE seems to require exactly 8 bytes, at least for the PBEWithMD5AndTripleDES algorithm...
			File file = new File( switches.getProperty(fileSaltTarget_key) );
			FileUtil.writeBytes(salt, file, false);	// false means we overwrite the file
		}
		else
			throw new IllegalArgumentException("failed to specify the salt file (\"" + fileSaltSource_key + "\" or \"" + fileSaltTarget_key + "\")");

		File file = switches.getMandatoryFile(fileIterationCount_key);
		int iterationCount = Integer.parseInt( FileUtil.readString(file) );
// +++ add a check that iterationCount is sufficiently large?  need a minimum of 1000 (see http://www.rsasecurity.com/products/bsafe/whitepapers/Article3-PBE.pdf)

		return new PBEParameterSpec(salt, iterationCount);
	}


	private static PBEKeySpec extractPbeKeySpec(Properties switches, PBEParameterSpec pbeParamSpec) throws IOException, InterruptedException {
		char[] password = null;
		try {
			if (switches.containsKey(password_key)) {
				password = switches.getProperty(password_key).toCharArray();	// WARNING: there is no way to null out the original password String here, so this is a bit of a security loss (see main javadoc)
				if (!isValidPassword(password)) throw new IllegalArgumentException("the supplied password (value suppressed for security) is invalid");
			}
			else {
				String passwordInstruction = switches.getProperty(passwordInstruction_key);
				//passwordInstruction = (passwordInstruction != null) ? " (" + passwordInstruction + ")" : "";
				//password = readConsoleSecure("Enter password" + passwordInstruction + ": ");
				// WARNING: the above console input is medicore (see StreamMasker javadocs below), so use a gui instead:
				password = DialogInputSecure.getInputSecure(null, "Password entry", true, passwordInstruction, "Enter password:", 32);
				if (!isValidPassword(password)) throw new IllegalStateException("the entered password (value suppressed for security) is invalid");
			}
			
			return new PBEKeySpec(password, pbeParamSpec.getSalt(), pbeParamSpec.getIterationCount());
		}
		finally {
			eraseChars(password);	// PBEKeySpec will store a clone of password internally, so must clear our local version before returning
		}
	}


	/**
	* Checks if the supplied password is valid.
* Current implementation (subject to change) is that only null or zero-length passwords are invalid.
* Should add stronger checks in the future (e.g. R5),
* but problem is that want this program to be able to decrypt files from any source,
* and the encrypting party may have different password policies in place.
	*/
	private static boolean isValidPassword(char[] password) {
		if ((password == null) || (password.length == 0)) return false;

// see:
// http://www.ftp.cl.cam.ac.uk/ftp/users/rja14/tr500.pdf
// http://www-10.lotus.com/ldd/today.nsf/0/098c9f7d4a0cccbd85256abc0011e4f0?OpenDocument

		return true;
	}


	// -------------------- getCipher --------------------


	/**
	* Returns a Password Based Encryption (PBE) Cipher.
	* <p>
	* The Cipher is initialized to algorithm and mode before being returned.
	* <p>
	* This method never returns null.
	* <p>
	* @throws Exception (many varieties) if any problem occurs
	* @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jce/JCERefGuide.html#PBEEx">Password based encryption code examples from JCE documentation</a>
	*/
	public static Cipher getPbeCipher(String algorithm, PBEKeySpec pbeKeySpec, PBEParameterSpec pbeParamSpec, int mode) throws IllegalArgumentException, NoSuchAlgorithmException, InvalidKeySpecException, NoSuchPaddingException, InvalidKeyException, InvalidAlgorithmParameterException {
		if (algorithm == null) throw new IllegalArgumentException("algorithm == null");
		if (pbeKeySpec == null) throw new IllegalArgumentException("pbeKeySpec == null");
		if (pbeParamSpec == null) throw new IllegalArgumentException("pbeParamSpec == null");

		SecretKey pbeKey = SecretKeyFactory.getInstance(algorithm).generateSecret(pbeKeySpec);
		Cipher pbeCipher = Cipher.getInstance(algorithm);
		pbeCipher.init(mode, pbeKey, pbeParamSpec);
		return pbeCipher;
	}


	// -------------------- readConsoleSecure, readLineSecure, checkBuffer --------------------


/*
+++ all this code will be obsoleted by the new java.io.Console class in jdk 1.6:
	http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4050435
	http://download.java.net/jdk6/docs/api/java/io/Console.html
	http://blogs.sun.com/roller/page/alanb/20051021

char[] password = System.console().readPassword("password: ");
*/

	/**
	* Reads and returns some sensitive piece of information (e.g. a password)
	* from the console (i.e. System.in and System.out) in a secure manner.
	* <p>
	* For top security, all console input is masked out while the user types in the password.
	* Once the user presses enter, the password is read via a call to {@link #readLineSecure readLineSecure(in)},
	* using a PushbackInputStream that wraps System.in.
	* <p>
	* This method never returns null.
	* <p>
	* @throws IOException if an I/O problem occurs
	* @throws InterruptedException if the calling thread is interrupted while it is waiting at some point
	* @see <a href="http://java.sun.com/features/2002/09/pword_mask.html">Password masking in console</a>
	*/
	public static char[] readConsoleSecure(String prompt) throws IOException, InterruptedException {
			// start a separate thread which will mask out all chars typed on System.in by overwriting them using System.out:
		StreamMasker masker = new StreamMasker(System.out, prompt);
		Thread threadMasking = new Thread( masker, "EncryptUtil_StreamMasker" );
		threadMasking.setPriority( Thread.NORM_PRIORITY );
		threadMasking.start();
		
			// Goal: block this current thread (allowing masker to mask all user input)
			// while the user is in the middle of typing the password.
			// This may be achieved by trying to read just the first byte from System.in,
			// since reading from System.in blocks until it detects that an enter has been pressed.
			// Wrap System.in with a PushbackInputStream because this byte will be unread below.
		PushbackInputStream pin = new PushbackInputStream(System.in);
		int b = pin.read();

		// When current thread gets here, the block on reading System.in is over (e.g. the user pressed enter, or some error occurred?)

			// signal threadMasking to stop and wait till it is dead:
		masker.stop();
		threadMasking.join();

			// check for stream errors:
		if (b == -1) throw new IOException("end-of-file was detected in System.in without any data being read");
		if (System.out.checkError()) throw new IOException("an I/O problem was detected in System.out");

			// pushback the first byte and supply the now unaltered stream to readLineSecure which will return the complete password:
		pin.unread(b);
		return readLineSecure(pin);
	}


	/**
	* Reads chars from in until an end-of-line sequence (EOL) or end-of-file (EOF) is encountered,
	* and then returns the data as a char[].
	* <p>
	* The EOL sequence may be any of the standard formats: '\n' (unix), '\r' (mac), "\r\n" (dos).
	* The EOL sequence is always completely read off the stream but is never included in the result.
	* <i>Note:</i> this means that the result will never contain the chars '\n' or '\r'.
	* In order to guarantee reading thru but not beyond the EOL sequence for all formats (unix, mac, dos),
	* this method requires that a PushbackInputStream and not a more general InputStream be supplied.
	* <p>
	* The code is secure: no Strings are used, only char arrays,
	* and all such arrays other than the result are guaranteed to be blanked out after last use to ensure privacy.
	* Thus, this method is suitable for reading in sensitive information such as passwords.
	* <p>
	* This method never returns null; if no data before the EOL or EOF is read, a zero-length char[] is returned.
	* <p>
	* @throws IllegalArgumentException if in == null
	* @throws IOException if an I/O problem occurs
	* @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jce/JCERefGuide.html#PBEEx">Password based encryption code examples from JCE documentation</a>
	*/
	public static char[] readLineSecure(PushbackInputStream in) throws IllegalArgumentException, IOException {
		if (in == null) throw new IllegalArgumentException("in == null");

		char[] buffer = null;
		try {
			buffer = new char[128];
			int offset = 0;

	loop:   while (true) {
				int c = in.read();
				switch (c) {
					case -1:
					case '\n':
						break loop;

					case '\r':
						int c2 = in.read();
						if ((c2 != '\n') && (c2 != -1)) in.unread(c2);	// guarantees that mac & dos line end sequences are completely read thru but not beyond
						break loop;

					default:
						buffer = checkBuffer(buffer, offset);
						buffer[offset++] = (char) c;
						break;
				}
			}

			char[] result = new char[offset];
			System.arraycopy(buffer, 0, result, 0, offset);
			return result;
		}
		finally {
			eraseChars(buffer);
		}
	}


	/**
	* Checks if buffer is sufficiently large to store an element at an index == offset.
	* If it is, then buffer is simply returned.
	* If it is not, then a new char[] of more than sufficient size is created and initialized with buffer's current elements and returned;
	* the original supplied buffer is guaranteed to be blanked out upon method return in this case.
	* <p>
	* @throws IllegalArgumentException if buffer == null; offset < 0
	*/
	public static char[] checkBuffer(char[] buffer, int offset) throws IllegalArgumentException {
		if (buffer == null) throw new IllegalArgumentException("buffer == null");
		if (offset < 0) throw new IllegalArgumentException("offset = " + offset + " is < 0");

		if (offset < buffer.length)
			return buffer;
		else {
			try {
				char[] bufferNew = new char[offset + 128];
				System.arraycopy(buffer, 0, bufferNew, 0, buffer.length);
				return bufferNew;
			}
			finally {
				eraseChars(buffer);
			}
		}
	}


	/** If buffer is not null, fills buffer with space (' ') chars. */
	public static void eraseChars(char[] buffer) {
		if (buffer != null)  Arrays.fill(buffer, ' ');
	}


	// -------------------- randomBytes --------------------


	/**
	* Returns a cryptographically strong pseudo-random sequence of bytes as an array.
	* A new {@link SecureRandom} instance, whose internal state is completely randomized, is used for each call.
	* <p>
	* @throws IllegalArgumentException if length < 0
	* @throws NoSuchAlgorithmException if {@link #SECURE_RANDOM_ALGORITHM} is unavailable
	*/
	public static byte[] randomBytes(int length) throws NoSuchAlgorithmException {
		if (length < 0) throw new IllegalArgumentException("length = " + length + " is < 0");

		return SecureRandom.getInstance(SECURE_RANDOM_ALGORITHM).generateSeed(length);
	}


	// -------------------- printAvailableAlgorithms --------------------


	/** Simply prints all the available algorithm names for seviceName to out. */
	public static void printAvailableAlgorithms(String serviceName, PrintStream out) throws IllegalArgumentException {
		if (serviceName == null) throw new IllegalArgumentException("serviceName == null");
		if (out == null) throw new IllegalArgumentException("out == null");

		out.println();
		out.println("Available " + serviceName + " algorithms:");
		SortedSet<String> algorithmNameSet = new TreeSet<String>( Security.getAlgorithms(serviceName) );
		for (String algorithmName : algorithmNameSet) {
			out.println("\t" + algorithmName);
		}
	}


	// -------------------- StreamMasker (static inner class) --------------------

	// see also the end of this class for another version which would work even better if ansi control sequences are supported

	/**
	* Masks an InputStream by overwriting blank chars to the PrintStream corresponding to its output.
	* A typical application is for password input masking.
	* <p>
	* @see <a href="http://forum.java.sun.com/thread.jsp?forum=9&thread=490728&tstart=0">My forum posting on password entry</a>
	* @see <a href="http://java.sun.com/features/2002/09/pword_mask.html">Password masking in console</a>
	*/
	public static class StreamMasker implements Runnable {

		private static final String TEN_BLANKS = StringUtil.repeatChars(' ', 10);
		private final PrintStream out;
		private final String promptOverwrite;
		private volatile boolean doMasking;	// MUST be volatile to ensure update by one thread is instantly visible to other threads

		/**
		* Constructor.
		* <p>
		* @throws IllegalArgumentException if out == null; prompt == null; prompt contains the char '\r' or '\n'
		*/
		public StreamMasker(PrintStream out, String prompt) throws IllegalArgumentException {
			if (out == null) throw new IllegalArgumentException("out == null");
			if (prompt == null) throw new IllegalArgumentException("prompt == null");
			if (prompt.contains("\n")) throw new IllegalArgumentException("prompt contains the new line ('\\n') char");
			if (prompt.contains("\r")) throw new IllegalArgumentException("prompt contains the carriage return ('\\r') char");

			this.out = out;
			String setCursorToStart = StringUtil.repeatChars('\b', prompt.length() + TEN_BLANKS.length());
			this.promptOverwrite =
				setCursorToStart +    // sets cursor back to beginning of line:
				prompt +              // writes prompt (critical: this reduces visual flicker in the prompt text that otherwise occurs if simply write blanks here)
				TEN_BLANKS +          // writes 10 blanks beyond the prompt to mask out any input; go 10, not 1, spaces beyond end of prompt to handle the (hopefully rare) case that input occurred at a rapid rate
				setCursorToStart +    // sets cursor back to beginning of line:
				prompt;               // writes prompt again; the cursor will now be positioned immediately after prompt (critical: overwriting only works if all input text starts here)
		}

		/**
		* Repeatedly overwrites the current line of out with prompt followed by blanks.
		* This effectively masks any chars coming on out, as long as the masking occurs fast enough.
		* <p>
		* To help ensure that masking occurs when system is in heavy use, the calling thread will have its priority
		* boosted to the max for the duration of the call (with its original priority restored upon return).
		* Interrupting the calling thread will eventually result in an exit from this method,
		* and the interrupted status of the calling thread will be set to true.
		* <p>
		* @throws RuntimeException (or some subclass) if any error occurs; this may merely wrap some other underlying Throwable
		*/
		public void run() throws RuntimeException {
			int priorityOriginal = Thread.currentThread().getPriority();
			try {
				Thread.currentThread().setPriority(Thread.MAX_PRIORITY);
				doMasking = true;	// do this assignment here and NOT at variable declaration line to allow this instance to be restarted if desired
				while (doMasking) {
					out.print(promptOverwrite);
						// call checkError, which first flushes out, and then lets us confirm that everything was written correctly:
					if (out.checkError()) throw new RuntimeException("an I/O problem was detected in out");	// should be an IOException, but that would break method contract

						// limit the masking rate to fairly share the cpu; interruption breaks the loop
					try {
						Thread.sleep(1);	// have experimentally found that sometimes see chars for a brief bit unless set this to its min value
					}
					catch (InterruptedException ie) {
						Thread.currentThread().interrupt();	// resets the interrupted status, which is typically lost when an InterruptedException is thrown, as per our method contract; see Doug Lea "Concurrent Programming in Java Second Edition" p. 171; in general should never swallow this; see http://www-128.ibm.com/developerworks/java/library/j-jtp05236.html
						return;	// return, NOT break, since now want to skip the lines below where write bunch of blanks since typically the user will not have pressed enter yet
					}
				}
					// erase any prompt that may have been spuriously written on the NEXT line after the user pressed enter (see reply 2 of http://forum.java.sun.com/thread.jsp?forum=9&thread=490728&tstart=0)
				out.print('\r');
				for (int i = 0; i < promptOverwrite.length(); i++) out.print(' ');
				out.print('\r');
			}
			finally {
				Thread.currentThread().setPriority(priorityOriginal);
			}
		}

		/** Signals any thread executing run to stop masking and exit run. */
		public void stop() { doMasking = false; }

	}


	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_printAvailableAlgorithms() {
			printAvailableAlgorithms("Cipher", System.out);
			printAvailableAlgorithms("KeyStore", System.out);
			printAvailableAlgorithms("Mac", System.out);
			printAvailableAlgorithms("MessageDigest", System.out);
			printAvailableAlgorithms("SecretKeyFactory", System.out);
			printAvailableAlgorithms("Signature", System.out);
		}
			
		@Test public void test_main() throws Exception {
			File filePlaintext = FileUtil.createTemp("../log/filePlainText.txt");
			File fileCiphertext = FileUtil.createTemp("../log/fileCiphertext.bytes");
			File fileDecryptedtext = FileUtil.createTemp("../log/fileDecryptedtext.txt");
			File fileSalt = FileUtil.createTemp("../log/fileSalt.bytes");
			File fileIterationCount = FileUtil.createTemp("../log/fileIterationCount.txt");
			
				// write sample plain text to filePlaintext
			FileUtil.writeString("sample plaintext...", filePlaintext, false);	// overwrite any existing contents
			
				// write sample iteration count to fileIterationCount
			FileUtil.writeString("1000", fileIterationCount, false);	// note that there is no line end; overwrite any existing contents
			// WARNING: in real life, use a iteration count value bigger than 1,000
			
				// encrypt filePlaintext to fileCiphertext
			System.out.println();
			main( new String[] {
				"-fileInput", filePlaintext.getPath(),
				"-fileOutput", fileCiphertext.getPath(),
				"-operation", encryptOperation,
				"-algorithm", ENCRYPTION_ALGORITHM_DEFAULT,
				"-fileSaltTarget", fileSalt.getPath(),	// generate and store salt here
				"-fileIterationCount",  fileIterationCount.getPath(),
				//"-passwordInstruction",  "Enter some password below.  Note that character echo is suppressed.",				
				"-password", "abracadabra"	// CRITICAL: must supply the password here, else a gui dialog will pop up, which will ruin the JUnit auto test
			} );
			
				// decrypt fileCiphertext to fileDecryptedtext
			System.out.println();
			main( new String[] {
				"-fileInput", fileCiphertext.getPath(),
				"-fileOutput", fileDecryptedtext.getPath(),
				"-operation", decryptOperation,
				"-algorithm", ENCRYPTION_ALGORITHM_DEFAULT,
				"-fileSaltSource", fileSalt.getPath(),	// use the same salt file used in encryption
				"-fileIterationCount",  fileIterationCount.getPath(),
				//"-passwordInstruction",  "Enter the same password here as entered for encryption.",				
				"-password", "abracadabra",	// CRITICAL: must supply the password here, else a gui dialog will pop up, which will ruin the JUnit auto test
				//"-ensureExit"	// do NOT do this unless executing this class manually, since it could prematurely kill a JUnit run
			} );
			
			Assert.assertTrue( FileUtil.compareContents(filePlaintext, fileDecryptedtext) == -1 );
			
// +++ confirm that can decrypt the file using a 3rd party program, especially if it is written in another language and on another platform
		}
	}


}


/*

The code below SHOULD in theory work with any console which supports Ansi control codes;
it is free of the bug found in the version of this class in the actual code above.
Unfortunately, I have not been able to test it, since win nt+ basicly does not support ansi control codes
(only if you specially configure it and only then if you use command.com instead of the default and better cmd.exe).
This is a darn shame, as it renders the code virtually useless, except maybe if you are restricted to unix.

	**
	* Masks an InputStream by overwriting blank chars to the PrintStream corresponding to its output.
	* A typical application is for password input.
	* <p>
	* @see <a href="http://java.sun.com/features/2002/09/pword_mask.html">Password masking in console</a>
	*
	public static class StreamMasker implements Runnable {

		// For more on ansi codes:
		//	http://www.computerhope.com/ansisys.htm#00
		//	http://www.evergreen.edu/biophysics/technotes/program/ansi_esc.htm
		//	http://forum.java.sun.com/thread.jsp?forum=31&thread=429395
		// Ansi code support on dos/windoze systems:
		//	requires having the ansi driver called ansi.sys installed
		//	for instructions see section titled "Enabling ANSI.SYS" of http://www.evergreen.edu/biophysics/technotes/program/ansi_esc.htm
		// Java curses packages:
		//	http://www.nongnu.org/jcurzez/
		private static final String CURSOR_SAVE = "\033[s";
		private static final String CURSOR_RESTORE = "\033[u";
		private static final String ERASE_TO_LINE_END = "\033[K";

		private final PrintStream out;
		private final String prompt;
		private volatile boolean doMasking;	// MUST be volatile to ensure update by one thread is instantly visible to other threads

		** Constructor.
		* <p>
		* @throws IllegalArgumentException if out == null; prompt == null; prompt contains the char '\r' or '\n'
		*
		public StreamMasker(PrintStream out, String prompt) throws IllegalArgumentException {
			if (out == null) throw new IllegalArgumentException("out == null");
			if (prompt == null) throw new IllegalArgumentException("prompt == null");
			if (prompt.contains('\r')) throw new IllegalArgumentException("prompt contains the char '\\r'");
			if (prompt.contains('\n')) throw new IllegalArgumentException("prompt contains the char '\\n'");

			this.out = out;
			this.prompt = prompt;
		}

		** Constantly overwrites out with prompt.
		* This effectively masks any chars coming on out, as long as the masking occurs fast enough.
		* <p>
		* To help ensure that masking occurs when system is in heavy use, the calling thread will have its priority
		* boosted to the max for the duration of the call (but its original priority will be restored upon return).
		* Interrupting the calling thread will eventually result in an exit from this method,
		* and the interrupted status of the calling thread will be set to true.
		* <p>
		* @throws RuntimeException (or some subclass) if any error occurs; this may merely wrap some other underlying Throwable
		*
		public void run() throws RuntimeException {
			int priorityOriginal = Thread.currentThread().getPriority();
			try {
				Thread.currentThread().setPriority(Thread.MAX_PRIORITY);
				out.print(prompt);
				out.print(CURSOR_SAVE);	// save cursor position just after prompt
				doMasking = true;	// do this assignment here and NOT at variable declaration line to allow this instance to be restarted if desired
				while (doMasking) {
					out.print(CURSOR_RESTORE);	// restore cursor position to just after prompt
					out.print(ERASE_TO_LINE_END);	// erase anything else on line
						// call checkError, which first flushes out, and then lets us confirm that everything was written correctly:
					if (out.checkError()) throw new RuntimeException("an I/O problem was detected in out");	// should be an IOException, but that would break method contract

						// limit the masking rate to fairly share the cpu; interruption breaks the loop
					try {
						Thread.sleep(1);	// have experimentally found that sometimes see chars for a brief bit unless set this to its min value
					}
					catch (InterruptedException ie) {
						Thread.currentThread().interrupt();	// resets the interrupted status, which is typically lost when an InterruptedException is thrown, as per our method contract; see Doug Lea "Concurrent Programming in Java Second Edition" p. 171; in general should never swallow this; see http://www-128.ibm.com/developerworks/java/library/j-jtp05236.html
						break;
					}
				}
					// once have exited loop, do one last mask followed by a newline to guarantee that the cursor is left in a consistent position
				out.print(CURSOR_RESTORE);	// restore cursor position to just after prompt
				out.print(ERASE_TO_LINE_END);	// erase anything else on line
				out.println();
			}
			finally {
				Thread.currentThread().setPriority(priorityOriginal);
			}
		}

		** Signals any thread executing run to stop masking and exit run. *
		public void stop() { doMasking = false; }

	}

*/