/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*
Programmer's notes:

+++ instead of doing explicit file writing, may want to consider this: http://www.prevayler.org/wiki.jsp

*/


package bb.io;


import java.io.*;
import java.util.*;

import org.junit.Assert;
import org.junit.Test;

import bb.util.StringUtil;


/**
* This class is similar to {@link FileWriter}, except that:
* <ol>
*  <li>writes to a rolling series of files instead of just one file</li>
*  <li>requires the user to specify the buffer size</li>
*  <li>is not multithread safe</li>
* </ol>
* At any moment in time, there is a single output file.
* When the number of chars written to that file is first detected as equalling or exceeding the specified limit,
* that file is transparently closed and a new file is opened for writing.
* <blockquote>
*	This char limit checking is the first action performed (after arg checking) by every one of the write methods.
*	One implication is that the chars supplied to a given write call always appear in the same file
*	(i.e. there are never write splits across 2 files).
*	Another implication is that the lengths of the rollover files are not necessarily exactly equal.
* </blockquote>
* <p>
* The names of the files all share the same prefix and suffix, but in the middle of their name is a series number.
* The number of digits of this series number may be specified.
* <blockquote>
*	For file sorting purposes, <i>it is crucial that the series numbers are represented with exactly the same number of digits</i>.
*	For example, suppose that there will 0-99 rollover files with prefix "log" and suffix ".txt".
*	Then the filenames should be log_#00.txt, log_#01.txt, ..., log_#99.txt to allow typical operating systems to sort them properly.
* </blockquote>
* <p>
* <i>Because this class does its own internal buffering,
* users should not wrap an instance of this class inside a {@link BufferedWriter}.</i>
* <p>
* One use of this class is to handle really large data sets.
* If a process only writes to a single big file, the application may hold a file lock for too long.
* Furthermore, one single big file may be too big to open in, say, a text editor.
* <p>
* This class is not multithread safe. 
* <p>
* @author Brent Boyer
*/
public class FileRollingWriter extends Writer {


	// -------------------- fields --------------------


	private final File directory;
	private final String filePrefix;
	private final int minDigitsInSeriesNumber;
	private final String fileSuffix;
	private final long nCharsTillRollover;
	private final int bufferSize;
	
	private boolean open = true;
	
	private File file;
	private Writer writer;
	private long nCharsWritten = 0;


	// -------------------- constructor --------------------


	/**
	* Constructor.
	* <p>
	* The filenames that will be produced by this instance will have the form
	* <blockquote><code>filePrefix_#<i>nnn</i>fileSuffix</code></blockquote>
	* where filePrefix and fileSuffix are the args to this method,
	* while <i>nnn</i> are the series numbers represented according to minDigitsInSeriesNumber.
	* <p>
	* @param directory the directory to write rollover files into
	* @param filePrefix the prefix to append to the begining of all rollover file names; may be null (in which case "" is substituted)
	* @param minDigitsInSeriesNumber if > 0, specifies the minimum number of digits that must appear in each filename's series number
	* (leading zeroes will be appended to a given number if necessary to make it achieve this length);
	* otherwise, must have the special value -1, which specifies that the series numbers are to be represented in the usual way,
	* that is, each number is written using precisely the minimum number of digits that are required
	* @param fileSuffix the suffix to append to the end of all rollover file names; may be null (in which case "" is substituted)
	* @param nCharsTillRollover the maximum number of chars written to the <i>current</i> rollover file
	* that a write method is to tolerate before closing this file and start writing to a new one
	* @throws IllegalArgumentException if directory fails {@link DirUtil#check DirUtil.check};
	* minDigitsInSeriesNumber < -1; nCharsTillRollover <= 0; bufferSize <= 0
	*/
	public FileRollingWriter(File directory, String filePrefix, int minDigitsInSeriesNumber, String fileSuffix, long nCharsTillRollover, int bufferSize) throws IllegalArgumentException {
		DirUtil.check(directory);
		if (minDigitsInSeriesNumber < -1) throw new IllegalArgumentException("minDigitsInSeriesNumber = " + minDigitsInSeriesNumber + " < -1");
		if (nCharsTillRollover <= 0) throw new IllegalArgumentException("nCharsTillRollover = " + nCharsTillRollover + " <= 0");
		if (bufferSize <= 0) throw new IllegalArgumentException("bufferSize = " + bufferSize + " is <= 0");

		this.directory = directory;
		this.filePrefix = (filePrefix != null) ? filePrefix + "_#" : "_#";
		this.minDigitsInSeriesNumber = minDigitsInSeriesNumber;
		this.fileSuffix = (fileSuffix != null) ? fileSuffix : "";
		this.nCharsTillRollover = nCharsTillRollover;
		this.bufferSize = bufferSize;
	}


	// -------------------- public api: write, flush, close --------------------


	/**
	* Simply calls
	*	<blockquote><code>{@link #write(String, int, int) write}( s, 0, s.length() )</code></blockquote>
	* <p>
	* @throws IllegalArgumentException if s == null
	* @throws SecurityException if a security manager exists and denies access to some resource
	* @throws IOException if an I/O problem occurs
	*/
	public void write(String s) throws IllegalArgumentException, SecurityException, IOException {
		if (s == null) throw new IllegalArgumentException("s == null");	// if do not check it here, then the call to s.length() below will result in a NullPointerException

		write( s, 0, s.length() );
	}


	/**
	* Checks whether or not to close the current rollover file and start a new one.
	* Writes the specified portion of s to the rollover file.
	* <p>
	* @param s the String to draw chars from
	* @param offset char index offset of s from which to start writing chars
	* @param  length number of chars of s to write
	* @throws IllegalArgumentException if s == null
	* @throws SecurityException if a security manager exists and denies access to some resource
	* @throws IOException if an I/O problem occurs
	*/
	public void write(String s, int offset, int length) throws IllegalArgumentException, SecurityException, IOException {
		if (s == null) throw new IllegalArgumentException("s == null");
		if (!open) throw new IOException("instance is closed");

		checkWriter();
		writer.write( s, offset, length );
		nCharsWritten += length;
	}


	/**
	* Simply calls
	*	<blockquote><code>{@link #write(char[], int, int) write}( cBuffer, 0, cBuffer.length )</code></blockquote>
	* <p>
	* @throws IllegalArgumentException if cBuffer == null
	* @throws SecurityException if a security manager exists and denies access to some resource
	* @throws IOException if an I/O problem occurs
	*/
	public void write(char[] cBuffer) throws IllegalArgumentException, SecurityException, IOException {
		if (cBuffer == null) throw new IllegalArgumentException("cBuffer == null");	// if do not check it here, then the call to cBuffer.length below will result in a NullPointerException

		write( cBuffer, 0, cBuffer.length );
	}


	/**
	* Checks whether or not to close the current rollover file and start a new one.
	* Writes the specified portion of cBuffer to the rollover file.
	* <p>
	* @param cBuffer the char[] to draw chars from
	* @param offset char index offset of cBuffer from which to start writing chars
	* @param  length number of chars of cBuffer to write
	* @throws IllegalArgumentException if cBuffer == null
	* @throws SecurityException if a security manager exists and denies access to some resource
	* @throws IOException if an I/O problem occurs
	*/
	public void write(char[] cBuffer, int offset, int length) throws IllegalArgumentException, SecurityException, IOException {
		if (cBuffer == null) throw new IllegalArgumentException("cBuffer == null");
		if (!open) throw new IOException("instance is closed");

		checkWriter();
		writer.write( cBuffer, offset, length );
		nCharsWritten += length;
	}


	/**
	* Checks whether or not to close the current rollover file and start a new one.
	* Writes c (cast to a char) to the rollover file.
	* <p>
	* @throws SecurityException if a security manager exists and denies access to some resource
	* @throws IOException if an I/O problem occurs
	*/
	public void write(int c) throws SecurityException, IOException {
		if (!open) throw new IOException("instance is closed");

		checkWriter();
		writer.write( c );
		nCharsWritten += 1;
	}


	/**
	* Flushes any output to the current rollover file.
	* <p>
	* @throws IOException if an I/O problem occurs
	*/
	public void flush() throws IOException {
		if (!open) throw new IOException("instance is closed");
		
		if (writer != null) writer.flush();
	}


	/** Closes all resources used by this instance. */
	public void close() {
		if (open) {
			StreamUtil.close(writer);
			file = null;
			writer = null;
		}
	}


	// -------------------- internal helper methods --------------------


	/** Determines if need to close the current writer and create a new one. */
	private void checkWriter() throws IllegalArgumentException, SecurityException, IOException {
		if ((file == null) || (nCharsWritten >= nCharsTillRollover)) {
			StreamUtil.close(writer);	// can handle writer == null

			file = createRolloverFile( nextFileCount() );
			writer = new BufferedWriter( new OutputStreamWriter( new FileOutputStream(file, false) ), bufferSize );
			nCharsWritten = 0;
		}
	}


	/** Creates the nth rollover file. */
	private File createRolloverFile(int n) {
		String nDigits = String.valueOf(n);
		if ((minDigitsInSeriesNumber > 0) && (nDigits.length() < minDigitsInSeriesNumber)) {
			nDigits = StringUtil.toLength( n, minDigitsInSeriesNumber );
		}

		return new File( directory, filePrefix + nDigits + fileSuffix );
	}


	/**
	* Finds any previously written files that both start with {@link #filePrefix}
	* and end with {@link #fileSuffix},
	* and then returns the next available file count.
	*/
	private int nextFileCount() throws SecurityException, IOException {
		FileFilter fileFilter = new FileFilter() {
			public boolean accept(File file) {
				String name = file.getName();
				return name.startsWith(filePrefix) && name.endsWith(fileSuffix);
			}
		};
		File[] files = DirUtil.getContents(directory, fileFilter, false);
		if (files.length == 0) return 0;
		
		Comparator<File> comparator = new Comparator<File>() {
			public int compare(File file1, File file2) throws IllegalArgumentException {
				// all args checked by calls to extractNumberFromFilename below
				
				return extractNumberFromFilename(file1) - extractNumberFromFilename(file2);
			}
		};
		Arrays.sort(files, comparator);

		File fileLast = files[ files.length - 1 ];
		return extractNumberFromFilename(fileLast) + 1;
	}


	/**
	* Parses and returns the series number from file.
	* Returns -1 if the series number is unparsable.
	* <b>Warning: does no arg or state checking, since is meant only be called by nextFileCount.</b>
	* <p>
	* @throws IllegalArgumentException if file == null
	*/
	private int extractNumberFromFilename(File file) throws IllegalArgumentException  {
		if (file == null) throw new IllegalArgumentException("file == null");
		
		String name = file.getName();
		int index1 = filePrefix.length();
		int index2 = name.length() - fileSuffix.length();
		try {
			return Integer.parseInt( name.substring(index1, index2) );
		}
		catch (NumberFormatException nfe) {
			return -1;
		}
	}


	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_all() throws Exception {
			File dir = new File("../log/testDirectoryFileRollingWriter");
			try {
				DirUtil.ensureExists(dir);
				int n = 10;
				for (int i = 1; i <= n; i++) {
					writeRolling(i, dir);
				}
				int nFilesExpected = n * (n + 1) / 2;	// familiar formula for the sum from i = 1 to n
				Assert.assertTrue( DirUtil.getTree(dir).length == nFilesExpected );
			}
			finally {
				DirUtil.delete(dir);
			}
		}
		
		/**
		* Constructs a new FileRollingWriter instance with {@link #nCharsTillRollover} equal to n.
		* Then writes a String of length n to the FileRollingWriter n times.
		* This should result in n new output fils being produced.
		*/
		private void writeRolling(int n, File dir) throws Exception {
			FileRollingWriter frw = null;
			try {
				frw = new FileRollingWriter(dir, "frw", 2, ".txt", n, 1024);
				String nChars = StringUtil.repeatChars((char) n, n);
				for (int i = 0; i < n; i++) {
					frw.write(nChars);
				}
			}
			finally {
				StreamUtil.close(frw);
			}
		}

	}


}

