/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*
Programmer notes:

+++ CHANGE THE TAR LIBRARY THAT THIS CLASS USES
	--currently am using apache's commons-compress
	--but this code is buggy (see various comments below...), and worse, the project seems to be dead in the water
	--look into using this code which seems to be actively developed:
		https://truezip.dev.java.net/
*/


package bb.io;


import java.awt.Toolkit;
import java.io.*;
import java.util.*;
import java.util.logging.Level;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.compress.Archive;
import org.apache.commons.compress.ArchiveException;
import org.apache.commons.compress.ArchiverFactory;
import org.apache.commons.compress.UnpackException;
import org.apache.commons.compress.archivers.tar.TarArchive;
import org.apache.commons.compress.archivers.tar.TarEntry;
import org.apache.commons.compress.archivers.tar.TarInputStream;
import org.apache.commons.compress.archivers.tar.TarOutputStream;

import org.junit.Assert;
import org.junit.Ignore;
import org.junit.Test;

import bb.util.CommandLineUtil;
import bb.util.DateUtil;
import bb.util.Properties2;
import bb.util.StringUtil;
import bb.util.logging.LogUtil;


/**
* This class provides static utility methods for dealing with TAR Files.
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @author Brent Boyer
* @see <a href="http://en.wikipedia.org/wiki/Tar_%28file_format%29">Wikipedia TAR file format</a>
* @see <a href="http://people.freebsd.org/~kientzle/libarchive/tar.5.txt">BSD TAR file format</a>
* @see <a href="http://www.mkssoftware.com/docs/man4/tar.4.asp">TAR file format</a>
* @see <a href="http://forum.java.sun.com/thread.jspa?threadID=757876">Forum posting on TAR and Java</a>
* @see <a href="http://jakarta.apache.org/commons/sandbox/compress/">jakarta compression library</a>
* @see <a href="http://wiki.apache.org/jakarta-commons/Compress">how to use the jakarta compression library</a>
* @see <a href="http://svn.apache.org/viewvc/jakarta/commons/sandbox/compress/trunk/src/examples/org/apache/commons/compress/examples/TarExample.java?view=markup">how to do TAR compression library</a>
*/
public class TarUtil {
	
	
	// -------------------- switch constants --------------------
	
	
	public static final String pathToArchive_key = "-pathToArchive";
	public static final String tarFile_key = "-tarFile";
	public static final String appendBackup_key = "-appendBackup";
	public static final String appendTimeStamp_key = "-appendTimeStamp";
	public static final String appendExtension_key = "-appendExtension";
	public static final String appendAll_key = "-appendAll";
	public static final String filter_key = "-filter";
	public static final String directoryExtraction_key = "-directoryExtraction";
	
	/** Specifies all the switch keys which, in theory, can legally appear as command line arguments to {@link #main main}. */
	private static final Set<String> TarUtil_keys = new HashSet<String>( Arrays.asList(
		pathToArchive_key,
		tarFile_key,
		appendBackup_key, appendTimeStamp_key, appendExtension_key, appendAll_key,
		filter_key,
		directoryExtraction_key
	) );
	
	
	// -------------------- misc constants --------------------
	
	
	/**
	* Records the maximum number of chars that can appear in a path...
	*/
//	private static final long pathCharsMax = 255;
/*
Commented out because this class is configured to always use GNU tar extensions which support unlimited length paths:
	Early versions of tar did not support path lengths greater than 100 characters. Modern versions of tar do so, but in incompatible ways... a GNU tar file which can have arbitrary length paths. Note however, that the resulting archive will only be able to be untarred with GNU tar.
	http://ant.apache.org/manual/CoreTasks/tar.html
	
	GNU tar as no file size limit. Even path names have no length limit!
	http://www.codecomments.com/archive309-2005-11-691598.html

See also
	Torture-testing Backup and Archive Programs
	http://www.coredumps.de/doc/dump/zwicky/testdump.doc.html
*/

	
	
	/**
	* Records the maximum file length supported by the classic TAR format.
	* The value is 8 GB = (2^33) - 1 = 8,589,934,591 bytes:
	* <blockquote>
	*	For historical reasons numerical values are encoded in octal with leading zeroes.
	*	The final character is either a null or a space.
	*	Thus although there are 12 bytes reserved for storing the file size, only 11 octal digits can be stored.
	*	This gives a maximum file size of 8 gigabytes on archived files.
	*	To overcome this limitation some versions of tar, including the GNU implementation,
	*	support an extension in which the file size is encoded in binary.<br>
	*	<a href="http://en.wikipedia.org/wiki/Tar_%28file_format%29">Tar (file format)</a>
	* </blockquote>
	*/
	private static final long fileLengthLimit = (8L * 1024L * 1024L * 1024L) - 1L;
// +++ WARNING: Commons Compress apache tar code that I am currently using actually seems to be broken for sizes above 2 GB due to some internal uses of int, see: https://issues.apache.org/jira/browse/SANDBOX-170
	
	
	private static final boolean giveConsoleFeedback = true;
	
	
	// -------------------- main --------------------
	
	
	/**
	* Either backs up a path (either a normal file or directory)
	* to a TAR archive file using a call to the {@link #archive(File, File, FileFilter) archive} method,
	* or extracts a TAR archive into a directory.
	* <p>
	* If archiving, the path to be archived and the TAR archive file are always specified as the command line switches
	* <code>-pathToArchive <i>insertPathHere</i></code> and <code>-tarFile <i>insertPathHere</i></code> respectively.
	* <p>
	* The following optional switches may also be supplied:
	* <ol>
	*  <li><code>-appendBackup</code> appends <code>_backup</code> to the TAR file path</li>
	*  <li><code>-appendTimeStamp</code> appends <code>_</code> followed by a timestamp to the TAR file path</li>
	*  <li><code>-appendExtension</code> appends <code>.tar</code> to the TAR file path</li>
	*  <li><code>-appendAll</code> is equivalent to supplying all the above append options</li>
	*  <li>
	*		<code>-filter <i>fullyQualifiedClassName</i></code>
	*		specifies the name of a FileFilter which limits what gets archived.
	*		Since this FileFilter class will be instantiated by a call to Class.forName,
	*		it must have a no-arg constructor.
	*  </li>
	* </ol>
	* <p>
	* If extracting, the archive's path and the directory to extract into are always specified as the command line switches
	* <code>-tarFile <i>insertPathHere</i></code> and <code>-directoryExtraction <i>insertPathHere</i></code> respectively.
	* <p>
	* Optional GZIP compression/decompression may also be done when archiving/extracting a tar file.
	* Normally, the value for the <code>-tarFile</code> switch must be a path
	* which ends in a ".tar" (case insensitive) extension.
	* However, this program will also accept either ".tar.gz" or ".tgz" extensions,
	* in which case it will perform GZIP compression/decompression on the tar file.
	* <p>
	* @throws Exception (or some subclass) if any problem occurs (e.g. see {@link #archive(File[], File, FileFilter)})
	*/
	public static void main(String[] args) throws Exception {
		try {
			Properties2 switches = CommandLineUtil.parseArgs(args);
			CommandLineUtil.checkSwitchKeys(switches, TarUtil_keys, "for " + TarUtil.class.getName());
			
			System.out.println();
			System.out.println("TarUtil start: " + DateUtil.getTimeStamp());
			
			if (switches.containsKey(pathToArchive_key)) {
if (switches.containsKey(directoryExtraction_key)) throw new IllegalArgumentException("args parsed into a set of switches that contains both " + pathToArchive_key + " and " + directoryExtraction_key);
// +++ really need to modify Properties2 to enforce proper subsets of keys which can be specified together...
				
				File pathToArchive = switches.getMandatoryFile(pathToArchive_key).getCanonicalFile();
				File tarFile = getTarFile(switches);
				FileFilter filter = getFileFilter(switches);
				
				archive( pathToArchive, tarFile, filter );
			}
			else if (switches.containsKey(directoryExtraction_key)) {
if (switches.containsKey(pathToArchive_key)) throw new IllegalArgumentException("args parsed into a set of switches that contains both " + directoryExtraction_key + " and " + pathToArchive_key);
if (switches.containsKey(appendBackup_key)) throw new IllegalArgumentException("args parsed into a set of switches that contains both " + directoryExtraction_key + " and " + appendBackup_key);
if (switches.containsKey(appendTimeStamp_key)) throw new IllegalArgumentException("args parsed into a set of switches that contains both " + directoryExtraction_key + " and " + appendTimeStamp_key);
if (switches.containsKey(appendExtension_key)) throw new IllegalArgumentException("args parsed into a set of switches that contains both " + directoryExtraction_key + " and " + appendExtension_key);
if (switches.containsKey(appendAll_key)) throw new IllegalArgumentException("args parsed into a set of switches that contains both " + directoryExtraction_key + " and " + appendAll_key);
if (switches.containsKey(filter_key)) throw new IllegalArgumentException("args parsed into a set of switches that contains both " + directoryExtraction_key + " and " + filter_key);
				
				File directoryExtraction = switches.getMandatoryFile(directoryExtraction_key).getCanonicalFile();
				File tarFile = getTarFile(switches);
				
				extract( tarFile, directoryExtraction );
			}
		}
		finally {
			System.out.println("TarUtil end: " + DateUtil.getTimeStamp());
			Toolkit.getDefaultToolkit().beep();
		}
	}
	
	
	private static File getTarFile(Properties2 switches) throws Exception {
		String path = switches.getMandatoryProperty(tarFile_key);
		if (switches.containsKey(appendBackup_key) || switches.containsKey(appendAll_key))
			path += "_backup";
		if (switches.containsKey(appendTimeStamp_key) || switches.containsKey(appendAll_key))
			path += "_" + DateUtil.getTimeStampForFile();
		if (switches.containsKey(appendExtension_key) || switches.containsKey(appendAll_key))
			path += ".tar";
		
		return new File(path).getCanonicalFile();
	}
	
	
	private static FileFilter getFileFilter(Properties switches) throws Exception {
		if (switches.containsKey(filter_key))
			return (FileFilter) Class.forName(switches.getProperty(filter_key)).newInstance();
		else
			return null;
	}
	
	
	// -------------------- isTarable --------------------
	
	
	/**
	* If file is actually a directory, then returns true.
	* <p>
	* Else if file is a normal file, then returns true if file's length (in bytes) is <= {@link #fileLengthLimit},
	* false otherwise.
	* <p>
	* Else returns false.
	* <p>
	* @throws IllegalArgumentException if file == null or file does not exist
	* @throws SecurityException if a security manager exists and its SecurityManager.checkRead method denies read access to file
	*/
	public static boolean isTarable(File file) throws IllegalArgumentException, SecurityException {
		if (file == null) throw new IllegalArgumentException("file == null");
		if (!file.exists()) throw new IllegalArgumentException("file = " + file.getPath() + " does not exist");
		
		if (file.isDirectory()) {
			return true;
		}
		else if (file.isFile()) {
			return (file.length() <= fileLengthLimit);
		}
		else return false;
	}
	
	
	// -------------------- archive and helper methods --------------------
	
	
	/** Simply calls {@link #archive(File[], File, FileFilter) archive(new File[] {pathToArchive}, tarFile, filter)}. */
	public static void archive(File pathToArchive, File tarFile, FileFilter filter) throws Exception {
		archive( new File[] {pathToArchive}, tarFile, filter );
	}
	
	
	/**
	* Writes each element of files to a new TAR format archive file specified by tarFile.
	* If any element is a directory, the entire contents of the directory tree will be archived.
	* An optional FileFilter can be supplied to screen out paths that would otherwise be archived.
	* <p>
	* In general, the path stored in the archive
	* is the path relative to the <i>parent</i> of the relevant element of files.
	* For example, let <code>file[i] = new File("D:/someDirectory/anotherDirectory")</code>
	* so <code>D:/someDirectory</code> is the parent of <code>file[i]</code>.
	* Then the path of <code>file[i]</code> relative to its parent is simply <code>anotherDirectory</code>,
	* which will be the path used for <code>file[i]</code> in the archive.
	* Furthermore, if <code>file[i]</code> contains the child path
	* <code>D:/someDirectory/anotherDirectory/aChild</code>,
	* then that child will have the path <code>anotherDirectory/aChild</code> in the archive.
	* <p>
	* One complication with the above scheme is paths which are file system roots: they have no parents.
	* Examples include the windows path <code>C:</code> or the unix path <code>/</code>.
	* In cases like these, this method uses an imaginary parent name
	* of the form <code>rootXXX</code> (where XXX is an integer).
	* For example, on a windows machine, if files contains the paths <code>C:</code> and <code>D:</code>,
	* then the contents of <code>C:</code> might be stored in the archive
	* with a path that starts with <code>root1</code>, and the contents of <code>D:</code>
	* may have an archive path that starts with <code>root2</code>.
	* This behavior ensures that the archive preserves the separate origins of the 2 sources.
	* <p>
	* The TAR archive witten by this method will use GNU TAR rules for the entry headers
	* if long path names are encountered.
	* <i>This means that standard POSIX compliant programs that do not support the GNU TAR extension
	* will be unable to extract the contents.</i>
	* <p>
	* Optional GZIP compression may also be done.
	* Normally, tarFile must be a path which ends in a ".tar" (case insensitive) extension.
	* However, this method will also accept either ".tar.gz" or ".tgz" extensions,
	* in which case it will perform GZIP compression on tarFile as part of archiving.
	* <p>
	* @param files the File[] to archive
	* @param tarFile the TAR File that will write the archive data to
	* @param filter a FileFilter that can use to screen out certain files from being written to the archive;
	* may be null (so everything specified by the files arg gets archived)
	* @throws Exception if any Throwable is caught; the Throwable is stored as the cause, and the message stores the path of tarFile;
	* here are some of the possible causes:
	* <ol>
	*  <li>
	*		IllegalArgumentException if files == null; files.length == 0;
	*		tarFile == null;
	*		tarFile already exists and either is not a normal file or is but already has data inside it;
	*		tarFile has an invalid extension;
	*		any element of files is null, does not exist, cannot be read, is equal to tarFile, its path contains tarFile,
	*		or it fails {@link #isTarable isTarable}
	*  </li>
	*  <li>SecurityException if a security manager exists and its SecurityManager.checkRead method denies read access to file</li>
	*  <li>IOException if an I/O problem occurs</li>
	* </ol>
	*/
	public static void archive(File[] files, File tarFile, FileFilter filter) throws Exception {
// +++ should use varargs...
		try {
			if (files == null) throw new IllegalArgumentException("files == null");
			if (files.length == 0) throw new IllegalArgumentException("files.length == 0");
			if (tarFile == null) throw new IllegalArgumentException("tarFile == null");
			if (tarFile.exists()) {
				if (!tarFile.isFile()) throw new IllegalArgumentException("tarFile = " + tarFile.getPath() + " exists but is not a normal file");
				if (tarFile.length() != 0) throw new IllegalArgumentException("tarFile = " + tarFile.getPath() + " already exists and already has data inside it; this method will not overwrite it");
			}
			for (int i = 0; i < files.length; i++) {
				if ( files[i] == null ) throw new IllegalArgumentException("files[" + i + "] == null");
				if ( !files[i].exists() ) throw new IllegalArgumentException("files[" + i + "] = " + files[i].getPath() + " is a non-existent path");
				if ( !files[i].canRead() ) throw new IllegalArgumentException("files[" + i + "] = " + files[i].getPath() + " cannot be read by this application");
				if ( files[i].equals(tarFile) ) throw new IllegalArgumentException("files[" + i + "] = " + files[i].getPath() + " is the same File as arg tarFile = " + tarFile.getPath());
				if ( files[i].isDirectory() && DirUtil.contains(files[i], tarFile) ) throw new IllegalArgumentException("the directory corresponding to arg files[" + i + "] = " + files[i].getCanonicalPath() + " will contain the path corresponding to arg tarFile = " + tarFile.getCanonicalPath() );
			}
			
			TarOutputStream tos = null;
			try {
				tos = new TarOutputStream( getOutputStream(tarFile) );
				tos.setLongFileMode(TarOutputStream.LONGFILE_GNU);
				for (File file : files) {
					archive( file, new FileParent(file), tos, filter );
				}
			}
			finally {
				if (giveConsoleFeedback) ConsoleUtil.eraseLine();
				StreamUtil.close(tos);
			}
		}
		catch (Throwable t) {
			throw new Exception("See cause for the underlying Throwable; happened for tarFile = " + (tarFile != null ? tarFile.getPath() : "<null>"), t);
		}
	}
	
	
	/**
	* If file's extension is simply tar, then returns a new FileOutputStream.
	* Else if file's extension is tar.gz or tgz,
	* then returns a new GZIPOutputStream wrapping a new FileOutputStream.
	* <p>
	* Note: the result is never buffered, since the TarOutputStream which will use the result
	* always has an internal buffer.
	* <p>
	* @throws IllegalArgumentException if encounter an extension that are unable to handle
	* @throws IOException if an I/O problem occurs
	* @see <a href="http://www.gzip.org/">gzip home page</a>
	* @see <a href="http://www.brouhaha.com/~eric/tgz.html">.tar.gz file format FAQ</a>
	*/
	private static OutputStream getOutputStream(File tarFile) throws IllegalArgumentException, IOException {
		String extension = FileUtil.getExtension(tarFile);
		if (extension.equalsIgnoreCase("tar")) {
			return new FileOutputStream(tarFile);
		}
		else if (extension.equalsIgnoreCase("gz")) {
			String name = FileUtil.getNameMinusExtension(tarFile);
			
			int i = name.lastIndexOf('.');
			if (i == -1) throw new IllegalArgumentException("tarFile = " + tarFile.getPath() + " has a gz final extension but, its complete extension is not tar.gz");
			String extensionInitial = name.substring(i + 1);
			
			if (extensionInitial.equalsIgnoreCase("tar")) {
				return new GZIPOutputStream( new FileOutputStream(tarFile) );
			}
			else {
				throw new IllegalArgumentException("tarFile = " + tarFile.getPath() + " has a gz final extension but, its complete extension is not tar.gz");
			}
		}
		else if (extension.equalsIgnoreCase("tgz")) {
			return new GZIPOutputStream( new FileOutputStream(tarFile) );
		}
		else {
			throw new IllegalArgumentException("tarFile = " + tarFile.getPath() + " has an invalid extension for a tar or tar/gzip file");
		}
	}
	
	
	/**
	* This method writes file as a new TarEntry to the TarOutputStream,
	* followed by file's data (which is 0 bytes if it is a directory).
	* <p>
	* If the File is a directory, then this method additionally calls itself on the contents
	* (thus recursing thru the entire directory tree).
	* <p>
	* <b>Warning</b>: several popular programs (e.g. winzip) fail to display mere directory entries.
	* Furthermore, if just a directory entry is present (i.e. it is empty), they also may fail
	* to create a new empty directoy when extracting the TAR file's contents.
	* These are bugs in their behavior.
	* <p>
	* An optional FileFilter can be supplied to screen out paths that would otherwise be archived.
	* <p>
	* <i>This method does not close the supplied TarOutputStream:</i> that is the responsibility of the caller.
	* <p>
	* The caller also must take on the responsibility to not do anything stupid, like write
	* file more than once, or have the file be the same File that out is writing to.
	* <p>
	* @param file the File to archive
	* @param fileParent the FileParent for file
	* @param out the TarOutputStream that will write the archive data to
	* @param filter a FileFilter that can use to screen out certain files from being written to the archive; may be null (so everything specified by the file arg gets archived)
	* @throws Exception if any Throwable is caught; the Throwable is stored as the cause, and the message stores the path of file;
	* here are some of the possible causes:
	* <ol>
	*  <li>IllegalArgumentException if file fails {@link #isTarable isTarable}</li>
	*  <li>SecurityException if a security manager exists and its SecurityManager.checkRead method denies read access to file</li>
	*  <li>IOException if an I/O problem occurs</li>
	* </ol>
	*/
	private static void archive(File file, FileParent fileParent, TarOutputStream out, FileFilter filter) throws Exception {
		try {
			if ( (filter != null) && !filter.accept(file) ) return;
			if (!isTarable(file)) throw new IllegalArgumentException("file = " + file.getPath() + " failed the isTarable method");
			if (giveConsoleFeedback) ConsoleUtil.overwriteLine("TarUtil.archive: " + file.getPath());
			
			TarEntry entry = new TarEntry( file );
			entry.setName( fileParent.getRelativePath(file, '/') );	// Note: getRelativePath will ensure that '/' is the path separator and that directories end with '/'
			out.putNextEntry( entry );
			writeFileData(file, out);
			out.closeEntry();
			
			if ( file.isDirectory() ) {
				for (File fileChild : DirUtil.getContents(file, null)) {	// supply null, since we test at beginning of this method (supplying filter here which just add a redundant test)
					archive( fileChild, fileParent, out, filter );
				}
			}
		}
		catch (Throwable t) {
			String preface = "See cause for the underlying Throwable; happened for file = ";
			if ( (t instanceof Exception) && (t.getMessage().startsWith(preface)) ) throw (Exception) t;	// CRITICAL: see if t was actually the wrapping Exception generated by a subsequent recursive call to this method, and if so simply rethrow it unmodified to avoid excessive exception chaining
			throw new Exception(preface + (file != null ? file.getPath() : "<null>"), t);
		}
	}
	
	
	/**
	* Returns immediately if file is not a normal file (e.g. is a directory).
	* Otherwise, reads all the bytes from file and writes them to out.
	* <p>
	* @throws IOException if an I/O problem occurs
	*/
	private static void writeFileData(File file, OutputStream out) throws IOException {
		if (!file.isFile()) return;
		
		InputStream in = null;
		try {
			in = new FileInputStream(file);	// no buffering, because StreamUtil.transfer below will do that itself
			StreamUtil.transfer(in, out);
		}
		finally {
			StreamUtil.close(in);
		}
	}
	
	
	// -------------------- close --------------------
	
	
	/**
	* This method attempts to close tarArchive if it is not null.
	* <p>
	* Any Throwable that occurs is logged using {@link LogUtil#logRobustly LogUtil.logRobustly}.
	* Thus, <i>this method will never throw a Throwable</i>.
	*/
	public static void close(TarArchive tarArchive) {
		try {
			if (tarArchive != null) tarArchive.close();
		}
		catch (Throwable t) {
			LogUtil.logRobustly("TarUtil", "close", "a problem occurred while closing tarArchive", t);
		}
	}
	
	
	// -------------------- getEntries, extract --------------------
	
	
	/**
	* Returns all the {@link TarEntry}s that can next be read by tarInputStream.
	* <p>
	* Nothing should have been previously read from tarInputStream if the full result is desired,
	* and nothing more can be read from tarInputStream when this method returns.
	* This method will close tarInputStream as its final action.
	* <p>
	* @param tarInputStream the {@link TarInputStream} to get the entries from
	* @param sortResult if true, then the result TarEntry[] is first sorted by each entry's name before return;
	* otherwise the order is the sequence read from tarInputStream
	* @throws IllegalArgumentException if tarInputStream == null
	* @throws IOException if an I/O problem occurs
	*/
	public static TarEntry[] getEntries(TarInputStream tarInputStream, boolean sortResult) throws IllegalArgumentException, IOException {
		if (tarInputStream == null) throw new IllegalArgumentException("tarInputStream == null");
		
		try {
			List<TarEntry> entries = new LinkedList<TarEntry>();
			while (true) {
				TarEntry entry = tarInputStream.getNextEntry();
				if (entry == null) break;
				else entries.add(entry);
			}
			
			if (sortResult) {
				Collections.sort(entries, new Comparator<TarEntry>() {
					public int compare(TarEntry te1, TarEntry te2) throws IllegalArgumentException {
						if (te1 == null) throw new IllegalArgumentException("te1 == null");
						if (te2 == null) throw new IllegalArgumentException("te2 == null");
						
						return te1.getName().compareTo( te2.getName() );
					}
				} );
			}
			
			return entries.toArray( new TarEntry[entries.size()] );
		}
		finally {
			StreamUtil.close(tarInputStream);
		}
	}
// +++ note that TarArchive has a getEntryIterator method from its superclass; might that be quicker?
	
	
	/**
	* Extracts tarFile to directoryExtraction.
	* <p>
	* It is an error if tarFile does not exist, is not a normal file, or is not in the proper TAR format.
	* In contrast, directoryExtraction need not exist, since it (and any parent directories)
	* will be created if necessary
	* (but directoryExtraction must be a proper directory path in this case).
	* <p>
	* @param tarFile the TAR archive File
	* @param directoryExtraction the directory that will extract the contents of tarFile into
	* @throws IllegalArgumentException if tarFile fails {@link FileUtil#check FileUtil.check};
	* if directoryExtraction fails {@link DirUtil#ensureExists DirUtil.ensureExists}
	* @throws SecurityException if a security manager exists and its SecurityManager.checkRead method
	* denies read access to tarFile or directoryExtraction
	* @throws IllegalStateException if directoryExtraction failed to be created or is not an actual directory but is some other type of file
	* @throws ArchiveException if an archive problem occurs
	* @throws UnpackException if an unpcking problem occurs
	*/
/*
	public static void extract(File tarFile, File directoryExtraction) throws IllegalArgumentException, SecurityException, ArchiveException, UnpackException {
		FileUtil.check(tarFile);
		DirUtil.ensureExists(directoryExtraction);
		
		TarArchive archive = null;
		try {
			Archive archiver = ArchiverFactory.getInstance(tarFile);
			archiver.unpack(directoryExtraction);
		}
		finally {
			close(archive);
		}
	}
*/

/*
The simple code above OUGHT to work, but fails because Apache's code is broken; see:
	https://issues.apache.org/jira/browse/SANDBOX-168

Furthermore, am not sure if the above will handle gzip files like the code below does.

Note: to see all issues with Apache's compress package:
	https://issues.apache.org/jira/secure/IssueNavigator.jspa?reset=true&mode=hide&pid=12310491&sorter/order=DESC&sorter/field=priority&resolution=-1&component=12311183

To deal with this, we use lower level code; see
	http://forum.java.sun.com/thread.jspa?threadID=573800&messageID=3115774
*/

	/**
	* Extracts tarFile to directoryExtraction.
	* <p>
	* It is an error if tarFile does not exist, is not a normal file, or is not in the proper TAR format.
	* In contrast, directoryExtraction need not exist, since it (and any parent directories)
	* will be created if necessary
	* (but directoryExtraction must be a proper directory path in this case).
	* <p>
	* Optional GZIP decompression may also be done.
	* Normally, tarFile must be a path which ends in a ".tar" (case insensitive) extension.
	* However, this method will also accept either ".tar.gz" or ".tgz" extensions,
	* in which case it will perform GZIP decompression on tarFile as part of extracting.
	* <p>
	* @param tarFile the TAR archive File
	* @param directoryExtraction the directory that will extract the contents of tarFile into
	* @throws IllegalArgumentException if tarFile fails {@link FileUtil#check FileUtil.check};
	* if directoryExtraction fails {@link DirUtil#ensureExists DirUtil.ensureExists};
	* tarFile has an invalid extension
	* @throws SecurityException if a security manager exists and its SecurityManager.checkRead method
	* denies read access to tarFile or directoryExtraction
	* @throws IllegalStateException if directoryExtraction failed to be created or is not an actual directory but is some other type of file
	* @throws IOException if an I/O problem occurs
	*/
	public static void extract(File tarFile, File directoryExtraction) throws IllegalArgumentException, SecurityException, IOException {
		FileUtil.check(tarFile);
		DirUtil.ensureExists(directoryExtraction);
		
		String rootPath = directoryExtraction.getPath() + File.separatorChar;
		
		TarInputStream tis = null;
		try {
			tis = new TarInputStream( getInputStream(tarFile) );
			
			for (TarEntry tarEntry = tis.getNextEntry(); tarEntry != null; tarEntry = tis.getNextEntry()) {
				File path = new File(rootPath + tarEntry.getName() );
// +++ how should overwriting be handled?
// ZipUtil.extract takes an overwrite param that specifies whether to allow overwriting or to crash instead,
// but I am inclined to ssay that maybe always crashing is the best thing to do?
//if (!overwrite && file.exists()) throw new IllegalStateException("overwrite == false and " + file.getPath() + " already exists");
				if (tarEntry.isDirectory()) {
					DirUtil.ensureExists(path);
				}
				else {
					DirUtil.ensureExists( path.getParentFile() );	// CRITICAL: the TAR format does not necessarily store all the directories as entries, so must make sure that they are created
					
					OutputStream out = null;
					try {
						out = new FileOutputStream(path);
						tis.copyEntryContents(out);
					}
					finally {
						StreamUtil.close(out);
					}
				}
			}
		}
		finally {
			StreamUtil.close(tis);
		}
	}
	
	
	/**
	* If file's extension is simply tar, then returns a new FileInputStream.
	* Else if file's extension is tar.gz or tgz,
	* then returns a new GZIPInputStream wrapping a new FileInputStream.
	* <p>
	* Note: the result is never buffered, since the TarInputStream which will use the result
	* always has an internal buffer.
	* <p>
	* @throws IllegalArgumentException if encounter an extension that are unable to handle
	* @throws IOException if an I/O problem occurs
	* @see <a href="http://www.gzip.org/">gzip home page</a>
	* @see <a href="http://www.brouhaha.com/~eric/tgz.html">.tar.gz file format FAQ</a>
	*/
	private static InputStream getInputStream(File tarFile) throws IllegalArgumentException, IOException {
		String extension = FileUtil.getExtension(tarFile);
		if (extension.equalsIgnoreCase("tar")) {
			return new FileInputStream(tarFile);
		}
		else if (extension.equalsIgnoreCase("gz")) {
			String name = FileUtil.getNameMinusExtension(tarFile);
			
			int i = name.lastIndexOf('.');
			if (i == -1) throw new IllegalArgumentException("tarFile = " + tarFile.getPath() + " has a gz final extension but, its complete extension is not tar.gz");
			String extensionInitial = name.substring(i + 1);
			
			if (extensionInitial.equalsIgnoreCase("tar")) {
				return new GZIPInputStream( new FileInputStream(tarFile) );
			}
			else {
				throw new IllegalArgumentException("tarFile = " + tarFile.getPath() + " has a gz final extension but, its complete extension is not tar.gz");
			}
		}
		else if (extension.equalsIgnoreCase("tgz")) {
			return new GZIPInputStream( new FileInputStream(tarFile) );
		}
		else {
			throw new IllegalArgumentException("tarFile = " + tarFile.getPath() + " has an invalid extension for a tar or tar/gzip file");
		}
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private TarUtil() {}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		//private static final String root = "../log/";
		private static final String root = "C:/";
		
		private static int count = 0;
		
		private static int countNext() { return count++; }
		
		// -------------------- utility methods (public static, because are also used by ZipUtil.UnitTest) --------------------
		
		/**
		* Result is always a temporary file.
		* It contains the alphabet continuously repeated.
<!--		* It contains random bytes. -->
		* Its length is given by lengthFile.
		* Its name includes an absolute count number as well as its length in bytes, 
		* which makes it unique (for this JVM session).
		*/
		public static File makeDataFile(File parent, long lengthFile) throws IllegalArgumentException, IOException {
			if (lengthFile < 0) throw new IllegalArgumentException("lengthFile = " + lengthFile + " < 0");
			
			OutputStream out = null;
			try {
				DirUtil.ensureExists(parent);	// this is NOT done by FileUtil.createTemp below
				File file = FileUtil.createTemp( new File(parent, "dataFile#" + countNext() + "_length" + lengthFile + ".txt") );
				
				int bufferSize = (int) Math.min( lengthFile, 128 * 1024 * 1024 );
				byte[] bytes = new byte[bufferSize];
					// repeat the ascii values a thru z--do this if want to look at the contents by eye and verify:
				for (int i = 0; i < bytes.length; i++) {
					bytes[i] = (byte) (97 + (i % 26));
				}
					// produce random values in the array--do this if want uncompressable bytes:
				//Random random = new Random();
				//random.nextBytes(bytes);
				
				out = new FileOutputStream(file, false);
				for (long lengthWritten = 0; lengthWritten < lengthFile; ) {
					int lengthToWrite = (int) Math.min( lengthFile - lengthWritten, bytes.length );
					out.write(bytes, 0, lengthToWrite);
					lengthWritten += lengthToWrite;
				}
				
				return file;
			}
			finally {
				StreamUtil.close(out);
			}
		}
		
		/**
		* Result is always a temporary file whose name is unique (for this JVM session) that contains no data.
		* Its name includes an absolute count number, which makes it unique (for this JVM session).
		*/
		public static File makeArchiveFile(String type) {
			return FileUtil.createTemp( root + type + "File#" + countNext() + "." + type );
		}
		
		/**
		* Result is always a directory that does not currently exist (hence, contains no data).
		* Its name includes an absolute count number, which makes it unique (for this JVM session).
		*/
		public static File makeDirectoryExtraction(String type) {
			File directoryExtraction = new File( root + type + "Extract#" + countNext() );
			if (directoryExtraction.exists()) throw new IllegalStateException(directoryExtraction.getPath() + " already exists");
			return directoryExtraction;
		}
		
		// -------------------- JUnit methods --------------------
		
		@Test public void test_archive_extract() throws Exception {
			File directoryExtraction = null;
			try {
				File dirToArchive = new File( "../src" );
				File tarFile = makeArchiveFile("tar");
				archive( dirToArchive, tarFile, null );
				printEntries(tarFile);
				
				directoryExtraction = makeDirectoryExtraction("tar");
				extract( tarFile, directoryExtraction );
				
				File directoryOneLevelDown = directoryExtraction.listFiles()[0];	// need to drill down one level in order to get the proper directory in which to compare with dirToArchive
				Assert.assertNull( DirUtil.compareContents(dirToArchive, directoryOneLevelDown) );	// DirUtil.compareContents ONLY returns null if 2 directories have exactly equally contents in all their elements (both normal files and subdirectories)
			}
			finally {
				//FileUtil.deleteIfExists(...);	// no need to worry about the files: they are all temporary
				DirUtil.deleteIfExists(directoryExtraction);	// unfortunately, cannot make temp directories--deleteOnExit fails if the directory has content--so must explicitly clean it up
			}
		}
		
		@Test public void test_archive_extract_gzip() throws Exception {
			File directoryExtraction = null;
			try {
				File dirToArchive = new File( "../class" );
				File tarFile = makeArchiveFile("tgz");
				archive( dirToArchive, tarFile, null );
				printEntries(tarFile);
				
				directoryExtraction = makeDirectoryExtraction("tar");
				extract(tarFile, directoryExtraction);
				
				File directoryOneLevelDown = directoryExtraction.listFiles()[0];	// need to drill down one level in order to get the proper directory in which to compare with dirToArchive
				Assert.assertNull( DirUtil.compareContents(dirToArchive, directoryOneLevelDown) );	// DirUtil.compareContents ONLY returns null if 2 directories have exactly equally contents in all their elements (both normal files and subdirectories)
			}
			finally {
				//FileUtil.deleteIfExists(...);	// no need to worry about the files: they are all temporary
				DirUtil.deleteIfExists(directoryExtraction);
			}
		}
		
@Ignore("Not running because this method does no cleanup since it requires manual inspection of its output tar files")
		@Test public void test_tarFormat() throws Exception {
			File directoryExtraction = null;
			try {
				File dirToArchive = new File( "../doc" );
				//File tarFile = makeArchiveFile("tar");
File tarFile = new File( root + "tarFile#" + countNext() + ".tar" );
				archive( dirToArchive, tarFile, null );
				
System.out.println("created the PERSISTENT test tar file " + tarFile.getPath() + "; open with some OTHER program like winzip to confirm that the format is correct");
// +++ hmm, is there a way that I could programmatically call something like jzip from java and automatically confirm compatibility?

// +++ i also need to go the other way: use external program to tar, and then untar here
				//directoryExtraction = makeDirectoryExtraction("tar");
				//extract(tarFile, directoryExtraction);
				
				//File directoryOneLevelDown = directoryExtraction.listFiles()[0];	// need to drill down one level in order to get the proper directory in which to compare with dirToArchive
				//Assert.assertNull( DirUtil.compareContents(dirToArchive, directoryOneLevelDown) );	// DirUtil.compareContents ONLY returns null if 2 directories have exactly equally contents in all their elements (both normal files and subdirectories)
			}
			finally {
				//FileUtil.deleteIfExists(...);	// no need to worry about the files: they are all temporary
				DirUtil.deleteIfExists(directoryExtraction);
			}
		}
		
		

@Ignore("Not running because this method fails due to a BUG in the TAR library that I am using; plus it takes too long...")
// +++ when last ran this method on 2008/6/18, it failed when tried to tar the large files:
//Caused by: java.io.IOException: request to write '32768' bytes exceeds size in header of '-1' bytes
//	at org.apache.commons.compress.archivers.tar.TarOutputStream.write(TarOutputStream.java:341)
		@Test public void test_archive_extract_fileLengthLimit_shouldPass() throws Exception {
			File directoryExtraction = null;
			File dirToArchive = null;
			try {
				dirToArchive = new File(root + "/test_archive_extract_fileLengthLimit_shouldPass");
				File fileHalf = makeDataFile(dirToArchive, fileLengthLimit / 2L);
				File fileMax = makeDataFile(dirToArchive, fileLengthLimit);
				File tarFile = makeArchiveFile("tar");
				archive( dirToArchive, tarFile, null );
				
				directoryExtraction = makeDirectoryExtraction("tar");
				extract( tarFile, directoryExtraction );
				
				File directoryOneLevelDown = directoryExtraction.listFiles()[0];	// need to drill down one level in order to get the proper directory in which to compare with dirToArchive
				Assert.assertNull( DirUtil.compareContents(dirToArchive, directoryOneLevelDown) );	// DirUtil.compareContents ONLY returns null if 2 directories have exactly equally contents in all their elements (both normal files and subdirectories)
			}
			finally {
				//FileUtil.deleteIfExists(...);	// no need: all the files are temporary
				DirUtil.deleteIfExists(dirToArchive);	// CRITICAL: do this here because for this method, dirToArchive is a temp directory
				DirUtil.deleteIfExists(directoryExtraction);
			}
		}
		
@Ignore("Not running because this method fails due to a BUG in the TAR library that I am using; plus it takes too long...")
// same bug as above...
		@Test(expected=IllegalArgumentException.class) public void test_archive_extract_fileLengthLimit_shouldFail() throws Exception {
			File dirToArchive = null;
			try {
				dirToArchive = new File(root + "/test_archive_extract_fileLengthLimit_shouldFail");
				File fileTooLarge = makeDataFile(dirToArchive, fileLengthLimit + 1L);
				File tarFile = makeArchiveFile("tar");
				archive( fileTooLarge, tarFile, null );
			}
			finally {
				//FileUtil.deleteIfExists(...);	// no need: all the files are temporary
				DirUtil.deleteIfExists(dirToArchive);	// CRITICAL: do this here because for this method, dirToArchive is a temp directory
			}
		}
		
		@Test public void test_archive_extract_pathLengthLimit() throws Exception {
			File dirToArchive = null;
			File directoryExtraction = null;
			try {
				dirToArchive = new File(root + "/test_archive_extract_pathLengthLimit");
				File subdirLongName = new File( StringUtil.toLength(dirToArchive.getPath() + "/X", 200, false, 'X') );
				File fileData = makeDataFile(subdirLongName, 2048);
				File tarFile = makeArchiveFile("tar");
				archive( dirToArchive, tarFile, null );
				
				directoryExtraction = makeDirectoryExtraction("tar");
				extract( tarFile, directoryExtraction );
				
				File directoryOneLevelDown = directoryExtraction.listFiles()[0];	// need to drill down one level in order to get the proper directory in which to compare with dirToArchive
				Assert.assertNull( DirUtil.compareContents(dirToArchive, directoryOneLevelDown) );	// DirUtil.compareContents ONLY returns null if 2 directories have exactly equally contents in all their elements (both normal files and subdirectories)
			}
			finally {
				//FileUtil.deleteIfExists(...);	// no need: all the files are temporary
				DirUtil.deleteIfExists(dirToArchive);	// CRITICAL: do this here because for this method, dirToArchive is a temp directory
				DirUtil.deleteIfExists(directoryExtraction);
			}
		}
		
		private void printEntries(File tarFile) throws Exception {
			System.out.println( "Below are the entries of the TAR file " + tarFile.getPath() + ":" );
			TarInputStream tarInputStream = null;
			try {
				tarInputStream = new TarInputStream( getInputStream(tarFile) );
				for (TarEntry entry : getEntries(tarInputStream, false)) {
					System.out.println( '\t' + entry.getName() );			
				}
			}
			finally {
				StreamUtil.close(tarInputStream);
			}
		}
		
	}
	
	
}


/*

If look at this webpage, they describe a really simple way to archive/extract using their generic Archiver functionality:
	http://wiki.apache.org/jakarta-commons/Compress

I use the above technique in the current extract method, but it is NOT used in archive because:
	1) it is not obvious what custom TAR settings are used or how they can be specified
	2) I need to guarantee that archive will use GNU TAR extensions in the event of long paths
	3) I want control over how paths are stored in the archive relative to the root dierectory being archived
Because of these issues, this class skips the simple Archiver programming
and instead uses the TAR specific and low level classes like the TarInputStream and TarEntry for archiving.

But below is the code that COULD use if wanted to archive using their Archiver functionality:

	public static void archive(File pathToArchive, File tarFile, FileFilter filter) throws IllegalArgumentException, SecurityException, IOException, ArchiveException {
		archive( new File[] {pathToArchive}, tarFile, filter );
	}
	
	
	public static void archive(File[] files, File tarFile, FileFilter filter) throws IllegalArgumentException, SecurityException, IOException, ArchiveException {
		if (files == null) throw new IllegalArgumentException("files is null");
		if (files.length == 0) throw new IllegalArgumentException("files.length == 0");
		if (tarFile == null) throw new IllegalArgumentException("tarFile is null");
		if (tarFile.exists()) {
			if (!tarFile.isFile()) throw new IllegalArgumentException("tarFile = " + tarFile.getPath() + " exists but is not a normal file");
			if (tarFile.length() != 0) throw new IllegalArgumentException("tarFile = " + tarFile.getPath() + " already exists and already has data inside it; this method will not overwrite it");
		}
		for (int i = 0; i < files.length; i++) {
			if ( files[i] == null ) throw new IllegalArgumentException("files[" + i + "] == null");
			if ( !files[i].exists() ) throw new IllegalArgumentException("files[" + i + "] = " + files[i].getPath() + " is a non-existent path");
			if ( !files[i].canRead() ) throw new IllegalArgumentException("files[" + i + "] = " + files[i].getPath() + " cannot be read by this application");
			if ( files[i].equals(tarFile) ) throw new IllegalArgumentException("files[" + i + "] = " + files[i].getPath() + " is the same File as arg tarFile = " + tarFile.getPath());
			if ( files[i].isDirectory() && DirUtil.contains(files[i], tarFile) ) throw new IllegalArgumentException("the directory corresponding to arg files[" + i + "] = " + files[i].getCanonicalPath() + " will contain the path corresponding to arg tarFile = " + tarFile.getCanonicalPath() );
		}
		
		try {
			Archive archiver = ArchiverFactory.getInstance("tar");
			for (File file : files) {
				archive(file, archiver, filter);
			}
			archiver.save(tarFile);
		}
		finally {
			if (giveConsoleFeedback) ConsoleUtil.eraseLine();
		}
	}
	
	
	private static void archive(File file, Archive archiver, FileFilter filter) throws IllegalArgumentException, SecurityException, IOException {
		if ( (filter != null) && !filter.accept(file) ) return;
		if (!isTarable(file)) throw new IllegalArgumentException("file = " + file.getPath() + " failed the isTarable method");
		if (giveConsoleFeedback) ConsoleUtil.overwriteLine("TarUtil.archive: " + file.getPath());
		
		archiver.add( file );
		
		if ( file.isDirectory() ) {
			for (File fileChild : DirUtil.getContents(file, null)) {	// supply null, since we test at beginning of this method (supplying filter here which just add a redundant test)
				archive( fileChild, archiver, filter );
			}
		}
	}

*/