/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*

Programmer notes:

+++ this class mainly currently only archives; could expand it to do extractions from archives as well
	--exception is the getEntries method, which can be used for zip reading purposes

*/


package bb.io;


import java.awt.Toolkit;
import java.io.*;
import java.util.*;
import java.util.logging.Level;
import java.util.zip.*;

import org.junit.Assert;
import org.junit.Ignore;
import org.junit.Test;

import bb.util.CommandLineUtil;
import bb.util.DateUtil;
import bb.util.Properties2;
import bb.util.StringUtil;
import bb.util.logging.LogUtil;


/**
* This class provides static utility methods for dealing with ZIP Files.
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @see <a href="http://www.pkware.com/support/appnote.html">The ZIP file specification</a>
* @see <a href="http://www.ddj.com/articles/1997/9712/9712e/9712e.htm?topic=compression">DDJ article on Java and ZIP</a>
* @see <a href="http://dogma.net/markn/articles/JavaZip/ZipCreate.java">The ZipCreate program</a>
* @see <a href="http://www.javaworld.com/javaworld/javatips/jw-javatip120.html">Self-extracting JAR files</a>
* @see "The O'Reilly Java I/O book"
* @author Brent Boyer
*/
public class ZipUtil {
	
	
	// -------------------- switch constants --------------------
	
	
	public static final String pathToArchive_key = "-pathToArchive";
	public static final String zipFile_key = "-zipFile";
	public static final String appendBackup_key = "-appendBackup";
	public static final String appendTimeStamp_key = "-appendTimeStamp";
	public static final String appendExtension_key = "-appendExtension";
	public static final String appendAll_key = "-appendAll";
	public static final String filter_key = "-filter";
	
	/** Specifies all the switch keys which, in theory, can legally appear as command line arguments to {@link #main main}. */
	private static final Set<String> ZipUtil_keys = new HashSet<String>( Arrays.asList(
		pathToArchive_key,
		zipFile_key,
		appendBackup_key, appendTimeStamp_key, appendExtension_key, appendAll_key,
		filter_key
	) );
	
	
	// -------------------- misc constants --------------------
	
	
	/** Is 4 * 2^30 = 2^32 = . */
	private static final long fourGibi = 4L * 1024L * 1024L * 1024L;
	
	
	/**
	* Records the maximum length of a file that can be zip archived by Java.
	* This value is valid as of 2005/1/16 under JDK 1.5.0.
	* <p>
	* @see <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4681995">bug report #1</a>
	* @see <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4418997">bug report #2</a>
	* @see "The file D:\software\java\proposalsQuestionsPostingsEtc\zipFileSupportInJava.txt"
	*/
	private static final long zippableFileLengthLimit = fourGibi - 1L;
	
	
	/**
	* Records the maximum length of any ZIP file that can be read by Java.
	* This value is valid as of 2005/1/16 under JDK 1.5.0.
	* <p>
	* @see <a href="http://www.backupforall.com/slicing.php">Discussion #1 of the official ZIP standard</a>
	* @see <a href="http://www.miclasificado.com.ar/cgi/News/i14/Zip_files_in_windows_xp.php">Discussion #2 of the official ZIP standard</a>
	*/
	private static final long zipFileLengthLimit = fourGibi - 1L;
	
	
	private static final boolean giveConsoleFeedback = true;
	
	
	// -------------------- main --------------------
	
	
// +++ main currently only archives; could expand it to do extractions from archives as well
//See the code in TarUtil for how to do this.
// Have now added the extract method to this class, so it should not be too much work...
	/**
	* Backs up a path (either a normal file or directory)
	* to a ZIP archive file using a call to the {@link #archive(File, File, FileFilter) archive} method.
	* <p>
	* The path to be archived and the ZIP archive file are specified as the command line switches
	* <code>-pathToArchive <i>insertPathHere</i></code> and <code>-zipFile <i>insertPathHere</i></code> respectively.
	* <p>
	* The following optional switches may also be supplied:
	* <ol>
	*  <li><code>-appendBackup</code> appends <code>_backup</code> to the ZIP file path</li>
	*  <li><code>-appendTimeStamp</code> appends <code>_</code> followed by a timestamp to the ZIP file path</li>
	*  <li><code>-appendExtension</code> appends <code>.zip</code> to the ZIP file path</li>
	*  <li><code>-appendAll</code> is equivalent to supplying all the above append options</li>
	*  <li>
	*		<code>-filter <i>fullyQualifiedClassName</i></code>
	*		specifies the name of a FileFilter which limits what gets archived.
	*		Since this FileFilter class will be instantiated by a call to Class.forName,
	*		it must have a no-arg constructor.
	*  </li>
	* </ol>
	* <p>
	* @throws Exception (or some subclass) if any problem occurs (e.g. see {@link #archive(File[], File, FileFilter)})
	*/
	public static void main(String[] args) throws Exception {
		try {
			Properties2 switches = CommandLineUtil.parseArgs(args);
			CommandLineUtil.checkSwitchKeys(switches, ZipUtil_keys, "for " + ZipUtil.class.getName());
			
			File pathToArchive = switches.getMandatoryFile(pathToArchive_key).getCanonicalFile();
			File zipFile = getZipFile(switches);
			FileFilter filter = getFileFilter(switches);
			
			System.out.println();
			System.out.println("ZipUtil start: " + DateUtil.getTimeStamp());
			archive( pathToArchive, zipFile, filter );
		}
		finally {
			System.out.println("ZipUtil end: " + DateUtil.getTimeStamp());
			Toolkit.getDefaultToolkit().beep();
		}
	}
	
	
	private static File getZipFile(Properties2 switches) throws Exception {
		String path = switches.getMandatoryProperty(zipFile_key);
		if (switches.containsKey(appendBackup_key) || switches.containsKey(appendAll_key))
			path += "_backup";
		if (switches.containsKey(appendTimeStamp_key) || switches.containsKey(appendAll_key))
			path += "_" + DateUtil.getTimeStampForFile();
		if (switches.containsKey(appendExtension_key) || switches.containsKey(appendAll_key))
			path += ".zip";
		
		return new File(path).getCanonicalFile();
	}
	
	
	private static FileFilter getFileFilter(Properties switches) throws Exception {
		if (switches.containsKey(filter_key))
			return (FileFilter) Class.forName(switches.getProperty(filter_key)).newInstance();
		else
			return null;
	}
	
	
	// -------------------- isZippable --------------------
	
	
	/**
	* Immediately returns true if file is a directory.
	* Else if file is a normal file,
	* then returns true if file's length is <= {@link #zippableFileLengthLimit}, false otherwise.
	* Else returns false.
	* <p>
	* @throws IllegalArgumentException if file == null or file does not exist
	* @throws SecurityException if a security manager exists and its SecurityManager.checkRead method denies read access to file
	*/
	public static boolean isZippable(File file) throws IllegalArgumentException, SecurityException {
		if (file == null) throw new IllegalArgumentException("file == null");
		if (!file.exists()) throw new IllegalArgumentException("file = " + file.getPath() + " does not exist");
		
		if (file.isDirectory()) return true;
		else if (file.isFile()) return (file.length() <= zippableFileLengthLimit);
		else return false;
	}
	
	
	// -------------------- archive and helper methods --------------------
	
	
	/** Simply calls {@link #archive(File[], File, FileFilter) archive(new File[] {pathToArchive}, zipFile, filter)}. */
	public static void archive(File pathToArchive, File zipFile, FileFilter filter) throws Exception {
		archive( new File[] {pathToArchive}, zipFile, filter );
	}
	
	
	/**
	* Writes each element of files to a new ZIP format archive file specified by zipFile.
	* If any element is a directory, the entire contents of the directory tree will be archived.
	* An optional FileFilter can be supplied to screen out paths that would otherwise be archived.
	* <p>
	* In general, the path stored in the archive
	* is the path relative to the <i>parent</i> of the relevant element of files.
	* For example, let <code>file[i] = new File("D:/someDirectory/anotherDirectory")</code>
	* so <code>D:/someDirectory</code> is the parent of <code>file[i]</code>.
	* Then the path of <code>file[i]</code> relative to its parent is simply <code>anotherDirectory</code>,
	* which will be the path used for <code>file[i]</code> in the archive.
	* Furthermore, if <code>file[i]</code> contains the child path
	* <code>D:/someDirectory/anotherDirectory/aChild</code>,
	* then that child will have the path <code>anotherDirectory/aChild</code> in the archive.
	* <p>
	* One complication with the above scheme is paths which are file system roots: they have no parents.
	* Examples include the windows path <code>C:</code> or the unix path <code>/</code>.
	* In cases like these, this method uses an imaginary parent name
	* of the form <code>rootXXX</code> (where XXX is an integer).
	* For example, on a windows machine, if files contains the paths <code>C:</code> and <code>D:</code>,
	* then the contents of <code>C:</code> might be stored in the archive
	* with a path that starts with <code>root1</code>, and the contents of <code>D:</code>
	* may have an archive path that starts with <code>root2</code>.
	* This behavior ensures that the archive preserves the separate origins of the 2 sources.
	* <p>
	* @param files the File[] to archive
	* @param zipFile the ZIP File that will write the archive data to
	* @param filter a FileFilter that can use to screen out certain files from being written to the archive;
	* may be null (so everything specified by the files arg gets archived)
	* @throws Exception if any Throwable is caught; the Throwable is stored as the cause, and the message stores the path of zipFile;
	* here are some of the possible causes:
	* <ol>
	*  <li>
	*		IllegalArgumentException if files == null; files.length == 0;
	*		zipFile == null;
	*		zipFile already exists and either is not a normal file or is but already has data inside it;
<!--
+++ should i add this check? TarUtil does...

*		zipFile has an invalid extension;
-->
	*		any element of files is null, does not exist, cannot be read, is equal to zipFile, its path contains zipFile,
	*		or it fails {@link #isZippable isZippable}
	*  </li>
	*  <li>SecurityException if a security manager exists and its SecurityManager.checkRead method denies read access to file</li>
	*  <li>IOException if an I/O problem occurs</li>
	* </ol>
	*/
	public static void archive(File[] files, File zipFile, FileFilter filter) throws Exception {
// +++ should use varargs...
		try {
			if (files == null) throw new IllegalArgumentException("files == null");
			if (files.length == 0) throw new IllegalArgumentException("files.length == 0");
			if (zipFile == null) throw new IllegalArgumentException("zipFile == null");
			if (zipFile.exists()) {
				if (!zipFile.isFile()) throw new IllegalArgumentException("zipFile = " + zipFile.getPath() + " is not a normal file");
				if (zipFile.length() != 0) throw new IllegalArgumentException("zipFile = " + zipFile.getPath() + " already exists and already has data inside it");
			}
			for (int i = 0; i < files.length; i++) {
				if ( files[i] == null ) throw new IllegalArgumentException("files[" + i + "] == null");
				if ( !files[i].exists() ) throw new IllegalArgumentException("files[" + i + "] = " + files[i].getPath() + " is a non-existent path");
				if ( !files[i].canRead() ) throw new IllegalArgumentException("files[" + i + "] = " + files[i].getPath() + " cannot be read by this application");
				if ( files[i].equals(zipFile) ) throw new IllegalArgumentException("files[" + i + "] = " + files[i].getPath() + " is the same File as arg zipFile = " + zipFile.getPath());
				if ( files[i].isDirectory() && DirUtil.contains(files[i], zipFile) ) throw new IllegalArgumentException("the directory corresponding to arg files[" + i + "] = " + files[i].getCanonicalPath() + " will contain the path corresponding to arg zipFile = " + zipFile.getCanonicalPath() );
			}
			
			ZipOutputStream out = null;
			try {
				out = new ZipOutputStream( new BufferedOutputStream( new FileOutputStream(zipFile), 32*1024 ) );
				for (File file : files) {
					archive( file, new FileParent(file), out, filter );
				}
				out.flush();	// must flush before can do the file size test below, since data may be buffered
				if (zipFile.length() > zipFileLengthLimit) throw new ZipException("this call to archive wrote " + zipFile.length() + " bytes to zipFile = " + zipFile.getPath() + " however the maximum valid ZIP file size = " + zipFileLengthLimit);
			}
			finally {
				if (giveConsoleFeedback) ConsoleUtil.eraseLine();
				if (out != null) out.close();
			}
		}
		catch (Throwable t) {
			throw new Exception("See cause for the underlying Throwable; happened for zipFile = " + (zipFile != null ? zipFile.getPath() : "<null>"), t);
		}
	}
	
	
	/**
	* This method writes file as a new ZipEntry to the ZipOutputStream,
	* followed by file's data (which is 0 bytes if it is a directory).
	* This ZipEntry is always written compressed, and at the highest compression level.
	* <p>
	* If the File is a directory, then this method additionally calls itself on the contents
	* (thus recursing thru the entire directory tree).
	* <p>
	* <b>Warning</b>: several popular programs (e.g. winzip) fail to display mere directory entries.
	* Furthermore, if just a directory entry is present (i.e. it is empty), they also may fail
	* to create a new empty directoy when extracting the ZIP file's contents.
	* These are bugs in their behavior.
	* <p>
	* An optional FileFilter can be supplied to screen out paths that would otherwise be archived.
	* <p>
	* <i>This method does not close the supplied ZipOutputStream:</i> that is the responsibility of the caller.
	* <p>
	* The caller also must take on the responsibility to not do anything stupid, like write
	* file more than once, or have the file be the same File that out is writing to.
	* <p>
	* @param file the File to archive
	* @param fileParent the FileParent for file
	* @param out the ZipOutputStream that will write the archive data to
	* @param filter a FileFilter that can use to screen out certain files from being written to the archive; may be null (so everything specified by the file arg gets archived)
	* @throws Exception if any Throwable is caught; the Throwable is stored as the cause, and the message stores the path of file;
	* here are some of the possible causes:
	* <ol>
	*  <li>IllegalArgumentException if file fails {@link #isZippable isZippable}</li>
	*  <li>SecurityException if a security manager exists and its SecurityManager.checkRead method denies read access to file</li>
	*  <li>IOException if an I/O problem occurs</li>
	* </ol>
	*/
	private static void archive(File file, FileParent fileParent, ZipOutputStream out, FileFilter filter) throws Exception {
		try {
			if ( (filter != null) && !filter.accept(file) ) return;
			if (!isZippable(file)) throw new IllegalArgumentException("file = " + file.getPath() + " has a length (" + file.length() + " bytes) which exceeds the ZIP format file size limit = " + zippableFileLengthLimit);
			if (giveConsoleFeedback) ConsoleUtil.overwriteLine(file.getPath());
			
			ZipEntry entry = new ZipEntry( fileParent.getRelativePath(file, '/') );	// Note: getRelativePath will ensure that '/' is the path separator and that directories end with '/'
				// Note: no need to call any of the setXXX methods of ZipEntry -- the ones I would care about will all get called automaticly when write to the stream
			out.setMethod( ZipOutputStream.DEFLATED );
			out.setLevel( Deflater.BEST_COMPRESSION );				 
// +++ may be faster to use Deflater.NO_COMPRESSION on known compressed files (e.g. .zip, .jpg, etc);
// on the other hand, I am still sometimes seeing a tiny bit of compression (say 1-5%) even on already compressed types like zip and jpg
			out.putNextEntry( entry );
			writeFileData(file, out);
			out.closeEntry();
			
			if ( file.isDirectory() ) {
				for (File fileChild : DirUtil.getContents(file, null)) {	// supply null, since we test at beginning of this method (supplying filter here which just add a redundant test)
					archive( fileChild, fileParent, out, filter );
				}
			}
		}
		catch (Throwable t) {
			String preface = "See cause for the underlying Throwable; happened for file = ";
			if ( (t instanceof Exception) && (t.getMessage().startsWith(preface)) ) throw (Exception) t;	// CRITICAL: see if t was actually the wrapping Exception generated by a subsequent recursive call to this method, and if so simply rethrow it unmodified to avoid excessive exception chaining
			throw new Exception(preface + (file != null ? file.getPath() : "<null>"), t);
		}
	}
	
	
	/**
	* Returns immediately if file is not a normal file (e.g. is a directory).
	* Otherwise, reads all the bytes from file and writes them to out.
	* <p>
	* @throws IOException if an I/O problem occurs
	*/
	private static void writeFileData(File file, OutputStream out) throws IOException {
		if (!file.isFile()) return;
		
		InputStream in = null;
		try {
			in = new FileInputStream(file);
			StreamUtil.transfer(in, out);
		}
		catch (IOException ioe) {
			throw new IOException( "Error while processing file = " + file.getPath() + "; see cause for underlying problem", ioe );
		}
		finally {
			StreamUtil.close(in);
		}
	}
	
	
	// -------------------- close --------------------
	
	
	/**
	* This method attempts to close zipFile if it is not null.
	* <p>
	* Any Throwable that occurs is logged using {@link LogUtil#logRobustly LogUtil.logRobustly}.
	* Thus, <i>this method will never throw a Throwable</i>.
	*/
	public static void close(ZipFile zipFile) {
		try {
			if (zipFile != null) zipFile.close();
		}
		catch (Throwable t) {
			LogUtil.logRobustly("ZipUtil", "close", "a problem occurred while closing zipFile", t);
		}
	}
	
	
	// -------------------- getEntries --------------------
	
	

// +++ the methods in thiss section ought to take a normal File as an argument?  or have convenience versions...

	/**
	* Returns all the {@link ZipEntry}s in zipFile.
	* <p>
	* @param zipFile the {@link ZipFile} to get the entries from
	* @param sortResult if true, then the result ZipEntry[] is first sorted by each entry's name before return;
	* otherwise the order is the sequence read from zipFile
	* @throws IllegalArgumentException if zipFile == null
	* @throws IllegalStateException if zipFile has been closed
	*/
	public static ZipEntry[] getEntries(ZipFile zipFile, boolean sortResult) throws IllegalArgumentException, IllegalStateException {
		if (zipFile == null) throw new IllegalArgumentException("zipFile == null");
		
		ZipEntry[] result = new ZipEntry[ zipFile.size() ];
		int i = 0;
		for (Enumeration<? extends ZipEntry> e = zipFile.entries(); e.hasMoreElements(); ) {
			result[i++] = e.nextElement();
		}
		
		if (sortResult) {
			Arrays.sort(result, new Comparator<ZipEntry>() {
				public int compare(ZipEntry ze1, ZipEntry ze2) throws IllegalArgumentException {
					if (ze1 == null) throw new IllegalArgumentException("ze1 == null");
					if (ze2 == null) throw new IllegalArgumentException("ze2 == null");
					
					return ze1.getName().compareTo( ze2.getName() );
				}
			} );
		}
		
		return result;		
	}
	
	
	/**
	* Extracts zipFile to directoryExtraction.
	* <p>
	* Note that directoryExtraction need not exist, since it (and any parent directories)
	* will be created if necessary
	* (but directoryExtraction must be a proper directory path in this case).
	* <p>
	* @param zipFile the ZIP archive File
	* @param directoryExtraction the directory that will extract the contents of zipFile into
	* @param overwrite specifies whther or not extraction is allowed to o
	* @throws IllegalArgumentException if zipFile == null;
	* if directoryExtraction fails {@link DirUtil#ensureExists DirUtil.ensureExists}
	* @throws SecurityException if a security manager exists and its SecurityManager.checkRead method
	* denies read access to zipFile's path or directoryExtraction
	* @throws IllegalStateException if directoryExtraction failed to be created or is not an actual directory but is some other type of file;
	* if overwrite == false and one of the items to be extracted already exists
	* @throws IOException if an I/O problem occurs
	*/
	public static void extract(ZipFile zipFile, File directoryExtraction, boolean overwrite) throws IllegalArgumentException, SecurityException, IOException {
		if (zipFile == null) throw new IllegalArgumentException("zipFile == null");
		DirUtil.ensureExists(directoryExtraction);
		
		for (ZipEntry entry : getEntries(zipFile, false)) {
			File file = new File(directoryExtraction, entry.getName());
			if (!overwrite && file.exists()) throw new IllegalStateException("overwrite == false and " + file.getPath() + " already exists");
			
			if (entry.isDirectory()) {
				DirUtil.ensureExists(file);
			}
			else {
				DirUtil.ensureExists(file.getParentFile());
				InputStream in = null;
				OutputStream out = null;
				try {
					in = zipFile.getInputStream(entry);	// no buffering, because StreamUtil.transfer below will do that itself
					out = new FileOutputStream(file, false);	// no buffering, because StreamUtil.transfer below will do that itself
					StreamUtil.transfer(in, out);
				}
				finally {
					StreamUtil.close(in);
					StreamUtil.close(out);
				}
			}
		}
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private ZipUtil() {}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		//private static String root = "../log/";
		private static String root = "C:/";
		
		private static int count = 0;
		
		private static int countNext() { return count++; }
		
// +++ needs to be rewritten to be with all the methods of TarUtil's UnitTest
		
		@Test public void test_archive_extract() throws Exception {
			File directoryExtraction = null;
			try {
				File dirToArchive = new File( "../src" );
				File zipFile = TarUtil.UnitTest.makeArchiveFile("zip");
				archive( dirToArchive, zipFile, null );
				printEntries(zipFile);
				
				directoryExtraction = TarUtil.UnitTest.makeDirectoryExtraction("zip");
//				extract( zipFile, directoryExtraction );
extract( new ZipFile(zipFile), directoryExtraction, false );
				
				File directoryOneLevelDown = directoryExtraction.listFiles()[0];	// need to drill down one level in order to get the proper directory in which to compare with dirToArchive
				Assert.assertNull( DirUtil.compareContents(dirToArchive, directoryOneLevelDown) );	// DirUtil.compareContents ONLY returns null if 2 directories have exactly equally contents in all their elements (both normal files and subdirectories)
			}
			finally {
				//FileUtil.deleteIfExists(...);	// no need to worry about the files: they are all temporary
				DirUtil.deleteIfExists(directoryExtraction);	// unfortunately, cannot make temp directories--deleteOnExit fails if the directory has content--so must explicitly clean it up
			}
		}
		
@Ignore("Not running because it takes too long...")
		@Test public void test_archive_extract_fileLengthLimit_shouldPass() throws Exception {
			File directoryExtraction = null;
			File dirToArchive = null;
			try {
				dirToArchive = new File(root + "/test_archive_extract_fileLengthLimit_shouldPass");
				long practicalLimit = zippableFileLengthLimit - (2L * 1024L * 1024L);
				File fileHalf = TarUtil.UnitTest.makeDataFile(dirToArchive, practicalLimit / 2L);
				File fileMax = TarUtil.UnitTest.makeDataFile(dirToArchive, practicalLimit);
				File zipFile = TarUtil.UnitTest.makeArchiveFile("zip");
				archive( dirToArchive, zipFile, null );
				
				directoryExtraction = TarUtil.UnitTest.makeDirectoryExtraction("zip");
//				extract( zipFile, directoryExtraction );
extract( new ZipFile(zipFile), directoryExtraction, false );
				
				File directoryOneLevelDown = directoryExtraction.listFiles()[0];	// need to drill down one level in order to get the proper directory in which to compare with dirToArchive
				Assert.assertNull( DirUtil.compareContents(dirToArchive, directoryOneLevelDown) );	// DirUtil.compareContents ONLY returns null if 2 directories have exactly equally contents in all their elements (both normal files and subdirectories)
			}
			finally {
				//FileUtil.deleteIfExists(...);	// no need: all the files are temporary
				DirUtil.deleteIfExists(dirToArchive);	// CRITICAL: do this here because for this method, dirToArchive is a temp directory
				DirUtil.deleteIfExists(directoryExtraction);
			}
		}
		
@Ignore("Not running because it takes too long...")
		@Test(expected=Exception.class) public void test_archive_extract_fileLengthLimit_shouldFail() throws Exception {
			File dirToArchive = null;
			try {
				dirToArchive = new File(root + "/test_archive_extract_fileLengthLimit_shouldFail");
				File fileTooLarge = TarUtil.UnitTest.makeDataFile(dirToArchive, zippableFileLengthLimit + 1L);
				File zipFile = TarUtil.UnitTest.makeArchiveFile("zip");
				archive( fileTooLarge, zipFile, null );
			}
			finally {
				//FileUtil.deleteIfExists(...);	// no need: all the files are temporary
				DirUtil.deleteIfExists(dirToArchive);	// CRITICAL: do this here because for this method, dirToArchive is a temp directory
			}
		}
		
		private static void printEntries(File file) throws Exception {
			System.out.println('\t' + "below are the entries of ZIP file " + file.getPath() + ":");
			ZipFile zipFile = new ZipFile(file);
			for (ZipEntry entry : ZipUtil.getEntries(zipFile, true)) {
				System.out.println( '\t' + entry.toString() );			
			}
		}
		
	}
	
	
}
