/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.io.filefilter;


import java.io.*;

import org.junit.Assert;
import org.junit.Test;


/**
* This class is a meta filter which incorporates all the constraints of all the file filters
* that it was constructed with in its {@link #accept accept} method.
* <p>
* Note that {@link FilenameSuffixFilter} can be viewed as a specialized type of compound filter:
* it only considers filename suffixes, but it considers a list of them and performs a logical OR test against filenames.
* Using/subclassing FilenameSuffixFilter is therefore more convenient than this class if that is all that is desired.
* <p>
* In contrast, this class offers much more flexibility: any types of filters can be used, not just filename suffixes,
* and logical {@link Mode#and AND} of the filters can be specified as well logical {@link Mode#or OR}.
* <p>
* Note that this class is both a subclass of {@link javax.swing.filechooser.FileFilter javax.swing.filechooser.FileFilter}
* (so that it can be passed to {@link javax.swing.JFileChooser}),
* as well as an implementation of {@link java.io.FileFilter java.io.FileFilter}
* (so that it is suitable as an argument to {@link java.io.File#listFiles(java.io.FileFilter) File.listFiles}).
* <p>
* This class has no immedaite mutable state.
* This class is multithread safe if the elements that are assigned to its {@link #filters} field are all multithread safe.
* <p>
* @author Brent Boyer
*/
public class CompoundFilter extends javax.swing.filechooser.FileFilter implements java.io.FileFilter {
	
	
	// -------------------- instance constants --------------------
	
	
	private final Mode mode;
	
	
	/** FileFilters which will be used in {@link #accept accept} and {@link #getDescription getDescription}. */
	private final javax.swing.filechooser.FileFilter[] filters;
	// NOTE: specify them as javax.swing.filechooser.FileFilter instead of java.io.FileFilter because the former includes the getDescription method
	
	
	// -------------------- factory methods and constructors --------------------
	
	
	/**
	* Returns a new CompoundFilter instances that uses {@link Mode#and} to combine all the filters.
	* <p>
	* @param filters assigned to the {@link #filters} field;
	*	<i>for top performance, place the most frequently encountered patterns earliest in the array</i>
	* @throws IllegalArgumentException if filters == null; filters.length == 0
	*/
	public static CompoundFilter makeAnd(javax.swing.filechooser.FileFilter... filters) {
		return new CompoundFilter(Mode.and, filters);
	}
	
	
	/**
	* Returns a new CompoundFilter instances that uses {@link Mode#or} to combine all the filters.
	* <p>
	* @param filters assigned to the {@link #filters} field;
	*	<i>for top performance, place the most frequently encountered patterns earliest in the array</i>
	* @throws IllegalArgumentException if filters == null; filters.length == 0
	*/
	public static CompoundFilter makeOr(javax.swing.filechooser.FileFilter... filters) {
		return new CompoundFilter(Mode.or, filters);
	}
	
	
	/**
	* Constructor.
	* <p>
	* @param mode assigned to the {@link #mode} field
	* @param filters assigned to the {@link #filters} field;
	*	<i>for top performance, place the most frequently encountered patterns earliest in the array</i>
	* @throws IllegalArgumentException if mode == null; filters == null; filters.length == 0
	*/
	protected CompoundFilter(Mode mode, javax.swing.filechooser.FileFilter... filters) {
		if (mode == null) throw new IllegalArgumentException("mode == null");
		if (filters == null) throw new IllegalArgumentException("filters == null");
		if (filters.length == 0) throw new IllegalArgumentException("filters.length == 0");
		
		this.mode = mode;
		this.filters = filters;
	}
	
	
	// -------------------- accessors --------------------
	
	
	/** Returns a (shallow) clone of the array of javax.swing.filechooser.FileFilters that are used in {@link #accept accept}. */
	public javax.swing.filechooser.FileFilter[] getFilters() { return filters.clone(); }
	
	
	// -------------------- accept --------------------
	
	
	/**
	* Reports whether or not file is accepted by this instance.
	* Applies all the {@link #filters} according to {@link #mode}.
	* <p>
	* @throws IllegalArgumentException if arg file is null
	* @throws SecurityException if a security manager exists and its SecurityManager.checkRead(java.lang.String) method denies read access to the file
	* @throws IllegalStateException if the class finds itself in an unexpected state
	*/
	public boolean accept(File file) throws IllegalArgumentException, SecurityException, IllegalStateException {
		if (file == null) throw new IllegalArgumentException("arg file is null");
		
		switch (mode) {
			case and:
				for (javax.swing.filechooser.FileFilter filter : filters) {
					if ( !filter.accept(file) ) return false;
				}
				return true;
				
			case or:
				for (javax.swing.filechooser.FileFilter filter : filters) {
					if ( filter.accept(file) ) return true;
				}
				return false;
			
			default:
				throw new IllegalStateException("mode = " + mode + " is an illegal value");
		}
	}
	
	
	// -------------------- getDescription --------------------
	
	
	/** Returns a description of this filter. Is simply a concatenation of the descriptions of all the {@link #filters}. */
	public String getDescription() {
		StringBuilder sb = new StringBuilder( filters.length * 32 );
		sb.append( getModeDescription() );
		for (int i = 0; i < filters.length; i++) {
			if (i > 0) sb.append(", ");
			sb.append( filters[i].getDescription() );
		}
		return sb.toString();
	}
	
	
	private String getModeDescription() throws IllegalStateException {
		switch (mode) {
			case and: return "AND of: ";
			case or: return "OR of: ";
			default: throw new IllegalStateException("mode = " + mode + " is an illegal value");
		}
	}
	
	
	// -------------------- Mode (static inner enum) --------------------
	
	
	/**
	* Enum of all the supported modes for applying the individual filters.
	* <p>
	* This enum is multithread safe: it is stateless.
	* <p>
	* Like all java enums, this enum is Comparable and Serializable.
	* <p>
	* @see <a href="http://java.sun.com/j2se/1.5.0/docs/guide/language/enums.html">Enum documentation</a>
	*/
	public static enum Mode {
		and,	// specifies the logical AND of every individual filter
		or;	// specifies the logical OR of every individual filter
	}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		private static final String[] suffixes = new String[] {".test1", ".test2"};
		
		private static final File[] matchingFiles = new File[] {
			new File("abc.test1"), new File("abc.TEST1"),
			new File("def.test2"), new File("def.TEST2")
		};
		
		private static final File[] nonmatchingFiles = new File[] {
			new File("abc.test12"), new File("abc.TEST12"),
			new File("def.txt"), new File("def.jpg")
		};
		
		@Test public void test_getDescription() {
			TestFilter1 filter1 = new TestFilter1();
			TestFilter2 filter2 = new TestFilter2();
			CompoundFilter compoundFilter = CompoundFilter.makeOr(filter1, filter2);
			
			System.out.println("Here is the result from the getDescription method of a test filter:");
			System.out.println( "\t" + compoundFilter.getDescription() );
		}
		
		@Test public void test_filtering() throws Exception {
			TestFilter1 filter1 = new TestFilter1();
			TestFilter2 filter2 = new TestFilter2();
			CompoundFilter compoundFilter = CompoundFilter.makeOr(filter1, filter2);
			
			for (File matchingFile : matchingFiles) {
				String errMsg = "MISTAKENLY rejected " + matchingFile.getName();
				Assert.assertTrue( errMsg, compoundFilter.accept(matchingFile) );
			}
			for (File nonmatchingFile : nonmatchingFiles) {
				String errMsg = "FAILED to reject " + nonmatchingFile.getName();
				Assert.assertFalse( errMsg, compoundFilter.accept(nonmatchingFile) );
			}
		}
		
		private static class TestFilter1 extends FilenameSuffixFilter {
			private static String[] suffixes1 = new String[] {suffixes[0]};
			private static final String DESCRIPTION = FilenameSuffixFilter.makeDescription("Type 1 UnitTest files", suffixes1);
			
			private TestFilter1() {
				super(suffixes1, false, FilenameSuffixFilter.REJECT_DIRECTORIES);
			}
			
			public String getDescription() { return DESCRIPTION; }
		}
		
		private static class TestFilter2 extends FilenameSuffixFilter {
			private static String[] suffixes2 = new String[] {suffixes[1]};
			private static final String DESCRIPTION = FilenameSuffixFilter.makeDescription("Type 2 UnitTest files", suffixes2);
			
			private TestFilter2() {
				super(suffixes2, false, FilenameSuffixFilter.REJECT_DIRECTORIES);
			}
			
			public String getDescription() { return DESCRIPTION; }
		}
		
	}
	
	
}
