/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*
Future features:

--regular expression matches?
(if use this, then need NOT restrict ourselves to suffixes...)
Probably better to make another base class similar to this one

*/


package bb.io.filefilter;


import java.io.*;

import org.junit.Assert;
import org.junit.Test;


/**
* This abstract class is meant to provide all the base functionality for file filters
* which accept or reject {@link File}s depending on how their names end.
* <p>
* This class must be {@link #FilenameSuffixFilter constructed} with a String[] of file name suffix patterns.
* A normal file is always accepted if the suffix of its name matches one of the patterns in the array.
* The matching of the filename suffix to a pattern String is done char by char.
* A boolean arg to the constructor specifies whether or not the matching is case insensitive.
* <p>
* This class's constructor must also take an int arg that specifies how to treat directories.
* The possibilities are:
* <ul>
*  <li>always reject</li>
*  <li>apply the same name suffix test to them as normal files</li>
*  <li>always accept</li>
* </ul>
* This class defines constants for each of the above behaviors (see the {@link #FilenameSuffixFilter} for details).
* <p>
* Concrete subclasses typically just need to define (and supply to this class's constructor) an array of suffixes,
* as well as implement the getDescription method.
* <p>
* Note that this class is both a subclass of {@link javax.swing.filechooser.FileFilter javax.swing.filechooser.FileFilter}
* (so that it can be passed to {@link javax.swing.JFileChooser}),
* as well as an implementation of {@link java.io.FileFilter java.io.FileFilter}
* (so that it is suitable as an argument to {@link java.io.File#listFiles(java.io.FileFilter) File.listFiles}).
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @author Brent Boyer
*/
public abstract class FilenameSuffixFilter extends javax.swing.filechooser.FileFilter implements java.io.FileFilter {


	// -------------------- static constants --------------------


	/** Specifies that every directory will be rejected, regardless of its name. */
	public static final int REJECT_DIRECTORIES = 0;

	/** Specifies that every directory will be subjected to the same name suffix test as normal files. */
	public static final int FILTER_DIRECTORIES_BY_SUFFIX = 1;

	/** Specifies that every directory will be accepted, regardless of its name. */
	public static final int ACCEPT_DIRECTORIES = 2;


	// -------------------- instance constants --------------------


	private final String[] filenameSuffixes;
	private final boolean caseSensitive;
	private final int directoryTreatment;


	// -------------------- constructor --------------------


	/**
	* Constructs a new FilenameSuffixFilter instance.
	* <p>
	* @param filenameSuffixes String[] of filename suffix patterns that if a file matches, result in the file being accepted;
	*	<i>for top performance, place the most frequently encountered patterns earliest in the array</i>
	* @param caseSensitive specifies whether or not the filename suffix comparison is to be case sensitive
	* @param directoryTreatment an int that must be one of the following constants of this class:
	*	{@link #REJECT_DIRECTORIES},
	*	{@link #FILTER_DIRECTORIES_BY_SUFFIX}, or
	*	{@link #ACCEPT_DIRECTORIES}
	* @throws IllegalArgumentException if arg filenameSuffixes is null or 0-length, or if arg directoryTreatment is invalid
	*/
	public FilenameSuffixFilter(String[] filenameSuffixes, boolean caseSensitive, int directoryTreatment) {
		if (filenameSuffixes == null) throw new IllegalArgumentException("arg filenameSuffixes is null");
		if (filenameSuffixes.length == 0) throw new IllegalArgumentException("arg filenameSuffixes is 0-length");
		if ( !isValidDirectoryTreatment(directoryTreatment) ) throw new IllegalArgumentException("arg directoryTreatment = " + directoryTreatment + " is invalid");

		this.filenameSuffixes = filenameSuffixes;
		this.caseSensitive = caseSensitive;
		this.directoryTreatment = directoryTreatment;
	}


	// -------------------- accessors --------------------


	/** Returns a clone of the array of filenameSuffixes that are used to match files. */
	public String[] getSuffixes() { return filenameSuffixes.clone(); }


	// -------------------- isValidDirectoryTreatment --------------------


	/** Reports whether or not the directoryTreatment arg is a valid value. */
	protected boolean isValidDirectoryTreatment(int directoryTreatment) {
		return (
			(directoryTreatment == REJECT_DIRECTORIES) ||
			(directoryTreatment == FILTER_DIRECTORIES_BY_SUFFIX) ||
			(directoryTreatment == ACCEPT_DIRECTORIES)
		);
	}


	// -------------------- accept and helper methods --------------------


	/**
	* Reports whether or not the supplied File is accepted by this filter.
	* <p>
	* @throws IllegalArgumentException if arg file is null
	* @throws SecurityException if a security manager exists and its SecurityManager.checkRead(java.lang.String) method denies read access to the file
	*/
	public boolean accept(File file) {
		if (file == null) throw new IllegalArgumentException("arg file is null");

		if ( file.isDirectory() ) {
			switch (directoryTreatment) {
				case REJECT_DIRECTORIES:
					return false;

				case FILTER_DIRECTORIES_BY_SUFFIX:
					return suffixMatches(file.getName());

				case ACCEPT_DIRECTORIES:
					return true;

					// the following should NEVER happen as long as nobody changes directoryTreatment from private and final and leaves the constructor checks in place
				default:
					throw new IllegalStateException("field directoryTreatment = " + directoryTreatment + " is an invalid value");
			}
		}
		else
			return suffixMatches(file.getName());
	}


	/** Reports whether or not the supplied fileName's suffix matches one of the patterns. */
	protected boolean suffixMatches(String fileName) {
		boolean ignoreCase = !caseSensitive;

		for (String suffix : filenameSuffixes) {
			int suffixLength = suffix.length();
			int fileNameOffset = fileName.length() - suffixLength;
			if ( fileName.regionMatches(ignoreCase, fileNameOffset, suffix, 0, suffixLength) ) return true;
		}
		return false;
	}


	// -------------------- getDescription and helper methods --------------------


	/** Returns a description of this filter. */
	public abstract String getDescription();


	/**
	* Returns a String that consists of the intro arg followed by the Strings in the suffixes arg.
	* The suffixes are put into a single comma separated list that is enclosed by parentheses.
	* Each suffix in that list is preceded by an asterisk, to indicate that the initial filename chars are a wildchar match.
	* <p>
	* This method is a convenience method to help subclasses generate a description of what file types they filter.
	* <p>
	* @throws IllegalArgumentException if arg intro or arg suffixes is null
	*/
	public static String makeDescription(String intro, String[] suffixes) {
		if (intro == null) throw new IllegalArgumentException("arg intro is null");
		if (suffixes == null) throw new IllegalArgumentException("arg suffixes is null");

		int descriptionLengthGuess = 16*(1 + suffixes.length);
		StringBuilder sb = new StringBuilder( descriptionLengthGuess );

		sb.append(intro);
		if (suffixes.length > 0) {
			sb.append(" (");
			for (int i = 0; i < suffixes.length; i++) {
				if (i > 0) sb.append(", ");
				sb.append("*").append(suffixes[i]);
			}
			sb.append(")");
		}

		return sb.toString();
	}


	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		private static final String[] suffixes = new String[] {".test1", ".test2"};
		
		private static final File[] matchingFilesCaseSens = new File[] {
			new File("abc.test1"), new File("def.test2")
		};
		
		private static final File[] nonmatchingFilesCaseSens = new File[] {
			new File("abc.TEST1"), new File("def.TEST2"), new File("ghi.test12")
		};
		
		private static final File[] matchingFilesCaseInsens = new File[] {
			new File("abc.test1"), new File("abc.TEST1"),
			new File("def.test2"), new File("def.TEST2")
		};
		
		private static final File[] nonmatchingFilesCaseInsens = new File[] {
			new File("ghi.test12")
		};
		
		@Test public void test_caseSensitiveFilter() {
			TestFilter csFilter = new TestFilter(true, FilenameSuffixFilter.REJECT_DIRECTORIES);
			for (File file : matchingFilesCaseSens) {
				Assert.assertTrue( csFilter.accept(file) );
			}
			for (File file : nonmatchingFilesCaseSens) {
				Assert.assertFalse( csFilter.accept(file) );
			}
		}


		@Test public void test_nonCaseSensitiveFilter() {
			TestFilter ciFilter = new TestFilter(false, FilenameSuffixFilter.REJECT_DIRECTORIES);
			for (File file : matchingFilesCaseInsens) {
				Assert.assertTrue( ciFilter.accept(file) );
			}
			for (File file : nonmatchingFilesCaseInsens) {
				Assert.assertFalse( ciFilter.accept(file) );
			}
		}


		@Test public void test_directoryHandling() {
			File directory = bb.io.DirUtil.WD;
			
			TestFilter acceptDirFilter = new TestFilter(true, FilenameSuffixFilter.ACCEPT_DIRECTORIES);
			Assert.assertTrue( acceptDirFilter.accept(directory) );
			
			TestFilter rejectDirFilter = new TestFilter(true, FilenameSuffixFilter.REJECT_DIRECTORIES);
			Assert.assertFalse( rejectDirFilter.accept(directory) );
		}
		
		private static class TestFilter extends FilenameSuffixFilter {
			private static final String DESCRIPTION = FilenameSuffixFilter.makeDescription("UnitTest Files", suffixes);
			
			private TestFilter(boolean caseSensitive, int directoryTreatment) {
				super(suffixes, caseSensitive, directoryTreatment);
			}
			
			public String getDescription() { return DESCRIPTION; }
		}
		
	}
	
	
}
