/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.io.filefilter;


import java.io.*;
import java.util.*;


/**
* Direct subclass of {@link VisibleFilter} that accepts everything accepted by VisibleFilter
* as well as accepts files which are system roots.
* This filter is needed because it has been observed that Windows (XP on NTFS disks, at least) marks roots as hidden.
* <p>
* Note that this class is both a subclass of {@link javax.swing.filechooser.FileFilter javax.swing.filechooser.FileFilter}
* (so that it can be passed to {@link javax.swing.JFileChooser}),
* as well as an implementation of {@link java.io.FileFilter java.io.FileFilter}
* (so that it is suitable as an argument to {@link java.io.File#listFiles(java.io.FileFilter) File.listFiles}).
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @author Brent Boyer
*/
public class VisibleAndRootFilter extends VisibleFilter {
	
	
	protected static final String DESCRIPTION = "Visible and Root Files Only";
	
	
	private final File[] rootFiles;
// +++ a MAJOR problem here is with removable media: the result returned by File.listRoots() can change as media is inserted/ejected...


	/** Constructs a new VisibleAndRootFilter instance. */
	public VisibleAndRootFilter() {
		rootFiles = File.listRoots();
		Arrays.sort(rootFiles);
	}
	
	
	/** Reports whether or not the supplied file is accepted by this filter. */
	public boolean accept(File file) {
		return super.accept(file) || isRootFile(file);
	}
	
	
	private boolean isRootFile(File file) {
		return (Arrays.binarySearch(rootFiles, file) >= 0);	
	}
	
	
	/** Returns a description of this filter. */
	public String getDescription() { return DESCRIPTION; }


}
