/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*

Programmer Notes:

--In assigning a value to jdbcUrl for MySQL, I stumbled across the idea to use www.ericmarder.com (instead of localhost) when i did
	./bin/mysql mysql
	select * from user
and noticed that the root user had a row with www.ericmarder.com as the host.
Apparently, that is the domain name that our particular Linux box assigns us when we login.
For how to specify the params, see:
	mm.mysql.jdbc-1.2c/doc/mm.doc/c106.html
	http://www.worldserver.com/mm.mysql/doc/mm.doc/c106.htm#AEN118

--IN FUTURE, should get properties like driver & db url from a properties text file
and/or use JDK1.4's new preferences capability instead of hard coding them as is currently done;
then would not have to recompile each time...
	--would need to add code which confirms that stuff like poolSize, and waitLimit have valid values

--IN FUTURE, should use serious logging, such as what is now builtin to JDK1.4

--On connection pooling:
	good web articles:
		http://www.jguru.com/faq/view.jsp?EID=161
			(their implementation is mediocre):  http://webdevelopersjournal.com/columns/connection_pool.html

	check out JDBC 1.4: I believe that the latest jdbc spec has some sort of built in connection pool class or something?

--a more sophisticated pool implementation than what is currently in this class might:
	--have an initial number of connections, grow it if necessary to some max number, and remove connections if they are unused for a long while
	--keep references to all the currently loaned out connections in a separate List; this would enable it to signal the owner of the connection (via an InterruptedException?) if has kept the resource too long?
	--have a method called shutdown which sets a closed flag, closes all connections currently in the pool, and closes all connections that subsequently get returned
	--specify constants like poolSize and waitLimit in some other less hard-wired way

NEED TO READ DOUG LEA'S CONCURRENT BOOK; HAS A SECTION ON POOLS; HE PROBABLY HAS SOME IDEAS THAT SHOULD INCORPORATE HERE

*/


package bb.jdbc;


import java.sql.*;
import java.util.*;

import bb.util.ThrowableUtil;


/**
* This class manages access to database Connections.
* The end user need only know about and use the 2 public methods, getConnection and returnConnection,
* which simply do what their names say.
* <p>
* This class is multithread safe: most of its state is immutable (both its immediate state, as well as the deep state of its fields).
* The sole exception is the {@link #pool} field, and all uses of it first synchronize on it.
* <p>
* @author Brent Boyer
* @see #getConnection
* @see #returnConnection
*/
public class DBConnManager {

	
	private static final String jdcDriver = "org.gjt.mm.mysql.Driver";
	//private static final String jdcDriver = "org.apache.derby.jdbc.EmbeddedDriver";
	
	private static final String jdbcUrl = "jdbc:mysql://localhost/EMA?user=root ?default-character-set=big5";
	//private static final String jdbcUrl = "jdbc:derby:" + dbName + ";create=true";
	
	
	private static final int poolSize = 75;
	private static final List<Connection> pool = new ArrayList<Connection>(poolSize);
	private static final long waitLimit = 1*60*1000;


	/**
	* Static initializer that loads the driver class and adds new Connections to the pool.
	* <p>
	* @throws RuntimeException (or some subclass) if any error occurs; this may merely wrap some other underlying Throwable
	*/
	static {
		try {
			Class.forName(jdcDriver);	 // if have problems, may need to add  .newInstance()  here -- see mm.mysql.jdbc-1.2c/doc/mm.doc/c106.html
			// see also: http://blogs.sun.com/blogAmit/entry/jdbc_driver_loading_with_mustang

			for (int i = 0; i < poolSize; ++i) {
				pool.add( newConnection() );
			}
		}
		catch (Throwable t) {
			throw ThrowableUtil.toRuntimeException(t);
		}
	}


	/**
	* Returns a newly constructed Connection instance.
	* <p>
	* Note: the result has auto commit turned off and (since is new) should have
	* nothing to be committed and no warnings present.
	* <p>
	* Note: this method assumes that the driver class has already been loaded.
	* <p>
	* @throws SQLException	if a database access error occurs
	*/
	private static Connection newConnection() throws SQLException {
		Connection conn = DriverManager.getConnection(jdbcUrl);
		conn.setAutoCommit(false);
		return conn;
	}


	/**
	* Returns a Connection instance.
	* <p>
	* Note: the result has auto commit turned off and is clear of all warnings.
	* <p>
	* @throws InterruptedException	if the calling Thread is interrupted while waiting inside this method for a Connection to become available
	*/
	public static Connection getConnection() throws InterruptedException {
		synchronized (pool) {
			int loopCount = 0;
			while (true) {
				++loopCount;
				if (loopCount > 1) {	// if true, then you have already waited for more than waitLimit
// +++ you may wish to log the fact that people are waiting excessively long to get a connection
// +++ in fact, you may want alarm bells & pagers to go off
				}

				int endIndex = pool.size() - 1;
				if (endIndex >= 0) {
					return pool.remove(endIndex);   // critical for performance to remove from end of most List types (e.g. ArrayList & Vector, but NOT LinkedList); also, we do not need to check for auto commit off etc because of the contracts of newConnection and returnConnectionImpl
				}
				else {
					pool.wait(waitLimit);
				}
			}
		}
	}


	/**
	* Every Connection returned by the getConnection method must be sent back to this class by calling this method.
	* <p>
	* <i>This is critically important: failure to do this will ultimately deprive this class of any Connections to return,
	* which may stifle your application</i>.
	* Therefore, the caller of the getConnection method must use robust programming techniques to ensure that
	* this method receives back all Connections. Here is the typical usage:
	* <code><pre>
	*	Connection conn = null;
	*	try {
	*		conn = DBConnManager.getConnection();
	*		//do some work with the Connection
	*	}
	*	finally {
	*		if (conn != null) DBConnManager.returnConnection(conn);
	*	}</pre></code>
	* <p>
	* @throws IllegalArgumentException	if arg conn is null
	*/
	public static void returnConnection(Connection conn) throws IllegalArgumentException {
		returnConnectionImpl(conn);
	}


	/**
	* This method can be used as a full implementation of the returnConnection method.
	* It is separated from returnConnection because it specifies the following additional contract
	* that is relevant only to internal users of this class:
	* <p>
	* This method always trys to add a Connection back to the pool.
	* <p>
	* If the supplied Connection is already closed, then it will be ignored and
	* this method will add the result of calling newConnection to the pool.
	* <p>
	* Otherwise, this method will add the result of calling cleanupConnection on the supplied Connection to the pool.
	* <p>
	* After a Connection is added to the pool, this method will notify JUST ONE of any Threads waiting
	* on the pool instance.
	* <p>
	* @throws IllegalArgumentException	if arg conn is null
	*/
	private static void returnConnectionImpl(Connection conn) throws IllegalArgumentException {
		if (conn == null) throw new IllegalArgumentException("arg conn is null");

		try {
			if (conn.isClosed()) {		// if user has already clossed it, create a new one (the returned one will hopefully have no further references to it, so it can be garbage collected)
				conn = newConnection();
			}
			else {
				conn = cleanupConnection(conn);
			}

			synchronized (pool) {
				pool.add(conn);
				pool.notify();
			}
		}
		catch (Throwable t) {
t.printStackTrace(System.out);
// +++ it is critical that you log any failure to restock the pool!!!
// +++ alarm bells & pagers should go off
		}
	}


	/**
	* This method attempts to cleanup the supplied Connection by applying the following order of operations:
	* <ol>
	*  <li>turn off its auto commit</li>
	*  <li>call its commit method</li>
	*  <li>clear its warnings</li>
	* </ol>
	* <p>
	* If all the above operations succeed, the supplied Connection is returned.
	* Otherwise, if any Exception is thrown while trying the above operations,
	* an attempt will be made to close the supplied Connection and
	* this method will try to return the result of calling newConnection.
	* <p>
	* @throws IllegalArgumentException	if arg conn is null or already closed
	* @throws SQLException	if a database access error occurs
	*/
	private static Connection cleanupConnection(Connection conn) throws IllegalArgumentException, SQLException {
		if (conn == null) throw new IllegalArgumentException("arg conn is null");
		if (conn.isClosed()) throw new IllegalArgumentException("arg conn is closed");

		try {
			conn.setAutoCommit(false);
			conn.commit();
			conn.clearWarnings();
		}
		catch (Throwable t) {
t.printStackTrace(System.out);
// +++ you should probably log any problem that occurs here

				// before calling newConnection, attempt to close the existing one as the database may have a connection limit:
			try {
				conn.close();
			}
			catch (Throwable t2) {
t2.printStackTrace(System.out);
// +++ you should probably log any problem that occurs here
			}

			conn = newConnection();
		}

		return conn;
	}


}
