/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.net;


import java.io.*;
import java.util.*;
import java.util.logging.*;
import java.util.logging.Level;

import javax.mail.*;
import javax.mail.event.*;
import javax.mail.internet.*;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import bb.io.EncryptUtil;
import bb.io.FileUtil;
import bb.util.Properties2;
import bb.util.StringUtil;
import bb.util.logging.Logger2;
import bb.util.logging.LogUtil;


/**
* This class first stores a specific set of mail server resources
* (see the data that can be supplied to the constructors or the {@link #useSmtpSsl(String, String, int, String) useSmtpSsl} method).
* <p>
* Once configured thusly, an instance may send arbitrary emails using one of the send methods.
* <p>
* This class is not multithread safe.
* <p>
* @author Brent Boyer
* @see "The 2001/October/23 JDC newsletter for tips on JavaMail"
*/
public class Emailer {


	// -------------------- constants --------------------
	
	
	/**
	* Default value for the "mail.transport.protocol" property.
	* <p>
	* @see "JavaMail-1.2.pdf Appendix A"
	*/
	private static final String mailTransportProtocol_default = "smtp";
	
	
	/**
	* Default value for the "mail.debug" property.
	* <p>
	* @see "JavaMail-1.2.pdf Appendix A"
	*/
	private static final boolean mailDebug_default = false;
	
	
	/** Default value for {@link #parseStrictness}. */
	private static final boolean parseStrictness_default = true;
	
	
	private static final String sslFactory_default = "javax.net.ssl.SSLSocketFactory";
	
	
	// -------------------- instance fields --------------------
	
	
	/**
	* Stores the from email address(es) which will be used for all emails sent by this instance.
	* <p>
	* Contract: is never blank.
	*/
	private final String from;
	
	
	/**
	* Stores all the properties that will be used in {@link #session}.
	* <p>
	* Contract: is always non-null, is initially empty, but is assigned basic properties in the constructor.
	*/
	private final Properties2 properties = new Properties2();
	
	
	/**
	* Specifies the boolean value that will be supplied to {@link InternetAddress#parse(String, boolean)}
	* when parsing address Strings (e.g. from, to, cc, and bcc).
	*/
	private final boolean parseStrictness;


	/**
	* {@link Session} used by this instance.
	* <p>
	* Contract: is initially assigned in the constructor, but may be reassigned later (e.g. by {@link #useSmtpSsl useSmtpSsl}); should never be null.
	*/
	private Session session;


	/**
	* {@link Transport} used by this instance.
	* <p>
	* Contract: is initially null;
	* is assigned in {@link #send(String, String, String, String, String, Map)};	
	* may be nulled out later (e.g. by {@link #useSmtpSsl useSmtpSsl}).
	*/
	private Transport transport = null;
	
	
	/**
	* Set of all {@link ConnectionListener}s used by this instance.
	* <p>
	* Contract: is never null.
	*/
	private final Set<ConnectionListener> connectionListeners = new HashSet<ConnectionListener>();


	/**
	* Set of all {@link TransportListener}s used by this instance.
	* <p>
	* Contract: is never null.
	*/
	private final Set<TransportListener> transportListeners = new HashSet<TransportListener>();
	
	
	// -------------------- constructors --------------------
	
	
	public Emailer(String mailHost, String from) throws IllegalArgumentException {
		this(mailTransportProtocol_default, mailHost, from, mailDebug_default, parseStrictness_default);
	}
	
	
	/**
	* Fundamental (full arg) constructor.
	* <p>
	* @throws IllegalArgumentException if mailTransportProtocol is blank; mailHost is blank;
	* from fails {@link EmailUtil#checkHeaderFieldBody EmailUtil.checkHeaderFieldBody}
	*/
	public Emailer(String mailTransportProtocol, String mailHost, String from, boolean mailDebug, boolean parseStrictness) throws IllegalArgumentException {
		if (StringUtil.isBlank(mailTransportProtocol)) throw new IllegalArgumentException("mailTransportProtocol is blank");
		if (StringUtil.isBlank(mailHost)) throw new IllegalArgumentException("mailHost is blank");
		EmailUtil.checkHeaderFieldBody(from, "from");
				
		properties.setProperty( "mail.transport.protocol", mailTransportProtocol );
		properties.setProperty( "mail.host", mailHost );
		this.from = from;
		properties.setProperty( "mail.debug", Boolean.toString(mailDebug) );
		this.parseStrictness = parseStrictness;
		
		session = Session.getInstance(properties, null);
	}
	
	
	public Emailer(File file) throws IllegalArgumentException, SecurityException, IOException, MessagingException {
		FileUtil.check(file);
		
		properties.load( new FileInputStream(file) );

		properties.ensureProperty( "mail.transport.protocol", mailTransportProtocol_default );		
		String mailHost = properties.getNonBlankProperty("mail.host");		
		properties.ensureProperty( "mail.debug", Boolean.toString(mailDebug_default) );
		this.parseStrictness = properties.getBoolean( "parseStrictness", parseStrictness_default );

		String username = properties.getMandatoryProperty("username");
		String password = properties.getMandatoryProperty("password");
		int smtpPort = properties.getMandatoryInt("smtpPort");
		String sslFactory = properties.ensureProperty("sslFactory", sslFactory_default);
		useSmtpSsl(username, password, smtpPort, sslFactory);
		
		this.from = username + "@" + mailHost;
	}
	
	
	// -------------------- XXXListener --------------------
	
	
	/**
	* Attempts to add connectionListener to an internal Set of {@link ConnectionListener}s.
	* <p>
	* @return true if connectionListener was added by this call, false if was already present
	* @throws IllegalArgumentException if connectionListener == null
	*/
	public boolean addConnectionListener(ConnectionListener connectionListener) throws IllegalArgumentException {
		if (connectionListener == null) throw new IllegalArgumentException("connectionListener == null");
		
		boolean added = connectionListeners.add(connectionListener);
		if (added && (transport != null)) transport.addConnectionListener(connectionListener);
		return added;
	}
	
	
	/**
	* Attempts to remove connectionListener from an internal Set of {@link ConnectionListener}s.
	* <p>
	* @return true if connectionListener was removed by this call, false if was not present
	* @throws IllegalArgumentException if connectionListener == null
	*/
	public boolean removeConnectionListener(ConnectionListener connectionListener) throws IllegalArgumentException {
		if (connectionListener == null) throw new IllegalArgumentException("connectionListener == null");
		
		boolean removed = connectionListeners.remove(connectionListener);
		if (removed && (transport != null)) transport.removeConnectionListener(connectionListener);
		return removed;
	}
	
	
	/**
	* Attempts to add transportListener to an internal Set of {@link TransportListener}s.
	* <p>
	* @return true if transportListener was added by this call, false if was already present
	* @throws IllegalArgumentException if transportListener == null
	*/
	public boolean addTransportListener(TransportListener transportListener) throws IllegalArgumentException {
		if (transportListener == null) throw new IllegalArgumentException("transportListener == null");
		
		boolean added = transportListeners.add(transportListener);
		if (added && (transport != null)) transport.addTransportListener(transportListener);
		return added;
	}
	
	
	/**
	* Attempts to remove transportListener from an internal Set of {@link TransportListener}s.
	* <p>
	* @return true if transportListener was removed by this call, false if was not present
	* @throws IllegalArgumentException if transportListener == null
	*/
	public boolean removeTransportListener(TransportListener transportListener) throws IllegalArgumentException {
		if (transportListener == null) throw new IllegalArgumentException("transportListener == null");
		
		boolean removed = transportListeners.remove(transportListener);
		if (removed && (transport != null)) transport.removeTransportListener(transportListener);
		return removed;
	}
	
	
	// -------------------- useSmtpSsl --------------------
	
	
	public void useSmtpSsl(String username, String password, int smtpPort) throws IllegalArgumentException, MessagingException {
		useSmtpSsl(username, password, smtpPort, sslFactory_default);
	}
	
	
	/**
	* This method was inspired by this forum discussion (see the post by gary_mcm):
	*	http://forums.java.sun.com/thread.jspa?threadID=591321&messageID=3486033
	* See also:
	*	http://www.jscape.com/articles/sending_email_smtp_ssl_gmail.html
	*	http://www.javaworld.com/javatips/jw-javatip115.html
	*	http://g4j.sourceforge.net/
	*	http://gavamail.sourceforge.net/
	*	http://gavamail.sourceforge.net/phpwiki/ 
	* <p>
	* @throws IllegalArgumentException if some arg is illegal
	* @throws IllegalStateException if this instances mail.transport.protocol property is not smtp
	* @throws MessagingException if a messaging error occurs
	*/
	public void useSmtpSsl(final String username, final String password, int smtpPort, String sslFactory) throws IllegalArgumentException, IllegalStateException, MessagingException {
		if (StringUtil.isBlank(username)) throw new IllegalArgumentException("username is blank");
		if (StringUtil.isBlank(password)) throw new IllegalArgumentException("password is blank");
		if (!NetUtil.isValidPort(smtpPort)) throw new IllegalArgumentException("smtpPort = " + smtpPort + " is an illegal TCP/IP port value");
		if (StringUtil.isBlank(sslFactory)) throw new IllegalArgumentException("sslFactory is blank");
		if (!"smtp".equals(properties.getProperty("mail.transport.protocol"))) throw new IllegalStateException("illegal to call this method because this instance has mail.transport.protocol = " + properties.getProperty("mail.transport.protocol") + " which != to smtp");
				
		properties.setProperty( "mail.smtp.auth", "true" );
		//properties.setProperty( "mail.smtp.host", ... );	// not needed: already set more generally in the constructor via "mail.host" property
		properties.setProperty( "mail.smtp.port", String.valueOf(smtpPort) );
		
		properties.setProperty( "mail.smtp.socketFactory.class", sslFactory );
		properties.setProperty( "mail.smtp.socketFactory.port", String.valueOf(smtpPort) );
		properties.setProperty( "mail.smtp.socketFactory.fallback", "false" );	// prevents JavaMail from ever fallback to a default (nonsecure) connection
		properties.setProperty( "mail.smtp.starttls.enable","true" );
		// Note: TLS is an evolution of SSL (in fact, it is sometimes called SSL 3.1); see
		//	http://mrcorp.infosecwriters.com/Final-Pape-published.htm
		// or
		//	http://www.intellireach.com/products/MP_Secure_Messaging.htm
		// For more on the starttls command, see
		//	http://www.gordano.com/kb.htm?q=1450
		//	http://www.technoids.org/wwstarttls.html
		
		// Note: for a list of all the properties that can set for SMTP see
		//	http://java.sun.com/products/javamail/javadocs/com/sun/mail/smtp/package-summary.html
		//	http://www-306.ibm.com/software/webservers/appserv/doc/v40/ae/infocenter/was/0606370001.html
 
		Authenticator authenticator = new Authenticator() {
			protected PasswordAuthentication getPasswordAuthentication() {
				return new PasswordAuthentication( username, password );
			}
		};
		
		session = Session.getInstance(properties, authenticator);
		
		if (transport != null) {
			transport.close();
			transport = null;
		}			
	}
	
	
	// -------------------- send --------------------
	
	
	/**
	* Convenience method that calls the more general <code>send</code> method with null values for
	* the code>cc</code> and <code>bcc</code> arguments.
	*/
	public final void send(String subject, String messageBody, String to) throws IllegalArgumentException, MessagingException {
		send(subject, messageBody, to, null, null, null);
	}


	/**
	* This method sends an email using the info in the arguments.
	* <p>
	* <b>Warning</b>: the mail server may override the <code>mailhost</code> value to prevent spoofing?
	* <p>
	* @throws IllegalArgumentException if some arg is illegal
	* @throws MessagingException if some messaging error occurs
	* @see "The program <JavaMail install directory>/demo/msgsend.java"
	* @see <a href="http://www.stopspam.org/email/headers.html">All About Email Headers</a>
	*/
	public final void send(String subject, String messageBody, String to, String cc, String bcc, Map<String, String> headers) throws IllegalArgumentException, MessagingException {
		if ( StringUtil.isBlank(subject) && StringUtil.isBlank(messageBody) ) throw new IllegalArgumentException("supplied blank values for both subject & messageBody");
		if ( StringUtil.isBlank(to) && StringUtil.isBlank(cc) && StringUtil.isBlank(bcc) ) throw new IllegalArgumentException("supplied blank values for all of to, cc, & bcc");
		EmailUtil.checkHeaderFieldBody(to, "to");
		EmailUtil.checkHeaderFieldBody(cc, "cc");
		EmailUtil.checkHeaderFieldBody(bcc, "bcc");
				
		MimeMessage message = new MimeMessage(session);
		message.setSubject(subject);
		message.setText(messageBody);
		if (from != null)
			message.setFrom( new InternetAddress(from, parseStrictness) );
		if (to != null)
			message.setRecipients( Message.RecipientType.TO, InternetAddress.parse(to, parseStrictness) );
		if (cc != null)
			message.setRecipients( Message.RecipientType.CC, InternetAddress.parse(cc, parseStrictness) );
		if (bcc != null)
			message.setRecipients( Message.RecipientType.BCC, InternetAddress.parse(bcc, parseStrictness) );
		if (headers != null) {
			for (Map.Entry<String,String> entry : headers.entrySet()) {
				message.setHeader( entry.getKey(), entry.getValue() );
			}
		}
		message.setSentDate( new Date() );

		if (transport == null) {
			transport = session.getTransport();
			for (ConnectionListener cListener : connectionListeners) {
				transport.addConnectionListener(cListener);
			}
			for (TransportListener tListener : transportListeners) {
				transport.addTransportListener(tListener);
			}
		}			
		if (!transport.isConnected()) {
			transport.connect();
		}
				
		transport.sendMessage(message, message.getAllRecipients());
	}
// +++ add support for sending file attachments; see http://www-106.ibm.com/developerworks/forums/dw_thread.jsp?forum=262&thread=63790&cat=11
// +++ further action that could do: save a record of the email sent; see the msgsend.java sample code from Sun
	
	
	// -------------------- close --------------------
	
	
	/** Calls {@link #close close} but catches any Throwable that is raised and logs it to logger. */
	public final void close(Logger logger) {
		try {
			close();
		}
		catch (Throwable t) {
			logger.logp(Level.SEVERE, "Emailer", "close", "an error occured while attempting to close an Emailer instance", t);
		}
	}
	
	
	/**
	* Closes all messaging resources currently being used by this instance.
	* <p>
	* @throws MessagingException if some messaging error occurs
	*/
	public final void close() throws MessagingException {
		if (transport != null) transport.close();

		try {
			Thread.sleep( 10 );	// must sleep for a short amount of time if want a better chance of any ConnectionListeners receiving the close event
		}
		catch (InterruptedException ie) {
			Thread.currentThread().interrupt();	// never swallow this; see http://www-128.ibm.com/developerworks/java/library/j-jtp05236.html
		}
		
		for (ConnectionListener cListener : connectionListeners) {
			if (cListener instanceof Listener) ((Listener) cListener).close();
		}
		for (TransportListener tListener : transportListeners) {
			if (tListener instanceof Listener) ((Listener) tListener).close();
		}
	}
	
	
	// -------------------- Listener (static inner class) --------------------
	
	
	public static class Listener implements ConnectionListener, TransportListener {
		
		private final Logger logger;
		
		public Listener(Logger logger) throws IllegalArgumentException {
			if (logger == null) throw new IllegalArgumentException("logger == null");
			
			this.logger = logger;
		}
		
		public void opened(ConnectionEvent ce) { handleConnectionEvent(ce); }
		public void disconnected(ConnectionEvent ce) { handleConnectionEvent(ce); }
		public void closed(ConnectionEvent ce) { handleConnectionEvent(ce); }
		private void handleConnectionEvent(ConnectionEvent ce) {
			logger.log( Level.INFO, EmailUtil.eventToString(ce) );
		}

		public void messageDelivered(TransportEvent te) { handleTransportEvent(te); }
		public void messageNotDelivered(TransportEvent te) { handleTransportEvent(te); }
		public void messagePartiallyDelivered(TransportEvent te) { handleTransportEvent(te); }
		private void handleTransportEvent(TransportEvent te) {
			try {
				logger.log( Level.FINER, EmailUtil.eventToString(te) );
			}
			catch (Throwable t) {
				logger.logp(Level.SEVERE, "Emailer.Listener", "handleTransportEvent", "Caught an unexpected Throwable", t);
			}				
		}

		public void close() {
			LogUtil.close( logger );
// +++ when sun gets back to me on my bug report, let them know that i solved it if use close instead of flush; but why does logger not have this method?
		}
		
	}
	
	
	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		private static final File fileProperties = new File("../../../../../ellipticGroup/KEEP_PRIVATE_SENSITIVE_INFO/gmailInfo.txt");
		
		private static final String validAddresses = "ellipticgroupinc@gmail.com";
		private static final String imaginaryAddresses = "32423t44t@1gi9erg389.com, 32jksad8957h@fljhjsnk.com";
		private static final String malformedAddresses = "\u234389.com, \u73542@fljhjsnk.com";
		private static final String addressesGoodAndBad = validAddresses + ", " + imaginaryAddresses + ", " + malformedAddresses;
		
		private Emailer emailer;
		
		@Before public void setup() throws Exception {
			emailer = new Emailer(fileProperties);
		}
		
		@After public void teardown() throws Exception {
			if (emailer != null) emailer.close();
		}
		
		@Test public void test_send_shouldWork() throws Exception {
			Logger2 logger = LogUtil.makeLogger2( Emailer.UnitTest.class, "test_send_shouldWork" );
			Listener listener = new Listener(logger);
			emailer.addConnectionListener( listener );
			emailer.addTransportListener( listener );
			emailer.send("UnitTest message#1", "Sent by test_send_shouldWork ...", validAddresses);
		}
		
		@Test(expected=IllegalArgumentException.class) public void test_send_shouldFail() throws Exception {
			Logger2 logger = LogUtil.makeLogger2( Emailer.UnitTest.class, "test_send_shouldFail" );
			Listener listener = new Listener(logger);
			emailer.addConnectionListener( listener );
			emailer.addTransportListener( listener );
			emailer.send("UnitTest message#2", "Sent by test_send_shouldFail ...", addressesGoodAndBad);
		}
		
	}


}
