/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.net;


import java.io.*;
import java.net.*;
import java.util.*;


/**
* This class provides miscellaneous static utility methods for dealing with the http protocol.
* <p>
* This class is multithread safe: it is stateless.
* <p>
* @see HttpParameters
* @author Brent Boyer
*/
public class HttpUtil {
	
	
	// -------------------- sendPostMessage --------------------
	
	
	/**
	* This method:
	* <ol>
	*  <li>sets the request method of the connection arg to "POST"</li>
	*  <li>adds the request property "Content-Type"/"application/x-www-form-urlencoded"</li>
	*  <li>writes <code>parameters.toPostParameterString()</code> to connection's OutputStream</li>
	* </ol>
	* <p>
	* Note: this method performs no cleanup actions (e.g. closing the OutputStream).
	* <p>
	* @throws IllegalArgumentException if args connection or parameters are null
	* @throws IOException if an I/O problem occurs
	*/
	public static void sendPostMessage(HttpURLConnection connection, HttpParameters parameters) throws IllegalArgumentException, IOException {
		if (connection == null) throw new IllegalArgumentException("arg connection is null");
		if (parameters == null) throw new IllegalArgumentException("arg parameters is null");
		
		connection.setRequestMethod("POST");
		connection.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
		
		//DataOutputStream dos = new DataOutputStream( connection.getOutputStream() );
		//dos.writeBytes( parameters.toPostParameterString() );	// is safe to call writeBytes, which ignores high byte of each char -- see the HttpParameters.toFormUrlencodedString contract
		//dos.flush();
			// cannot use the lines above: FindBugs complains that dos is never closed; cannot close it because that would affect connection...
		OutputStream os = connection.getOutputStream();
		String s = parameters.toPostParameterString();
		int len = s.length();	// this and the next 3 lines were copied from DataOutputStream.writeBytes
		for (int i = 0 ; i < len ; i++) {
			os.write( (byte) s.charAt(i) );	// it is safe to ignore the high byte of each char -- see the HttpParameters.toFormUrlencodedString contract
		}
		os.flush();
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private HttpUtil() {}


}
