/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.net;


import java.io.*;
import java.net.*;
import java.text.ParseException;

import org.junit.Assert;
import org.junit.Test;

import bb.io.FileUtil;
import bb.io.StreamUtil;
import bb.util.StringUtil;
import bb.util.ThrowableUtil;
import bb.util.TimeLength;


/**
* This class provides static utility methods for dealing with {@link URL}s.
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @author Brent Boyer
*/
public class UrlUtil {
	
	
	// -------------------- constants --------------------
	
	
	private static final String fileProtocol = "file";
//	private static final String ftpProtocol = "ftp";
//	private static final String gopherProtocol = "gopher";
	private static final String httpProtocol = "http";
	private static final String httpsProtocol = "https";
	private static final String mailtoProtocol = "mailto";
//	private static final String newsProtocol = "news"; 
	
	
	// -------------------- extractUrl --------------------
	
	
	/**
	* This utility method first strips any quote marks from urlString via a call to removeQuotes,
	* and then returns a new URL instance based off of urlString and context.
	* <p>
	* @param urlString a String which should specify a URL
	* @param lineNumber the line number where urlString is found; used only if throw a ParseException
	* @param context the context URL; used to resolve relative URLs; may be null if is not to be specified
	* @throws IllegalArgumentException if urlString is null, or lineNumber < 0
	* @throws ParseException if a leading but no matching trailing quote mark is present, or vice versa, on urlString
	* @throws MalformedURLException if urlString cannot be parsed into a proper URL specification
	*/
	public static URL extractUrl(String urlString, int lineNumber, URL context) throws IllegalArgumentException, ParseException, MalformedURLException {
		urlString = StringUtil.removeQuotes( urlString, lineNumber );
		return new URL(context, urlString);
	}
	
	
	// -------------------- drain, drainIntoString, getInputStream --------------------
	
	
	/**
	* This method returns <code>{@link StreamUtil#drain StreamUtil.drain}( {@link #getInputStream getInputStream}(url) )</code>.
	* <p>
	* @throws IllegalArgumentException if url == null
	* @throws IllegalStateException if url's inputStream turns out to hold more than {@link Integer#MAX_VALUE} bytes (which cannot be held in a java array)
	* @throws IOException if an I/O problem occurs
	* @throws UnknownServiceException if the protocol does not support input
	*/
	public static byte[] drain(URL url) throws IllegalArgumentException, IllegalStateException, IOException, UnknownServiceException {
		return StreamUtil.drain( getInputStream(url) );	// Note: StreamUtil.drain closes the InputStream
	}
	
	
	/**
	* This method returns <code>{@link StreamUtil#drainIntoString StreamUtil.drainIntoString}( {@link #getInputStream getInputStream}(url) )</code>.
	* <p>
	* @throws IllegalArgumentException if url == null
	* @throws IllegalStateException if url's inputStream turns out to hold more than {@link Integer#MAX_VALUE} bytes (which cannot be held in a java array)
	* @throws IOException if an I/O problem occurs
	* @throws UnknownServiceException if the protocol does not support input
	*/
	public static String drainIntoString(URL url) throws IllegalArgumentException, IllegalStateException, IOException, UnknownServiceException {
		return StreamUtil.drainIntoString( getInputStream(url) );	// Note: StreamUtil.drainIntoString (via StreamUtil.drain) closes the InputStream
	}
	
	
	/**
	* This method creates a URLConnection for url, sets reasonably long connect and read timeouts on it,
	* opens an InputStream from it, and returns the InputStream.
	* <p>
	* @throws IllegalArgumentException if url == null
	* @throws IOException if an I/O problem occurs
	* @throws UnknownServiceException if the protocol does not support input
	*/
	public static InputStream getInputStream(URL url) throws IllegalArgumentException, IOException, UnknownServiceException {
		if (url == null) throw new IllegalArgumentException("arg url is null");
		
		URLConnection conn = url.openConnection();
		conn.setConnectTimeout( (int) (5 * TimeLength.MINUTE) );	// originally used Integer.MAX_VALUE, but found that some websites could stall forever, so aborted this
		conn.setReadTimeout( (int) (5 * TimeLength.MINUTE) );	// originally used Integer.MAX_VALUE, but found that some websites could stall forever, so aborted this
		conn.connect();
		return conn.getInputStream();
	}
	
	
	// -------------------- analyseUrl and helper methods --------------------
	
	
	/**
	* This method analyses the supplied URL. The exact analysis performed is completely
	* protocol dependent, but it typically involves trying to follow the link and see
	* if valid data is on the other end.

<!--
Note that most analyses typically begin with either "+Analysis outcome: "
or "-Analysis outcome: ". The "+" sign indicates success, while the "-" sign
indicates some kind of warning or problem. The reason why I do this is so
that you can do a search thru the output file for all the "-A" occurrences,
since those are generally the ones that you are concerned about.

In the future, need a better way of telling the user what are good and bad links.
-->

	* <p>
	* @param url the URL to analyse
	* @throws IllegalArgumentException if url == null
	* @throws IOException if an I/O problem occurs
	* @see <a href="http://lava.net/support/urls.html">List of some protocols</a>
	* @see <a href="http://www.faqs.org/rfcs/rfc2396.html">RFC2396 Uniform Resource Identifiers (URI): Generic Syntax</a>
	*/
	public static String analyseUrl(URL url) throws IllegalArgumentException {
		if (url == null) throw new IllegalArgumentException("url == null");
		
		String protocol = url.getProtocol();	// NOTE: hostnames and protocols of URLs are not case sensitive (http://webtips.dan.info/url.html), so the code below needs to handle this
		
		if ( fileProtocol.equalsIgnoreCase(protocol) )
			return analyseFileUrl(url);
		
		else if ( httpProtocol.equalsIgnoreCase(protocol) || httpsProtocol.equalsIgnoreCase(protocol) )
			return analyseHttpUrl(url);
// +++ is it right to handle https in the same way as http?
		
		else if ( mailtoProtocol.equalsIgnoreCase(protocol) )
			return analyseMailtoUrl(url);
		
		else
			return analyseUnsupportedUrl(url);
	}
	
	
	private static String analyseFileUrl(URL url) {
		try {
			URLConnection conn = url.openConnection();
			conn.connect();
			conn.getInputStream();
			//conn.getOutputStream();	// WRONG!  the file protocol does not support writing
			
			return	"+Analysis outcome: SUCCESS -- was able to open a URL connection, " +
					"and access an InputStream from the file";
		}
		catch (Throwable t) {
			return	"-Analysis outcome: INTERRUPTED -- the following Throwable was thrown during analysis " +
					"(will need to manually analyse):" + "\n" +
					ThrowableUtil.toString(t);
		}
	}
	
	
	private static String analyseHttpUrl(URL url) {
		try {
			HttpURLConnection conn = (HttpURLConnection) url.openConnection();
			
//conn.setInstanceFollowRedirects(false);
// +++ uncomment the above line if want a stricter check in which redirects are NOT to be followed
			
			conn.connect();
			int statusCode = conn.getResponseCode();
			
			if ( (100 <= statusCode) && (statusCode <= 199) )
				return	"-Analysis outcome: INDETERMINATE -- was able to open an HTTP connection, " +
						"but received a continuation notification from the server " +
						"(http status code = " + statusCode + ")";
						
			else if ( (200 <= statusCode) && (statusCode <= 299) )
				return	"+Analysis outcome: SUCCESS -- was able to open an HTTP connection, " +
						"and received a success notification from the server " +
						"(http status code = " + statusCode + ", http response message = " + conn.getResponseMessage() + ")";
// +++ if want to see the actual data sent back, use the drainIntoString method from above
			
			else if ( (300 <= statusCode) && (statusCode <= 399) )
				return	"-Analysis outcome: WARNING -- was able to open an HTTP connection, " +
						"but received a warning notification from the server " +
						"(http status code = " + statusCode + ", http response message = " + conn.getResponseMessage() + ")";
			
			else
				return	"-Analysis outcome: FAILURE -- was able to open an HTTP connection, " +
						"but received an error notification from the server " +
						"(http status code = " + statusCode + ", http response message = " + conn.getResponseMessage() + ")";
		}
		catch (Throwable t) {
			return	"-Analysis outcome: INTERRUPTED -- the following Throwable was thrown during analysis " +
					"(will need to manually analyse):" + "\n" +
					ThrowableUtil.toString(t);
		}
	}
	
	
	private static String analyseMailtoUrl(URL url) {
		return	"-Analysis outcome: UNKNOWN -- no analysis of the mailto protocol is performed";
	}
	
	
// +++ could also report back what the getContentLength() method says?
	private static String analyseUnsupportedUrl(URL url) {
		try {
			URLConnection conn = url.openConnection();
			conn.connect();
			
			return	"-Analysis outcome: UNCLEAR -- was to open a URL connection, " +
					"but no further analysis could be done because this URL's protocol " +
					"is not currently supported";
		}
		catch (Throwable t) {
			return	"-Analysis outcome: INTERRUPTED -- the following Throwable was thrown during analysis " +
					"(will need to manually analyse):" + "\n" +
					ThrowableUtil.toString(t);
		}
	}
	
	
	// -------------------- description --------------------
	
	
	/**
	* Returns a string that describes every aspect of url.
	* <p>
	* @param url a URL
	* @throws IllegalArgumentException if url == null
	*/
	public static String description(URL url) throws IllegalArgumentException {
		if (url == null) throw new IllegalArgumentException("url == null");
		
		return
			"Authority: " + url.getAuthority() + "\n" +
			"File: " + url.getFile() + "\n" +
			"Host: " + url.getHost() + "\n" +
			"Path: " + url.getPath() + "\n" +
			"Port: " + url.getPort() + "\n" +
			"Protocol: " + url.getProtocol() + "\n" +
			"Query: " + url.getQuery() + "\n" +
			"Ref: " + url.getRef() + "\n" +
			"UserInfo: " + url.getUserInfo() + "\n" +
			"url.toExternalForm: " + url.toExternalForm() + "\n" +
			"url.toString: " + url.toString();
//			"url.toURI: " + url.toURI();
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private UrlUtil() {}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_drainXXX_getInputStream() throws Exception {
			String path = "../src/bb/net/UrlUtil.java";
			String sFile = FileUtil.readString( new File(path) );
			
			URL url = extractUrl("file:" + path, 0, null);	// this choice yields repeatable results
			
			String sDrain = new String( drain(url) );
			Assert.assertEquals(sFile, sDrain);
			
			String sDrainIntoString = drainIntoString(url);
			Assert.assertEquals(sFile, sDrainIntoString);
		}
		
		@Test public void test_extractUrl_analyseUrl() throws Exception {
			System.out.println();
			System.out.println("----------Valid URLS----------");
			
			URL url = extractUrl("http://www.apple.com", 0, null);
			doAnalysis(url, "+Analysis outcome: ");
			
			System.out.println();
			File fileRelative = new File("../src/bb/net/UrlUtil.java");
			url = extractUrl("file:" + fileRelative.getPath(), 0, null);	// Note: if want to use relative paths for the file protocol, do NOT specify the // as then would need to also specify the host and port followed by an absolute path (see example below)
			doAnalysis(url, "+Analysis outcome: ");
			
			System.out.println();
			File fileCanonical = fileRelative.getCanonicalFile();
			url = extractUrl("file:///" + fileCanonical.getPath(), 0, null );	// Note: there is a third / immediately after the first 2 /s -- this means local host and default port; we must follow this with the start of an absolute path
			doAnalysis(url, "+Analysis outcome: ");
			
			System.out.println();
			System.out.println("----------INVALID URLS----------");
			
			System.out.println();
			url = new URL("http://www.ibm.com/kjdnssdvksdvsdvsdkkkj.html");
			doAnalysis(url, "-Analysis outcome: ");
			
			System.out.println();
			url = new URL("file:./dvdsvsvdvsdvsdvsdvsd");
			doAnalysis(url, "-Analysis outcome: ");
		}
		
		private void doAnalysis(URL url, String prefixExpected) throws Exception {
			System.out.println();
			System.out.println(url);
			String analysis = analyseUrl(url);
			System.out.println(analysis);
			Assert.assertTrue( analysis.startsWith(prefixExpected) );
		}
		
	}
	
	
}


/*
The code below was added when experienced all kinds of timeout issues with drainIntoString.
These issues seem to have gone away, so am removing this hack code for now...

	**
	* This method opens a socket to the host and port specified by url,
	* creates an input stream (wrapped in a PrintWriter) from it,
	* and sends the http headers appropriate for requesting the file specified by url.
	* In addition, the http header specifies that the response type must be one of the standard text types,
	* and that the connection will be closed when this method is done (i.e. no keep alive).
	* It then opens an output stream from the socket and drains all the bytes into a String which is returned.
	* <p>
	* This method was introduced because of all the time out problems that was having with Sun's URLConnection; see my report:
	*	http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6319814
	* <p>
	* It was inspired by:
	*	http://www.onjava.com/pub/a/onjava/2003/04/23/java_webserver.html
	* and
	*	http://computing.dcu.ie/~humphrys/Notes/Networks/java.html
	*	(see section on "The "URL" class hides the sockets")
	* <p>
	* See also
	*	http://jan.netcomp.monash.edu.au/distjava/socket/lecture.html
	* <p>
	* If Sun ever listens to my bug report
	* and fixes URLConnection, then this method becomes obsolete.
	* <p>
	* @throws IllegalArgumentException if url == null
	* @throws IllegalStateException if the socket's inputStream turns out to hold more than {@link Integer#MAX_VALUE} bytes (which cannot be held in a java array)
	* @throws IOException if an I/O problem occurs
	*
	public static String drainHttpIntoString(URL url) throws IllegalArgumentException, IllegalStateException, IOException {
		if (url == null) throw new IllegalArgumentException("arg url is null");
		
		Socket socket = null;
		try {
			int port = (url.getPort() != -1) ? url.getPort() : url.getDefaultPort();
//			socket = new Socket(url.getHost(), port );
// The above line often fails on my machine with a connect timeout, especially when many threads are trying to create sockets.
// This appears to be due to:
//		http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6209838
// But see also:
//		http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6224843
//		http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=5092063
//		http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4954495
// So, until the above gets fixed, try:
socket = new Socket(Proxy.NO_PROXY);
//socket.setPerformancePreferences(1, 0, -1);	// means prefer connectionTime over latency over bandwidth
InetSocketAddress socketAddress = new InetSocketAddress(url.getHost(), port );
socket.connect(socketAddress, Integer.MAX_VALUE);
// no dice: the above still suffers from ConnectExceptions, so the only (partial) solution is to reduce the number of threads...
// +++ if the above does not make any difference, then why am I still using it? (comment added on 2007/11/8 when relooked at this code...)
			
			PrintWriter pw = new PrintWriter( socket.getOutputStream(), false );
			pw.print("GET " + url.getFile() + " HTTP/1.0" + "\r\n");
			pw.print("Accept: text/plain, text/html, text/*" + "\r\n");
			pw.print("Connection: Close" + "\r\n");	// am not doing any keep alive
			pw.print("\r\n");
			pw.flush();
			
			return StreamUtil.drainIntoString( socket.getInputStream() );	// Note: StreamUtil.drainIntoString (via StreamUtil.drain) closes the InputStream
		}
		finally {
			NetUtil.close(socket);
		}	
	}
*/