/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*
Programmer notes:

+++ future evolution of this class:
	--DataStorer should be an interface instead
	--the implementation below is meant for speed: it has unsynchronized methods, data is unsorted but stored as added 
	--but other subclasses could make different guarantees, such as: thread safe, sorted data, resettable, etc
	--since it might be nice to subclass this class, more members should have protected access

+++ REALLY need to move all the bins code to its own top level Bins class,
since binning is more general than just this class.
This new class:
	--would obviously store all the data currently stored in the double[][] that is now being returned from toBins,
		altho it possibly might store it in a different underlying form, such as 2 individual double[]
		or some kind of point array
	--the current toBins methods below would be constructors in the new class
		--need to add javadocs to the rest of the toBins methods
	--would store what the bin interval width is, so would know what this is even if there is just 1 interval
	--would have all the utility methods inside it like binShift etc
	--should offer 3 normalization modes: none, normalize to probability, and normalize to probability density
		--currently normalize always means normalize to probability density
		--should introduce an enum with 3 states to specify this: none, toProbability, toProbabilityDensity

+++ REALLY need to move fitGaussian to Math2,
since that fitting is more general than just this class.
*/


package bb.science;


import java.util.*;
import java.io.File;

import org.junit.Test;

import jaolho.data.lma.LMA;
import jaolho.data.lma.LMAFunction;

import bb.io.FileUtil;
import bb.util.NumberUtil;
import bb.util.StringUtil;


/**
* Class which simply stores double values via the {@link #add add} method.
* <p>
* The current implementation uses a double[], which is resized as necessary, to hold all the values.
* This yields top performance, both for cpu and memory usage.
* A List implementation (e.g. an ArrayList) could be used instead of this class,
* however, performance would greatly suffer because Lists store objects (e.g. Doubles).
* So, this class is suitable for the most demanding applications.
* <p>
* One limitation of this class's use of an array
* is that the add method can only accept up to {@link Integer#MAX_VALUE} values.
* For discussions of the array size restrictions in Java, see
* <a href="http://java.sun.com/docs/books/jls/third_edition/html/expressions.html#15.10">the language specification</a>
* (<q>...this means, specifically, that the type of a dimension expression must not be long</q>) or 
* <a href="http://forum.java.sun.com/thread.jspa?forumID=31&threadID=598705">this forum discussion</a>.
<!--
+++ would a LinkedList implementation get around this (at the cost of vastly more memory used per node)?
-->
* <p>
* This class only accepts normal (non-NaN and non-infinite) doubles.
* This restriction means that the result returned by {@link #values values}
* may be safely supplied to other classes (e.g. the statistical routines inside {@link Math2}).
* <p>
* This class offers several methods useful for analysis:
* several versions of toBins
* (see the {@link #toBins(double, double, double, double, boolean) fundamental version} for documentation)
* and {@link #fitGaussian fitGaussian}.
* <p>
* This class is not multithread safe.
* <p>
* @author Brent Boyer
*/
public class DataStorer {
	
	
	// -------------------- constants --------------------
	
	
	/** Specifies the default value for the initial size of the values buffer. */
	private static final int sizeInitial_default = 1024;
	
	
	/** Specifies the default value for the normalization mode. */
	private static final boolean normalize_default = true;
	
	
	private static final int integerMaxDiv2 = Integer.MAX_VALUE / 2;
	
	
	// -------------------- fields --------------------
	
	
	/** Buffer which holds the data values. */
	private double[] values = null;
	
	
	/** Stores the number of data values recorded by this instance. Hence, n - 1 is the last valid index of {@link #values}. */
	private int n = 0;
	
	
	// -------------------- constructors --------------------
	
	
	/** Constructor which allocates an initial buffer with size {@link #sizeInitial_default}. */
	public DataStorer() {
		this(sizeInitial_default);
	}
	
	
	/**
	* Constructor which allocates an initial buffer with size specified by sizeInitial.
	* <p>
	* @throws IllegalArgumentException if sizeInitial <= 0
	*/
	public DataStorer(int sizeInitial) throws IllegalArgumentException {
		if (sizeInitial <= 0) throw new IllegalArgumentException("sizeInitial = " + sizeInitial + " <= 0");
		
		values = new double[sizeInitial];
	}
	
	
	// -------------------- add and helper methods --------------------
	
	
	/**
	* Adds d to the internal array which stores all the values.
	* This array will be resized if necessary in order to hold d.
	* <p>
	* @throws IllegalArgumentException if {@link NumberUtil#checkNormal NumberUtil.checkNormal} objects to d
	* @throws IllegalStateException if the buffer's length = {@link Integer#MAX_VALUE}
	*/
	public void add(double d) throws IllegalArgumentException, IllegalStateException {
		Check.arg().normal(d);
		
		if (n == values.length) increaseBuffer();
		values[n++] = d;
	}
	
	
	/**
	* Increases the capacity of the values buffer.
	* The current implementation doubles the capacity (within an upper bound of {@link Integer#MAX_VALUE}).
	* <p>
	* @throws IllegalStateException if the buffer's length = {@link Integer#MAX_VALUE}
	*/
	private void increaseBuffer() throws IllegalStateException {
		if (values.length == Integer.MAX_VALUE) throw new IllegalStateException("cannot resize the buffer, since its length already = Integer.MAX_VALUE = " + Integer.MAX_VALUE + " (Java restricts arrays to int addressing)");
		
		int lengthNext = (values.length < integerMaxDiv2) ? 2 * values.length : Integer.MAX_VALUE;
		double[] valuesNew = new double[ lengthNext ];
		System.arraycopy(values, 0, valuesNew, 0, values.length);
		values = valuesNew;
	}
	
	
	// -------------------- accessors and mutators --------------------		
	
	
	/**
	* Returns a new array which holds all the values that have been added to this instance.
	* <p>
	* The implementation here returns an array which has the values in the order that they were added.
	* <p>
	* Contract: the result is never null, but will be zero-length if no values have ever been added.
	*/
	public double[] values() {
		double[] valuesCopy = new double[n];
		System.arraycopy(values, 0, valuesCopy, 0, n);
		return valuesCopy;
	}
	
	
	/** Returns the number of values that have been added to this instance. */
	public int size() { return n; }
	
	
/*
	** Clears all previously added values from this instance. *
	public void reset() {
		n = 0;
	}
Have decided for now that it is a bad idea to have this method: user should simply create a new instance instead.
+++ see Programmer notes at beginning of class for other alternatives.

If ever have a subclass which has this method, then need to modify the javadocs of many of the above methods to read like this:
	* Returns a new array which holds all the values that are currently being stored by this instance.
	* This is the values added since the last call to {@link #reset reset}, if it has ever been called,
	* or since construction, if reset has never been called.
*/
	
	
	// -------------------- toBins and helper methods, binShift --------------------
	
	
	// FIRST group of toBins methods: the number of bins is specified; the bin width is calculated so as to fit this
	
	
	public double[][] toBins(int number) throws IllegalArgumentException, IllegalStateException {
		return toBins(number, normalize_default);
	}
	
	
	public double[][] toBins(int number, boolean normalize) throws IllegalArgumentException, IllegalStateException {
		if (n == 0) throw new IllegalStateException("no values have been added to this instance, so it is impossible to determine bin boundaries");
		
		double[] minMax = Math2.minMax( values() );
		double min = minMax[0];
		double max = minMax[1];
		return toBins(number, min, max, normalize);
	}
	
	
	public double[][] toBins(int number, double begin, double end) throws IllegalArgumentException, IllegalStateException {
		return toBins(number, begin, end, normalize_default);
	}
	
	
	/**
* ...
	* <p>
	* Note that the last bin interval should normally be <code>[end, end + width)</code>.
	* However, floating point errors could cause the last interval to be different than this.
	*/
	public double[][] toBins(int number, double begin, double end, boolean normalize) throws IllegalArgumentException, IllegalStateException {
		if (number < 2) throw new IllegalArgumentException("number = " + number + " < 2");
		if (begin == end) throw new IllegalArgumentException("begin = end = " + begin + ", which makes it impossible to determine where to place the bins");
		
		double width = (end - begin) / (number - 1);	// divide by number - 1 so that the last bin interval is [end, end + width)
		
		return toBins(begin, width, begin, end, normalize);
	}
	
	
	// SECOND group of toBins methods: the bin width is specified; the number of bins is calculated so as to fit this
	
	
	public double[][] toBins(double width) throws IllegalArgumentException, IllegalStateException {
		return toBins(width, normalize_default);
	}
	
	
	public double[][] toBins(double width, boolean normalize) throws IllegalArgumentException, IllegalStateException {
		double[] minMax = Math2.minMax( values() );
		double min = minMax[0];
		double max = minMax[1];
		return toBins(min, width, min, max, normalize);
	}
	
	
	public double[][] toBins(double offset, double width) throws IllegalArgumentException, IllegalStateException {
		return toBins(offset, width, normalize_default);
	}
	
	
	public double[][] toBins(double offset, double width, boolean normalize) throws IllegalArgumentException, IllegalStateException {
		if (n == 0) throw new IllegalStateException("no values have been added to this instance, so it is impossible to determine bin boundaries");
		
		double[] minMax = Math2.minMax( values() );
		double min = minMax[0];
		double max = minMax[1];
		return toBins(offset, width, min, max, normalize);
	}
	
	
	public double[][] toBins(double offset, double width, double includeStart, double includeFinish) throws IllegalArgumentException, IllegalStateException {
		return toBins(offset, width, includeStart, includeFinish, normalize_default);
	}
	
	
	/**
	* <h3>Comments for all the toBins methods</h3>
	* The various versions of the toBins method
	* all return a bin representation of the stored values which measures their distribution.
	
	* <h4>Bins</h4>
	* A bin is defined here to be an interval in the value space plus an associated number.
	
	* <h5>Bin intervals</h5>
	* The toBins methods always use the following semi-open bin interval convention:
	* each bin's interval is of the form <code>[x1, x2)</code>,
	* that is, it includes all values <code>x</code> that satisfy <code>x1 <= x < x2</code>.
	* So, the left boundary is included inside the interval, but the right boundary is not.
	* <p>
	* The toBins methods always use a finite set of contiguous intervals.
	* These can be written, in order, as <code>[x1, x2), [x2, x3), [x3, x4), ..., [xN, xN+1)</code>
	* where <code>N</code> is the total number of intervals.
	* <p>
	* A simplifying convention when speaking of contiguous bins like this
	* is to only list the beginning of each bin,
	* since each bin's end is listed as the beginning of the subsequent bin.
	* Thus, we designate the previous set of intervals as <code>{x1, x2, x3, ..., xN}</code>.
	* Taken as a whole, these bins cover the single larger semi-open interval <code>[x1, xN+1)</code>.
	* This designation convention is used by the result.
	* <p>
	* The toBins methods always guarantee that the set of bin intervals will cover every stored value.
	* <p>
	* Furthermore, the bins are equally sized (within rounding error).
	* So, the bin intervals can be designated as <code>{x1, x1 + width, x1 + 2*width), ..., x1 + N*width}</code>.
	
	* <h5>Bin numbers</h5>
	* The number associated with each bin by toBins is some measure of the bin's population.
	* <p>
	* If normalization is not requested, then the bin number is simply the bin count
	* (i.e. how many values fall inside the bin).
	* <p>
	* Otherwise, if normalization is requested, then the bin number is the bin's probability density,
	* which is the bin count times <code>(1/n)*(1/width)</code>.
	* Here, <code>n</code> is the number of stored values
	* (so <code>1/n</code> converts the bin count to a probabilty)
	* and width is the bin's width (so <code>1/width</code> converts the probability into a probability density).
	* Thus, a probability density function (<code>pdf</code>) may be defined as
	* <code>pdf(x) = the number associated with the ith bin, if x lies in it, and 0 if x lies outside the bins</code>.
	* The integral of <code>pdf(x)</code> over all <code>x</code> should be 1 (within rounding error).
	* <p>
	* The reason why normalization is so useful is that the <code>pdf</code> curve becomes approximately independent
	* both of the number of stored values (when sufficiently many have been added)
	* and of the bin width (when the bin width is sufficiently fine),
	* assuming that the values are drawn from a stable distribution.
	* <p>
	* Note that many versions of toBins have a normalize param to allow the user to specify the behavior.
	* All versions of toBins which lack this param have normalization behavior specified by {@link #normalize_default}.
	
	* <h4>Format of the result</h4>
	* The result is a <code>double[][]</code> that is always a <code>2 by N</code> matrix
	* with the x values in the first subarray array and the y values in the second subarray.
	* Written as code, if <code>double[][] bins = storer.toBins(...)</code>,
	* then <code>double[] xValues = bins[0]</code> and <code>double[] yValues = bins[1]</code>
	* and <code>xValues.length = yValues.length</code> always.
	* <p>
	* The x values in that first subarray use the simplifying designation convention described above
	* (i.e. the x values are the left boundary points of each interval).
	* The y values are the bin numbers described above.
	* <p>
	* If the user wants to use some other point inside the interval to characterise it instead of the left boundary,
	* (e.g. the midpoint or right boundary), then the {@link #binShift binShift} method should be called. 
	
	* <h4>The different toBins methods</h4>
	* The main difference between the various versions is in how they define the bin boundaries.
	* The first group of toBins methods specifies how many bins should be present;
	* the bin width is then a dependent variable, and is calculated so as to yield the desired number of bins.
	* The second group of toBins methods specifies the bin width;
	* the number of bins is then a dependent variable, and is calculated so as to yield the desired bin width.
	
	* <h4>Floating point error</h4>
	* Because doubles are used for all the calculations,
	* <i>it is possible that floating point error will cause unexpected results</i>.
	* For instance, with those toBins methods which specify how many bins should be present,
	* <i>errors could cause the length of the result to differ</i>.
	* Similarly, those toBins methods which specify the bin interval width
	* <i>may end up with intervals whose actual width varies</i>.
	* So, the user may need to inspect the result, especially if the numbers are extreme double values.
	
	* <h3>Comments for this particular toBins method</h3>
	* This is the fundamental version of toBins:
	* all the other versions eventually end up calling it to do most of their work.
	* <p>
	* The bin interval width is directly supplied as a param to this method.
	* <p>
	* To understand how the bin boundaries are determined,
	* first imagine that the value space is partitioned into an infinite number of bin intervals.
	* <i>The fundamental bin boundary is always specified by the offset param,</i>
	* while the remaining bin boundaries occur at integral multiples of width away from offset.
	* <p>
	* The actual bin intervals used by this method is that finite subset of the aforementioned infinite set
	* whose first interval contains the includeStart param
	* and whose last interval contains the includeFinish param.
	* <p>
	* Note that offset need not lie inside the actual intervals present in the result
	* (i.e. it can fall outside of [includeStart, includeFinish]).
	* Furthermore, includeStart and includeFinish may occur anywhere inside their respective intervals:
	* they need not be boundary points themselves.
	* <p>
	* The motivation for supplying all of these params to specify the intervals
	* is that it allows a program to always get a consistent set of intervals
	* that does not depend on the data that happens to have been stored.
	
	* <p>
	* @throws IllegalArgumentException if {@link NumberUtil#checkNormalAndPositive NumberUtil.checkNormalAndPositive} objects to width;
	* if {@link NumberUtil#checkNormal NumberUtil.checkNormal} objects to offset;
	* if {@link NumberUtil#checkNormal NumberUtil.checkNormal} objects to includeStart;
	* if includeStart > min;
	* if {@link NumberUtil#checkNormal NumberUtil.checkNormal} objects to includeFinish;
	* if includeFinish &lt; max;
	* if includeStart > includeFinish
	* @throws IllegalStateException if normalize == true and no values have ever been added to this instance;
	* any internal problem is encountered
	*/
	public double[][] toBins(double offset, double width, double includeStart, double includeFinish, boolean normalize) throws IllegalArgumentException, IllegalStateException {
		Check.arg().normalPositive(width);
		Check.arg().normal(offset);
		Check.arg().normal(includeStart);
		Check.arg().normal(includeFinish);
		if (includeStart > includeFinish) throw new IllegalArgumentException("includeStart = " + includeStart + " > includeFinish = " + includeFinish);
		if (normalize && (n == 0)) throw new IllegalStateException("normalize == true but no values have ever been added to this instance, so it is impossible to have the result cannot be a proper pdf (integrate to 1)");
		
		int markStart = mark(includeStart, offset, width);
		int markFinish = mark(includeFinish, offset, width);
		
		long nBins = ((long) markFinish) - ((long) markStart) + 1L;
		if (nBins > Integer.MAX_VALUE) throw new IllegalStateException("calculated numberBins = " + nBins + " which is too large to be represented as an int");
		int numberBins = (int) nBins;
		
		double[][] bins = new double[2][numberBins];
		for (int i = 0; i < numberBins; i++) {
			bins[0][i] = binBegin(i, markStart, offset, width);	// use the bin's left boundary
			bins[1][i] = 0;	// initialize the count to 0
		}
		
		for (int i = 0; i < n; i++) {
			int index = index(values, i, offset, width, markStart, markFinish, includeStart, includeFinish);
			bins[1][index] += 1;	// increment the count of the appropriate bin
		}
		
		if (normalize) {
			double factor = (1.0 / n) * (1.0 / width);	// will be infinite if n == 0, but the check above detects this and the loop below would never execute anyways
			for (int i = 0; i < numberBins; i++) {
				bins[1][i] *= factor;
			}
		}
		
		return bins;
	}
	
	
	/**
	* Returns the mark of the bin which contains d, given the specification of offset and width.
	* The mark of d is the index-like label of the bin which contains d.
	* Specifically, the label of the bin which starts at offset is 0,
	* and the labels of the bins immediately to the left and right are -1 and 1 respectively, etc.
	* <p>
	* <b>Warning: the mark cannot serve as a Java array index because it can be negative (i.e. when d < offset).</b>
	* <p>
	* @throws IllegalStateException if calculate a mark that cannot be represented as an int
	*/
	private int mark(double d, double offset, double width) throws IllegalStateException {
		double mark = Math.floor( (d - offset) / width );
		if (mark < Integer.MIN_VALUE) throw new IllegalStateException("calculated mark = " + mark + " which is too small to be represented as an int");
		else if (mark > Integer.MAX_VALUE) throw new IllegalStateException("calculated mark = " + mark + " which is too large to be represented as an int");
		return (int) mark;
	}
	
	
	/** Returns the begin value for the ith bin, given that bins start at offset, given the specification of offset and width. */
	private double binBegin(int i, int markStart, double offset, double width) {
		int mark = markStart + i;
		return offset + (mark * width);
	}
	
	
	/**
	* Returns the index of the bin which contains d, given the specification of offset, width, and markStart.
	* The markFinish, includeStart, and includeFinish params are used for bounds checking.
	* <p>
	* Note: the result can serve as a Java array index because it is never negative.
	* <p>
	* @throws IllegalArgumentException if calculate a mark < markStart or > markFinish
	*/
	private int index(double[] values, int i, double offset, double width, int markStart, int markFinish, double includeStart, double includeFinish) throws IllegalArgumentException {
		int mark = mark(values[i], offset, width);
		if (mark < markStart) throw new IllegalArgumentException("calculated mark = " + mark + " < markStart = " + markStart + "; happened for values[" + i + "] = " + values[i] + ", offset = " + offset + ", width = " + width + ", markStart = " + markStart + ", markFinish = " + markFinish + ", includeStart = " + includeStart + ", includeFinish = " + includeFinish);
		if (mark > markFinish) throw new IllegalArgumentException("calculated mark = " + mark + " > markFinish = " + markFinish + "; happened for values[" + i + "] = " + values[i] + ", offset = " + offset + ", width = " + width + ", markStart = " + markStart + ", markFinish = " + markFinish + ", includeStart = " + includeStart + ", includeFinish = " + includeFinish);
		
		int index = mark - markStart;
		if (index < 0) throw new IllegalStateException("calculated index = " + index + " < 0; happened for values[" + i + "] = " + values[i] + ", offset = " + offset + ", width = " + width + ", markStart = " + markStart + ", markFinish = " + markFinish + ", includeStart = " + includeStart + ", includeFinish = " + includeFinish);
		
		return mark - markStart;
	}
	
	
	/**
	* The bins returned as a double[][] by any of the toBins methods
	* all use the abbreviated designation convention described in the fundamental version of
	* {@link #toBins(double, double, double, double, boolean) that method}.
	* Namely, each element of the x subarray stores the left boundary point of the corresponding bin interval.
	* However, users of the bins may wish to characterise their intervals by some other point.
	* For instance, for curve fitting, it is more natural to characterise the interval by its midpoint.
	* <p>
	* This method allows any position within the bin interval to be selected as the interval's characteristic point.
	* The shiftPercent param specifies the location of this new point as follows:
	* it is restricted to the range [0, 1] and is the offset from the left boundary
	* expressed as a relative amount of the bin interval width.
	* So, a shiftPercent value of 0 (i.e. 0% of interval width) means use the left boundary,
	* 0.5 (i.e. 50% of interval width) means use the midpoint, and 1.0 (100%) means use the right boundary.
	* <p>
	* The elements of the x subarray of the bins param are usually modified to reflect their new characteristic point.
	* The sole exception is if the x subarray has only 1 element,
	* in which case nothing is done because it is impossible to determine the interval width in this case.
	* <p>
	* This method should only be called once on a particular bins instance.
	* <p>
	* @throws IllegalArgumentException if bins == null; bins.length != 2; bins[0].length != bins[1].length;
	* {@link NumberUtil#checkNormalAndNonNegative NumberUtil.checkNormalAndNonNegative} objects to shiftPercent;
	* if shiftPercent > 1
	* @throws IllegalStateException if any internal problem is encountered
	*/
	public void binShift(double[][] bins, double shiftPercent) throws IllegalArgumentException {
		if (bins == null) throw new IllegalArgumentException("bins == null");
		if (bins.length != 2) throw new IllegalArgumentException("bins.length = " + bins.length + " != 2");
		if (bins[0].length != bins[1].length) throw new IllegalArgumentException("bins[0].length = " + bins[0].length + " != bins[1].length = " + bins[1].length);
		NumberUtil.checkNormalAndNonNegative(shiftPercent, "shiftPercent");
		if (shiftPercent > 1) throw new IllegalArgumentException("shiftPercent = " + shiftPercent + " > 1");
		
		if (bins[0].length < 2) return;
		
		double width = bins[0][1] - bins[0][0];
		if (width < 0) throw new IllegalStateException("calculated width = " + width + " < 0");
		double shift = shiftPercent * width;
		
		for (int i = 0; i < bins[0].length; i++) {
			bins[0][i] += shift;
		}
	}
	
// +++ may wish to offer a binsToPoints method which converts double[][] to Point2D.Double[]
// so that other Java graphing packages could use, assuming that they take that format?
	
	
	// -------------------- fitGaussian --------------------
	
	
	/**
	* Determines that gaussian (i.e. normal distribution) which best fits the data inside bins.
	* <p>
	* The bins param must have the same format as the result of the toBins methods
	* <i>with normalization used</i> (i.e. it must represent a probability density function).
	* See the fundamental {@link #toBins(double, double, double, double, boolean) toBins} method for documentation.
	* <p>
	* The result of this method is always a double[][] which contains the fit params plus x and y values of a fit curve.
	* Written as code, if <code>double[][] fit = storer.fitGaussian(...)</code>,
	* then <code>double[] parameters = bins[0]</code>,
	* <code>double[] xValues = bins[1]</code>, and <code>double[] yValues = bins[2]</code>.
	* Here, <code>parameters</code> always has length 3,
	* with <code>parameters[0]</code> holding the mean of the gaussian,
	* <code>parameters[1]</code> holding the standard deviation of the gaussian,
	* and <code>parameters[2]</code> holding the r-squared of the gaussian fit (i.e. a goodness of fit measure).
<!--
+++ would also like to store other measures of the goodness of the fit like
	Degrees of Freedom Adjusted R-Square
	Root Mean Squared Error
-->
	* Also, <code>xValues.length = yValues.length</code>,
	* <code>xValues</code> is the same array as <code>bins[0]</code>,
	* and each element of <code>yValues</code>
	* is the y value of the fitted gaussian at the corresponding x.
	* Thus, one can supply both bins and the result to a graphing package for simultaneous display.
	* <p>
	* @throws IllegalArgumentException if bins == null; bins.length != 2; bins[0].length != bins[1].length;
	* @throws IllegalStateException if no values have been stored
	*/
	public double[][] fitGaussian(double[][] bins) throws IllegalStateException, IllegalArgumentException {
		if (bins == null) throw new IllegalArgumentException("bins == null");
		if (bins.length != 2) throw new IllegalArgumentException("bins.length = " + bins.length + " != 2");
		if (bins[0].length != bins[1].length) throw new IllegalArgumentException("bins[0].length = " + bins[0].length + " != bins[1].length = " + bins[1].length);
		
		double[] numbers = values();
		double meanData = Math2.mean(numbers);
		double sdData = Math2.sd(numbers, meanData);
		double[] parameters = new double[] {meanData, sdData};	// great guess as to the best fit gaussian's mean and sd is the data's corresponding values
		double[] weights = new double[bins[0].length];
		Arrays.fill(weights, 1);	// must use 1 since have no idea, in general, what the error weights for each point are
		
		LMA lma = new LMA(new GaussianLMAFunction(), parameters, bins, weights); 
		lma.fit();
		
		double meanFit = parameters[0];
		double sdFit = parameters[1];
			// uncomment these lines if want to see how close the least squares parameters are to the data's statistics:
		//System.out.println("meanFit = " + meanFit + ", meanData = " + meanData);
		//System.out.println("sdFit = " + sdFit + ", sdData = " + sdData);
		
			// calculated the r-squared, which is one goodness of fit measure:
		double sse = lma.chi2;	// see the "Least Squares Fitting" section in http://www.physics.ohio-state.edu/~gan/teaching/spring04/Chapter6.pdf
		double sst = Math2.sst(bins[1]);
		double rSquared = 1 - (sse/sst);	// see http://en.wikipedia.org/wiki/R-squared
		
		double[][] result = new double[3][];
		result[0] = new double[] {meanFit, sdFit, rSquared};
		result[1] = new double[ bins[0].length ];
		result[2] = new double[ bins[0].length ];
		for (int i = 0; i < bins[0].length; i++) {
			double x = bins[0][i];
			result[1][i] = x;
			result[2][i] = Math2.gaussianPdf(x, meanFit, sdFit);
		}
		return result;
	}
	/*
	General discussion of curve fitting, including nonlinear fitting:
		http://mathworld.wolfram.com/NonlinearLeastSquaresFitting.html
		http://www.aip.org/tip/INPHFA/vol-9/iss-2/p24.html
		http://www.wavemetrics.com/products/igorpro/dataanalysis/curvefitting.htm
		
	Levenberg-Marquardt algorithm for solving nonlinear least squares:
		http://en.wikipedia.org/wiki/Levenberg-Marquardt_algorithm
		
	Java code for Levenberg-Marquardt algorithm:
		http://users.utu.fi/jaolho/java/lma.zip
		http://www.ccp4.ac.uk/jwc/Xhtml/Jdl/JdlLib_class_list.html#s9
		http://www.idiom.com/~zilla/Computer/Javanumeric/LM.java
		http://www.amstat.org/publications/jse/secure/v7n2/taur.cfm
	
	Java numerical code:
		Colt open source library provides many high performance and high quality routines for linear algebra, sparse and dense matrices, statistical tools for data
		analysis, random number generators, array algorithms, mathematical functions and complex numbers.
		http://dsd.lbl.gov/~hoschek/colt/
		
		The Ninja project is working to make Java competitive with Fortran and C++ in the domain of technical computing.
		http://www.research.ibm.com/ninja/
	*/
	
	
	private static class GaussianLMAFunction extends LMAFunction {
		
		private GaussianLMAFunction() {}
		
		public double getY(double x, double[] a) {	// Note: no arg checking, since is only used by fitGaussian, and can assume that that is written correctly
			if (a[1] <= 0) return Double.NaN;	// CRITICAL: have seen LMA occaisionally generate non-positive values for a[1] (sigma), which Math2.gaussianPdf will correctly crash on; in order to not stop the fitting process, however, have found that if simply return NaN then LMA.fir will reduce the step size until a[1] becomes positive, and thus all is well (except for some error messages that it prints out to System.err)
			return Math2.gaussianPdf(x, a[0], a[1]);
		}
		
		public double getPartialDerivate(double x, double[] a, int parameterIndex) {	// Note: no arg checking, since is only used by fitGaussian, and can assume that that is written correctly
			if (a[1] <= 0) return Double.NaN;	// CRITICAL: have seen LMA occaisionally generate non-positive values for a[1] (sigma), which Math2.gaussianPdf will correctly crash on; in order to not stop the fitting process, however, have found that if simply return NaN then LMA.fir will reduce the step size until a[1] becomes positive, and thus all is well (except for some error messages that it prints out to System.err)
				// analytical version:
			double g = Math2.gaussianPdf(x, a[0], a[1]);
			double z = (x - a[0]) / a[1];
			switch (parameterIndex) {
				case 0: return g * (z / a[1]);
				case 1: return (g / a[1]) * ( (z * z) - 1 );
				default: throw new IllegalArgumentException("parameterIndex = " + parameterIndex + " is an illegal value");
			}
			/*
				// numerical version (works too, but has issues such as accuracy, choice of da below, etc):
			double da = 1e-10;	// tried 1e-20 and ran into singular matrix issues; what OUGHT to do is have a loop and make da bigger if it causes a singular matrix; da should probably be made into a field in this case to cache the value which works
			double[] aModified = new double[] {a[0], a[1]};
			aModified[parameterIndex] += da;
			double dy = getY(x, aModified) - getY(x, a);
			return dy/da;
			*/
		}
	}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
// +++ hmm, the tests below have no assertions and require manual inspection instead, which is not so cool...
		
		@Test public void test_whiteNoise() throws Exception {
			int n = 1 * 1000 * 1000;
			DataStorer dataStorer = new DataStorer(n);
			Random random = new Random();
			for (int i = 0; i < n; i++) {
				dataStorer.add( random.nextDouble() );
			}
			double[][] bins = dataStorer.toBins(0.0, 0.01, false);
			String binsText = StringUtil.arraysToTextColumns(bins, new String[] {"x", "count"});
			File file = new File("../log/binsOfWhiteNoise.txt");
			FileUtil.writeString(binsText, file, false);
			System.out.println("plot the file " + file.getPath() + " to see if the bins are approximately uniform in value");
		}
		
		@Test public void test_gaussianNoise() throws Exception {
			DataStorer dataStorer = new DataStorer();
			Random random = new Random();			
			for (int i = 0; i < 1000; i++) {
				dataStorer.add( random.nextGaussian() );
			}
			double[][] bins = dataStorer.toBins(0.0, 0.1, true);
			dataStorer.binShift(bins, 0.5);	// use bin midpoints			
			double[][] dataFit = dataStorer.fitGaussian(bins);
			double[][] dataAll = new double[][] { dataFit[0], bins[0], bins[1], dataFit[2]};
			String allText = StringUtil.arraysToTextColumns(dataAll, new String[] {"fitParams", "x", "dataPdf", "fitPdf"});
			File fileAll = new File("../log/binsOfGaussianFit.txt");
			FileUtil.writeString(allText, fileAll, false);
			System.out.println("plot the file " + fileAll.getPath() + " to see if the bins are approximately Gaussian");
		}
		
	}
	
	
}
