/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*
Programmer's notes:

--C code for a 160 bit maximal LFSR: http://www.cryogenius.com/software/lfsr/
	
*/


package bb.science;


import org.junit.Assert;
import org.junit.Test;

import bb.util.StringUtil;


/**
* Implements a <a href="http://en.wikipedia.org/wiki/Linear_feedback_shift_register">linear feedback shift register</a>.
* <p>
* This class is not multithread safe.
* <p>
* @author Brent Boyer
* @see <a href="http://forum.java.sun.com/thread.jspa?threadID=5276320&tstart=0">this forum posting</a>
*/
public class Lfsr {
	
	
	// -------------------- fields --------------------
	
	
	private final int n;
	private final int mask;
	private final int taps;
	private int register;
	
	
	// -------------------- constructor and helper methods --------------------
	
	
	/**
	* Constructor.
	* <p>
	* @throws IllegalArgumentException if n < 2 or n > 32
	*/
	public Lfsr(int n) throws IllegalArgumentException {
		// n checked by calls to makeMask and makeTaps below
		
		this.n = n;
		mask = makeMask(n);
		taps = makeTaps(n);
		register = 1;
	}
	
	
	/**
	* Generates a mask that will preserve the n low order bits and put 0's into the high order bits.
	* <p>
	* @throws IllegalArgumentException if n < 0 or n > 32
	*/
	private int makeMask(int n) throws IllegalArgumentException {
		if (n < 0) throw new IllegalArgumentException("n = " + n + " < 0");
		if (n > 32) throw new IllegalArgumentException("n = " + n + " > 32");
		
		int m = 0;
		for (int i = 0; i < n; i++) m |= 1 << i;
		return m;
	}
	
	
	/**
	* Returns an int which can function as the taps of n order maximal LFSR.
	* <p>
	* @throws IllegalArgumentException if n < 2 or n > 32
	*/
	private int makeTaps(int n) throws IllegalArgumentException {
		if (n < 2) throw new IllegalArgumentException("n = " + n + " < 2");
		if (n > 32) throw new IllegalArgumentException("n = " + n + " > 32");
		
			// There is no easy algorithm to generate the taps as a function of n.
			// Instead simply have to rely on known results:
		switch (n) {
			case 2: return  (1 << 1) | (1 << 0);	// i.e. 2 1
			case 3: return  (1 << 2) | (1 << 1);	// i.e. 3 2
				// for the ones below, see http://en.wikipedia.org/wiki/Linear_feedback_shift_register#Some_Polynomials_for_Maximal_LFSRs
			case 4: return  (1 << 3) | (1 << 2);	// i.e. 4 3
			case 5: return  (1 << 4) | (1 << 2);	// i.e. 5 3
			case 6: return  (1 << 5) | (1 << 4);	// i.e. 6 5
			case 7: return  (1 << 6) | (1 << 5);	// i.e. 7 6
			case 8: return  (1 << 7) | (1 << 5) | (1 << 4) | (1 << 3);	// i.e. 8 6 5 4
			case 9: return  (1 << 8) | (1 << 4);	// i.e. 9 5
			case 10: return  (1 << 9) | (1 << 6);	// i.e. 10 7
			case 11: return  (1 << 10) | (1 << 8);	// i.e. 11 9
			case 12: return  (1 << 11) | (1 << 10) | (1 << 9) | (1 << 3);	// i.e. 12 11 10 4
			case 13: return  (1 << 12) | (1 << 11) | (1 << 10) | (1 << 7);	// i.e. 13 12 11 8
			case 14: return  (1 << 13) | (1 << 12) | (1 << 11) | (1 << 1);	// i.e. 14 13 12 2
			case 15: return  (1 << 14) | (1 << 13);	// i.e. 15 14
			case 16: return  (1 << 15) | (1 << 13) | (1 << 12) | (1 << 10);	// i.e. 16 14 13 11
			case 17: return  (1 << 16) | (1 << 13);	// i.e. 17 14
			case 18: return  (1 << 17) | (1 << 10);	// i.e. 18 11
			case 19: return  (1 << 18) | (1 << 17) | (1 << 16) | (1 << 13);	// i.e. 19 18 17 14
				// for the ones below, see http://homepage.mac.com/afj/taplist.html
			case 20: return  (1 << 19) | (1 << 16);	// i.e. 20 17
			case 21: return  (1 << 20) | (1 << 18);	// i.e. 21 19
			case 22: return  (1 << 21) | (1 << 20);	// i.e. 22 21
			case 23: return  (1 << 22) | (1 << 17);	// i.e. 23 18
			case 24: return  (1 << 23) | (1 << 22) | (1 << 21) | (1 << 16);	// i.e. 24 23 22 17
			case 25: return  (1 << 24) | (1 << 21);	// i.e. 25 22
			case 26: return  (1 << 25) | (1 << 24) | (1 << 23) | (1 << 19);	// i.e. 26 25 24 20
			case 27: return  (1 << 26) | (1 << 25) | (1 << 24) | (1 << 21);	// i.e. 27 26 25 22
			case 28: return  (1 << 27) | (1 << 24);	// i.e. 28 25
			case 29: return  (1 << 28) | (1 << 26);	// i.e. 29 27
			case 30: return  (1 << 29) | (1 << 28) | (1 << 27) | (1 << 6);	// i.e. 30 29 28 7
			case 31: return  (1 << 30) | (1 << 27);	// i.e. 31 28
			case 32: return (1 << 31) | (1 << 30) | (1 << 29) | (1 << 9);	// i.e. 32 31 30 10
			
			default: throw new IllegalStateException("should never reach here (n = " + n + " case)");
		}
	}
	
	
	// -------------------- accessors --------------------
	
	
	/** Accessor for {@link #register}. */
	public int getRegister() { return register; }
	
	
	// -------------------- advance --------------------
	
	
	/** Convenience method that simply returns {@link #advance advance}(1). */
	public int advance() { return advance(1); }
	
	
	/**
	* Advances the internal state of this instance by numberTransitions.
	* <p>
	* Ignoring an initial argument check and a final return statement,
	* this method is essentially 2 lines of code: a loop head and its body.
	* The loop body involves 6 bitwise and/or unary int operators, <i>so it is very simple and fast</i>.
	* In spite of the simplicity of the computation, the LFSR internal state
	* (stored in the {@link #register} field) is pseudo-random.
	* It should be impossible for a smart compiler to cut many corners and avoid doing the computations.
	* <p>
	* @return the final value of the internal state after all the transitions have been carried out
	* @throws IllegalArgumentException if numberTransitions < 0
	*/
	public int advance(long numberTransitions) throws IllegalArgumentException {
		if (numberTransitions < 0) throw new IllegalArgumentException("numberTransitions = " + numberTransitions + " < 0");
		
		for (long i = 0; i < numberTransitions; i++) {
			register = ((register >>> 1) ^ (-(register & 1) & taps)) & mask;	// register & 1 selects the low order bit, and do - to create an int that is all 1's if was 1, or all 0's if was 0; the rest is easy to understand
		}
		return register;
	}
	
	
	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_period() {
			for (int i = 2; i <= 32; i++) checkPeriod(i);
		}
		
		private void checkPeriod(int n) {
			System.out.println("checkPeriod: n = " + n);
			Lfsr lfsr = new Lfsr(n);
			int registerInitial = lfsr.register;
			long period = 0;
			do {
				//System.out.println( toBinaryString(lfsr.register, n) );	// only uncomment if testing small n to confirm that results make sense
				lfsr.advance(1);
				++period;
			} while (lfsr.register != registerInitial);
			diagnosePeriod(period, n);
		}
		
		private String toBinaryString(int i, int length) {
			return StringUtil.toLength( Integer.toBinaryString(i), length, true, '0' );
		}
		
		private void diagnosePeriod(long period, int n) {
			long periodExpected = (1L << n) - 1L;
			String errMsg = "period = " + period + " IS NOT EQUAL TO THE MAXIMAL LENGTH VALUE of " + periodExpected + " for n = " + n;
			Assert.assertTrue( errMsg, period == periodExpected );
		}
		
	}
	
	
}
	
	
/*
Below is code for a Gray code, which is somewhat similar to a LFSR:

// http://en.wikipedia.org/wiki/Gray_code#Programming_algorithms
public class GrayCode implements Callable<Integer> {

	public Integer call() {
		int n = 3;
		int mask = makeMask(n);
		int gc = 0;
		int gcFinal = encode(1 << n);
		int period = 0;
		do {
			System.out.println("gc = " + gc + " <-> " + toBinaryString(gc, n));
			gc = encode( decode(gc) + 1 );
			++period;
		} while (gc != gcFinal);

		if (period == (1 << n)) System.out.println("period = " + period + " as expected for n = " + n);
		else System.out.println("period = " + period + " IS NOT EQUAL TO THE MAXIMAL LENGTH VALUE for n = " + n);

		return gc;
	}

	private int encode(int binary) {
		return binary ^ (binary >> 1);
	}

	private int decode(int gc) {
		int ish = 1;
		int ans = gc;
		while (true) {
			int idiv = ans >> ish;
			ans ^= idiv;
			if (idiv <= 1 || ish == 32) return ans;
			ish <<= 1; // double number of shifts next time
		}
	}

}
*/
