/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.util;


import java.io.*;


/**
* This is class provides miscellaneous static <code>char</code> utility methods.
* <p>
* Note that the argument type for many of the methods is of <code>int</code> type,
* as opposed to <code>char</code> type. This was deliberately chosen for two good reasons.
* First, all <code>Reader</code>s return <code>int</code> values when you do single character reads,
* in order to indicate EOF by returning -1. So, by taking <code>int</code> type arguments, the methods
* here will conveniently deal with EOF values. This frees the programmer from having to first check.
* Second, the internal comparison operations done in many of the methods here would promote a
* <code>char</code> argument to an <code>int</code> anyways, so it is actually quicker to avoid the
* cast operations in the first place.
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @author Brent Boyer
*/
public class CharUtil {
	
	
	// -------------------- constants --------------------
	
	
	/** The minimum (int) value that a valid Java char can have. */
	public static final int minValueAsInt = (int) Character.MIN_VALUE;
	
	
	/** The maximum (int) value that a valid Java char can have. */
	public static final int maxValueAsInt = (int) Character.MAX_VALUE;
	
	
	/**
	* The maximum (int) value that a valid US-ASCII (i.e. 7 bit) char can have.
	* <p>
	* @see <a href="http://en.wikipedia.org/wiki/Ascii">US-ASCII standard</a>
	*/
	public static final int maxAsciiValueAsInt = 127;
	
	
/*
		// chars in this array act as word separators and will also be skipped over (i.e. not counted as part of words):
	public static final char[] WORD_SEPARATOR_CHARS = {
			// whitespace chars:
		' ', '\t',
		
			// text formatting chars:
		'\b', '\f', '\n', '\r',
		
			// punctuation chars (going left to right, bottom to top of keyboard):
		',', '.', '/', '?', ';', ':', '\'', '\"', '[', '{', ']', '}', '\\', '|',
		'`', '~', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '-', '_', '=', '+'
	};
*/
	
	
	// -------------------- isXXX methods --------------------
	
	
	/** Determines whether or not the arg, which is an <code>int</code>, has a value in the <code>char</code> range. */
	public static boolean isCharValue(int charAsInt) {
		return (
			(minValueAsInt <= charAsInt) &&
			(charAsInt <= maxValueAsInt)
		);
	}
	
	
	/** Determines whether or not charAsInt represents a (lower or upper case) Roman Letter (i.e. 'a', 'b', ..., 'z', 'A', 'B', ..., 'Z' ). */
	public static boolean isLineEnd(int charAsInt) {
		return (charAsInt == '\n') || (charAsInt == '\r');
	}
	
	
	/** Determines whether or not charAsInt represents a (lower or upper case) Roman Letter (i.e. 'a', 'b', ..., 'z', 'A', 'B', ..., 'Z' ). */
	public static boolean isRomanLetter(int charAsInt) {
		return isRomanLetterLowerCase(charAsInt) ||	isRomanLetterUpperCase(charAsInt);	// speed heuristic: do the lower case test first, as it should be the most common			
	}
	
	
	/** Determines whether or not charAsInt represents a lower case Roman Letter (i.e. 'a', 'b', ..., 'z'). */
	public static boolean isRomanLetterLowerCase(int charAsInt) {
		return (97 <= charAsInt) && (charAsInt <= 122);
	}
	
	
	/** Determines whether or not charAsInt represents an upper case Roman Letter (i.e. 'A', 'B', ..., 'Z'). */
	public static boolean isRomanLetterUpperCase(int charAsInt) {
		return (65 <= charAsInt) && (charAsInt <= 90);
	}
	
	
	/** Determines whether or not charAsInt represents a decimal digit (i.e. '0', '1', ..., '9'). */
	public static boolean isDecimalDigit(int charAsInt) {
		return (48 <= charAsInt) && (charAsInt <= 57);
	}
	
	
	// -------------------- Escape methods --------------------
	
	
	/**
	* This method reports whether or not the supplied char has a "simple"
	* escape sequence that may appear in a Java char or String Literal.
	* (These chars are defined in section 3.10.6 of "The Java Language Specification Second Edition".
	* They are the ones which have 2-char escape sequences which begin with a backslash.)
	*/
	public static boolean hasEscapeForLiteral(char c) {
		switch (c) {
			case '\b':
			case '\t':
			case '\n':
			case '\f':
			case '\r':
			case '\"':
			case '\'':
			case '\\':
				return true;
			
			default:
				return false;
		}
	}
	
	
	/**
	* This method returns the "simple" (i.e. 2 char) escape sequence for the supplied char,
	* which is suitable for appearing in a Java char or String Literal.
	* <p>
	* @throws IllegalArgumentException if the char has no "simple" escape
	* (i.e. if CharUtil.hasEscapeForLiteral(c) returns false)
	*/
	public static String getEscapeForLiteral(char c) throws IllegalArgumentException {
		switch (c) {
			case '\b':
				return "\\b";
			
			case '\t':
				return "\\t";
			
			case '\n':
				return "\\n";
				
			case '\f':
				return "\\f";
				
			case '\r':
				return "\\r";
				
			case '\"':
				return "\\\"";
				
			case '\'':
				return "\\'";
				
			case '\\':
				return "\\\\";
				
			default:
				throw new IllegalArgumentException("arg c = " + c + "has no simple escape sequence");
		}
	}
	
	
	// -------------------- matches --------------------
	
	
	/**
	* This method determines whether or not c1 and c2 are matching char values.
	* <p>
	* It immediately returns true if <code>c1 == c2</code>.
	* <p>
	* If that test fails but <code>isCaseSensitive == false</code>, then this c1 and c2 have their case converted and are recompared.
	*/
	public static boolean matches(char c1, char c2, boolean isCaseSensitive) {
		if (c1 == c2) return true;
		
		if (!isCaseSensitive) {
				// see source code of String.regionMatches(boolean, int, String, int, int) for why this has to be so complicated:
			char up1 = Character.toUpperCase(c1);
			char up2 = Character.toUpperCase(c2);
			if (up1 == up2) return true;
			
			char low1 = Character.toLowerCase(c1);
			char low2 = Character.toLowerCase(c2);
			if (low1 == low2) return true;
		}
		
		return false;
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private CharUtil() {}
	
	
}