/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.util;


import org.junit.Assert;
import org.junit.Test;


/**
* This class provides static utility methods that deal with the {@link Class} of an object.
* <p>
* This class is multithread safe: it is stateless.
* <p>
* @author Brent Boyer
*/
public class ClassUtil {
	
	
	// -------------------- getNameSimple, getPackage --------------------
	
	
	/**
	* Returns the "simple" name of the fully quallified class name that is stored in nameFull,
	* that is, any leading package info is removed.
	* For example, if nameFull equals "java.lang.Object" then "Object" is returned.
	* <p>
	* The result is identical to what {@link Class#getSimpleName Class.getSimpleName} returns
	* except that this method works on a raw String.
	* <p>
	* @throws IllegalArgumentException if nameFull is blank
	*/
	public static String getNameSimple(String nameFull) throws IllegalArgumentException {
		if (StringUtil.isBlank(nameFull)) throw new IllegalArgumentException("nameFull is blank");
		
		int index = nameFull.lastIndexOf('.');
		return (index > -1) ? nameFull.substring(index + 1) : nameFull;
	}
	
	
	/**
	* Returns the any package name of the fully quallified class name that is stored in nameFull,
	* that is, the "simple" name is removed.
	* For example, if nameFull equals "java.lang.Object" then "java.lang" is returned.
	* If nameFull represents a class in the default (unnamed) package, then null is returned.
	* <p>
	* @throws IllegalArgumentException if nameFull is blank
	*/
	public static String getPackage(String nameFull) throws IllegalArgumentException {
		if (StringUtil.isBlank(nameFull)) throw new IllegalArgumentException("nameFull is blank");
		
		int index = nameFull.lastIndexOf('.');
		return (index > -1) ? nameFull.substring(0, index) : null;
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private ClassUtil() {}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_all() {
			Assert.assertEquals( "Object", getNameSimple("java.lang.Object") );
			Assert.assertEquals( "java.lang", getPackage("java.lang.Object") );
			Assert.assertEquals( null, getPackage("ClassInDefaultPackage") );
		}
		
	}
	
	
}
