/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*
Programmer Notes:

+++ is there ever a need for switch values to start with a hyphen, so that the parseArgs method should be changed?

+++ need to add methods for
	1) reading System.in input
	2) console interface programs, like what did for TC's ConsoleInterface class wherein used reflection to autogenerate interface

--this website discusses the POSIX conventions for command line arguments:
	http://www.iam.ubc.ca/guides/javatut99/essential/attributes/_posix.html

--using reflection to handle command line arguments:
	http://www-128.ibm.com/developerworks/java/library/j-dyn0715.html

--using hash tables for command line interfaces:
	http://www.ibm.com/developerworks/linux/library/l-gperf.html?S_TACT=105AGX59&S_CMP=GR&ca=dgr-lnxw01GPERF

--see also this project:
	http://commons.apache.org/cli/

--there is also the unix bourne shell command getopts which does command line argument processing; maybe should look at it for inspiration
	--see the "Interpreting Options With getopts Command" section of http://www.esscc.uq.edu.au/~ksteube/Bshell/#page_14

--on using java for even small systems programming tasks:
	http://weblogs.java.net/blog/evanx/archive/2006/05/bin_bash_java_c.html
	http://weblogs.java.net/blog/evanx/archive/2006/05/a_fools_errand.html
	http://weblogs.java.net/blog/evanx/archive/2006/05/java_is_all_you.html
see also a quick way to execute java:
	http://www.martiansoftware.com/nailgun/index.html
*/


package bb.util;


import java.util.*;

import org.junit.Assert;
import org.junit.Test;


/**
* This class provides utility methods for command line programs.
* <p>
* This class is multithread safe: it is stateless.
* <p>
* @author Brent Boyer
*/
public class CommandLineUtil {
	
	
	// -------------------- property methods: parseArgs, isSwitchKey, checkSwitchKeys, findBadSwitchKeys --------------------
	
	
	/**
	* Parses key/value pairs out of args, returning these as a new {@link Properties2} instance.
	* <p>
	* The key/value pairs in args must be a series of <i>command line switches</i>
	* (since args is typically supplied on the command line to a program's main).
	* The precise format is:
	* <ol>
	*  <li>a command line switch is a key optionally followed by a value</li>
	*  <li>every key must pass {@link #isSwitchKey isSwitchKey} and every value must <i>fail</i> isSwitchKey</li>
	*  <li>each key and (optional) value is a separate consecutive element of args</li>
	*  <li>a value element of args must always be preceded by an element of args that is its key</li>
	*  <li>corollary: args[0] must be a switch key</li>
	*  <li>since switch values are optional, there may be multiple consecutive elements of args which are just keys</li>
	*  <li>if a switch lacks a value element, this method associates the switch's key with the empty string ""</li>
	*  <li>a given switch key may only appear once (i.e. it is an error to have duplicate switches)</li>
	* </ol>
	* <p>
	* @throws IllegalArgumentException if args is null or violates a format rule
	* @see <a href="http://www.scit.wlv.ac.uk/cbook/chap9.command.line.html">Command line programs</a>
	*/
	public static Properties2 parseArgs(String[] args) throws IllegalArgumentException {
		if (args == null) throw new IllegalArgumentException("args == null");
		
		Properties2 properties = new Properties2();
		
		for (int i = 0; i < args.length; i++) {	// NOTE: the loop below expects i to always index the next switch key
			String switchKey = args[i];
			if (!isSwitchKey(switchKey)) throw new IllegalArgumentException("format error: args[" + i + "] = " + switchKey + " is not switch key");
			if (properties.containsKey(switchKey)) throw new IllegalArgumentException("format error: args[" + i + "] = " + switchKey + " has already been specified as a key");
			
			String switchValue = "";
			int nextIndex = i + 1;
			if (nextIndex < args.length) {	// if there is at least one more String after current i
				if (!isSwitchKey(args[nextIndex])) {	// and it is not a switch key
					switchValue = args[nextIndex];	// then it is this switch's value
					i = nextIndex;	// must set i to the value index, so that the loop increment will advance it past
				}
			}
			
			properties.setProperty(switchKey, switchValue);
		}
		
		return properties;
	}
	
	
	/**
	* Determines whether or not s is a key for a command line switch.
	* The sole criteria is whether or not s starts with a hyphen (i.e. '-') char.
	* <p>
	* @throws IllegalArgumentException if arg s is null
	*/
	public static boolean isSwitchKey(String s) throws IllegalArgumentException {
		if (s == null) throw new IllegalArgumentException("arg s is null");
		
		return s.startsWith("-");
	}
	
	
// +++ this method finds bad switch keys,
// but need more complicated method which checks that a given valid key has a valid value
	/**
	* Finds any invalid switch keys by calling the {@link #findBadSwitchKeys findBadSwitchKeys} method.
	* If any are found, then an IllegalArgumentException is thrown with a message in the following format:
	* <ol>
	*  <li>the text "The following switch key(s) are illegal"</li>.
	*  <li>followed by the trimmed <code>description</code> arg (if it is not null or empty)</li>.
	*  <li>a list of all the offending switch keys</li>.
	* </ol>
	* If none are found, then this method returns normally.
	* <p>
	* This method is useful for command line argument syntax checking.
	* <p>
	* @throws IllegalArgumentException if any offending switch keys are found,
	* or if findBadSwitchKeys objects to any arg when it is called
	*/
	public static void checkSwitchKeys(Properties switches, Set<String> legalKeys, String description) throws IllegalArgumentException {
		// for arg checking, rely on that done in findBadSwitchKeys call below
		
		String[] badKeys = findBadSwitchKeys(switches, legalKeys);
		if (badKeys.length > 0) {
			StringBuilder sb = new StringBuilder(64 + 16*badKeys.length);
			
			sb.append("The following switch key(s) are illegal");
			if ( !StringUtil.isBlank(description) ) sb.append(" " + description.trim());
			sb.append(":");
			for (String badKey : badKeys) sb.append(" ").append(badKey);
			throw new IllegalArgumentException(sb.toString());
		}
	}
	
	
	/**
	* Finds any switch keys in the supplied Properties arg that are not members of the legalKeys Set,
	* and returns them as a String[].
	* If all the switch keys are valid, then a zero-length array is returned.
	* The bad switch keys array is sorted first before being returned for user convenience.
	* <p>
	* This method is useful for command line argument syntax checking.
	* <p>
	* @throws IllegalArgumentException if arg switches or legalKeys is null
	*/
	public static String[] findBadSwitchKeys(Properties switches, Set<String> legalKeys) {
		if (switches == null) throw new IllegalArgumentException("arg switches is null");
		if (legalKeys == null) throw new IllegalArgumentException("arg legalKeys is null");
		
		List<String> badKeys = new ArrayList<String>();
		for (Object key : switches.keySet()) {	// note that, unfortunately, the Properties class in java 1.5+ extends Hashtable<Object,Object> and not Hashtable<String,String> so must do an additional cast...
			String switchKey = (String) key;
			if (!legalKeys.contains(switchKey)) {
				badKeys.add(switchKey);
			}
		}
		
		String[] result = badKeys.toArray( new String[badKeys.size()] );
		Arrays.sort(result);	// must sort, since an Iterator from a Properties will spit stuff out in essentially random order
		return result;
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private CommandLineUtil() {}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_parseArgs_pass() {
			Properties2 propertiesExpected = new Properties2();
			propertiesExpected.setProperty("-a", "a's value");
			propertiesExpected.setProperty("-b", "b's value");
			propertiesExpected.setProperty("-c", "");
			
			String[] args = new String[] {"-a", "a's value", "-b", "b's value", "-c"};
			Properties2 propertiesParsed = parseArgs(args);
			
			Assert.assertEquals(propertiesExpected, propertiesParsed);
			
			System.out.println();
			System.out.println("test_parseArgs_pass:");
			propertiesParsed.list(System.out);
		}
		
		@Test(expected=IllegalArgumentException.class) public void test_parseArgs_fail1() {
			String[] args = new String[] {"a's value", "-b", "b's value", "-c"};	// does not start with a key
			parseArgs(args);
		}
		
		@Test(expected=IllegalArgumentException.class) public void test_parseArgs_fail2() {
			String[] args = new String[] {"-a", "a's value", "b's value", "-c"};	// is missing a key (2 values in a row occur)
			parseArgs(args);
		}
		
		@Test(expected=IllegalArgumentException.class) public void test_parseArgs_fail3() {
			String[] args = new String[] {"-a", "-a"};	// duplicates a key
			parseArgs(args);
		}
		
		@Test public void test_checkSwitchKeys_pass() {
			Set<String> validSwitchKeys = new HashSet<String>( Arrays.asList( "-a", "-b", "-c" ) );
			Properties2 propertiesParsed = parseArgs( new String[] {"-a", "-b", "-c"} );
			checkSwitchKeys(propertiesParsed, validSwitchKeys, null);
		}
		
		@Test(expected=IllegalArgumentException.class) public void test_checkSwitchKeys_fail() {
			Set<String> validSwitchKeys = new HashSet<String>( Arrays.asList( "-a", "-b", "-c" ) );
			Properties2 propertiesParsed = parseArgs( new String[] {"-d", "-e", "-f"} );
			checkSwitchKeys(propertiesParsed, validSwitchKeys, "in test mode");
		}
		
	}
	
	
}
