/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*

Programmer Notes:

--see the article "Inspect Your Java Objects" in Java Pro March 2002

+++ another use for this class would be to do generic deep copying of objects?

+++ it would be nice if could write a version of this class which autoparses C structs;
	--see the problem faced by this forum posting: http://forum.java.sun.com/thread.jspa?threadID=5276697&tstart=0

--note that can also access the Abstract Syntax Tree using the new Compiler Tree API to maybe do similar stuff?
	http://today.java.net/pub/a/today/2008/04/10/source-code-analysis-using-java-6-compiler-apis.html

*/


package bb.util;


import java.io.Serializable;
import java.lang.reflect.*;
import java.util.*;
import java.util.logging.Level;

import org.junit.Test;

import bb.util.logging.LogUtil;


/**
* Determines the state of some other object.
* <p>
* By default, the constructor finds the <i>complete</i> state:
* every Field in object's Class, all of its superclasses,
* and its complete interface hierarchy.
* (Recall: interfaces can declare static final fields,
* altho this is <a href="http://java.sun.com/j2se/1.5.0/docs/guide/language/static-import.html">not a good idea</a>).
* <p>
* An optional number of {@link Filter}s may be provided to restrict this state.
* Many users will find the {@link #filtersTypical} field convenient to use.
* <p>
* This data is stored as a Class --> Field[] map, which can be retrieved via {@link #getClassToFields getClassToFields}.
* <p>
* This class offers convenience {@link #toString} methods,
* which return nicely formatted Strings and make it trivial for any class to write a generic toString method:
* <code><pre>
*	    &#64;Override public String toString() {
*		return new {@link #ObjectState(Object, Filter[]) ObjectState}( this, {@link ObjectState#filtersTypical} ).{@link #toString() toString}();
*	}</pre></code>
* should suffice.
* Note: this class's toString results use the convention of placing angle brackets ("&lt;" and ">")
* around any text that is unusual (e.g. when describing a null value or an empty Collection).
* <p>
* If this generic toString approach is inadequate, {@link #generateToStringCode generateToStringCode}
* may be used to automatically generate java source code which can then be customized.
* A call like:
* <code><pre>
*	System.out.println( new {@link #ObjectState(Object, Filter[]) ObjectState}( this, {@link ObjectState#filtersTypical} ).{@link #generateToStringCode generateToStringCode}() );</pre></code>
* possibly followed by some hand editing of the result might be useful.
* <p>
* This class is useful in debugging, and is heavily used by the {@link d.p p} and {@link d.g g} classes.
* <p>
* This class uses reflection to get the Field data.
* Since reflection is much slower than direct field access,
* care should be taken in using this class if high performance is needed.
* <p>
* This class is not multithread safe.
<!--
In its current implementation, all its immediate state is immutable.
Furthermore, its static fields are all also deeply immutable or multithread safe.
However, the state of its instance fields is not.
For instance, classToFields's LinkedHashMap is probably not multithread safe.
(Hmm, nothing is added to it after construction, however, am not sure if it supports concurrent access such as iteration or not...)
Most users of this class, however, will probably be single thread users, which makes this not too important.
-->
* <p>
* @author Brent Boyer
*/
public class ObjectState {
	
	
	// -------------------- constants --------------------
	
	
	private static final FieldComparator fieldComparator = new FieldComparator();
	
	
	/**
	* This class's {@link #toString(int)} method faces the exact same problem as object serialization:
	* certain object graphs can have circular reference chains that lead to infinite loops.
	* The usual way to prevent this is to keep track of objects encountered so far,
	* and use a placeholder if ever encounter a known object.
	* That is what this field does:
	* it allows each thread to know what objects it has encountered so far that use ObjectState.toString.
	* It is a Map from an encountered Object to the number of times that it has been encountered.
	*/
	private static final ThreadLocal<Map<Object,Integer>> threadToIdentitySet = new ThreadLocal<Map<Object,Integer>>() {
		protected synchronized Map<Object,Integer> initialValue() {
			return new HashMap<Object,Integer>();
		}
	};
	
	
	/**
	* An array of {@link Filter}s that most users of this class will typically want.
	* The current implementation rejects:
	* <ol>
	*  <li>Object.class, since a) as of jdk 1.6 at least, it has no fields and b) is always implicitly understood as the final superclass</li>
	*  <li>all interfaces, since they can only have constants (and rarely at that), whose state is usually uninteresting</li>
	*  <li>all static fields, since they is shared among many instances and are usually uninteresting</li>
	* </ol>
	*/
	public static final Filter[] filtersTypical = new Filter[] {
		//new AcceptOnlyImmediateClass(),
		new RejectObjectClass(),
		new RejectInterfaces(),
		//new AcceptOnlyPublicProtectedFields(),
		new RejectStaticFields()
	};
	
	
	// -------------------- instance fields --------------------
	
	
	/**
	* Records the Object that this instance is concerned with.
	* <p>
	* Contract: none; may even be null.
	*/
	private final Object object;
	
	
	/**
	* Records all the {@link Filter}s that this instance uses.
	* <p>
	* Contract: is never null, but may be zero-length.
	*/
	private final Filter[] filters;
	
	
	/**
	* For each Class in {@link #object}'s hierarchy (i.e. object's direct Class, its superclass, and the interface hierarchy),
	* stores a mapping from the Class to an array of every declared Field of that class.
	* <p>
	* Contract:
	* <ol>
	*  <li>is never null</li>
	*  <li>is empty if and only if {@link #object} is null</li>
	*  <li>
	*		iteration returns Classes in the following order: first the Class for object itself,
	*		then the Class hierarchy for each interface that is directly implemented by object.
	*		This process is then recursively repeated for each superclass above object.
	*		If an interface appears multiple times in the interface hierarchy,
	*		then it will obviously only appear once as a key in this field;
	*		its iteration order is determined by its first appearance.
	*  </li>
	*  <li>
	*		each Field[] that is mapped to will a) never be null
	*		but b) will be zero length if the Class/Interface has no fields
	*		and c) is sorted using a {@link FieldComparator}.
	*  </li>
	* </ol>
	*/
	private final Map<Class,Field[]> classToFields = new LinkedHashMap<Class,Field[]>();
	
	
	// -------------------- static ObjectGraph methods --------------------
	
	
	private static boolean isInObjectGraph(Object obj) {
		return threadToIdentitySet.get().containsKey(obj);
	}
	
	
	private static boolean isObjectGraphEmpty() {
		return threadToIdentitySet.get().isEmpty();
	}
	
	
	private static void incrementObjectGraph(Object obj) {
		Map<Object,Integer> map = threadToIdentitySet.get();
		Integer count = map.get(obj);
		if (count == null) {
			map.put(obj, 1);
		}
		else if (count.intValue() <= 0) {
			throw new IllegalStateException("count.intValue() <= 0 when incrementObjectGraph called--this should never happen");
		}
		else {
			map.put(obj, count.intValue() + 1);
		}
	}
	
	
	private static void decrementObjectGraph(Object obj) {
		Map<Object,Integer> map = threadToIdentitySet.get();
		Integer count = map.get(obj);
		if (count == null) {
			throw new IllegalStateException("count == null when decrementObjectGraph called--this should never happen");
		}
		else if (count.intValue() <= 0) {
			throw new IllegalStateException("count.intValue() <= 0 when decrementObjectGraph called--this should never happen");
		}
		else if (count.intValue() == 1) {
			map.remove(obj);	// CANNOT HAVE 0 counts in map in order for the logic inside isInObjectGraph/isObjectGraphEmpty to work
		}
		else {
			map.put(obj, count.intValue() - 1);
		}
	}
	
	
	// -------------------- static helper methods --------------------
	
	
	/**
	* For top performance, switches on indentLevel and returns a String constant if indentLevel is small (which it usually will be).
	* <p>
	* @throws IllegalArgumentException if indentLevel is < 0
	*/
	private static String getTabs(int indentLevel) throws IllegalArgumentException {
		// indentLevel is checked by StringUtil.repeatChars below
		
		switch (indentLevel) {
			case 0: return "";
			case 1: return "\t";
			case 2: return "\t\t";
			case 3: return "\t\t\t";
			case 4: return "\t\t\t\t";
			case 5: return "\t\t\t\t\t";
			default: return StringUtil.repeatChars('\t', indentLevel);
		}
	}
	
	
	// -------------------- constructor and helper methods --------------------
	
	
	/**
	* Constructor.
	* Determines object's state as restricted by filters, and places the result into {@link #classToFields}.
	* <p>
	* @param object may be anything, even null
	* @param filters may be anything, even null;
	* if more than one Filter is present, then each of them is connected by an implicit AND
	* (i.e. the Class/Field must be accepted by every Filter in order to be be present in classToFields)
	* @throws SecurityException if {@link Class#getDeclaredFields Class.getDeclaredFields} has an isssue
	*/
	public ObjectState(Object object, Filter... filters) throws SecurityException {
		filters = (filters != null) ? filters : new Filter[0];
		
		this.object = object;
		this.filters = filters;
		
		if (object == null) return;
		
		for (Class c = object.getClass(); c != null; c = c.getSuperclass()) {
			handleClass(c, object, filters);
		}
	}
	
	
	private void handleClass(Class c, Object object, Filter... filters) {
		if (classAccepted(c, object, filters)) {
			classToFields.put( c, extractFields(c, object, filters) );
		}
		
		for (Class interf : c.getInterfaces()) {
			handleClass(interf, object, filters);
		}
	}
	
	
	private boolean classAccepted(Class c, Object object, Filter... filters) {
		for (Filter filter : filters) {
			if (!filter.acceptClass(c, object)) return false;
		}
		return true;
	}
	
	
	private Field[] extractFields(Class c, Object object, Filter... filters) {
		Field[] fields = c.getDeclaredFields();		// According to its contract, getDeclaredFields should never return null
		Arrays.sort(fields, fieldComparator);
		
		List<Field> list = new ArrayList<Field>( fields.length );
		for (Field field : fields) {
			if (fieldAccepted(field, object, filters)) list.add(field);
		}
		return list.toArray( new Field[list.size()] );
	}
	
	
	private boolean fieldAccepted(Field field, Object object, Filter... filters) {
		for (Filter filter : filters) {
			if (!filter.acceptField(field, object)) return false;
		}
		return true;
	}
	
	
	// -------------------- accessors and mutators --------------------
	
	
	/** Accessor for {@link #object}. */
	public Object getObject() { return object; }
	
	
	/** Accessor for {@link #filters}.  Returns a shallow copy of the field. */
	public Filter[] getFilters() { return filters.clone(); }
	
	
	/** Accessor for {@link #classToFields}. */
	public Map<Class,Field[]> getClassToFields() { return classToFields; }
	
	
	// -------------------- convenience methods --------------------
	
	
	/**
	* Returns {@link #getObject getObject}'s class name.
	* In general, the result is getObject's class name, however, the following are special cases:
	* <ol>
	*  <li>if getObject returns null, then the String "&lt;NO TYPE: null Object reference>" is returned</li>
	*  <li>
	*		if getObject returns an array, then the result is formatted just like it would be declared in source code
	*		(i.e. as the element type followed by braces, e.g. "byte[]";
	*		this contrasts with the class name produced by the JVM for a byte array, which is the cryptic "[B")
	*  </li>
	* </ol>
	*/
	public String getType() {
		if (getObject() == null) return "<NO TYPE: null Object reference>";
		if (getObject().getClass().isArray()) return getObject().getClass().getComponentType().getName() + "[]";
		
		return getObject().getClass().getName();
	}
	
	
	/**
	* Returns 0 if <code>{@link #getObject}</code> returns null.
	* Else returns <code>getObject().hashCode()</code>.
	*/
	public int getHashCode() {
		if (getObject() == null) return 0;
		
		return getObject().hashCode();
	}
	
	
	/**
	* Returns the total number of fields stored in all the values of {@link #classToFields}.
	* <p>
	* @throws IllegalStateException if {@link #getClassToFields getClassToFields} raises it
	*/
	public int getNumberOfFields() throws IllegalStateException {
		int numberFields = 0;
		for (Class c : getClassToFields().keySet()) {
			numberFields += getClassToFields().get(c).length;
		}
		return numberFields;
	}
	
	
	// -------------------- toString and helper methods --------------------
	
	
	/** Simply returns <code>{@link #toString toString}(0)</code>. */
	@Override public String toString() throws RuntimeException {
		return toString(0);
	}
	
	
	/**
	* Returns a nicely formatted String which describes the data in this class
	* (chiefly, the state of {@link #object}).
	* <p>
	* Contract: the result is never blank and always ends with a newline.
	* <p>
	* @param indentLevel specifies how many levels of indentation to prepend to each line in the result
	* @throws RuntimeException (or some subclass) if any problem occurs; this RuntimeException may wrap some underlying Throwable, so be sure to inspect its cause
	*/
	public String toString(int indentLevel) throws RuntimeException {
		String tabs = getTabs(indentLevel);
		
		if (isInObjectGraph(getObject())) {
			return tabs + "<OBJECT ALREADY ENCOUNTERED by this thread in this ObjectState.toString call chain; see hashCode = " + getHashCode() + ">" + '\n';
		}
		
		StringBuilder sb = new StringBuilder( 64 + (getNumberOfFields() * 32) );	// guesstimate that each field averages 32 chars in length
		
		if (isObjectGraphEmpty()) {
			sb.append(tabs).append("viewing thread = ").append( Thread.currentThread().toString() ).append(", hashcode = ").append( getHashCode() ).append('\n');
		}
		
		try {
			incrementObjectGraph( getObject() );
			
			if (getObject() == null) {
				sb.append(tabs).append("<null object reference>").append('\n');
			}
			else {
				for (Class c : getClassToFields().keySet()) {
					sb.append(tabs).append( c.getName() ).append('\n');
					appendFields(c, sb, indentLevel + 1);	// + 1 because want the field lines to be indented relative to the type line
				}
			}
		}
		finally {
			decrementObjectGraph( getObject() );
		}
		
		return sb.toString();
	}
	
	
	private void appendFields(Class c, StringBuilder sb, int indentLevel) throws RuntimeException {
		String tabs = getTabs(indentLevel);
		
		Field[] fields = getClassToFields().get(c);
		if (fields.length == 0) {
			sb.append(tabs).append("<no fields (either none declared, or none pass all the filters)>").append('\n');
			return;
		}
		
		for (Field field : fields) {
			appendField(field, sb, tabs, indentLevel);
		}
	}
	
	
	private void appendField(Field field, StringBuilder sb, String tabs, int indentLevel) throws RuntimeException {
		boolean isAccessible = field.isAccessible();
		try {
			if (!isAccessible) field.setAccessible(true);	// ensure accessibility for the rest of this method
			Object value = field.get( getObject() );
			sb.append(tabs).append( field.getName() ).append(" = ").append( toString(value, indentLevel + 1) ).append('\n');	// + 1 because if need to use multilines, then want them indented relative to the field name line
		}
		catch (Throwable t) {
			throw ThrowableUtil.toRuntimeException(t);
		}
		finally {
			if (!isAccessible) {
				try {
					field.setAccessible(false);	// restore non-accessibility if that is what it originally was
				}
				catch (Throwable t) {
					LogUtil.getLogger2().logp(Level.WARNING, "ObjectState.FieldInfo", "constructor", "a problem occured while trying to restore non-accessibility of a field", t);
				}
			}
		}
	}
	
	
	/**
	* The following special cases are first checked in the following order:
	* <ol>
	*  <li>if obj is null, then "&lt;null>" is returned</li>
	*  <li>if obj is an array of any primitive type, then a comma separated list containing each element is returned</li>
	*  <li>
	*		if obj is an Object array, {@link Collection}, or {@link Map}, then each element is printed on its own indented line
	*		(altho if it is empty, then the text "&lt;EMPTY: XXX has no elements>" is returned,
	*		where XXX is either array, Collection, or Map as appropriate)
	*  </li>
	*  <li>
	*		if obj is in the <code>java.lang</code> package
	*		and/or is an instance of {@link CharSequence},
	*		then obj.toString() is returned
	*		(the idea here being that the toString method for these types is likely to produce a simple 1 line result)
	*  </li>
	* <ol>
	* For all other classes, this method returns a newline
	* followed by <code>new ObjectState( obj, {@link #getFilters} ).toString(indentLevel)</code> (i.e. a recursive call).
	*/
	private String toString(Object obj, int indentLevel) {
		if (obj == null) return "<null>";

		String separatorMultiLine = "\n" + getTabs(indentLevel);
		if (obj.getClass().isArray()) {
			if (obj.getClass().getComponentType().isPrimitive()) {
				if (obj.getClass().getComponentType() == boolean.class) {
					boolean[] booleans = (boolean[]) obj;
					if (booleans.length == 0) return "<EMPTY: array has no elements>";
					return StringUtil.toString(booleans, ", ");
				}
				if (obj.getClass().getComponentType() == byte.class) {
					byte[] bytes = (byte[]) obj;
					if (bytes.length == 0) return "<EMPTY: array has no elements>";
					return StringUtil.toString(bytes, ", ");
				}
				if (obj.getClass().getComponentType() == char.class) {
					char[] chars = (char[]) obj;
					if (chars.length == 0) return "<EMPTY: array has no elements>";
					return StringUtil.toString(chars, ", ");
				}
				if (obj.getClass().getComponentType() == double.class) {
					double[] doubles = (double[]) obj;
					if (doubles.length == 0) return "<EMPTY: array has no elements>";
					return StringUtil.toString(doubles, ", ");
				}
				else if (obj.getClass().getComponentType() == float.class) {
					float[] floats = (float[]) obj;
					if (floats.length == 0) return "<EMPTY: array has no elements>";
					return StringUtil.toString(floats, ", ");
				}
				else if (obj.getClass().getComponentType() == int.class) {
					int[] ints = (int[]) obj;
					if (ints.length == 0) return "<EMPTY: array has no elements>";
					return StringUtil.toString(ints, ", ");
				}
				else if (obj.getClass().getComponentType() == long.class) {
					long[] longs = (long[]) obj;
					if (longs.length == 0) return "<EMPTY: array has no elements>";
					return StringUtil.toString(longs, ", ");
				}
				else if (obj.getClass().getComponentType() == short.class) {
					short[] shorts = (short[]) obj;
					if (shorts.length == 0) return "<EMPTY: array has no elements>";
					return StringUtil.toString(shorts, ", ");
				}
					// boy does the above make me hate Java's mixed type system--get rid of primitives!
				else {
					throw new IllegalStateException("unable to handle the new Java primitive type " + obj.getClass().getComponentType());	// this should never happen
				}
			}
			else {
				return separatorMultiLine + StringUtil.toString((Object[]) obj, separatorMultiLine);
			}
		}
		else if (obj instanceof Collection) {
			Collection collection = (Collection) obj;
			if (collection.size() == 0) return "<EMPTY: Collection has no elements>";
			return separatorMultiLine + StringUtil.toString(collection, separatorMultiLine);
		}
		else if (obj instanceof Map) {
			Map map = (Map) obj;
			if (map.size() == 0) return "<EMPTY: Map has no elements>";
			return separatorMultiLine + StringUtil.toString(map, separatorMultiLine);
		}

		if (
			obj.getClass().getName().startsWith("java.lang") ||	// among others, this will pick up the primitive wrapper classes plus fundamental ones like Class, Object, String, StringBuffer, StringBuilder, Thread, etc
			(obj instanceof CharSequence)
		)
			return obj.toString();
		
		return "\n" + (new ObjectState( obj, getFilters() ).toString(indentLevel));
	}
	
	
	// -------------------- generateToStringCode --------------------
	
	
	/**
	* Returns a String which is valid java source code for a toString implementation for {@link #object}.
	* <p>
	* The result contains every field present in {@link #classToFields}.
	* Each such field is given its own line in the result,
	* both in the source code as well as the result that the source code produces.
	* <p>
	* If a field's access level is such that its direct access will cause an IllegalAccessException,
	* then the line for that field is actually commented out and a diagnostic error line is printed beneath it
	* which explains the issue and possibly offers resolution (the programmer will have to manually intervene).
	* <p>
	* If a field is a static, then a warning comment is placed at the end of its line
	* suggesting that the line be removed.
	* <p>
	* The programer is highly likely to need to hand edit the result before using it in an actual class:
	* the warnings and errors mentioned above should be dealt with,
	* custom decisions will likely need be made about how to toString individual fields, etc.
	* This method merely helps automate the process of writing a custom toString method
	* for someone who does not want to use the generic {@link #toString(int) toString}
	* (e.g. to obtain more customization or performance).
	* <p>
	* Contract: the result is never blank and always ends with a newline.
	*/
	public String generateToStringCode() {
		StringBuilder sb = new StringBuilder( 128 + (getNumberOfFields() * 32) );	// guesstimate that each field averages 32 chars in length
		
		sb.append("\t").append("public String toString() {").append('\n');
		
		if (getNumberOfFields() == 0) {
			sb.append("\t\t").append("return \"\";").append("// Note: this object either has no state, or none that pass the current filters").append('\n');
		}
		else {
			sb.append("\t\t").append("return").append('\n');
			for (Class c : getClassToFields().keySet()) {
				appendCodeForClass(c, sb);
			}
			sb.append("\t\t\t").append("\"\"").append('\t').append("// MANUAL IMPROVEMENT: remove this line as well as the final field line + symbol above").append('\n');	// need to add a String target of the final + symbol from a field line above in order to produce compilable code, but the user should hand edit this off
			sb.append("\t\t").append(';').append('\n');
		}
		
		sb.append("\t").append("}").append('\n');
		
		return sb.toString();
	}
	
	
	private void appendCodeForClass(Class c, StringBuilder sb) {
		sb.append("\t\t\t\t").append("// state from ").append( c.getName() ).append(':').append('\n');
		
		Field[] fields = getClassToFields().get(c);
		if (fields.length == 0) {
			sb.append("\t\t\t").append("// NONE: this class either declares no state, or declares none that pass the current filters").append('\n');
			return;
		}
		
		for (Field field : fields) {
			String errors = fieldErrors(field);
			if (errors.length() == 0) {
				sb.append("\t\t\t").append("\"").append( field.getName() ).append(" = \" + ").append( field.getName() ).append(" + ").append("\"\\n\" +").append( fieldWarnings(field) ).append('\n');
			}
			else {
				sb.append("//").append("\t\t\t").append("\"").append( field.getName() ).append(" = \" + ").append( field.getName() ).append(" + ").append("\"\\n\" +").append( fieldWarnings(field) ).append('\n');
				sb.append( errors ).append('\n');
			}
		}
	}
	
	
	private String fieldErrors(Field field) {
		StringBuilder sb = null;
		
		if (field.getClass() != getObject().getClass()) {
			if (
				ReflectionUtil.isDefault(field) &&
				!field.getClass().getPackage().equals(object.getClass().getPackage())
			) {
				if (sb == null) sb = new StringBuilder();
				sb.append("// +++ INACCESSABLE: field above has default access and belongs to a different Class that is in a different package than object's");
			}
			else if (ReflectionUtil.isPrivate(field)) {
				if (sb == null) sb = new StringBuilder();
				sb.append("// +++ INACCESSABLE: field above has private access and belongs to a different Class than object's");
			}
			
			if (sb != null) sb.append("; usual fix: replace direct field access with an accessor method").append('\n');
		}
		
		return (sb != null) ? sb.toString() : "";
	}
	
	
	private String fieldWarnings(Field field) {
		StringBuilder sb = null;
		
		if (ReflectionUtil.isStatic(field)) {
			if (sb == null) sb = new StringBuilder();
			sb.append('\t').append("// WARNING: field is static; may wish to skip it");
		}
		
		return (sb != null) ? sb.toString() : "";
	}
	
	
	// -------------------- serializeState, deserializeState, checkForChanges --------------------
	
	
// +++ write code which checks to see if a class has changed its state
// probably need 3 methods:
//	1: serializeState
//		serializes classToFields to an OutputStream
//	2: deserializeState
//		deserializes classToFields from an InputStream
//	2: checkForChanges
//		compares 2 classToFields (e.g. a previously serialzed one and the current one); need to
//		a) find all classes/interfaces which are in one but not the other,
//		and on classes/interfaces whose modifiers have changed
//		b) for classes/interfaces which are in both, need to report on fields which are in one but not the other,
//		and on fields whose modifiers and types changed

// When write the 2 methods above, should change generateToStringCode above
// so that it the first time that it is executed, it makes calls to deserializeState and checkForChanges
// which detect if it is noww obsolete
	
	
	// -------------------- FieldComparator (static inner class) --------------------
	
	
	/**
	* Imposes an ordering on Fields from the same Class.
	* <p>
	* This class is multithread safe: it is stateless.
	*/
	private static class FieldComparator implements Comparator<Field>, Serializable {
		
		private static final long serialVersionUID = 1;
		
		/**
		* Compares this 2 Fields (which must be from the same Class) using the following tests:
		* <ol>
		*  <li>if f1 is static and f2 is not, f1 is less than f2, and vice versa</li>
		*  <li>if f1 is final and f2 is not, f1 is less than f2, and vice versa</li>
		*  <li>comparison of their names</li>
		* </ol>
		* <p>
		* This ordering is consistent with equals.
		* <p>
		* @throws IllegalArgumentException if f1 or f2 == null; f1.getClass() != f2.getClass()
		*/
		public int compare(Field f1, Field f2) throws IllegalArgumentException {
			if (f1 == null) throw new IllegalArgumentException("f1 == null");
			if (f2 == null) throw new IllegalArgumentException("f2 == null");
			if (f1.getClass() != f2.getClass()) throw new IllegalArgumentException("f1.getClass() != f2.getClass()");
			
			if (ReflectionUtil.isStatic(f1) && !ReflectionUtil.isStatic(f2)) return -1;
			if (!ReflectionUtil.isStatic(f1) && ReflectionUtil.isStatic(f2)) return 1;
			
			if (ReflectionUtil.isFinal(f1) && !ReflectionUtil.isFinal(f2)) return -1;
			if (!ReflectionUtil.isFinal(f1) && ReflectionUtil.isFinal(f2)) return 1;
			
			int nameComparison = f1.getName().compareTo( f2.getName() );
			if (nameComparison != 0) return nameComparison;
			
			throw new IllegalStateException("unable to rank f1 = " + f1 + " and f2 = " + f2);	// should never reach here
		}
		
	}
	
	
	// -------------------- Filter and some standard implementations (static inner interface and classes) --------------------
	
	
	/**
	* Interface for types which accept/reject Classes/Fields.
	* May be used to limit what gets put into {@link #classToFields} during construction.
	* Implementations should be multithread safe.
	*/
	public static interface Filter {
		
		/** Returns true if this instance accepts c for object. */
		public boolean acceptClass(Class c, Object object);
		
		/** Returns true if this instance accepts field for object. */
		public boolean acceptField(Field field, Object object);
		
	}
	
	
	// Class-only filters:
	
	
	/** Base class for Filters which care only about Classes.  Implements {@link #acceptField acceptField} to always return true. */
	public static abstract class ClassFilter implements Filter {
		
		/** Always returns true. */
		public boolean acceptField(Field field, Object object) { return true; }
		
	}
	
	
	/** Accepts a Class only if it is object's class.  Otherwise rejects everything else. This class is multithread safe: it is stateless. */
	public static class AcceptOnlyImmediateClass extends ClassFilter {
		
		/** Returns <code>c == object.getClass()</code>. */
		public boolean acceptClass(Class c, Object object) { return (c == object.getClass()); }
		
	}
	
	
	/** Rejects a Class if it is Object.class.  Otherwise accepts everything else. This class is multithread safe: it is stateless. */
	public static class RejectObjectClass extends ClassFilter {
		
		/** Returns <code>!(c == Object.class)</code>. */
		public boolean acceptClass(Class c, Object object) { return !(c == Object.class); }
		
	}
	
	
	/** Rejects a Class if it is an interface.  Otherwise accepts everything else. This class is multithread safe: it is stateless. */
	public static class RejectInterfaces extends ClassFilter {
		
		/** Returns <code>!c.isInterface()</code>. */
		public boolean acceptClass(Class c, Object object) { return !c.isInterface(); }
		
	}
	
	
	// Field-only filters:
	
	
	/** Base class for Filters which care only about Fields.  Implements {@link #acceptClass acceptClass} to always return true. */
	public static abstract class FieldFilter implements Filter {
		
		/** Always returns true. */
		public boolean acceptClass(Class c, Object object) { return true; }
		
	}
	
	
	/** Accepts a Field only if it is public or protected.  Otherwise rejects everything else. This class is multithread safe: it is stateless. */
	public static class AcceptOnlyPublicProtectedFields extends FieldFilter {
		
		/** Returns <code>!{@link ReflectionUtil#isPublic ReflectionUtil.isPublic}(field)</code>. */
		public boolean acceptField(Field field, Object object) { return ReflectionUtil.isPublic(field) || ReflectionUtil.isProtected(field); }
		
	}
	
	
	/** Rejects a Field if it is static.  Otherwise accepts everything else. This class is multithread safe: it is stateless. */
	public static class RejectStaticFields extends FieldFilter {
		
		/** Returns <code>!{@link ReflectionUtil#isStatic ReflectionUtil.isStatic}(field)</code>. */
		public boolean acceptField(Field field, Object object) { return !ReflectionUtil.isStatic(field); }
		
	}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_toString() {
			System.out.println();
			System.out.println("Here is how null is handled:");
			System.out.println( new ObjectState(null).toString(1) );
			
			System.out.println();
			System.out.println("Here is how an object with no fields is handled:");
			System.out.println( new ObjectState( new NoFields() ).toString(1) );
			
			System.out.println();
			System.out.println("Here is how a complex class is handled:");
			System.out.println( new ObjectState( new SuperClass() ).toString(1) );
			
			System.out.println();
			System.out.println("Here is how an object with a complex superclass is handled:");
			System.out.println( new ObjectState( new SubClass() ).toString(1) );
			
			System.out.println();
			System.out.println("Here is how an object which has fields are non-trivial types is handled:");
			System.out.println( new ObjectState( new CompoundClass() ).toString(1) );
			
			ReferringClass num1 = new ReferringClass();
			ReferringClass num2 = new ReferringClass();
			ReferringClass num3 = new ReferringClass();
			num1.other = num2;
			num2.other = num3;
			num3.other = num1;
			System.out.println();
			System.out.println("Here is how an object whose object graph has a circular chain of references is handled:");
			System.out.println( new ObjectState(num1).toString(1) );
			
			System.out.println();
			System.out.println("Here is how an object with a complex interface hierarchy  is handled:");
			System.out.println( new ObjectState( new InterfaceImplementingClass() ).toString(1) );
		}
		
		@Test public void test_Filters() {
			System.out.println();
			System.out.println("Here is the effect of the AcceptOnlyImmediateClass filter:");
			System.out.println( new ObjectState( new SubClass(), new AcceptOnlyImmediateClass() ).toString(1) );
			
			System.out.println();
			System.out.println("Here is the effect of the RejectObjectClass filter:");
			System.out.println( new ObjectState( new NoFields(), new RejectObjectClass() ).toString(1) );
			
			System.out.println();
			System.out.println("Here is the effect of the RejectInterfaces filter:");
			System.out.println( new ObjectState( new InterfaceImplementingClass(), new RejectInterfaces() ).toString(1) );
			
			System.out.println();
			System.out.println("Here is the effect of the AcceptOnlyPublicProtectedFields plus the RejectStaticFields filter:");
			System.out.println( new ObjectState( new SuperClass(), new AcceptOnlyPublicProtectedFields(), new RejectStaticFields() ).toString(1) );
		}
		
		@Test public void test_generateToStringCode() {
			System.out.println();
			System.out.println("Here is the output of generateToStringCode for a complex class:");
			System.out.println( new ObjectState( new SubClass() ).generateToStringCode() );
		}
		
// +++ the code above has no assertions, which is not so good; it relies on manual inspection at the moment;
// should I compare literal string output agreement?  or is that too rigid?
		
// +++ the classes below offer a decent variety of challenges for my code,
// but what really ought to do in order to stress test ObjectState
// is to randomly generate Classes and interfaces and fields with all permutations of modifiers and types
// and then confirm that no bugs are present (how? see if can recreate the original source code?).
// Not sure how to create Java classes on the fly
//	--need to look into the new compiler api
//	--or try hotpatching: http://www.fasterj.com/articles/hotpatch1.shtml
		
		private static class NoFields {}
		
		/**
		* Has fields which provide complete coverage of:
		* <pre>
			all modifier permutations
				public protected <none> private
				static
				final
			many field types
				primitives
				arrays, Collection, Map
				java.lang and CharSequence</pre>
		* Also, all the values are distinct, both within this class and from {@link SubClass} below.
		*/
		private static class SuperClass {
			public static final boolean boolean_1 = false;
			protected static final byte byte_1 = (byte) (1*100 + 1);
			static final char char_1 = (char) (1*100 + 2);
			private static final double double_1 = (double) (1*100 + 3);
			
			public static float float_1 = (float) (1*100 + 4);
			protected static int int_1 = 1*100 + 5;
			static long long_1 = (long) (1*100 + 6);
			private static short short_1 = (short) (1*100 + 7);
			
			public final double[] doubleArray_1 = new double[] {(double) (1*110 + 1), (double) (1*110 + 2), (double) (1*110 + 3)};
			protected final Object[] ObjectArray_1 = new Object[] {new Object(), new Object()};
			final Collection<Object> ObjectCollection_1 = Arrays.asList( new Object(), new Object() );
			private final Map<Integer,Character> intsToCharsMap_1 = new TreeMap<Integer,Character>();
			{
				intsToCharsMap_1.put(1*120 + 1, 'a');
				intsToCharsMap_1.put(1*120 + 2, 'b');
			}
			
			public Class Class_1 = SuperClass.class;
			protected Object Object_1 = null;
			String String_1 = "String_1";
			private StringBuffer StringBuffer_1 = new StringBuffer("StringBuffer_1");
		}
		
		/** Has fields which provide similar complete coverage and distinct values as {@link SuperClass}. */
		private static class SubClass extends SuperClass {
			public static final boolean boolean_2 = true;
			protected static final byte byte_2 = (byte) (2*100 + 1);
			static final char char_2 = (char) (2*100 + 2);
			private static final double double_2 = (double) (2*100 + 3);
			
			public static float float_2 = (float) (2*100 + 4);
			protected static int int_2 = 2*100 + 5;
			static long long_2 = (long) (2*100 + 6);
			private static short short_2 = (short) (2*100 + 7);
			
			public final float[] floatArray_2 = new float[0];
			protected final Object[] ObjectArray_2 = null;
			final Collection<Object> ObjectCollection_2 = new LinkedList<Object>();
			private final Map<Integer,Character> intsToCharsMap_2 = new TreeMap<Integer,Character>();
			
			public Class Class_2 = SubClass.class;
			protected Object Object_2 = new Object();
			String String_2 = "String_2";
			private StringBuffer StringBuffer_2 = new StringBuffer("StringBuffer_2");
		}
		
		private static class CompoundClass {
			static InnerClass1 innerClass1 = new InnerClass1();
			InnerClass2 innerClass2 = new InnerClass2();
			
			static class InnerClass1 {
				Object field1 = new Object();
			}
			
			class InnerClass2 {
				Object field2 = new Object();
			}
		}
		
		private static class ReferringClass {
			private Object other;
		}
		
		private static interface SuperInterface1 {
			public static final Object interfaceConstant1 = new Object();
		}
		
		private static interface SubInterface2 extends SuperInterface1 {
			public static final Object interfaceConstant1 = new Object();
		}
		
		private static interface SuperInterface3 extends SuperInterface1 {	// NOTE: am doing extends SuperInterface1 so that SuperInterface1 appears twice in the interface hierarchy of InterfaceImplementingClass below in order to test how this situation is handled (SuperInterface1 should only appear once in the printout, just after SubInterface2)
			public static final Object interfaceConstant3 = new Object();
		}
		
		private static interface SubInterface4 extends SuperInterface3 {
			public static final Object interfaceConstant4 = new Object();
		}
		
		private static class InterfaceImplementingClass implements SubInterface2, SubInterface4 {}
		
	}
	
	
}