/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.util;


import java.io.File;
import java.io.InputStream;
import java.io.IOException;

import org.junit.Assert;
import org.junit.Test;

import bb.io.StreamUtil;


/**
* Provides static utility methods relating to the operating system that the JVM is running on.
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @author Brent Boyer
*/
public class OsUtil {
	
	
	// -------------------- constants --------------------
	
	
	/** Official name of the Microsoft Windows 98 operating system. */
	private static final String windows98_osName = "Windows 98";
	
	/** Official name of the Microsoft Windows NT operating system. */
	private static final String windowsNt_osName = "Windows NT";
	
	/** Official name of the Microsoft Windows 2000 operating system. */
	private static final String windows2000_osName = "Windows 2000";
	
	/** Official name of the Microsoft Windows XP operating system. */
	private static final String windowsXp_osName = "Windows XP";
	
	
	/** Name of the command processor program to execute DOS commands under Windows 98 (and below). */
	private static final String windows98_commandProcessor = "command.com /c";	// Note: the /c makes it end after executing a single command; see http://www.microsoft.com/resources/documentation/windows/xp/all/proddocs/en-us/cmd.mspx?mfr=true
	
	/** Name of the command processor program to execute DOS commands under Windows NT (and above). */
	private static final String windowsNt_commandProcessor = "cmd /c";	// Note: the /c makes it end after executing a single command; see http://www.microsoft.com/resources/documentation/windows/xp/all/proddocs/en-us/cmd.mspx?mfr=true
	
	
	// -------------------- getXXX --------------------
	
	
	/**
	* Returns a known <i>invalid</i> command for the current operating system.
	* This method is used during testing.
	* <p>
	* @throws IllegalStateException if unable to determine an invalid command for this operating system
	*/
	public static String getCommandInvalid() throws IllegalStateException {
		if (isMicrosoft()) return getDosCommandProcessor() + " " + "sdvnjsddvjsj" + "\n";
		else if (isUnix()) return "sdvnjsddvjsj" + "\n";
		else throw new IllegalStateException("unable to determine an invalid command for this operating system, since it is of unknown type; its name is: " + getOsName());
	}
	
	
	/**
	* Returns a known valid command for the current operating system.
	* This method is used during testing.
	* <p>
	* @throws IllegalStateException if unable to determine a valid command for this operating system
	*/
	public static String getCommandValid() throws IllegalStateException {
		if (isMicrosoft()) return getDosCommandProcessor() + " " + "dir" + "\n";
		else if (isUnix()) return "ls -a" + "\n";
		else throw new IllegalStateException("unable to determine a valid command for this operating system, since it is of unknown type; its name is: " + getOsName());
	}
	
	
	/**
	* Returns the name of the command processor program used to execute DOS commands if this is a Microsoft operating system.
	* <p>
	* @throws IllegalStateException if unable to determine what DOS command processor to use for this operating system
	*/
	public static String getDosCommandProcessor() throws IllegalStateException {
		if (isWindows98()) return windows98_commandProcessor;
		else if (isWindowsNT() || isWindows2000() || isWindowsXP()) return windowsNt_commandProcessor;
		else throw new IllegalStateException("unable to determine what DOS command processor to use for this operating system, as it may not even be from Microsoft; its name is: " + getOsName());
	}
	
	
	/** Returns the current operating system name. */
	public static String getOsName() {
		return System.getProperties().getProperty("os.name");
	}
	
	
	/** Returns the current operating system version. */
	public static String getOsVersion() {
		return System.getProperties().getProperty("os.version");
	}
	
	
	/** Returns a concatenation (with a space char separator) of the current operating system name and version. */
	public static String getOsNameAndVersion() {
		return getOsName() + " " + getOsVersion();
	}
	
	
	// -------------------- isXXX --------------------
	
	
	/**
	* Determines whether or not the operating system is from Microsoft (e.g. DOS, Windows).
	* All of these tests must be passed:
	* <ol>
	*  <li>{@link File#separatorChar File.separatorChar} == '\\'</li>
	*  <li>{@link #getOsName getOsName} contains the text "windows" or "msdos" (case irrelevant)</li>
	*  <li>the dos command <code>ver</code> was sucessfully executed</li>
	* </ol>
	*/
	public static boolean isMicrosoft() {
		String name = getOsName().toLowerCase();
		return
			(File.separatorChar == '\\') &&
			(name.contains("windows")) || (name.contains("msdos")) &&
			execSucceeded(getDosCommandProcessor() + " " + "ver" + "\n");	// CRITICAL: do not call getCommandValid here, since that would involve circular calls between these 2 methods
	}
	
	
	/**
	* Determines whether or not the operating system is Windows 98
	* by comparing {@link #windows98_osName} to the result of getOsName.
	*/
	public static boolean isWindows98() {
		return windows98_osName.equals(getOsName());
	}
	
	
	/**
	* Determines whether or not the operating system is Windows NT
	* by comparing {@link #windowsNt_osName} to the result of getOsName.
	*/
	public static boolean isWindowsNT() {
		return windowsNt_osName.equals(getOsName());
	}
	
	
	/**
	* Determines whether or not the operating system is Windows 2000
	* by comparing {@link #windows2000_osName} to the result of getOsName.
	*/
	public static boolean isWindows2000() {
		return windows2000_osName.equals(getOsName());
	}
	
	
	/**
	* Determines whether or not the operating system is Windows 2000
	* by comparing {@link #windowsXp_osName} to the result of getOsName.
	*/
	public static boolean isWindowsXP() {
		return windowsXp_osName.equals(getOsName());
	}
	
	
	/**
	* Determines whether or not the operating system is Unix (including Linux and OS/X).
	* All of these tests must be passed:
	* <ol>
	*  <li>{@link File#separatorChar File.separatorChar} == '/'</li>
	*  <li>the unix command <code>uname</code> was sucessfully executed</li>
	* </ol>
	*/
	public static boolean isUnix() {
		return
			(File.separatorChar == '/') &&
			execSucceeded("uname" + "\n");	// CRITICAL: do not call getCommandValid here, since that would involve circular calls between these 2 methods
	}
	
	
	// -------------------- exec --------------------
	
	
/*
+++ Accessing Windows DLL with pure Java (no JNI):
	https://jna.dev.java.net/
*/
	/**
	* Executes a native (operating system specific) command.
	* The mode is synchronous: this method does not return until the command has finished.
	* <p>
	* <i>The command must not require any input, since this method does not know how to supply it with any;
	* it will hang forever waiting for input if you mistakenly call this method with such a command.</i>
	* The command may, however, produce arbitrary output on either stream (its output or err streams)
	* which will properly be drained by this method.
	* <p>
	* Note: the current implementation spawns the new process by calling {@link Runtime#exec(String)}
	* which means that the new process inherits env vars and working directory from this java process.
	* See also
	* <a href="http://java.sun.com/developer/JDCTechTips/2001/tt1204.html">Accessing the Environment from Java Applications and Working With Number Bases</a>.
	* <p>
	* @return a byte[] containing everything written to the native process's standard output stream
	* @throws IllegalArgumentException if command is blank
	* @throws SecurityException if a security manager exists and its checkExec method doesn't allow creation of a subprocess 
	* @throws InterruptedException if calling thread is interrupted while waiting for the dispatch thread to finish excecuting task.run 
	* @throws IOException if an I/O error occurs
	* @throws RuntimeException if the native process terminates abnormally or wrote any data to its error stream,
	* or if some other error was detected (will be stored in the cause field of this RuntimeException)
	* @see "Brent's nativeExecutor project for more sophisticated code"
	*/
	public static byte[] exec(String command) throws IllegalArgumentException, SecurityException, InterruptedException, IOException, RuntimeException {
		if (StringUtil.isBlank(command)) throw new IllegalArgumentException("command is blank");
		
			// execute command:
		Process p = Runtime.getRuntime().exec(command);
// +++ see also java.lang.ProcessBuilder
		
			// simultaneously and independently drain p's in and err outputs as it executes:
		StreamDrainer inDrainer = new StreamDrainer( p.getInputStream() );
		Thread tIn = new Thread( inDrainer, "SystemUtil_InStreamDrainer");
		tIn.setPriority( Thread.NORM_PRIORITY );
		tIn.start();
		
		StreamDrainer errDrainer = new StreamDrainer( p.getErrorStream() );
		Thread tErr = new Thread( errDrainer, "SystemUtil_ErrStreamDrainer" );
		tErr.setPriority( Thread.NORM_PRIORITY );
		tErr.start();
		
			// wait for all processes/threads to finish:
		p.waitFor();
		tIn.join();
		tErr.join();
		
			// throw a RuntimeException if p did not sucessfully terminate or some error occured:
		if (p.exitValue() != 0) throw new RuntimeException("Execution of command = " + command + " resulted in abnormal termination (exit code = " + p.exitValue() + ")");
		if (errDrainer.getBytes().length != 0) throw new RuntimeException("Execution of command = " + command + " produced in the following error output: " + new String(errDrainer.getBytes()));
		if (inDrainer.getThrowable() != null) throw new RuntimeException("Execution of command = " + command + " encountered a Throwable (see cause)", inDrainer.getThrowable());
		if (errDrainer.getThrowable() != null) throw new RuntimeException("Execution of command = " + command + " encountered a Throwable (see cause)", errDrainer.getThrowable());
		
			// return p's output if it did successfully terminate:
		return inDrainer.getBytes();
	}
	
	
	private static class StreamDrainer implements Runnable {
		
		private final InputStream in;
		private byte[] bytes = null;
		private Throwable throwable = null;
		
		private StreamDrainer(InputStream in) throws IllegalArgumentException {
			if (in == null) throw new IllegalArgumentException("in == null");
			
			this.in = in;
		}
		
		public void run() {
			try {
				byte[] bytes = StreamUtil.drain(in);	// Note: StreamUtil.drain closes the InputStream
				setBytes(bytes);	// CRITICAL: setBytes is called on a separate line from StreamUtil.drain (as opposed to wrapping it around the drain) to avoid holding this instance's lock during the entire draining process
			}
			catch (Throwable t) {
				setThrowable(t);
			}
		}
		
		/**
		* Returns all the bytes drained from the InputStream.
		* Must only be called after draining is finished.
		* <b>Warning:</b> if an error occurred during draining, only part of the information may have been recorded.
		* @throws IllegalStateException if called while draining is still going on
		*/
		private synchronized byte[] getBytes() {
			if (bytes == null) throw new IllegalStateException("draining is still going on");
			
			return bytes;
		}
		
		private synchronized void setBytes(byte[] bytes) { this.bytes = bytes; }
		
		private synchronized Throwable getThrowable() { return throwable; }
		
		private synchronized void setThrowable(Throwable throwable) { this.throwable = throwable; }
		
	}
/*
this version of the class would be very useful if ever write a version of exec which executes asynchronously,
since then the streams could be continuously drained to files or whatever

	private static class StreamDrainer implements Runnable {
		
		private final BufferedReader reader;
		private final Writer writer;
		private long charCount = 0;
		private Throwable throwable = null;
		
		**
		* @throws FileNotFoundException if file exists but is a directory rather than a regular file, does not exist but cannot be created, or cannot be opened for any other reason 
    	* @throws SecurityException if a security manager exists and its checkWrite method denies write access to file
		*
		private StreamDrainer(InputStream in, File file) throws FileNotFoundException, SecurityException {
			this( new BufferedReader(new InputStreamReader(in)), new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file))) );
		}
		
		private StreamDrainer(BufferedReader reader, Writer writer) throws IllegalArgumentException {
			if (reader == null) throw new IllegalArgumentException("reader == null");
			if (writer == null) throw new IllegalArgumentException("writer == null");
			
			this.reader = reader;
			this.writer = writer;
		}
		
		public void run() {
			try {
				for (String line = reader.readLine(); line != null; line = reader.readLine()) {
// +++ hmm, would really like a reader class which preserved the line termination chars
// both so could accurate count as well as could perfectly write them to writer
// should add a readLine version to ParseReader which preserves them?
					incrementCharCount(line.length());
					writer.write(line);
					writer.write('\n');
					writer.flush();
				}
			}
			catch (Throwable t) {
				setThrowable(t);
			}
		}
		
		**
		* Returns the count of the chars drained from the InputStream.
		* Note that line terminator chars are not included in the count.
		*
		public synchronized long getCharCount() { return charCount; }
		
		private synchronized void incrementCharCount(int n) { charCount += n; }
		
		public synchronized Throwable getThrowable() { return throwable; }
		
		private synchronized void setThrowable(Throwable throwable) { this.throwable = throwable; }
		
	}
*/
	
	
	/**
	* Calls {@link #exec exec}(command).
	* Returns true if no Exception of any type was thrown, false otherwise.
	*/
	private static boolean execSucceeded(String command) {
		try {
			exec(command);
			return true;
		}
		catch (Exception e) {
			return false;
		}
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private OsUtil() {}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_getXXX() {
			System.out.println("getOsName() = " + getOsName());
			System.out.println("getOsVersion() = " + getOsVersion());
		}
		
		@Test public void test_exec_shouldPass() throws Exception {
			System.out.println("Output from executing the valid command \"" + getCommandValid().replace("\n", "") + "\":");
			System.out.println( new String( exec( getCommandValid() ) ) );
		}
		
		@Test(expected=RuntimeException.class) public void test_exec_shouldFail() throws Exception {
			exec(getCommandInvalid());
		}
		
	}
	
	
}