/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*
Programmer Notes:

+++ should NOT extend the broken Properties class?
	--that class directly exposes its HashTable, which is not an api that want exposed?

+++ really need to modify Properties2 to support what proper subsets of keys can be specified together
	--some programs have switches that only make sense to allow certain subsets to appear together
	--as an example, see TarUtil.main where archiving and extraction have different requirements

--could add more convenience methods like getShort

--need to introduce support for multiple files, track overwritting of values, etc
*/


package bb.util;


import java.io.*;
import java.util.Properties;

import bb.io.FileUtil;
import bb.io.StreamUtil;
import bb.util.ThrowableUtil;


/**
* This class extends Properties to offer more convenience methods.
* <p>
* This class is multithread safe: every public method is synchronized.
* <p>
* @author Brent Boyer
*/
public class Properties2 extends Properties {
	
	
	// -------------------- constants --------------------
	
	
	private static final long serialVersionUID = 1;
	
	
	// -------------------- static methods --------------------
	
	
	/**
	* Parses a new {@link Properties} instance from the contents of file.
	* <p>
	* This method never throws a checked Exception for the reasons described in {@link #Properties2(File)}.
	* <p>
	* @throws IllegalArgumentException if {@link FileUtil#check FileUtil.check} objects to file,
    * file contains a malformed Unicode escape sequence
    * @throws SecurityException if a security manager exists and its checkRead method denies read access to file
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static Properties extractProperties(File file) throws IllegalArgumentException, SecurityException, RuntimeException {
		FileUtil.check(file);
		
		InputStream in = null;
		try {
			Properties properties = new Properties();
			in = new BufferedInputStream( new FileInputStream(file) );
			properties.load(in);
			return properties;
		}
		catch (Throwable t) {
			throw ThrowableUtil.toRuntimeException(t); 
		}
		finally {
			StreamUtil.close(in);
		}
	}
	
	
	// -------------------- constructors --------------------
	
	
	public Properties2() {
		super();
	}
	
	
	public Properties2(Properties defaults) {
		super(defaults);
	}
	
	
	/**
	* Constructs a new instance from the contents of file.
	* <p>
	* This method never throws a checked Exception in order to support 1-line field assignments like
	* <blockquote><code>private final int port = new Properties2(file).getInt("Port", 7462);</code></blockquote>
	* <p>
	* @throws IllegalArgumentException if {@link FileUtil#check FileUtil.check} objects to file,
    * file contains a malformed Unicode escape sequence
    * @throws SecurityException if a security manager exists and its checkRead method denies read access to file
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public Properties2(File file) throws IllegalArgumentException, SecurityException, RuntimeException {
		this( extractProperties(file) );
	}
	
	
	// -------------------- ensureProperty --------------------
	
	
	/**
	* If no value is associated with key, assigns defaultValue as the property's value.
	* So, upon normal method return, key will always be associated with a value.
	* <p>
	* @return the value associated with key (either its already existing value, or defaultValue if none previously existed)
	* @throws IllegalArgumentException if key == null or defaultValue == null
	*/
	public synchronized String ensureProperty(String key, String defaultValue) throws IllegalArgumentException {
		if (key == null) throw new IllegalArgumentException("key == null");
		if (defaultValue == null) throw new IllegalArgumentException("defaultValue == null");

		String value = getProperty(key);
		if (value == null) {
			setProperty(key, defaultValue);
			return defaultValue;		
		}
		else
			return value;
	}


	// -------------------- getXXX --------------------
	
	
	/**
	* Returns <code>Boolean.parseBoolean( ensureProperty(key, String.valueOf(defaultValue)) )</code>.
	* <p>
	* @throws IllegalArgumentException if key == null
	*/
	public synchronized boolean getBoolean(String key, boolean defaultValue) {
		return Boolean.parseBoolean( ensureProperty(key, String.valueOf(defaultValue)) );
	}


	/**
	* Returns <code>Double.parseDouble( ensureProperty(key, String.valueOf(defaultValue)) )</code>.
	* <p>
	* @throws IllegalArgumentException if key == null
	* @throws NumberFormatException if key's value is not a parsable double
	*/
	public synchronized double getDouble(String key, double defaultValue) throws NumberFormatException {
		return Double.parseDouble( ensureProperty(key, String.valueOf(defaultValue)) );
	}


	/**
	* If defaultValue != null, returns <code>new File( ensureProperty(key, defaultValue.getPath()) )</code>.
	* Otherwise, sees if key has any existing value and returns a new File constructed from that value if so,
	* else it returns null.
	* <p>
	* @throws IllegalArgumentException if key == null
	*/
	public synchronized File getFile(String key, File defaultValue) throws NumberFormatException {
		if (defaultValue != null) {
			return new File( ensureProperty(key, defaultValue.getPath()) );
		}
		else {
			String value = getProperty(key);
			if (value != null) return new File(value);
			else return null;
		}
	}


	/**
	* Returns <code>Integer.parseInt( ensureProperty(key, String.valueOf(defaultValue)) )</code>.
	* <p>
	* @throws IllegalArgumentException if key == null
	* @throws NumberFormatException if key's value is not a parsable int
	*/
	public synchronized int getInt(String key, int defaultValue) throws NumberFormatException {
		return Integer.parseInt( ensureProperty(key, String.valueOf(defaultValue)) );
	}


	/**
	* Returns <code>Long.parseLong( ensureProperty(key, String.valueOf(defaultValue)) )</code>.
	* <p>
	* @throws IllegalArgumentException if key == null
	* @throws NumberFormatException if key's value is not a parsable long
	*/
	public synchronized long getLong(String key, long defaultValue) throws NumberFormatException {
		return Long.parseLong( ensureProperty(key, String.valueOf(defaultValue)) );
	}


	// -------------------- getMandatoryXXX  --------------------
	
	
	/**
	* Returns the value associated with key, first checking that the value is non-null; null will never be returned.
	* <p>
	* @throws IllegalArgumentException if key == null; no value found for key
	*/
	public synchronized String getMandatoryProperty(String key) throws IllegalArgumentException {
		if (key == null) throw new IllegalArgumentException("key == null");
		
		String value = getProperty(key);
		if (value == null) throw new IllegalArgumentException("no value found for key = " + key);
		
		return value;
	}


	/**
	* Returns <code>Double.parseDouble( getMandatoryProperty(key) )</code>.
	* <p>
	* @throws IllegalArgumentException if key == null; no value found for key
	* @throws NumberFormatException if the property's value is not a parsable double
	*/
	public synchronized double getMandatoryDouble(String key) throws IllegalArgumentException, NumberFormatException {
		return Double.parseDouble( getMandatoryProperty(key) );
	}


	/**
	* Returns <code>new File( getMandatoryProperty(key) )</code>.
	* <p>
	* @throws IllegalArgumentException if key == null; no value found for key
	*/
	public synchronized File getMandatoryFile(String key) throws IllegalArgumentException, NumberFormatException {
		return new File( getMandatoryProperty(key) );
	}


	/**
	* Returns <code>Integer.parseInt( getMandatoryProperty(key) )</code>.
	* <p>
	* @throws IllegalArgumentException if key == null; no value found for key
	* @throws NumberFormatException if the property's value is not a parsable int
	*/
	public synchronized int getMandatoryInt(String key) throws IllegalArgumentException, NumberFormatException {
		return Integer.parseInt( getMandatoryProperty(key) );
	}


	/**
	* Returns <code>Long.parseLong( getMandatoryProperty(key) )</code>.
	* <p>
	* @throws IllegalArgumentException if key == null; no value found for key
	* @throws NumberFormatException if the property's value is not a parsable long
	*/
	public synchronized long getMandatoryLong(String key) throws IllegalArgumentException, NumberFormatException {
		return Long.parseLong( getMandatoryProperty(key) );
	}


	// -------------------- getNonBlankProperty  --------------------	
	
	
	/**
	* Returns the value associated with key, first checking that the value is non-null; null will never be returned.
	* <p>
	* @throws IllegalArgumentException if key == null; no value found for key
	*/
	public synchronized String getNonBlankProperty(String key) throws IllegalArgumentException {
		if (key == null) throw new IllegalArgumentException("key == null");
		
		String value = getProperty(key);
		if (value == null) throw new IllegalArgumentException("no value found for key = " + key);
		if (StringUtil.isBlank(value)) throw new IllegalArgumentException("the value found for key = " + key + " is all whitespace");
		
		return value;
	}


}
