/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.util;


import java.lang.reflect.*;


/**
* This class provides various static utility methods useful when doing reflection.
* <p>
* This class is multithread safe: it is stateless.
* <p>
* @author Brent Boyer
*/
public class ReflectionUtil {
	
	
	// -------------------- isXXX --------------------
	
	
	/** Returns <code>{@link Modifier#isPublic Modifier.isPublic}( f.{@link Field#getModifiers} )</code>. */
	public static boolean isPublic(Field f) { return Modifier.isPublic( f.getModifiers() ); }
	
	
	/** Returns <code>{@link Modifier#isProtected Modifier.isProtected}( f.{@link Field#getModifiers} )</code>. */
	public static boolean isProtected(Field f) { return Modifier.isProtected( f.getModifiers() ); }
	
	
	/**
	* Returns true if f has default (unnamed) access.
	* This is determined by the logic <code>!{@link #isPublic isPublic}(f) && !{@link #isProtected isProtected}(f) && >!{@link #isPrivate isPrivate}(f)</code>.
	*/
	public static boolean isDefault(Field f) { return !isPublic(f) && !isProtected(f) && !isPrivate(f); }
	
	
	/** Returns <code>{@link Modifier#isPrivate Modifier.isPrivate}( f.{@link Field#getModifiers} )</code>. */
	public static boolean isPrivate(Field f) { return Modifier.isPrivate( f.getModifiers() ); }
	
	
	/** Returns <code>{@link Modifier#isStatic Modifier.isStatic}( f.{@link Field#getModifiers} )</code>. */
	public static boolean isStatic(Field f) { return Modifier.isStatic( f.getModifiers() ); }
	
	
	/** Returns <code>{@link Modifier#isFinal Modifier.isFinal}( f.{@link Field#getModifiers} )</code>. */
	public static boolean isFinal(Field f) { return Modifier.isFinal( f.getModifiers() ); }
	
	
	// -------------------- diagnoseGetProblem --------------------
	
	
	/**
	* Returns a useful diagnostic message for a Throwable generated when calling the get method of Field.
	* <p>
	* @throws IllegalArgumentException if arg t is null
	* @see Field#get Field.get
	*/
	public static String diagnoseGetProblem(Throwable t) throws IllegalArgumentException {
		if (t == null) throw new IllegalArgumentException("arg t is null");
		
		if (t instanceof IllegalAccessException)
			return "Field is inaccessible";
			
		else if (t instanceof IllegalArgumentException)
			return "the Object supplied to the Field's get method is not an instance of the class or interface corresponding to the Field";
			
		else if (t instanceof NullPointerException)
			return "the Object supplied to the Field's get method is null, but the Field is an instance member";
			
		else if (t instanceof ExceptionInInitializerError)
			return "the initialization provoked by the Field's get method failed";
			
		else
			return "an unexpected problem," + ThrowableUtil.getTypeAndMessage(t) + ", occurred";
	}
	
	
	// -------------------- findSignature --------------------
	
	
	/**
	* Returns a Class[] where each element is the Class of the corresponding Object in the arguments array.
	* If arguments is null, a zero-element array is returned.
	*/
	public static final Class[] findSignature(Object[] arguments) {
		if (arguments == null)
			return new Class[0];
// +++ this is probably wrong--null in should return null back...
		else {
			Class[] signature = new Class[arguments.length];
			for (int i = 0; i < signature.length; i++) {
				signature[i] = arguments[i].getClass();
			}
			return signature;
		}
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private ReflectionUtil() {}
	
	
}
