/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*
--many of these methods should be changed to take the CharSequence interface introduced in java 1.4
*/


package bb.util;


import java.io.*;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CodingErrorAction;
import java.nio.charset.MalformedInputException;
import java.nio.charset.UnmappableCharacterException;
import java.text.ParseException;
import java.util.*;

import org.junit.Assert;
import org.junit.Test;

import bb.science.Math2;


/**
* This class provides various static String (and sometimes char[]) utility methods.
* <p>
* This class is multithread safe: it is immutable (both its immediate state, as well as the deep state of its fields).
* <p>
* @author Brent Boyer
*/
public class StringUtil {
	
	
	// -------------------- constants --------------------
	
	
	/**
	* Records the platform's standard for the char sequence that should separate lines.
	* Is assigned by a call to <code>{@link System#getProperty System.getProperty}("line.separator")</code>.
	* See this <a href="http://en.wikipedia.org/wiki/Newline">wikipedia entry</a> for more discussion.
	*/
	public static final String newline = System.getProperty("line.separator");
	
	
	// -------------------- newString, bytesToChars --------------------
	
	
	/**
	* Immediately returns null if s == null.
	* Otherwise, returns a new String instance whose underlying char[] precisely represents the chars of s
	* (i.e. the length of that char[] equals the length of s).
	* This guarantees that the result uses no more memory than required.
	* <p>
	* This minimal memory guarantee can be crucial in many situations.
	* For example, consider the case of a String which was parsed
	* as a substring from a long line of text from some file.
	* Then a reference to all the chars of the original line of text
	* must be maintained because of how the substring method operates,
	* and so excess memory is used.
	*/
	public static String newString(String s) {
		if (s == null) return null;
//		else return new String( s );
// +++ use the line above, which is potentially faster, if sun ever commits to a contract in that String constructor that it will behave as I require; I submitted a RFE to them about this in 2007/8...
else return new String( s.toCharArray() );
// for now, to be absolutely safe, must use the slower line above...
	}
	
	
	/**
	* Converts bytes into a char[] which is returned.
	* <p>
	* The conversion is done using a new {@link CharsetDecoder}
	* created from this platform's {@link Charset#defaultCharset default Charset}.
	* Strict conversion mode is used
	* (i.e. all errors result in Exceptions; no char substitutions or other silent error handling is performed).
	* So, bytes must be perfectly encoded using this platform's default encoding.
	* <p>
	* @throws IllegalArgumentException if bytes == null
	* @throws MalformedInputException if an illegal byte sequence for this charset is encountered 
	* @throws UnmappableCharacterException if a byte sequence is encountered which cannot be mapped to an equivalent character sequence
	*/
	public static char[] bytesToChars(byte[] bytes) throws IllegalArgumentException, IllegalStateException, SecurityException, IOException {
		if (bytes == null) throw new IllegalArgumentException("bytes == null");
		
		//return new String( bytes ).toCharArray();
		//return Charset.defaultCharset().decode( ByteBuffer.wrap(bytes) ).array();
			// do NOT use either of the above: first is bad performance due to extra array generation and copy
			// and both have the bad choice of CodingErrorAction.REPLACE for onMalformedInput/onUnmappableCharacter.
			// Instead, custom create a CharsetDecoder that throws Exceptions:
		CharsetDecoder decoder = Charset.defaultCharset().newDecoder();
		decoder.onMalformedInput( CodingErrorAction.REPORT );
		decoder.onUnmappableCharacter( CodingErrorAction.REPORT );
		
		return decoder.decode( ByteBuffer.wrap(bytes) ).array();
	}
	
	
	// -------------------- equals, equalChars --------------------
	
	
	/**
	* Determines whether or not s1 is equal to s2.
	* The motivation for this method is that it correctly handles nulls, which frees the user from having to do that repetitive check.
	* Specifically, it first checks for <code>s1 == null</code>, and if that is true returns the value of <code>s2 == null</code>).
	* Otherwise, if s1 is not null, then it returns <code>s1.equals(s2)</code>.
	*/
	public static boolean equals(String s1, String s2) {
		if (s1 == null) return (s2 == null);
		else return s1.equals(s2);
	}
// +++ this method is actually generic--should use a generic type T--and so it should be put in some other class...
	
	
	/**
	* Determines whether a <code>char[]</code> has exactly the same chars as a <code>String</code>.
	* If one of the args, is null, then the other must be null as well for true to be returned.
	*/
	public static boolean equalChars(char[] chars, String s) {
		if (chars == null) return (s == null);
		else if (s == null) return false;
		
		if (chars.length != s.length()) return false;
		
		for (int i = 0; i < chars.length; i++) {
			if (chars[i] != s.charAt(i)) return false;
		}
		
		return true;
	}
	
	
	// -------------------- toLength, repeatChars, keepWithinLength --------------------
	
	
	/**
	* Returns a String which represents number and consists of exactly length digits, with leading 0's padded if necessary.
	* <p>
	* @throws IllegalArgumentException if number is such that it cannot be represented by length digits
	*/
	public static String toLength(int number, int length) throws IllegalArgumentException {
		return toLength( Integer.toString(number), length, true, '0' );
	}
	
	
	/**
	* Returns a String of exactly length chars.
	* If s.length() == length, then s is returned.
	* Else if s.length() < length, then s is returned with a sufficient number of chars added to make the have length chars.
	* The char added on is always c, and it is either prepended or postpended to s as specified by prepend.
	* <p>
	* @throws IllegalArgumentException if s == null; s.length() > length
	*/
	public static String toLength(String s, int length, boolean prepend, char c) throws IllegalArgumentException {
		if (s == null) throw new IllegalArgumentException("s == null");
		if (s.length() > length) throw new IllegalArgumentException("s.length() = " + s.length() + " > length = " + length);
		
		StringBuilder sb = new StringBuilder(length);
		int numberToAdd = length - s.length();
		if (prepend) {
			for (int i = 0; i < numberToAdd; i++) {
				sb.append(c);
			}
		}
		sb.append(s);
		if (!prepend) {
			for (int i = 0; i < numberToAdd; i++) {
				sb.append(c);
			}
		}
		return sb.toString();
	}


	/**
	* Returns a String of the specified length which consists of entirely of the char c.
	* <p>
	* @throws IllegalArgumentException if length < 0
	*/
	public static String repeatChars(char c, int length) throws IllegalArgumentException {
		if (length < 0) throw new IllegalArgumentException("length = " + length + " is < 0");

		StringBuilder sb = new StringBuilder(length);
		for (int i = 0; i < length; i++) {
			sb.append(c);
		}
		return sb.toString();
	}


	/**
	* Returns s if its length is less than limitLength.
	* Otherwise tries to return as much from both the beginning and end of s as it can
	* with an ellipsis "..." in the middle of the result to indicate that the middle was cut out.
	* For example, <code>keepWithinLength("abcdefghi", 7)</code> returns <code>"ab...hi"</code>.
	* <p>
	* @throws IllegalArgumentException if s == null; limitLength < 5
	*/
	public static String keepWithinLength(String s, int limitLength) throws IllegalArgumentException {
		if (s == null) throw new IllegalArgumentException("s == null");
		if (limitLength < 5) throw new IllegalArgumentException("limitLength = " + limitLength + " is < 5");

		if (s.length() <= limitLength) return s;

		int numbersFromStart = (int) Math.ceil( (limitLength - 3) / 2.0 );
		String start = s.substring(0, numbersFromStart);
		int numbersFromEnd = (int) Math.floor( (limitLength - 3) / 2.0 );
		String end = s.substring(s.length() - numbersFromEnd, s.length());
		return start + "..." + end;
	}
	
	
	// -------------------- arraysToTextColumns --------------------
	
	
	/**
	* Returns a String representation of arrays as a series of text columns.
	* <p>
	* Since arrays is a double[][], then arrays[0], arrays[1], etc are double[] subarrays.
	* The result takes each subarray and uses it as a column of text:
	* arrays[0] is the first column, arrays[1] is the second column, etc.
	* <p>
	* Each row of the result is tab delimited: tab chars separate the numbers from each other.
	* Note that the subarrays need not have equal length:
	* this method will simply leave blanks in cells that have no data.
	* <p>
	* The header arg is optional (may be null),
	* but if present, must have the same length as the number of subarrays (i.e. arrays.length)
	* because it will be used to label each column.
	* <p>
	* To illustrate, the following code
	* <code><pre>
	*	double[][] arrays = new double[][] {
	*		new double[] {1},
	*		new double[] {1, 2},
	*		new double[] {1, 2, 3}
	*	};
	*	String[] header = new String[] {"A", "B", "C"};
	*	System.out.println( arraysToTextColumns(arrays, header) );</pre></code>
	* should produce this output:
	* <code><pre>
	*	A       B       C
	*	1.0     1.0     1.0
	*			2.0     2.0
	*					3.0</pre></code>
	* <p>
	* @throws IllegalArgumentException if arrays == null;
	* header != null && header.length != arrays.length
	*/
	public static String arraysToTextColumns(double[][] arrays, String[] header) throws IllegalArgumentException {
		if (arrays == null) throw new IllegalArgumentException("arrays == null");
		if ((header != null) && (header.length != arrays.length)) throw new IllegalArgumentException("header != null && header.length = " + header.length + " != arrays.length = " + arrays.length);
		
		int numberColumns = arrays.length;
		int numberDataRows = 0;
		for (double[] subarray : arrays) {
			if (numberDataRows < subarray.length) numberDataRows = subarray.length;
		}
		
		StringBuilder sb = new StringBuilder( numberDataRows * numberColumns * 32 );
		
		if (header != null) {
			for (int k = 0; k < header.length; k++) {
				if (k > 0) sb.append('\t');
				sb.append(header[k]);
			}
			sb.append('\n');
		}
		
		for (int j = 0; j < numberDataRows; j++) {
			for (int i = 0; i < numberColumns; i++) {
				if (i > 0) sb.append('\t');
				if (j < arrays[i].length) sb.append(arrays[i][j]);
			}
			sb.append('\n');
		}
		
		return sb.toString();
	}
	
	
	// -------------------- isAllAsciiChars, asciiBytesToChars --------------------


	/**
	* Determines whether or not the <code>String</code> consists exclusively of ASCII chars.
	* (The US-ASCII standard is used.)
	* <p>
	* @throws IllegalArgumentException if s is null
	*/
	public static boolean isAllAsciiChars(String s) {
		if (s == null) throw new IllegalArgumentException("arg s is null");

		for (int i = 0; i < s.length(); i++) {
			if ( s.charAt(i) > CharUtil.maxAsciiValueAsInt)
				return false;
		}

		return true;
	}


	/**
	* Converts each element of bytes to a char[] which is returned.
	* Every element of bytes must be a US-ASCII byte (i.e. high bit should be 0, so the only permissible values are [0, 127]).
	* This method is much faster than going thru the standard java ascii converter.
	* <p>
	* @throws IllegalArgumentException if bytes == null; if bytes contains a non-ascii byte (i.e. a negative value)
	* @see "Sun's ascii converter: goto <java home>\jre\lib\rt.jar, then extract all the US_ASCII classes, then decompile with jad"
	*/
	public static char[] asciiBytesToChars(byte[] bytes) throws IllegalArgumentException {
		if (bytes == null) throw new IllegalArgumentException("bytes == null");
		
		char[] chars = new char[bytes.length];
		for (int i = 0; i < bytes.length; i++) {
			byte b = bytes[i];
			if (b < 0) throw new IllegalArgumentException("bytes[" + i + "] = " + b + " is a non-ascii byte (i.e. is < 0)" + diagnoseProblem(bytes, i));
			chars[i] = (char) b;
		}
		return chars;
	}
	
	
	/** Meant for use just by asciiBytesToChars to provide a better diagnostic message to the user. */
	private static String diagnoseProblem(byte[] bytes, int i) {
		return
			'\n' +
			"----------" + '\n' +
			"Context of chars about the offending byte:" + '\n' +
			getAsciiContext(bytes, i) + '\n' +
			"----------" + '\n' +
			"Hamming distance ranking of all US-ASCII chars from the offending byte:" + '\n' +
			describeAsciiCharsByHammingDistance(bytes[i]);
	}
	
	
	private static final int numCharsContext = 16;
	
	/** Meant for use just by diagnoseProblem to provide a better diagnostic message to the user. */
	private static String getAsciiContext(byte[] bytes, int i) {
		int start = Math.max( i - numCharsContext, 0 );
		int end = Math.min( i + numCharsContext, bytes.length - 1 );
		char[] chars = new char[end - start + 1];
		for (i = 0; i < chars.length; i++) {
			byte b = bytes[start + i];
			chars[i] = (b >= 0) ? (char) b : '?';
		}
		return new String(chars);
	}
	
	
	/** Meant for use just by diagnoseProblem to provide a better diagnostic message to the user. */
	private static String describeAsciiCharsByHammingDistance(byte b) {
		SortedMap<Integer,List<Character>> map = rankAsciiCharsByHammingDistance(b);
		StringBuilder sb = new StringBuilder(512);
		for (int distance : map.keySet()) {
			sb.append(distance).append(": ");
			boolean previous = false;
			for (char c : map.get(distance)) {
				if (previous) sb.append(", ");
				else previous = true;
				sb.append(c);				
			}
			sb.append('\n');			
		}
		return sb.toString();
	}
	
	
	/** Meant for use just by describeAsciiCharsByHammingDistance to provide a better diagnostic message to the user. */
	private static SortedMap<Integer,List<Character>> rankAsciiCharsByHammingDistance(byte b) {
		int bb = Math2.byteToUnsignedInt(b);
		SortedMap<Integer,List<Character>> map = new TreeMap<Integer,List<Character>>();
		//for (int i = 32; i <= 126; i++) {	// the printable ASCII chars
		for (int i = 0; i <= 127; i++) {	// the complete ASCII chars
			int distance = Math2.hammingDistance(i, bb);
			List<Character> list = map.get(distance);
			if (list == null) {
				list = new ArrayList<Character>();
				map.put(distance, list);
			}
			list.add((char) i);
		}
		return map;
	}
	
	
	// -------------------- isBlank --------------------
	
	
	/** Determines whether or not s is "blank" (i.e. is either null, zero-length, or solely consists of whitespace). */
	public static boolean isBlank(String s) {
		if (s == null) return true;
		if (s.length() == 0) return true;
		
		for (int i = 0; i < s.length(); i++) {
			if ( !Character.isWhitespace( s.charAt(i) ) )
				return false;
		}
		return true;
	}
	
	
	// -------------------- isNewLineEnd and ensureEndsInNewLine --------------------


	/**
	* This method determines whether or not the supplied String ends in a newline char.
	* <p>
	* The {@link #newline} constant of this class is the newline char sequence that is used.
	* <p>
	* @throws IllegalArgumentException if arg s is null
	*/
	public static boolean isNewLineEnd(String s) {
		if (s == null) throw new IllegalArgumentException("arg s is null");

		return s.endsWith(newline);
	}


	/**
	* This method returns the supplied String if it already ends in a newline char sequence.
	* Otherwise, the result is a newline char sequence appended to the original.
	* <p>
	* The {@link #newline} constant of this class is the newline char sequence that is used.
	* <p>
	* @throws IllegalArgumentException if arg s is null
	*/
	public static String ensureEndsInNewLine(String s) {
		if (s == null) throw new IllegalArgumentException("arg s is null");

		if ( isNewLineEnd(s) )
			return s;
		else
			return s + newline;
	}


	// -------------------- normalizeWhitespace --------------------


	/**
	* Performs <i>whitespace normalization</i>, as per the XML spec.
	* <p>
	* @throws IllegalArgumentException if s is null
	* @see <a href="http://www.w3.org/TR/html401/struct/text.html#h-9.1">the HTML whitespace normalization spec</a>
	* @see <a href="www.w3.org/TR/REC-xml#AVNormalize">the XML whitespace normalization spec</a>
<!-- when i checked the above url on 2004/6/17, it now refers to a new spec and i was unable to find the section that i wanted... -->
	*/
	public static String normalizeWhitespace(String s) {
		if (s == null) throw new IllegalArgumentException("arg s is null");

		StringBuilder result = new StringBuilder( s.length() );
		boolean lastWasWhitespace = false;
		for (int i = 0; i < s.length(); i++) {
			char c = s.charAt(i);
			if (Character.isWhitespace(c)) {
				lastWasWhitespace = true;
			}
			else {
				if (lastWasWhitespace) {
					if (result.length() > 0) result.append(' ');	// need result.length() > 0 to enforce skipping of leading whitespace
					lastWasWhitespace = false;
				}
				result.append(c);
			}
		}

		return result.toString();
	}


	// -------------------- parseLines, concatLines --------------------


	/**
	* Parses individual lines out of the supplied String.
	* Each line is an entry in the returned String[].
	* The end of line char(s) are not included in the lines.
	* The result will be a zero-length array if the supplied text String is blank (i.e. "").
	* This method is the inverse of the {@link #concatLines concatLines} method.
	* <p>
	* @throws IllegalArgumentException if arg text is null
	*/
	public static String[] parseLines(String text) throws IllegalArgumentException {
		if (text == null) throw new IllegalArgumentException("arg text is null");

		try {
			BufferedReader br = new BufferedReader( new StringReader(text) );
			List<String> lines = new ArrayList<String>(16);
			while (true) {
				String line = br.readLine();
				if (line == null)
					break;
				else
					lines.add(line);
			}
			return lines.toArray( new String[ lines.size() ] );
		}
		catch (IOException ioe) {
			// this should never occur, given that we have a String as the io source...
			throw new RuntimeException("An IOException was caught, but this was unanticipated and cannot be handled by this method", ioe);
		}
	}


	/**
	* Adds all the individual text lines into a single String, which is returned.
	* Each line will have a (unix format) end of line char (i.e. '\n') appended to it before it is added to the result
	* if it does not already end with an end of line char (i.e. '\n' or '\r').
	* If the input array is zero-length, so will be the result.
	* This method is the inverse of the {@link #parseLines parseLines} method.
	* <p>
	* @throws IllegalArgumentException if arg linesOfText is null
	*/
	public static String concatLines(String[] linesOfText) throws IllegalArgumentException {
		if (linesOfText == null) throw new IllegalArgumentException("arg linesOfText is null");

			// pre-compute the text's length (this avoids array resizing sb below, which is inefficient):
		int textLength = 0;
		for (String line : linesOfText) {
			textLength += line.length() + 1;	// + 1 to account for maybe having to add \n below
		}

		StringBuilder sb = new StringBuilder(textLength);
		for (String line : linesOfText) {
			sb.append(line);
			if ( !line.endsWith("\n") && !line.endsWith("\r") ) sb.append('\n');
		}

		return sb.toString();
	}
	
	
	// -------------------- splitByChar --------------------
	
	
	/**
	* Splits a String into tokens, which are returned as an array.
	* <p>
	* This method has the following major restrictions:
	* <ol>
	*  <li>the token delimter must be a single char (no multiple char delimiters allowed)</li>
	*  <li>
	*		the user must specify <i>exactly</i> how many tokens should be present
	*		(this method only works on rigid, known in advance formats)
	*  </li>
	* </ol>
	* <p>
	* The above constraints, however, allow a highly optimized parsing algorithm to be used.
	* Calling this method is much faster than, say, calling {@link String#split(String, int) String.split}:
	* when this was last benchmarked on 2006/5/10
	* by running {@link UnitTest#benchmark_splitByChar UnitTest.benchmark_splitByChar},
	* it was found to be over 10 times faster:
	* <pre>
	*	StringUtil.splitByChar takes approximately 1.645660400390625E-7 s
	*	String.split takes approximately 2.05078125E-6 s</pre>
	* <p>
	* Contract: the result is never null nor zero-length.
	* <p>
	* @throws IllegalArgumentException if s == null; n < 1; s fails to split into exactly n tokens
	*/
	public static final String[] splitByChar(String s, char delimiter, int n) throws IllegalArgumentException {
		if (s == null) throw new IllegalArgumentException("s == null");
		if (n < 1) throw new IllegalArgumentException("n = " + n + " < 1");
		
		String[] tokens = new String[n];
		int last = 0;
		for (int i = 0; i < n - 1; i++) {
			int next = s.indexOf(delimiter, last);
			if (next == -1) throw new IllegalArgumentException("s has " + (i + 1) + " tokens, which is < the required n = " + n + " tokens");
			tokens[i] = s.substring(last, next);
			last = next + 1;
		}
		tokens[n - 1] = s.substring(last);
		
			// if the final token contains the delimiter char, determine exactly how many tokens in total are present in s, so can give an accurate error message:
		if (tokens[n - 1].indexOf(delimiter) >= 0) {
			int numberTokensInitial = n - 1;	// number of all the tokens before tokens[n - 1]
			int numberDelimitersInFinal = 0;
			for (int j = 0; j < tokens[n - 1].length(); j++) {
				if (tokens[n - 1].charAt(j) == delimiter) ++numberDelimitersInFinal;
			}
			int numberTokensInFinal = numberDelimitersInFinal + 1;
			int numberTokensTotal = numberTokensInitial + numberTokensInFinal;
			throw new IllegalArgumentException("s has " + numberTokensTotal + " tokens, which is > the required n = " + n + " tokens");
		}
		
		return tokens;
	}
/*
I wonder if the following alternative algorithm would be faster:
	--immediately allocate int[] indices = int[n - 1]
	--race thru all the chars of s in a single loop, and store the indices where delimiter occurs inside that array
	--assuming the correct number of tokens were found, then do a loop for substring(indices[i-1], indices[i]) to get all the tokens
This algorithm does involve an additional object instantiation, but it would avoid the method overhead of invoking indexOf.
Unfortunately, the String class keeps its char[] encapsulated, so this algorithm cannot get quick access to it...
*/
	
	
	// -------------------- quoteWhitespaceTokenize --------------------


	/**
	* This method breaks up the supplied String into tokens.
	* Tokens are delimited either by a sequence of double quote (i.e. '"') chars,
	* or by whitespace chars.
	* <p>
	* The procedure followed is to find the next occurring double quote char.
	* The substring of all untokenized chars before that double quote char is then tokenized
	* using whitespace chars as delimiters. Then, the next token is the substring which
	* consists of all chars from the current double quote till the next double quote.
	* This procedure is repeated until the source String is exhausted.
	* <p>
	* It is an error if the source String has an odd number of double quote chars
	* (i.e. they must occur in pairs).
	* <p>
	* An example application of this method is to parse command lines.
	* Here, command line arguments are normally separated by spaces.
	* However, double quotes are used to enclose those args which have spaces inside them.
	* <p>
	* @param source the String to be tokenized
	* @param includeQuotes specifies whether or not to include double quote marks with those tokens that are delimted by them
	* @throws IllegalArgumentException if arg source is null or if it contains an odd number of double quote chars
	*/
	public static String[] quoteWhitespaceTokenize(String source, boolean includeQuotes) {
		if (source == null) throw new IllegalArgumentException("arg source is null");
		
		List<String> tokens = new ArrayList<String>( Math.max(source.length()/8, 16) );		// i.e. guesstimate that the average token is 8 chars in length, and insist that our minimum initial capacity be 16

		int index = 0;
		while (index < source.length()) {
			int startQuoteIndex = source.indexOf('"', index);

			if (startQuoteIndex == -1) {
				wsTokensize(source.substring(index), tokens);
				break;
			}
			else {
				wsTokensize( source.substring(index, startQuoteIndex), tokens );

				int endQuoteIndex = source.indexOf('"', startQuoteIndex + 1);
				if (endQuoteIndex == -1) throw new IllegalArgumentException("there is a double quote char at index = " + startQuoteIndex + " which does not have a subsequent matching quote in the source String = " + source);

				if (includeQuotes)
					tokens.add( source.substring(startQuoteIndex, endQuoteIndex + 1) );
				else
					tokens.add( source.substring(startQuoteIndex + 1, endQuoteIndex) );

				index = endQuoteIndex + 1;
			}
		}

		return tokens.toArray( new String[tokens.size()] );
	}


	private static void wsTokensize(String source, List<String> tokens) {
		int index = 0;
		while (true) {
				// skip over all whitespace chars; when done, index points to next non-whitespace char (or to source.length()):
			for ( ; index < source.length(); index++) {
				if ( !Character.isWhitespace( source.charAt(index) ) )
					break;
			}
			if (index == source.length()) return;

				// skip over all non-whitespace chars; when done, index points to next whitespace char (or to source.length()):
			int tokenStartIndex = index;
			for ( ; index < source.length(); index++) {
				if ( Character.isWhitespace( source.charAt(index) ) )
					break;
			}
			tokens.add( source.substring(tokenStartIndex, index) );
			if (index == source.length()) return;

				// may as well increment index so that do not retest the current char in the first for loop above:
			++index;
		}
	}


	// -------------------- removeQuotes --------------------


	/**
	* This utility method removes a <i>matching leading and trailing pair</i> of quote marks, if present,
	* from the supplied String and returns the substring inside the quotes.
	* If no such pair of leading or trailing quotes are present, then the original String is returned.
	* <p>
	* Any leading and trailing quote marks must be either both single or both double quote chars
	* in order to match.
	* <p>
	* @param lineNumber the line number where the String was found; used only if throw a ParseException
	* @throws IllegalArgumentException if s is null, or lineNumber < 0
	* @throws ParseException if a leading but no matching trailing quote mark is present, or vice versa
	*/
	public static String removeQuotes(String s, int lineNumber) throws IllegalArgumentException, ParseException {
		if (s == null) throw new IllegalArgumentException("arg s is null");
		if (lineNumber < 0) throw new IllegalArgumentException("arg lineNumber = " + lineNumber + " is < 0");

			// handle the zero-length case before proceed with remaining code (which assumes length > 0):
		if (s.length() == 0)
			return s;

		int lastIndex = s.length() - 1;

			// strip matching single quotes if present, bomb if starts with single quote but does not end with one:
		if (s.charAt(0) == '\'') {
			if (s.charAt(lastIndex) == '\'')
				return s.substring(1, lastIndex);	// Note: lastIndex is exclusive in substring, so it is not included
			else
				throw new ParseException("the following String starts with a single quote, but does not end with one: " + s, lineNumber);
		}
			// bomb if ends with a single quote mark but does not start with one:
		else if (s.charAt(lastIndex) == '\'')
			throw new ParseException("the following String ends with a single quote, but does not start with one: " + s, lineNumber);

			// strip matching double quotes if present, bomb if starts with double quote but does not end with one:
		else if (s.charAt(0) == '"') {
			if (s.charAt(lastIndex) == '"')
				return s.substring(1, lastIndex);	// Note: lastIndex is exclusive in substring, so it is not included
			else
				throw new ParseException("the following String starts with a double quote, but does not end with one: " + s, lineNumber);
		}
			// bomb if ends with a double quote mark but does not start with one:
		else if (s.charAt(lastIndex) == '"')
			throw new ParseException("the following String ends with a double quote, but does not start with one: " + s, lineNumber);

		else
			return s;
	}
	
	
	// -------------------- toString --------------------
	
	
// +++ wonder if should add versions of the methods below in which have a boolean printIndex param
// which would label each line with the count (e.g. #1:, #2:, etc)
	
	
	/**
	* Returns a String representation of array.
	* <p>
	* The result consists of the elements of array in sequence, with separator between each element.
	* <p>
	* @throws IllegalArgumentException if array == null; separator == null
	*/
	public static String toString(boolean[] array, String separator) throws IllegalArgumentException {
		if (array == null) throw new IllegalArgumentException("array == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(16 * array.length, 1 * 1024) );
		for (boolean element : array) {
			if (sb.length() > 0) sb.append(separator);
			sb.append(element);
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of array.
	* <p>
	* The result consists of the elements of array in sequence, with separator between each element.
	* <p>
	* @throws IllegalArgumentException if array == null; separator == null
	*/
	public static String toString(byte[] array, String separator) throws IllegalArgumentException {
		if (array == null) throw new IllegalArgumentException("array == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(16 * array.length, 1 * 1024) );
		for (byte element : array) {
			if (sb.length() > 0) sb.append(separator);
			sb.append(element);
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of array.
	* <p>
	* The result consists of the elements of array in sequence, with separator between each element.
	* <p>
	* @throws IllegalArgumentException if array == null; separator == null
	*/
	public static String toString(char[] array, String separator) throws IllegalArgumentException {
		if (array == null) throw new IllegalArgumentException("array == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(16 * array.length, 1 * 1024) );
		for (char element : array) {
			if (sb.length() > 0) sb.append(separator);
			sb.append(element);
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of array.
	* <p>
	* The result consists of the elements of array in sequence, with separator between each element.
	* <p>
	* @throws IllegalArgumentException if array == null; separator == null
	*/
	public static String toString(double[] array, String separator) throws IllegalArgumentException {
		if (array == null) throw new IllegalArgumentException("array == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(16 * array.length, 1 * 1024) );
		for (double element : array) {
			if (sb.length() > 0) sb.append(separator);
			sb.append(element);
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of array.
	* <p>
	* The result consists of the elements of each array[i] on their own line,
	* with these elements in sequence, with separator between each element.
	* <p>
	* @throws IllegalArgumentException if array == null; separator == null
	*/
	public static String toString(double[][] array, String separator) throws IllegalArgumentException {
		if (array == null) throw new IllegalArgumentException("array == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		int m = array.length;
		int n = (m > 0) ? array[0].length : 0;
		StringBuilder sb = new StringBuilder( Math.max(16 * m * n, 1 * 1024) );
		for (double[] arrayInner : array) {
			if (sb.length() > 0) sb.append("\n");
			for (int i = 0; i < arrayInner.length; i++) {
				if (i > 0) sb.append(separator);
				sb.append(arrayInner[i]);
			}
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of array.
	* <p>
	* The result consists of the elements of array in sequence, with separator between each element.
	* <p>
	* @throws IllegalArgumentException if array == null; separator == null
	*/
	public static String toString(float[] array, String separator) throws IllegalArgumentException {
		if (array == null) throw new IllegalArgumentException("array == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(16 * array.length, 1 * 1024) );
		for (float element : array) {
			if (sb.length() > 0) sb.append(separator);
			sb.append(element);
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of array.
	* <p>
	* The result consists of the elements of array in sequence, with separator between each element.
	* <p>
	* @throws IllegalArgumentException if array == null; separator == null
	*/
	public static String toString(long[] array, String separator) throws IllegalArgumentException {
		if (array == null) throw new IllegalArgumentException("array == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(16 * array.length, 1 * 1024) );
		for (long element : array) {
			if (sb.length() > 0) sb.append(separator);
			sb.append(element);
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of array.
	* <p>
	* The result consists of the elements of array in sequence, with separator between each element.
	* <p>
	* @throws IllegalArgumentException if array == null; separator == null
	*/
	public static String toString(short[] array, String separator) throws IllegalArgumentException {
		if (array == null) throw new IllegalArgumentException("array == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(16 * array.length, 1 * 1024) );
		for (short element : array) {
			if (sb.length() > 0) sb.append(separator);
			sb.append(element);
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of array.
	* <p>
	* The result consists of the elements of array in sequence, with separator between each element.
	* <p>
	* @throws IllegalArgumentException if array == null; separator == null
	*/
	public static String toString(int[] array, String separator) throws IllegalArgumentException {
		if (array == null) throw new IllegalArgumentException("array == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(16 * array.length, 1 * 1024) );
		for (int element : array) {
			if (sb.length() > 0) sb.append(separator);
			sb.append(element);
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of array.
	* <p>
	* The result consists of the elements of array in sequence, with separator between each element.
	* If any element == null, then it is represented by the text "null".
	* <p>
	* @throws IllegalArgumentException if array == null; separator == null
	*/
	public static <T> String toString(T[] array, String separator) throws IllegalArgumentException {
		if (array == null) throw new IllegalArgumentException("array == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(16 * array.length, 1 * 1024) );
		for (T element : array) {
			if (sb.length() > 0) sb.append(separator);
			sb.append( element );
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of collection.
	* <p>
	* The result consists of the elements of collection as returned by its Iterator, with separator between each element.
	* If any element == null, then it is represented by the text "null".
	* <p>
	* @throws IllegalArgumentException if collection == null; separator == null
	*/
	public static String toString(Collection<?> collection, String separator) throws IllegalArgumentException {
		if (collection == null) throw new IllegalArgumentException("collection == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(16 * collection.size(), 1 * 1024) );
		for (Object element : collection) {
			if (sb.length() > 0) sb.append(separator);
			sb.append( element );
		}
		return sb.toString();
	}
	
	
	/**
	* Returns a String representation of map.
	* <p>
	* The result consists of map's key/value pairs, with separator between each pair.
	* The keys are obtained by map's keySet Iterator, and the text " --> " appears between the key and its value.
	* If any key or value == null, then it is represented by the text "null".
	* <p>
	* @throws IllegalArgumentException if map == null; separator == null
	*/
	public static String toString(Map<?,?> map, String separator) throws IllegalArgumentException {
		if (map == null) throw new IllegalArgumentException("map == null");
		if (separator == null) throw new IllegalArgumentException("separator == null");
		
		StringBuilder sb = new StringBuilder( Math.max(32 * map.size(), 1 * 1024) );
		for (Map.Entry<?,?> entry : map.entrySet()) {
			if (sb.length() > 0) sb.append(separator);
			sb.append( entry.getKey() ).append(" --> ").append( entry.getValue() );
		}
		return sb.toString();
	}
	
	
	// -------------------- toStringLiteral --------------------
	
	
	/**
	* This method converts a String into a series of chars that constitute a Java String literal.
	* In particular, double quotes are placed at the beginning and end, and any special chars
	* inside are properly escaped.
	* <p>
	* One use is that you could take the result and directly paste it into a Java source file.
	* Another use for this method is to handle filepaths with spaces and '\' chars
	* (which commonly occur in dos).
	* <p>
	* @throws IllegalArgumentException if s == null
	*/
	public static String toStringLiteral(String s) throws IllegalArgumentException {
		if (s == null) throw new IllegalArgumentException("s == null");
		
		StringBuilder sb = new StringBuilder(s.length() + 16);
		
		sb.append('"');
		for (int i = 0; i < s.length(); i++) {
			char c = s.charAt(i);
			if ( CharUtil.hasEscapeForLiteral( c ) )	// this case will handle quotes, line terminator chars, and the backslash char (among others) that must be escaped
				sb.append( CharUtil.getEscapeForLiteral( c ) );
			else
				sb.append( c );
		}
		sb.append('"');
		
		return sb.toString();
	}
	
	
	// -------------------- constructor --------------------


	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private StringUtil() {}


	// -------------------- UnitTest (static inner class) --------------------


	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
// +++ need to add more tests, such as for bytesToChars...
		
		@Test public void test_toLength() {
			Assert.assertEquals( "000123", toLength(123, 6) );
			Assert.assertEquals( "helloWorldxxxxxxxxxx", toLength("helloWorld", 20, false, 'x') );
		}
		
		@Test public void test_repeatChars() {
			Assert.assertEquals( "aaaaa", repeatChars('a', 5) );
		}
		
		@Test public void test_keepWithinLength() {
			Assert.assertEquals( "ab...i", keepWithinLength("abcdefghi", 6) );
		}
		
		@Test public void test_arraysToTextColumns() {
			String textExpected =
				"A" + '\t' + "B" + '\t' + "C" + '\n' +
				"1.0" + '\t' + "1.0" + '\t' + "1.0" + '\n' +
				'\t' + "2.0" + '\t' + "2.0" + '\n' +
				'\t' + '\t' + "3.0" + '\n';
			
			double[][] arrays = new double[][] {
				new double[] {1},
				new double[] {1, 2},
				new double[] {1, 2, 3}
			};
			String[] header = new String[] {"A", "B", "C"};
			
			Assert.assertEquals( textExpected, arraysToTextColumns(arrays, header) );
		}
		
		@Test public void test_normalizeWhitespace() {
			String s = "\t\r\n  abc\n\t\r\n def \t\r\n";
			String normalizationExpected = "abc def";
			Assert.assertEquals( normalizationExpected, normalizeWhitespace(s) );
		}
		
		@Test public void test_splitByChar_pass() {
			String s = "a" + '|' + "b" + '|' + "c";
			String[] tokensExpected = new String[] {"a", "b", "c"};
			Assert.assertArrayEquals( tokensExpected, splitByChar(s, '|', 3) );
		}
		
		@Test(expected=IllegalArgumentException.class) public void test_splitByChar_fail1() {
			splitByChar(null, '|', 3);	// should fail (null supplied)
		}
		
		@Test(expected=IllegalArgumentException.class) public void test_splitByChar_fail2() {
			String s = "a" + '|' + "b" + '|' + "c";
			String[] tokensExpected = new String[] {"a", "b", "c"};
			Assert.assertArrayEquals( tokensExpected, splitByChar(s, '|', 0) ); // should fail (zero tokens specified)
		}
		
		@Test(expected=IllegalArgumentException.class) public void test_splitByChar_fail3() {
			String s = "a" + '|' + "b" + '|' + "c";
			String[] tokensExpected = new String[] {"a", "b", "c"};
			Assert.assertArrayEquals( tokensExpected, splitByChar(s, '|', 2) ); // should fail (too few tokens specified)
		}
		
		@Test(expected=IllegalArgumentException.class) public void test_splitByChar_fail4() {
			String s = "a" + '|' + "b" + '|' + "c";
			String[] tokensExpected = new String[] {"a", "b", "c"};
			Assert.assertArrayEquals( tokensExpected, splitByChar(s, '|', 4) ); // should fail (too many tokens specified)
		}
		
		/**
		* Results on 2007/11/28 on my 2.0 GHz Core 2 Duo laptop (server jvm):
		* <pre><code>
			splitByChar: mean = 141.662 ns (CI deltas: -52.206 ps, +55.192 ps), sd = 434.742 ns (CI deltas: -98.615 ns, +131.005 ns)
			String.split: mean = 1.150 us (CI deltas: -784.688 ps, +953.076 ps), sd = 2.475 us (CI deltas: -490.545 ns, +507.577 ns)</code></pre>
		*/
		@Test public void benchmark_splitByChar() throws Exception {
			final String s = "a" + '|' + "b" + '|' + "c";
			
			Runnable task1 = new Runnable() {
				private int state;	// needed to prevent DCE since this is a Runnable
				@Override public String toString() { return String.valueOf(state); }	// needed to prevent DCE since this is a Runnable
				public void run() {
					String[] tokens = splitByChar(s, '|', 3);
					state ^= tokens.length;
				}
			};
			System.out.println("splitByChar: " + new Benchmark(task1));
			
			Runnable task2 = new Runnable() {
				private int state;	// needed to prevent DCE since this is a Runnable
				@Override public String toString() { return String.valueOf(state); }	// needed to prevent DCE since this is a Runnable
				public void run() {
					String[] tokens = s.split("|", -1);
					state ^= tokens.length;
				}
			};
			System.out.println("String.split: " + new Benchmark(task2));
		}
		
		@Test public void test_quoteWhitespaceTokenize() {
			String source = "abc d e f \"ghi\" \"j k l\"";
			boolean includeQuotes = false;
			String[] tokensExpected = new String[] {"abc", "d", "e", "f", "ghi", "j k l"};
			Assert.assertArrayEquals( tokensExpected, quoteWhitespaceTokenize(source, includeQuotes) );
			
			source = "123 4 5 6 \"789\" \"10 11 12\"";
			includeQuotes = true;
			tokensExpected = new String[] {"123", "4", "5", "6", "\"789\"", "\"10 11 12\""};
			Assert.assertArrayEquals( tokensExpected, quoteWhitespaceTokenize(source, includeQuotes) );
		}
		
		@Test public void test_diagnoseProblem() {
			byte[] bytes = new byte[] {(byte) 'a', (byte) -55, (byte) 'c'};
			System.out.println(diagnoseProblem(bytes, 1));
		}
		
		@Test public void investigate_bitFlip() {
			byte b = (byte) '9';
			System.out.println( describeAsciiCharsByHammingDistance(b) );
		}
// when last ran this program on 2006/5/11, confirmed that an Interval number read in as 5.88 was actually 5.89 in the file, and that bit flipping must have happened because 8 is 1 bit off from 9
		
		@Test public void test_toString() {
			double[] doubleArray = new double[] {1, 2, 3};
			String doubleArrayTextExpected = "1.0, 2.0, 3.0";
			Assert.assertEquals( doubleArrayTextExpected, StringUtil.toString(doubleArray, ", ") );
			
			double[][] doubleMatrix = new double[][] { {1, 2, 3}, {4, 5, 6}, {7, 8, 9} };
			String doubleMatrixTextExpected =
				"1.0, 2.0, 3.0" + '\n' +
				"4.0, 5.0, 6.0" + '\n' +
				"7.0, 8.0, 9.0";
			Assert.assertEquals( doubleMatrixTextExpected, StringUtil.toString(doubleMatrix, ", ") );
			
			String[] stringArray = new String[] {"a", "b", "c", null};
			String stringArrayTextExpected = "a;b;c;null";
			Assert.assertEquals( stringArrayTextExpected, StringUtil.toString(stringArray, ";") );
			
			Collection<String> collection = Arrays.asList( "a", "b", "c", null );
			String collectionTextExpected = "a  b  c  null";
			Assert.assertEquals( collectionTextExpected, StringUtil.toString(collection, "  ") );
			
			Map<Integer,Character> map = new LinkedHashMap<Integer,Character>();	// use LinkedHashMap because HashMap permits nulls yet need guaranteed iteration order for the assertEquals to work
			map.put(1, 'a');
			map.put(2, 'b');
			map.put(3, 'c');
			map.put(null, null);
			String mapTextExpected = "1 --> a, 2 --> b, 3 --> c, null --> null";
			Assert.assertEquals( mapTextExpected, StringUtil.toString(map, ", ") );
		}
		
	}
	
	
}
