/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.util;


import java.util.logging.*;

import org.junit.Test;

import static bb.util.StringUtil.newline;
import bb.util.logging.Logger2;
import bb.util.logging.LogUtil;


/**
* Simple implementation of the Thread.UncaughtExceptionHandler interface which logs all uncaught Throwables to a Logger.
* <p>
* This class is multithread safe: the types of its state are individually multithread safe, and are used such that only their individual multithread safety matters.
* <p>
* @author Brent Boyer
* @see <a href="http://java.sun.com/developer/JDCTechTips/2006/tt0211.html#2">Catching Uncaught Exceptions</a>
*/
public class UncaughtThrowableHandler implements Thread.UncaughtExceptionHandler {
	
	
	// -------------------- fields --------------------
	
	
	/**
	* Logger where all Throwables will get logged to.
	* <p>
	* Contract: this field is never null.
	*/
	private final Logger logger;
	
	
	// -------------------- constructor --------------------
	
	
	/**
	* Constructor.
	* <p>
	* @throws IllegalArgumentException if logger == null	
	*/
	public UncaughtThrowableHandler(Logger logger) {
		if (logger == null) throw new IllegalArgumentException("logger == null");
		
		this.logger = logger;
	}
	
	
	// -------------------- Thread.UncaughtExceptionHandler api --------------------
	
	
	public void uncaughtException(Thread thread, Throwable throwable) {
		StringBuilder sb = new StringBuilder(256);
		sb.append("AN UNCAUGHT Throwable HAS BEEN DETECTED.").append(newline);
		
		String threadInfo = (thread != null) ? ThreadUtil.toString(thread) : "UNKNOWN (null was supplied)" + newline;
		sb.append("Thread which raised the Throwable: ").append(threadInfo);
		
		String throwableInfo = (throwable != null) ? ThrowableUtil.toString(throwable) : "UNKNOWN (null was supplied)" + newline;
		sb.append("Throwable: ").append(throwableInfo);
		
		String message = sb.toString();
		
		logger.log(Level.SEVERE, message);
	}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test(expected=Exception.class) public void test_uncaughtException() throws Exception {
			System.out.println();
			System.out.println("Creating a new UncaughtThrowableHandler instance that uses a file Logger...");
			Logger2 logger2 = LogUtil.makeLogger2( UncaughtThrowableHandler.UnitTest.class );
			UncaughtThrowableHandler handler = new UncaughtThrowableHandler(logger2);
			
			Exception e = new Exception("deliberately generated test Exception");
			
			System.out.println();
			System.out.println("Calling uncaughtException with the current thread and a deliberately generated Exception:");
			handler.uncaughtException(Thread.currentThread(), e);
			
			System.out.println();
			System.out.println("Calling uncaughtException with the current thread and a null Throwable:");
			handler.uncaughtException(Thread.currentThread(), null);
			
			System.out.println();
			System.out.println("Calling uncaughtException with a null thread and a deliberately generated Exception:");
			handler.uncaughtException(null, e);
			
			System.out.println();
			System.out.println("Calling uncaughtException with a null thread and a null Throwable:");
			handler.uncaughtException(null, null);
			
			System.out.println();
			System.out.println("Calling Thread.setDefaultUncaughtExceptionHandler with the current UncaughtThrowableHandler instance,");
			Thread.setDefaultUncaughtExceptionHandler( handler );
			System.out.println("and now deliberately throwing our test Exception to see if handler processes it:");
			System.out.println("if see any output below, then handler was correctly passed this Exception:");
			throw e;
		}
		
	}
	
	
}