/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package bb.util.logging;


import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.FileHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.LogManager;

import org.junit.Test;

import bb.io.FileUtil;
import bb.io.DirUtil;
import bb.io.StreamUtil;
import bb.util.DateUtil;
import bb.util.StringUtil;
import bb.util.ThrowableUtil;


/**
* Provides static constants and utility methods relating to the <code>java.util.logging</code> package.
* <p>
* Two powerful features of the <code>java.util.logging</code> package
* are that much behavior can be inherited from parent Loggers,
* and that much behavior can be specified outside of code.
* The most important means to do this is via the logging properties file
* that is specified by the <code>java.util.logging.config.file</code> {@link System} property.
* A sample file is the <code>.../script/logging.properties</code> file in this project.
* <p>
* This class adds support for the following logging properties:
* <ol>
*  <li>
*		A property named <code>logDirectory-NS</code>.
*		This defines the path of the directory where any log files should be written to.
*		The directory will be created if currently unexisting.
*		Note: it is an error to specify a path which is impossible to create.
*  </li>
*  <li>
*		A property named <code>&lt;logger>.handlers-NS</code>
*		where <code>&lt;logger></code> is some Logger name <i>prefix</i>.
*		This property is identical to the <code>&lt;logger>.handlers</code> property
*		defined in {@link LogManager} except for these differences:
*		<ul>
*		 <li>
*			<i>it is inherited:</i> it causes the list of handlers to be instantiated at <i>each</i> matching Logger.
*			(Recall: <code>&lt;logger>.handlers</code> causes the list of Handlers to only be created at <code>logger</code> itself.
*			Any child Logger of <code>&lt;logger></code>, unless further configured, will not have its own version of those Handlers
*			and will only use those Handlers if it is set to use its parent's Handlers.)
*		 </li>
*		 <li>
*			<i>it is a prefix:</i> in determining whether or not a Logger matches the <code>&lt;logger></code> pattern,
*			the pattern is assumed to end with an implicit * wildcard.
*			For example, specifying <code>com.handlers-NS</code> will cause both a Logger named <code>com.xyz</code>
*			as well as a Logger named <code>com123</code> to match.
*			(Recall: <code>com.handlers</code> would only cause <code>com.xyz</code> to match, since it is a proper child Logger,
*			whereas </code>com123<code> would not match, since it is not a child Logger but is a peer.)
*		 </li>
*		</ul>
*		One example where <code>&lt;logger>.handlers-NS</code> is essential
*		is when you desire to have many Classes each log to their own file.
*		Here, it is inconvenient and buggy to list each class's Logger name in the logging properties file.
* 		Instead, you should simply specify some Logger name prefix
*		and have each matching Logger inherit that behavior.
*		See {@link #nsHandlers nsHandlers} for how this class uses this property.
*		Note: it is an error to specify the same Handler class more than once in the list.
*  </li>
*  <li>
*		A property named <code>&lt;logger>.level-NS</code>
*		where <code>&lt;logger></code> is some Logger name prefix.
*		This property is identical to the <code>&lt;logger>.level</code> property defined in LogManager
*		except that, like <code>&lt;logger>.handlers-NS</code> above, its pattern is a prefix.
*  </li>
*  <li>
*		A property named <code>&lt;logger>.useParentHandlers-NS</code>
*		where <code>&lt;logger></code> is some Logger name prefix.
*		This property is identical to the <code>&lt;logger>.useParentHandlers</code> property defined in LogManager
*		except that, like <code>&lt;logger>.handlers-NS</code> above, it is inherited and its pattern is a prefix.
*		(Recall: Loggers do not inherit the <code>useParentHandlers property</code> from the last ancestor Logger which sets it.
*		Instead, the <code>useParentHandlers</code> property is true by default unless explicitly set to false.)
*  </li>
* </ol>
* These properties must be defined in the same logging properties file
* that is used to specify the <code>java.util.logging</code> properties for your project.
* The suffix of every property name described above ends with <code>-NS</code>
* because they are <i>non-standard</i> properties:
* the normal JDK logging code (e.g. LogManager) does not understand them,
* only a special class like this one does.
* This class's static initializer will reread any logging properties file and process the above properties.
* Only Loggers created by the {@link #makeLogger2(String) makeLogger2} method
* will have these non-standard properties applied to them.
* <p>
* Every Logger returned by this class is a {@link Logger2} instance, because of that subclass's advanced features.
* <p>
* <a href="http://www.oracle.com/technology/pub/articles/hunter_logging.html">Jason Hunter's JDK logging tutorial</a>
*  notes the following:
* <blockquote>
*	In Java we first saw variable substitution used for internationalization and localization,
*	because in those cases simple string concatentation isn't possible.
*	With simple log() methods you could rely on string concatentation like this:
*	<code style="display: block; margin-left: 2em">logger.log(Level.INFO, "Your file was " + size + " bytes long");</code>
*	The problem with this approach is that the runtime must perform the string concatentation
*	before executing the log() method call.
*	That wastes CPU cycles and memory when the logger's going to reject the message
*	for being at too low a level.
*	The solution is to use variable substitution and avoid all concatentation:
*	<code style="display: block; margin-left: 2em">logger.log(Level.INFO, "Your file was {0} bytes long", size);</code>
*	This call executes directly and the substitution only happens after the message
*	has passed the logger's quick level check.
* </blockquote>
* The above seems like a really clever use of param substitution,
* but it is not obvious that will actually reduce the cpu impact of logging very much.
* Reason: it is almost impossible to use params in one of the log/logp methods of Logger
* without causing an Object[] to be created.
* (The sole exception is the rare situation when all the params that you need are already in an array.)
* Thus, you always take a hit from creating that Object[], which is just as bad as creating a new String.
* Furthermore, if the level check is passed, then the work that needs to be done to actually carry out
* the param substitution is somewhat greater than doing simple String concatenation.
* So, you are unlikely to gain much performance.
* Since param substitution makes your code look even worse,
* should only use param substitution for what it was originally designed for,
* namely, to support pre-localized generic messages that can subsequently be localized
* and then have params substituted in.
* This is discussed further <a href="http://blogs.sun.com/CoreJavaTechTips/entry/logging_localized_message">here</a>.
* <p>
* This class is multithread safe: most of its immediate state is immutable,
* and the deep state of its fields is also either immutable or is multithread safe.
* The sole mutable field is {@link #logger2_default}, which is protected by synchronized access.
* <p>
* @author Brent Boyer
* @see <a href="http://java.sun.com/j2se/1.5.0/docs/guide/logging/overview.html">JavaTM Logging Overview</a>
* @see <a href="http://java.sun.com/j2se/1.5.0/docs/api/java/util/logging/package-summary.html">API Specification: java.util.logging</a>
*/
public class LogUtil {
	
	
	// -------------------- constants --------------------
	
	
	private static final Level[] levels;
	static {
		levels = new Level[] {
			Level.ALL, Level.CONFIG, Level.FINE, Level.FINER, Level.FINEST,
			Level.INFO, Level.OFF, Level.SEVERE, Level.WARNING
		};
		Arrays.sort(levels, new Comparator<Level>() {
			public int compare(Level a, Level b) {
				return a.intValue() - b.intValue();
			}
		} );
// +++ in the future, if can get sun to make Level implement Comparable, could simply do Arrays.sort(levels)...
// see http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6405743
	}
	
	
	private static final String logDirectory_key = "logDirectory-NS";
	
	private static final String logDirectory_valueDefault = "../log/";
	
	/**
	* Directory where log files are to be stored.
	* <p>
	* Internally used by this class's {@link #getPath(String)} method,
	* and therefore implicitly used by the various makeLogger2 methods.
	* May be used by external users in arbitrary ways.
	* <p>
	* Contract: is never null, and always exists.
	*/
	private static final File logDirectory;
	
	
	private static final String nsSuffix = "-NS";
	
	private static final ConcurrentMap<String,NodeProp> nameToNodeProps = new ConcurrentHashMap<String,NodeProp>();
	
	
	static {
		try {
			Properties properties = loadProperties();
			logDirectory = extract_logDirectory(properties);
			init_nameToNodeProps(properties);
		}
		catch (Throwable t) {
			throw ThrowableUtil.toRuntimeException(t); 
// +++ should call System.exit instead?
		}
	}
	
	
	// -------------------- static fields --------------------
	
	
	/**
	* A default Logger2 instance that any user of this class may draw on
	* via the {@link #getLogger2 getLogger2} accessor if they do not want to bother creating their own Logger.
	* This Logger is always given the name "default".
	* <p>
	* Note that while {@link Logger#global} is a similar default Logger,
	* there are several reasons why this field was introduced:
	* <ol>
	*  <li>
	*		as of JDK 1.5, sun, in their infinite wisdom, chose to not initialize Logger.global;
	*		the programmer must manually initialize it via a call to Logger.getLogger("global") which is annoying;
	*		in contrast, getLogger2 lazy initializes this field
<!--
+++ Note: jdk 1.6 deprecates Logger.glogal and jdk 1.7 introduces Logger.getGlobal():
	http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6476146
so eliminate this when 1.7 comes out?
-->
	*  </li>
	*  <li>
	*		since getLogger2 initializes this field by a call to <code>{@link #makeLogger2 makeLogger2}("default")</code>,
	*		this means that the logging properties file can specify <code>-NS</code> properties
	*		for this field (use the Logger name "default");
	*		this is particularly valuable if want it to write to a log file
	*		located in {@link #logDirectory}, and with a timestamp in its file name.
	*  </li>
	*  <li>since this field is a Logger2 instance, it has the extra functionality of that class</li>
	* </ol>
	*/
	private static Logger2 logger2_default;
	
	
	// -------------------- loadProperties, extract_logDirectory, init_nameToNodeProps --------------------
	
	
	private static Properties loadProperties() throws Exception {
		InputStream in = null;
		try {
			Properties properties = new Properties();
			
			String configFile = System.getProperty("java.util.logging.config.file");
			if (configFile != null) {
				File file = new File(configFile);
				FileUtil.check(file);
				in = new BufferedInputStream( new FileInputStream(file) );
				properties.load(in);
			}
				
			return properties;
		}
		finally {
			StreamUtil.close(in);
		}
	}
	
	
	private static File extract_logDirectory(Properties properties) throws SecurityException, IllegalStateException {
		String path = properties.getProperty(logDirectory_key, logDirectory_valueDefault);
		File directory = new File( path );
		DirUtil.ensureExists(directory);
		return directory;
	}
	
	
	private static void init_nameToNodeProps(Properties properties) throws IllegalStateException, ClassNotFoundException {
		for (Object obj : properties.keySet()) {
			String key = (String) obj;
			if (key.equals(logDirectory_key)) continue;
			if (!key.endsWith(nsSuffix)) continue;
			
			int lengthRetain = key.length() - nsSuffix.length();
			String keyEssential = key.substring(0, lengthRetain);	// strip off nsSuffix
			if (properties.containsKey(keyEssential)) throw new IllegalStateException("error/duplication: the logging properties specifies key = " + key + ", which has keyEssential = " + keyEssential + ", which is already defined in the properties");
			
			int indexLastDot = keyEssential.lastIndexOf('.');
			if (indexLastDot == -1) throw new IllegalStateException("the logging properties specifies the key " + key + " which fails to contain a period inside it");
			String name = keyEssential.substring(0, indexLastDot);
			String type = keyEssential.substring(indexLastDot + 1);
			
			NodeProp nodeProp = nameToNodeProps.get(name);
			if (nodeProp == null) {
				nodeProp = new NodeProp();
				nameToNodeProps.put(name, nodeProp);
			}
			
			if (type.equals("handlers")) {
				String value = properties.getProperty(key).trim();
				String[] classNames = value.split("(\\s+|\\s*,\\s*)", -1);
				Set<Class<? extends Handler>> classes = new HashSet<Class<? extends Handler>>( classNames.length );
				for (String className : classNames) {
					boolean added = classes.add( Class.forName(className).asSubclass(Handler.class) );
					if (!added) throw new IllegalStateException("className = " + className + " was specified more than once in the list for " + key + " in the logging properties");
				}
				nodeProp.handlerClasses = classes;
			}
			else if (type.equals("level")) {
				String value = properties.getProperty(key).trim();
				nodeProp.level = parseLevel(value);
			}
			else if (type.equals("useParentHandlers")) {
				String value = properties.getProperty(key).trim();
				nodeProp.useParentHandlers = Boolean.valueOf(value);
			}
			else throw new IllegalStateException("the logging properties specifies the key " + key + " whose final type part (" + type + ") is an illegal value");
		}
	}
	
	
	// -------------------- accessors --------------------
	
	
	/**
	* Returns an array of all the possible Levels, sorted into ascending order (according to {@link Level#intValue Level.intValue}).
	* <p>
	* Contract: the result is a clone of an internal field, so the caller may safely modify the result. 
	*/
	public static Level[] getLevels() { return levels.clone(); }
	
	
// +++ in the future, add a getConfiguration method like the currentConfiguration method described here:
// http://www.forward.com.au/javaProgramming/javaGuiTips/javaLoggingDebugging.html
	
	
	/** Accessor for {@link #logDirectory}. */
	public static File getLogDirectory() { return logDirectory; }
	
	
	/**
	* Accessor for {@link #logger2_default}.
	* Lazy initializes it if necessary.
	* <p>
	* @throws IllegalStateException if a Logger with the name "default" already exists
	* @throws SecurityException if a security manager exists and if the caller does not have LoggingPermission("control")
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static synchronized Logger2 getLogger2() throws IllegalStateException, SecurityException, RuntimeException {
		if (logger2_default == null) logger2_default = makeLogger2("default");
		return logger2_default;
	}
	
	
	// -------------------- makeLogger2 --------------------
	
	
	/**
	* Returns <code>{@link #makeLogger2(Class, String) makeLogger2}(c, null) )</code>.
	* <p>
	* @throws IllegalArgumentException if c == null
	* @throws IllegalStateException if a Logger with getName(c, null) already exists
	* @throws SecurityException if a security manager exists and if the caller does not have LoggingPermission("control")
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static Logger2 makeLogger2(Class c) throws IllegalArgumentException, IllegalStateException, SecurityException, RuntimeException {		
		return makeLogger2(c, null);
	}
	
	
	/**
	* Returns <code>{@link #makeLogger2(String) makeLogger2}( {@link #getName getName}(c, suffix) )</code>.
	* <p>
	* @throws IllegalArgumentException if c == null
	* @throws IllegalStateException if a Logger with getName(c, suffix) already exists
	* @throws SecurityException if a security manager exists and if the caller does not have LoggingPermission("control")
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static Logger2 makeLogger2(Class c, String suffix) throws IllegalArgumentException, IllegalStateException, SecurityException, RuntimeException {		
		return makeLogger2( getName(c, suffix) );
	}
	
	
	/**
	* Returns <code>{@link #makeLogger2(String, String) makeLogger2}( name, null )</code>.
	* <p>
	* @throws IllegalArgumentException if name is blank
	* @throws IllegalStateException if a Logger with name already exists
	* @throws SecurityException if a security manager exists and if the caller does not have LoggingPermission("control")
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static Logger2 makeLogger2(String name) throws IllegalArgumentException, IllegalStateException, SecurityException, RuntimeException {		
		return makeLogger2( name, null );
	}
	
	
	/**
	* Returns a new Logger2 created by a call to
	* <code>{@link Logger2#getLogger2(String, String) Logger2.getLogger2}(name, resourceBundleName)</code>.
	* So, only if this is the first Logger with name known by LogManager is it created.
	* <p>
	* If the Logger2 is created, then {@link #nsHandlers nsHandlers},
	* {@link #nsLevel nsLevel}, and {@link #nsUseParentHandlers nsUseParentHandlers} are called on it.
	* <p>
	* No other configuration is subsequently performed
	* (e.g. for the Level or Formatter of Logger2's handlers; all are left with their default settings,
	* such as whatever has been specified by some logging properties file).
	* <p>
	* This method only throws unchecked Exceptions so that it can be used to initialize a field at its declaration.
	* <p>
	* @throws IllegalArgumentException if name is blank
	* @throws IllegalStateException if a Logger with name already exists
	* @throws SecurityException if a security manager exists and if the caller does not have LoggingPermission("control")
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static Logger2 makeLogger2(String name, String resourceBundleName) throws IllegalArgumentException, IllegalStateException, SecurityException, RuntimeException {
		// name, resourceBundleName checked by new Logger2 below
		
		try {
			Logger2 logger2 = Logger2.getLogger2(name, resourceBundleName);
			nsHandlers(logger2);
			nsLevel(logger2);
			nsUseParentHandlers(logger2);
			return logger2;
		}
		catch (Throwable t) {
			throw ThrowableUtil.toRuntimeException(t); 
		}
	}
	
	
	// -------------------- getName, getPath --------------------
	
	
	/**
	* Determines an appropriate Logger name for the arguments.
	* The fully qualified name of c (i.e. package included) always appears first.
	* If suffix is non-null, then a '_' followed by suffix is appended.
	* Examples:
	* <ul>
	*  <li>c = java.lang.Double and suffix = null ==> "java.lang.Double"</li>
	*  <li>c = java.lang.Float and suffix = "diagnostics" ==> "java.lang.Float_diagnostics"</li>
	* </ul>
	* <p>
	* Motivation: Loggers are often static fields of a class,
	* so it is convenient to simply supply the Class object and have this method determine the name.
	* The suffix arg allows optional further discrimination
	* (e.g. a class may have multiple loggers for different kinds of events,
	* or, if each instance has its own Logger, then each one's name should have some instance specific part).
	* <p>
	* @throws IllegalArgumentException if c == null
	*/
	public static String getName(Class c, String suffix) throws IllegalArgumentException {
		if (c == null) throw new IllegalArgumentException("c == null");
		
		if (suffix != null) return c.getName() + "_" + suffix;
		else return c.getName();
	}
	
	
	/**
	* Returns <code>{@link #getPath getPath}( {@link #getName getName}(c, suffix) )</code>.
	* <p>
	* @throws IllegalArgumentException if c == null
	* @throws SecurityException if a security manager exists and its SecurityManager.checkWrite method does not permit directoryLogDefault and all necessary parent directories to be created
	* @throws IllegalStateException if an attempt to create directoryLogDefault is made but fails
	*/
	public static String getPath(Class c, String suffix) throws IllegalArgumentException, SecurityException, IllegalStateException {
		return getPath( getName(c, suffix) );
	}
	
	
	/**
	* Returns an appropriate path to a log file for name.
	* <p>
	* This is typically used with {@link #makeLogger2(Class, String)}.
	* The result is located in the directory returned by {@link #getLogDirectory getLogDirectory}
	* (see javadocs for possible side effects), and consists of name, a '_', a timestamp, and then ".log".
	* <p>
	* @throws IllegalArgumentException if name is blank
	* @throws SecurityException if a security manager exists and its SecurityManager.checkWrite method does not permit directoryLogDefault and all necessary parent directories to be created
	* @throws IllegalStateException if an attempt to create directoryLogDefault is made but fails
	*/
	public static String getPath(String name) throws IllegalArgumentException, SecurityException, IllegalStateException {
		if (StringUtil.isBlank(name)) throw new IllegalArgumentException("name is blank");
		
		String filename = name + "_" + DateUtil.getTimeStampForFile() + ".log";
		File path = new File(getLogDirectory(), filename);
		return path.getPath();
	}
	
	
	// -------------------- nsXXX --------------------
	
	
	/**
	* Searches thru the keys of {@link #nameToNodeProps} for the closest match to logger
	* which has an associated NodeProp value whose handlerClasses field is non-null.
	* The match is the longest key which logger's name starts with,
	* and is not necessarily a strict ancestor Logger name for logger.
	* For example, while <code>com.xyz</code> is an ancestor Logger of <code>com.xyz.foo</code>,
	* <code>com.xys.f</code> is not an ancestor of it but would still be a match.
	* This behavior was chosen to allow property files more power in specifying wildcard matches.
<!--
+++ would really like to have added true regular expression support,
but am not sure how, if multiple regular expressions match, can determine which is the best match.
Maybe a compromise would be to broaden the pattern into a substring instead of a prefix
(i.e. that it is implicitly *SSS* instead of just SSS*).
-->
	* <p>
	* If such a match is found, then its list of Handler Classes is retrieved
	* and a new Handler instance is created for each Class and is added to logger.
	* <p>
	* Note: if a {@link FileHandler} Class is encountered, it is treated specially:
	* its String arg constructor is called instead of its default constructor.
	* The param passed to that constructor is a call to {@link #getPath getPath}( logger.getName() ).
	* <p>
	* @throws IllegalArgumentException if logger == null
	* @throws IllegalAccessException if the Handler class or its nullary constructor is not accessible
	* @throws NoSuchMethodException if a matching method is not found
	* @throws InstantiationException if the Handler class represents an abstract class, an interface, an array class, a primitive type, or void; or if the class has no nullary constructor; or if the instantiation fails for some other reason
	* @throws InvocationTargetException if the underlying constructor throws an exception
	* @throws ExceptionInInitializerError - if the initialization provoked by this method fails
	* @throws SecurityException if a security manager, s, is present and finds some issue
	*/
	public static void nsHandlers(Logger logger) throws IllegalArgumentException, IllegalAccessException, NoSuchMethodException, InstantiationException, InvocationTargetException, ExceptionInInitializerError, SecurityException {
		if (logger == null) throw new IllegalArgumentException("logger == null");
		
		String keyBest = null;
		for (String key : nameToNodeProps.keySet()) {
			if (
				logger.getName().startsWith(key) &&
				(nameToNodeProps.get(key).handlerClasses != null)
			) {
				if (keyBest == null) keyBest = key;
				else if (keyBest.length() < key.length()) keyBest = key;
			}
		}
		if (keyBest == null) return;
		
		for (Class<? extends Handler> c : nameToNodeProps.get(keyBest).handlerClasses) {
			if (c == FileHandler.class) {
				String path = getPath( logger.getName() );
				Constructor<? extends Handler> constructor = c.getConstructor(String.class);
				logger.addHandler( constructor.newInstance(path) );
			}
			else {
				logger.addHandler( c.newInstance() );
			}
		}
	}
	
	
	/**
	* Uses the exact same search algorithm as {@link #nsHandlers nsHandlers}
	* except that looks for NodeProps with non-null level fields.
	* <p>
	* If an appropriate match is found, then its level is assigned to logger.
	* <p>
	* @throws IllegalArgumentException if logger == null
	*/
	public static void nsLevel(Logger logger) throws IllegalArgumentException {
		if (logger == null) throw new IllegalArgumentException("logger == null");
		
		String keyBest = null;
		for (String key : nameToNodeProps.keySet()) {
			if (
				logger.getName().startsWith(key) &&
				(nameToNodeProps.get(key).level != null)
			) {
				if (keyBest == null) keyBest = key;
				else if (keyBest.length() < key.length()) keyBest = key;
			}
		}
		if (keyBest == null) return;
		
		logger.setLevel( nameToNodeProps.get(keyBest).level );
	}
	
	
	/**
	* Uses the exact same search algorithm as {@link #nsHandlers nsHandlers}
	* except that looks for NodeProps with non-null useParentHandlers fields.
	* <p>
	* If an appropriate match is found, then its useParentHandlers boolean value is assigned to logger.
	* <p>
	* @throws IllegalArgumentException if logger == null
	*/
	public static void nsUseParentHandlers(Logger logger) throws IllegalArgumentException {
		if (logger == null) throw new IllegalArgumentException("logger == null");
		
		String keyBest = null;
		for (String key : nameToNodeProps.keySet()) {
			if (
				logger.getName().startsWith(key) &&
				(nameToNodeProps.get(key).useParentHandlers != null)
			) {
				if (keyBest == null) keyBest = key;
				else if (keyBest.length() < key.length()) keyBest = key;
			}
		}
		if (keyBest == null) return;
		
		logger.setUseParentHandlers( nameToNodeProps.get(keyBest).useParentHandlers );
	}


/*
+++ tell sun that they need to:
 --vastly improve their javadocs on logging; issues:
	discussion of root logger
	naming and inheritance
		some props like Level are pushed down; others like handlers are defined just at that node and rely on usePar to work
	discussion of global Logger
	param subst and what Object[] params really mean
	key/values that can appear in the logging.properties file
		--their JDK example should cite that the Handler stuff is in the approp Handler javadocs
		--give them my file as a better example
package description needs to be the equiv of hunter article

FileHandler autonaming using the pattern does not do what want; what want is the classname/Loggername
had to write subclass, but would not have had to if pattern did not have to be specified at cons


StreamHandler bugs:
	--many methods lack synch

Handler bugs:
	setFormatter uses newFormatter.getClass() to check for null instead of == null?
		http://jonasboner.com/2006/04/04/the-optimized-null-check-pattern/


java.util.logging.MemoryHandler
	javadocs say:
		A subclass overrides the log method
	what log nmethod?  do they mean publish?


Logger
    private synchronized ResourceBundle findResourceBundle(String name) {
needs to be public

	javadocs about subclass need to warn about
		setLevel
	because Logger fails to always use the getLevel accessor
	unlike
	where log(LogRecord record) uses getUseParentHandlers() instead of useParentHandlers


*/
	
	
	// -------------------- logRobustly --------------------
	
	
	/** Simply calls <code>{@link #logRobustly(Logger, Level, String, String, String, Throwable) logRobustly}(null, null, className, methodName, message, t)</code>. */
	public static void logRobustly(String className, String methodName, String message, Throwable t) {
		logRobustly(null, null, className, methodName, message, t);
	}
	
	
	/**
	* Uses logger to record level, className, methodName, message, and t.
	* <p>
	* There are no restrictions on any of the parameters.
	* They may even be null, altho if logger is null then {@link #getLogger2 getLogger2} is assigned to it
	* and if level is null then {@link Level#SEVERE Level.SEVERE} is assigned to it.
	* <p>
	* If another Throwable occurs during the logging process,
	* then an attempt is made to print the information to System.err followed by that second Throwable.
	* If a third Throwable occurs during this last step, it is caught and ignored and no further actions are taken.
	* <p>
	* Contract: this method never throws a Throwable.
	* This robustness makes it especially suitable for use in close methods (no try-catch must encase it).
	*/
	public static void logRobustly(Logger logger, Level level, String className, String methodName, String message, Throwable t) {
		try {
			if (logger == null) logger = getLogger2();
			if (level == null) level = Level.SEVERE;
			logger.logp(level, className, methodName, message, t);
		}
		catch (Throwable t2) {
			try {
				System.err.println();
				System.err.println("className = " + className + ", methodName = " + methodName + ", message = " + message);
				if (t != null) t.printStackTrace(System.err);
				System.err.println("Note: when first tried to log the above to a Logger, this Throwable occurred:");
				t2.printStackTrace(System.err);
			}
			catch (Throwable t3) {
				// can do nothing if reach here...
			}
		}
	}
	
	
	// -------------------- parseLevel --------------------
	
	
	/**
	* Returns the Level that corresponds to s.
	* <p>
	* @throws IllegalArgumentException if s is blank or no Level corresponds to s
	*/
	public static Level parseLevel(String s) throws IllegalArgumentException {
		if (StringUtil.isBlank(s)) throw new IllegalArgumentException("s is blank");
		
		for (int i = 0; i < levels.length; i++) {
			if (s.equals(levels[i].toString())) return levels[i];
		}
		throw new IllegalArgumentException("s = " + s + " does not correspond to any known Level instance");
	}
	
	
	// -------------------- close, flush, removeHandlers --------------------
	
	
	/**
	* Retrieves all of Logger's Handlers and closes them.
	* Does nothing if logger == null.
	* <p>
	* Any Throwable that occurs is simply logged using {@link #getLogger2 getLogger2}.
	* Thus, <i>this method should never throw a Throwable</i> (unless the logging process generates one).
	*/
	public static void close(Logger logger) {
		try {
			if (logger == null) return;
			
			for (Handler handler : logger.getHandlers()) handler.close();
		}
		catch (Throwable t) {
			getLogger2().logp(Level.SEVERE, "LogUtil", "close", "unexpected error caught", t);
		}
	}
	
	
	/**
	* Retrieves all of Logger's Handlers and flushes them.
	* <p>
	* @throws IllegalArgumentException if logger == null
	*/
	public static void flush(Logger logger) throws IllegalArgumentException {
		if (logger == null) throw new IllegalArgumentException("logger == null");
		
		for (Handler handler : logger.getHandlers()) handler.flush();
	}
	
	
	/**
	* Removes all of Logger's Handlers.
	* <p>
	* @throws IllegalArgumentException if logger == null
	*/
	public static void removeHandlers(Logger logger) throws IllegalArgumentException {
		if (logger == null) throw new IllegalArgumentException("logger == null");
		
		for (Handler handler : logger.getHandlers()) logger.removeHandler( handler );
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private LogUtil() {}
	
	
	// -------------------- NodeProp (static inner class) --------------------
	
	
	private static class NodeProp {
		private Set<Class<? extends Handler>> handlerClasses;
		private Level level; 
		private Boolean useParentHandlers;	// has to be Boolean and not boolean because need to detect null 
	}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_getLogger2() {    		
			Logger2 logger = null;
			try {
				logger = getLogger2();
				logger.logp(Level.WARNING, "LogUtil.UnitTest", "test_getLogger2", null, new Exception("test if LogUtil.getLogger2() can handle a null message; if see this in a log entry, then it does"));
				logger.logp(Level.INFO, "LogUtil.UnitTest", "test_getLogger2", "test if LogUtil.getLogger2() can handle a null Throwable; if see this in a log entry, then it does", (Throwable) null);
			}
			finally {
				flush(logger);
			}
		}
		
		@Test public void test_makeLogger2() {    		
			Logger2 logger = null;
			try {
				logger = makeLogger2( LogUtil.UnitTest.class, "test_makeLogger2" );
				
				logger.entering("LogUtil.UnitTest", "test_makeLogger2");
				logger.entering("LogUtil.UnitTest", "test_makeLogger2", new Object[] {"param1", "param2", "param3"});	// do this to test and see how method params can be printed using the entering method
				logger.exiting("LogUtil.UnitTest", "test_makeLogger2");
				logger.throwing("LogUtil.UnitTest", "test_makeLogger2", new Exception("deliberately generated Exception", new Exception("with a cause too")));
				
				logger.finest("step #1");
				logger.finer("step #2");
				logger.fine("step #3");
				logger.config("step #4");
				logger.info("step #5");
				logger.warning("step #6");
				logger.severe("step #7");
			}
			finally {
				close(logger);
			}
		}
		
	}
	
	
}
