/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


package d;


import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.InvocationTargetException;
import java.util.concurrent.atomic.AtomicLong;

import javax.swing.*;

import bb.gui.LineLayout;
import bb.util.ObjectState;
import bb.util.ThrowableUtil;


/**
* This class provides static methods which <i>send</i> output to a <i>GUI</i>.
* <p>
* <i>This class was designed to require as little typing as possible to use</i>.
* In particular, all public names (package, class, and method) are lower case and very short.
* So, you can avoid import statements and easily invoke all methods using their fully qualified form.
* Here is a typical use:
* <code><pre>
*    StringBuilder sb = new StringBuilder();
*    d.g.s(sb);</pre></code>
* <p>
* Mnemonic: the package name d stands for debug; the class name g stands for GUI; the methods named s stand for send.
* <p>
* This class is stateless, therefore it is multithread safe.
* Furthermore, this class ensures that the only thread which touches the gui code is EventQueue's dispatch thread.
* <p>
* @author Brent Boyer
*/
public class g {


	// -------------------- constructor --------------------


	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private g() {}


	// -------------------- s --------------------


	/**
	* This method calls {@link EventQueue} with a background task
	* which sends the obj arg's state to a GUI widget.
	* <p>
	* Specificly, it pops up a new window which first shows the label and then displays the state of obj.
	* It is perfectly acceptable to supply null for obj.
	* <p>
	* If pauseExecution is true, then the calling Thread waits until the user
	* either closes the window or clicks on the "Continue Execution" button before returning from this method.
	* <p>
	* @throws RuntimeException (or some subclass) if any other error occurs; this may merely wrap some other underlying Throwable
	*/
	public static void s(final String label, final Object obj, boolean pauseExecution) throws RuntimeException {
		try {
			if (pauseExecution) {
				final Object waitObject = new Object();
				synchronized (waitObject) {
						// see http://java.sun.com/docs/books/tutorial/uiswing/misc/threads.html for why must do ALL swing gui in event dispatch thread (the earlier advice given in http://java.sun.com/developer/JDCTechTips/2003/tt1208.html#1 to only do the displaying in the event dispatch thread is actually wrong)
					EventQueue.invokeLater( new Runnable() {
						public void run() {
							new ObjectStateDisplay(label, obj, waitObject);
						}
					} );
					waitObject.wait();
				}
			}
			else {
					// see http://java.sun.com/docs/books/tutorial/uiswing/misc/threads.html for why must do ALL swing gui in event dispatch thread (the earlier advice given in http://java.sun.com/developer/JDCTechTips/2003/tt1208.html#1 to only do the displaying in the event dispatch thread is actually wrong)
				EventQueue.invokeLater( new Runnable() {
					public void run() {
						new ObjectStateDisplay(label, obj, null);
					}
				} );
			}
		}
		catch (Throwable t) {
			throw ThrowableUtil.toRuntimeException(t);
		}
	}


	/** Simply calls <code>s(label, obj, true)</code>. */
	public static void s(String label, Object obj) throws RuntimeException {
		s(label, obj, true);
	}


	/** Simply calls <code>s(null, obj, pauseExecution)</code>. */
	public static void s(Object obj, boolean pauseExecution) throws RuntimeException {
		s(null, obj, pauseExecution);
	}


	/** Simply calls <code>s(null, obj, true)</code>. */
	public static void s(Object obj) throws RuntimeException {
		s(null, obj, true);
	}


	// -------------------- ObjectStateDisplay (inner class) --------------------


// +++ this class is very preliminary, AND NEEDS A LOT OF WORK...
	private static class ObjectStateDisplay extends JFrame {
		
		private static final long serialVersionUID = 1;
		
		/** Specifies how many spaces a tab char will expand to when display Strings. */
		protected static final int TAB_SIZE = 4;
		
		/**
		* Records the next ObjectStateDisplay instance's {@link #instanceId} field.
		* <p>
		* Contract: is initialized to 0, and if this field has that value, it means that no instances have been created.
		*/
		private static final AtomicLong instanceIdNext = new AtomicLong();
		
		/** Every time any ObjectStateDisplay instance is moved, this field stores its latest x position. */
		private static int latestX = 50;
		/** Every time any ObjectStateDisplay instance is moved, this field stores its latest y position. */
		private static int latestY = 50;
// +++ it would be nice to store the above to disk so the app would remember your preferences...


		/** Records this instance's Id. */
		private final long instanceId = instanceIdNext.incrementAndGet();
		
		private Object notifyObject;
		private final ObjectState objectState;
		private JButton continueButton;
		private JButton closeButton;
		private boolean firstActivation = true;
		
		
		private ObjectStateDisplay(String label, Object displayObject, Object notifyObject) throws IllegalStateException {
			if (!EventQueue.isDispatchThread()) throw new IllegalStateException("calling thread (" + Thread.currentThread().toString() + ") is not EventQueue's dispatch thread");

			this.objectState = new ObjectState(displayObject);
			this.notifyObject = notifyObject;

			setTitle( makeTitle(label) );
			synchronized (ObjectStateDisplay.class) {
				setLocation(latestX, latestY);
			}
			//setSize(600, 500);	// obsolete -- call pack below (will do an optimal setting)
			getContentPane().setLayout( new FlowLayout() );
			getContentPane().setBackground( new Color(225, 225, 225) );

			addComponentListener( new ComponentAdapter() {
				public void componentMoved(ComponentEvent ce) {
					Component c = ce.getComponent();
					synchronized (ObjectStateDisplay.class) {
						ObjectStateDisplay.latestX = c.getX();
						ObjectStateDisplay.latestY = c.getY();
					}
				}
			} );

			addWindowListener( new WindowAdapter() {
				public void windowClosing(WindowEvent we) { ObjectStateDisplay.this.handleWindowClose(); }

				/**
				* A call to <someComponent>.requestFocus() made when <someComponent> is initially created
				* usually fails to work.
				* The reason why is complicated and has to do with how Java initializes GUI stuff
				* and deals with native widgets. The workaround is to call requestFocus in the method below.
				* <p>
				* For more discussion, see:
				*	http://developer.java.sun.com/developer/bugParade/bugs/4103289.html
				*	http://www.jguru.com/faq/view.jsp?EID=45866
				*/
				public void windowActivated(WindowEvent we) {
					if (firstActivation) {
						firstActivation = false;
						if (continueButton.isEnabled()) continueButton.requestFocus();
					}
				}
			} );

			getContentPane().add( buildGui() );
			pack();

			setVisible(true);	// this call should always be last (e.g. http://developer.java.sun.com/developer/bugParade/bugs/4103289.html)
		}
		
		
		private String makeTitle(String label) {
			label = (label != null) ? " -- " + label : "";
			return "ObjectStateDisplay #" + instanceId + label;
		}
		
		
		private JComponent buildGui() {
			JPanel jpanel = new JPanel( LineLayout.makeVertical() );
			jpanel.setBorder( BorderFactory.createEmptyBorder(20, 20, 20, 20) );
			
			jpanel.add( new JLabel("Current Thread: " + Thread.currentThread().toString()) );
			
			jpanel.add( Box.createVerticalStrut(10) );
			if (objectState.getObject() == null) {
				jpanel.add( new JLabel(objectState.getType() ) );	// note that getType works even if the underlying Object is null
			}
			else {
				jpanel.add( new JLabel("Hashcode: " + objectState.getHashCode() ) );
				jpanel.add( buildFieldsDisplay() );
			}
			
			jpanel.add( Box.createVerticalStrut(10) );
			jpanel.add( buildButtons() );
			
			return jpanel;
		}
		
		
		private JComponent buildFieldsDisplay() {
// +++ do something better than the below in the future, like a JTable or something, with one row per item
// and have general objects be links (buttons?) so that can click on them and bring up their states, etc

// issues with JTable:
//	-- seems to be column centric
//	-- i do NOT want column labels; can i avoid them by using some of the constructors?
//	-- i do NOT want every cell in the right column to be rendered the same way
// (altho i probably DO want every cell in the left column to be a JLabel that is italic text, top centered & right justified)


//JTextArea jtextArea = new JTextArea(fieldInfos.length, 50);
JTextArea jtextArea = new JTextArea(35, 50);
jtextArea.setEditable(false);
jtextArea.setLineWrap(false);
jtextArea.setTabSize(TAB_SIZE);
jtextArea.append( objectState.toString() );

//return jtextArea;
JScrollPane jscrollPane = new JScrollPane(jtextArea);
jscrollPane.setPreferredSize( getScreenLimitedSize(jtextArea) );
return jscrollPane;
/*
*/

// +++ should allow the user to edit some of the table's values, like the field data?  that might be cool...

		}


		private JComponent buildButtons() {
			JPanel jpanel = new JPanel( LineLayout.makeHorizontal() );

			continueButton = new JButton("Continue Execution");
			continueButton.addActionListener( new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					handleContinueExecution();
				}
			} );
			jpanel.add( continueButton );

			jpanel.add( Box.createHorizontalStrut(20) );

			closeButton = new JButton("Close");
			closeButton.addActionListener( new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					handleWindowClose();
				}
			} );
			jpanel.add( closeButton );

			if (notifyObject != null) {
				this.getRootPane().setDefaultButton(continueButton);
				//continueButton.requestFocus();	// this will typically fail -- see the addWindowListener...windowActivated code in the constructor
			}
			else {
				handleContinueExecution();	// i.e. do a virtual click of continueButton at this point
			}

			return jpanel;
		}





		private Dimension getScreenLimitedSize(Component c) {
			Toolkit toolkit = java.awt.Toolkit.getDefaultToolkit();
			Dimension screensize = toolkit.getScreenSize();
// the above may not be quite what i need: in windoze, the dock at the bottom may reduce usable screen space...

// need to compare what screen offers to what c wants

return new Dimension(500, 400);
		}




		/**
		* This method always disables continueButton and sets its label to indicate that execution has proceeded.
		* Then it makes the close button be the default button.
		* Finally, if notifyObject is not null, it calls notifyObject.notifyAll() and then
		* dereferences notifyObject (assigns it to null) so that the notification code never executes twice.
		* This method is multithread safe.
		*/
		private synchronized void handleContinueExecution() {
			continueButton.setEnabled(false);
			continueButton.setText("Execution has proceeded");

			getRootPane().setDefaultButton(closeButton);
			closeButton.requestFocus();

			if (notifyObject != null) {
				synchronized (notifyObject) {
					notifyObject.notifyAll();
				}
				notifyObject = null;
			}
		}


		/** This method calls handleContinueExecution(), and then calls dispose(). */
		private void handleWindowClose() {
			handleContinueExecution();
			dispose();
		}

	}
/*

Programmer Notes:

--how should arrays be handled?  autoexpand all the entries?

-- could add Field metadata (e.g. a string of all the Modifiers for the field);
this would be most useful if do the above gui thing, whereby you would have to click
on a button to bring up all the meta data; to do it by default in the getState method is too much

-- may also be interested in the fields from superclasses too?
(i.e. need a method called getCompleteState)
In this case, need an algorithm that recursively goes up the inheritance tree to find fields

*/


	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		public static void main(String[] args) throws Exception {
			StringBuilder sb = new StringBuilder();
			sb.append("This is another...");
			d.g.s( "Some test StringBuilder", sb, true );
			d.g.s( "Some test Object", new Object(), true );
			d.g.s( null, null, false );
		}
		
	}
	
	
}