/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


/*
Programmer notes:

--should i add ability to stop the program just like the g class has in its s methods?
the user would need to hit something on the keyboard to continue execution...

--in my current code, Strings are always treated as primitives in that they just get simply printed out,
as opposed to having their complete internal state printed out; any need to ever print out a String's internal state?
if ever wanted this functionality, would need to define a separate s(String) method to handle the simple case,
and reserve the s(String, Object) & s(Object) methods to always print out the state of the Object arg.
*/


package d;


import java.io.*;

import org.junit.Test;

import bb.util.*;


/**
* This class provides static methods which <i>send</i> output to a {@link PrintStream}.
* You can use the various s methods to simply print out primitive types or Strings,
* or print out complete Object states.
* <p>
* The getps/setps methods are the accessor/mutator for the PrintStream used.
* By default, System.out is the initial PrintStream.
* The setps method is conveniently overloaded with a version that takes a file path String and
* generates a PrintStream that appends to the specified file.
* <p>
* <i>This class was designed to require as little typing as possible to use</i>.
* In particular, all public names (package, class, and method) are lower case and very short.
* So, you can avoid import statements and easily invoke all methods using their fully qualified form.
* Here are some typical uses:
* <code><pre>
*	d.p.s();	// prints a blank line
*	...
*	d.p.s("reached point #1");	// use to note that your program reached a certain point
*	...
*	d.p.s(i);	// i is an int, previously defined and assigned, that you now want to print
*	...
*	d.p.s("l = " + l);	// print out some descriptive text followed by some long value
*	...
*	OneOfMyTypes myObject = new OneOfMyTypes();
*	d.p.s("Complete state of OneOfMyTypes", myObject);	// print out an object's fields</pre></code>
* <p>
* Mnemonic: the package name d stands for debug; the class name p stands for PrintStream; the methods named s stand for send.
* <p>
* This class is multithread safe.
* <p>
* @author Brent Boyer
*/
public class p {


	// -------------------- static fields --------------------


	private static PrintStream printStream = null;	// always initialize this field in the static initializer method instead


	// -------------------- static initializer --------------------


	/**
	* Calls <code>setps(System.out)</code></li>.
	* Then registers a shutdown hook that calls {@link #onShutdown onShutdown}.
	* <p>
	* @throws IllegalArgumentException if setps finds a problem with System.out
	*/
	static {
		setps(System.out);

		Thread shutdownHook = new Thread(
			new Runnable() {
				public void run() {
					onShutdown();
				}
			},
			"p_ShutdownHook"
		);
		shutdownHook.setPriority( Thread.NORM_PRIORITY );
		Runtime.getRuntime().addShutdownHook( shutdownHook );
	}


	// -------------------- getps and setps --------------------


	/** Accessor for the <code>PrintStream</code> used by the <code>s</code> methods. */
	public static synchronized PrintStream getps() { return printStream; }


	/**
	* Mutator for the <code>PrintStream</code> used by the <code>s</code> methods.
	* Before the existing <code>PrintStream</code> is swapped out, it is first flushed.
	* <p>
	* @throws IllegalArgumentException if printStream == null; printStream is in the error state
	*/
	public static synchronized void setps(PrintStream printStream) throws IllegalArgumentException {
		if (printStream == null) throw new IllegalArgumentException("printStream is null");
		if (printStream.checkError()) throw new IllegalArgumentException("printStream is in the error state");
		
		if (p.printStream != null) p.printStream.flush();
		p.printStream = printStream;
	}


	/**
	* Calls the setps(PrintStream) with a new PrintStream which writes to the file specified by the filePath arg.
	* If that file does not exist, it will be created.  If it exists, the writing will append to it, not overwrite it.
	* The PrintStream is configured to autoflush (this conveniently helps ensure that data gets written to the file,
	* at a cost in performance).
	* <p>
	* @throws IllegalArgumentException if filePath is blank
	* @throws FileNotFoundException if filePath specifies a directory rather than a regular file,
	* or if it specifies a file that does not exist but cannot be created or opened
	* @throws SecurityException if a security manager exists and its checkWrite method denies write access to the file
	*/
	public static synchronized void setps(String filePath) throws IllegalArgumentException, FileNotFoundException, SecurityException {
		if ( StringUtil.isBlank(filePath) ) throw new IllegalArgumentException("filePath is blank");
		
		setps(
			new PrintStream( new FileOutputStream(filePath, true), true )
		);
	}
	
	
	// -------------------- s --------------------
	
	
	/**
	* Prints a blank line.
	* <p>
	* @throws IllegalStateException if printStream is in the error state
	*/
	public static synchronized void s() throws IllegalStateException {
		if (printStream.checkError()) throw new IllegalStateException("printStream is in the error state");
		
		printStream.println();
	}
	
	
	/**
	* Prints b on a line.
	* <p>
	* @throws IllegalStateException if printStream is in the error state
	*/
	public static synchronized void s(boolean b) throws IllegalStateException {
		if (printStream.checkError()) throw new IllegalStateException("printStream is in the error state");
		
		printStream.println(b);
	}
	
	
	/**
	* Prints c on a line.
	* <p>
	* @throws IllegalStateException if printStream is in the error state
	*/
	public static synchronized void s(char c) throws IllegalStateException {
		if (printStream.checkError()) throw new IllegalStateException("printStream is in the error state");
		
		printStream.println(c);
	}
	
	
	/**
	* Prints l on a line.
	* It should be used to print out any integral-type primitive (i.e. byte, short, int, long).
	* <p>
	* @throws IllegalStateException if printStream is in the error state
	*/
	public static synchronized void s(long l) throws IllegalStateException {
		if (printStream.checkError()) throw new IllegalStateException("printStream is in the error state");
		
		printStream.println(l);
	}
	
	
	/**
	* Prints d on a line.
	* It should be used to print out any floating point-type primitive (i.e. float, double).
	* <p>
	* @throws IllegalStateException if printStream is in the error state
	*/
	public static synchronized void s(double d) throws IllegalStateException {
		if (printStream.checkError()) throw new IllegalStateException("printStream is in the error state");
		
		printStream.println(d);
	}
	
	
	/**
	* Prints charSequence on a line.
	* <p>
	* @throws IllegalStateException if printStream is in the error state
	*/
	public static synchronized void s(CharSequence charSequence) throws IllegalStateException {
		if (printStream.checkError()) throw new IllegalStateException("printStream is in the error state");
		
		printStream.println(charSequence);
	}
	
	
	/**
	* Prints out label followed by obj.
	* Both args are optional (either may be null). 
	* <p>
	* If label is non-null, it is printed followed by ": " on one line.
	* Next, the complete state of obj is printed using {@link ObjectState}.
	* <p>
	* @throws IllegalStateException if printStream is in the error state
	*/
	public static synchronized void s(String label, Object obj) throws IllegalStateException {
		if (printStream.checkError()) throw new IllegalStateException("printStream is in the error state");
		
		if (label != null) printStream.println(label + ": ");
		printStream.println( (new ObjectState(obj).toString()) );
	}
	
	
	// -------------------- onShutdown --------------------
	
	
	/** Performs actions that should only be done before the JVM shuts down, such as flushing printStream. */
	private static synchronized void onShutdown() {
		//printStream.close();
		//printStream = null;
			// do NOT do the above, since other Threads during shutdown could conceivably still call methods of this class (see the javadocs on Runtime.addShutdownHook)
			// so just flush printStream if it still exists:
		if (printStream != null) printStream.flush();
	}
	
	
	// -------------------- constructor --------------------
	
	
	/** This private constructor suppresses the default (public) constructor, ensuring non-instantiability. */
	private p() {}
	
	
	// -------------------- UnitTest (static inner class) --------------------
	
	
	/** See the Overview page of the project's javadocs for a general description of this unit test class. */
	public static class UnitTest {
		
		@Test public void test_setps() throws IllegalArgumentException, FileNotFoundException, SecurityException {
			try {
				String filepath = "../log/testoutput.txt";
				d.p.setps(filepath);
				d.p.s("If the setps(String) method is working, then this text will be written to the file " + filepath);
			}
			finally {
				d.p.setps(System.out);
			}
		}
		
		@Test public void test_s_boolean() throws IllegalStateException {
			d.p.s(true);
			d.p.s(false);
		}
		
		@Test public void test_s_char() throws IllegalStateException {
			d.p.s('a');
			d.p.s('b');
			d.p.s('c');
		}
		
		@Test public void test_s_long() throws IllegalStateException {
			d.p.s();
			d.p.s("byte:");
			byte b = 1;
			d.p.s(b);
			
			d.p.s();
			d.p.s("short:");
			short sh = 12;
			d.p.s(sh);
			
			d.p.s();
			d.p.s("int:");
			int i = 123;
			d.p.s(i);
			
			d.p.s();
			d.p.s("long:");
			long l = 1234;
			d.p.s(l);
		}
		
		@Test public void test_s_double() throws IllegalStateException {
			d.p.s();
			d.p.s("float:");
			float f = 1234.5f;
			d.p.s(f);
			
			d.p.s();
			d.p.s("double:");
			double dbl = 12345.56;
			d.p.s(dbl);
		}
		
		@Test public void test_s_CharSequence() throws IllegalStateException {
			d.p.s();
			d.p.s("Pass a String to the s(CharSequence) method");
			
			d.p.s();
			d.p.s( new StringBuffer("Pass a StringBuffer to the s(CharSequence) method") );
			
			d.p.s();
			d.p.s( new StringBuilder("Pass a StringBuilder to the s(CharSequence) method") );
		}
		
		@Test public void test_s_StringObject() throws IllegalStateException {
			d.p.s();
			d.p.s("Pass a null reference as the 2nd arg of the s(String, Object) method", null);
			d.p.s();
			d.p.s("Pass a String as the 2nd arg of the s(String, Object) method", "Mary had a little lamb...");
			d.p.s();
			d.p.s("Pass an int[] as the 2nd arg of the s(String, Object) method", new int[0]);
		}
		
	}


}